/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * @author UNSW 2002 Robocup (David Wang)
 *
 * Last modification background information
 * $Id: RoboCommander.java 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 *
**/

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.util.*;


public class RoboCommander extends JFrame {

    private static final boolean debugMsg      = false;
    private static final boolean delayDebugMsg = false;
    private static final boolean hasTestButton = false;
    private static final boolean showInfoPanel = true;
    private static final boolean replaceLongCommandDebugMsg = false;

    // world model display constants
    private static final int MAX_NUM_PLAYER                = 4;
    private static final int WORLD_MODEL_DISPLAY_PORT_BASE = 5019;
    private int[] WORLD_MODEL_DISPLAY_PORTS                = new int[MAX_NUM_PLAYER];
    private static final String DEFAULT_PLAYER             = "1";

    private static final boolean LATENCY_TEST3      = false;
    private static final boolean LATENCY_TEST4      = false;
    private static int trial3                       = 0;
    private static int trial4                       = 0;

    CommandSender commandSender = new CommandSender("RoboCommander");

    JScrollPane scrollPane    = new JScrollPane();
    JTextArea dataInfoArea    = new JTextArea();
    JPanel userPanel          = new JPanel();
    
    JMenuBar menuBar        = new JMenuBar();
    JMenu menu              = new JMenu("RoboCommander");
    JMenuItem cplaneControl = new JMenuItem("CPlane Control Interface");
    JMenuItem humanControl  = new JMenuItem("Human Control Interface");

    // GUI for the commands
    JPanel commandPanel        = new JPanel();
    TitledBorder commandBorder = new TitledBorder("Command Panel");
    JLabel nameLabel           = new JLabel("Name:");
    JLabel valueLabel          = new JLabel("Value:");
    JTextField nameField       = new JTextField(RoboConstant.MODE_TRIGGER);
    JTextField valueField      = new JTextField("0");
    JButton sendButton         = new JButton("Send");
    JButton motionButton       = new JButton("Motion");
    JButton testButton         = new JButton("Test");
    JCheckBox controlCheckBox  = new JCheckBox("Manual Control", false);
    
    // GUI for the command top panels (shortcuts)
    JPanel quickKeyPanel        = new JPanel();
    TitledBorder quickKeyBorder = new TitledBorder("Quick Key Panel");
    
    JPanel quickKeyPanel1       = new JPanel();
    JButton pauseLocaliseButton = new JButton("Pause Game and localise");
    JButton startGameButton     = new JButton("Start Game");
    JButton stopGameButton      = new JButton("Stop Game");

    JPanel quickKeyPanel2       = new JPanel();
    JLabel playerLabel          = new JLabel("Player: ");
    JComboBox playerList        = new JComboBox();
    JButton worldModelButton    = new JButton("World Model");
    JButton cplaneButton        = new JButton("CPlane");
    JButton burstButton         = new JButton("BURST");
    JButton yuvplaneButton      = new JButton("YUVPlane");
    JButton oplaneButton        = new JButton("OPlane");

    // GUI for the log
    TitledBorder logBorder      = new TitledBorder("Log Panel");
    JPanel logPanel             = new JPanel();
    
    JPanel oplaneLogPanel           = new JPanel();
    JCheckBox oplaneLogFileCheckBox = new JCheckBox("Log Oplane?", false);
    JLabel oplaneLogFileLabel       = new JLabel("Log:");
    JTextField oplaneLogFileField   = new JTextField(RoboConstant.DEFAULT_OPLANE_LOG_FILE, 30);
    JButton oplaneChooseButton      = new JButton("Browse...");


    // stores forward, left, turn, tilt, pan values
    private float speeds[] = {0,0,0,0,0};
    
    RoboWirelessBase parent;
    int sendPort, receivePort;
    OutputStream out;
    PrintWriter oplaneLogFile;
    PrintWriter startTime, endTime;
    int currentCommand;

    public RoboCommander(RoboWirelessBase parent, int sendPort, int receivePort) {

        super("RoboCommander");
        this.parent      = parent;
        this.sendPort    = sendPort;
        this.receivePort = receivePort;
        try {

            commandSender.start();
            (new FrameRateDisplay()).start();

            // initialise port numbers for world model display
            for (int i = 0; i < MAX_NUM_PLAYER; i++) {
                WORLD_MODEL_DISPLAY_PORTS[i] = WORLD_MODEL_DISPLAY_PORT_BASE + i;
            }
        
            WorldModelDisplay[] wmDisplays = new WorldModelDisplay[MAX_NUM_PLAYER];

            // establish threads for world model display
            for (int i = 0; i < MAX_NUM_PLAYER; i++) {
                Integer portNumber = new Integer(WORLD_MODEL_DISPLAY_PORTS[i]);
                wmDisplays[i] = new WorldModelDisplay(portNumber.toString());
            }

            // start threads for world model display
            for (int i = 0; i < MAX_NUM_PLAYER; i++) {
                wmDisplays[i].start();
            }

            init();
            if (showInfoPanel) {
                setSize(650,600);
            } else {
                pack();
            }
            show();
        } catch(Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }
    
    private void init() throws Exception {
        menu.setMnemonic(KeyEvent.VK_R);
        menu.setFont(new Font("Dialog", 1, 24));
        
        cplaneControl.setMnemonic(KeyEvent.VK_C);
        cplaneControl.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, ActionEvent.CTRL_MASK));
        cplaneControl.setToolTipText("Show the CPlane Control interface");
        cplaneControl.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                showCPlaneControl();
            }
        });
        humanControl.setMnemonic(KeyEvent.VK_H);
        humanControl.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_H, ActionEvent.CTRL_MASK));
        humanControl.setToolTipText("Show the HumanControl RoboCommander interface");
        humanControl.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                showHumanControl();
            }
        });
        
        menu.add(cplaneControl);
        menu.add(humanControl);
        
        menuBar.add(menu);

        if (showInfoPanel) {
            dataInfoArea.setFont(new Font("Monospaced", 0, 12));
            scrollPane.getViewport().add(dataInfoArea);
            scrollPane.setAutoscrolls(true);
        }
        
        nameField.setMinimumSize(new Dimension(60, 21));
        nameField.setPreferredSize(new Dimension(60, 21));
        nameField.addFocusListener(new FocusAdapter() {
            public void focusGained(FocusEvent e) {
                nameField.selectAll();
            }
        });
        nameField.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send();
            }
        });
        valueField.setPreferredSize(new Dimension(100, 21));
        valueField.addFocusListener(new FocusAdapter() {
            public void focusGained(FocusEvent e) {
                valueField.selectAll();
            }
        });
        valueField.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send();
            }
        });
        sendButton.setMnemonic(KeyEvent.VK_S);
        sendButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send();
            }
        });
        motionButton.setMnemonic(KeyEvent.VK_M);
        motionButton.addKeyListener(new KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                moveDog(e);
            }
        });
        controlCheckBox.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (controlCheckBox.isSelected()) {
                    // take control
                    commandSender.send(RoboConstant.CONTROL_TRIGGER, RoboConstant.CONTROL_START_VALUE);
                } else {
                    // release control
                    commandSender.send(RoboConstant.CONTROL_TRIGGER, RoboConstant.CONTROL_END_VALUE);
                }
            }
        });
        if (hasTestButton) {
            if (LATENCY_TEST4) {
                try {
                    startTime = new PrintWriter(new BufferedWriter(new FileWriter("04start")));
                } catch (Exception ex1) {
                    System.out.println("Can't create file 04start");
                }
                try {
                    endTime = new PrintWriter(new BufferedWriter(new FileWriter("04end")));
                } catch (Exception ex) {
                    System.out.println("Can't write file 04end");
                }
            }
            testButton.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    int numTimes = 10;
                    int[] speeds = {1,2,3,4,5,6,7,8,7,6,5,4,3,2,1,0};

                    if (LATENCY_TEST4) {
                        long time = System.currentTimeMillis();
                        startTime.println("t4start, trial #"+trial4+"="+time);
                        startTime.flush();
                    }

                    for (int j=0 ; j<numTimes ; j++) {
                        for (int i=0; i<speeds.length ; i++) { 
                            send("f",speeds[i]+"");
                            if (debugMsg) {
                                System.out.println("Sent: "+"f " + speeds[i]);
                            }
                        }
                    }
                    if (LATENCY_TEST4) {
                        long time = System.currentTimeMillis();
                        endTime.println("t4end, trial #"+trial4+"="+time);
                        endTime.flush();
                        trial4++;
                    }
                }
            });
        }

        commandBorder.setTitleFont(new Font("Helvetica", Font.BOLD, 14));
        commandPanel.setBorder(commandBorder);
        commandPanel.add(nameLabel);
        commandPanel.add(nameField);
        commandPanel.add(valueLabel);
        commandPanel.add(valueField);
        commandPanel.add(sendButton);
        commandPanel.add(motionButton);
        commandPanel.add(controlCheckBox);
        if (hasTestButton) {
            commandPanel.add(testButton);
        }
        
        pauseLocaliseButton.setMnemonic(KeyEvent.VK_P);
        pauseLocaliseButton.setToolTipText("(" + RoboConstant.MODE_TRIGGER + ", 5)");
        pauseLocaliseButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send(RoboConstant.MODE_TRIGGER, "5");
            }
        });
        startGameButton.setMnemonic(KeyEvent.VK_G);
        startGameButton.setToolTipText("(" + RoboConstant.MODE_TRIGGER + ", 0)");
        startGameButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send(RoboConstant.MODE_TRIGGER, "0");
            }
        });
        stopGameButton.setMnemonic(KeyEvent.VK_T);
        stopGameButton.setToolTipText("(" + RoboConstant.MODE_TRIGGER + ", 1)");
        stopGameButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send(RoboConstant.MODE_TRIGGER, "1");
            }
        });
        
        quickKeyPanel1.add(pauseLocaliseButton);
        quickKeyPanel1.add(startGameButton);
        quickKeyPanel1.add(stopGameButton);
        
        worldModelButton.setMnemonic(KeyEvent.VK_W);
        worldModelButton.setToolTipText("(" + RoboConstant.WORLDMODEL_TRIGGER + ", " + DEFAULT_PLAYER + ")");
        worldModelButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send(RoboConstant.WORLDMODEL_TRIGGER, (String) playerList.getSelectedItem());
                switchVisiblePlayer();
            }
        });
        cplaneButton.setMnemonic(KeyEvent.VK_C);
        cplaneButton.setToolTipText("(" + RoboConstant.CPLANE_TRIGGER + ", " + DEFAULT_PLAYER + ")");
        cplaneButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send(RoboConstant.CPLANE_TRIGGER, (String) playerList.getSelectedItem());
                switchVisiblePlayer();
            }
        });
        burstButton.setToolTipText("(" + RoboConstant.BURST_TRIGGER + ", " + DEFAULT_PLAYER + ")");
        burstButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send(RoboConstant.BURST_TRIGGER, (String) playerList.getSelectedItem());
            }
        });
        yuvplaneButton.setMnemonic(KeyEvent.VK_Y);
        yuvplaneButton.setToolTipText("(" + RoboConstant.YUVPLANE_TRIGGER + ", " + DEFAULT_PLAYER + ")");
        yuvplaneButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send(RoboConstant.YUVPLANE_TRIGGER, (String) playerList.getSelectedItem());
            }
        });
        oplaneButton.setMnemonic(KeyEvent.VK_O);
        oplaneButton.setToolTipText("(" + RoboConstant.OPLANE_TRIGGER + ", " + DEFAULT_PLAYER + ")");
        oplaneButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                commandSender.send(RoboConstant.OPLANE_TRIGGER, (String) playerList.getSelectedItem());
                showHumanControl();
                switchVisiblePlayer();
            }
        });
        for (int i=0; i<=MAX_NUM_PLAYER; i++) {
            playerList.addItem(""+i);
        }
        playerList.setSelectedItem(DEFAULT_PLAYER);
        playerList.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                worldModelButton.setToolTipText("(" + RoboConstant.WORLDMODEL_TRIGGER + ", " + (String) playerList.getSelectedItem() + ")");
                cplaneButton.setToolTipText("(" + RoboConstant.CPLANE_TRIGGER + ", " + (String) playerList.getSelectedItem() + ")");
                oplaneButton.setToolTipText("(" + RoboConstant.OPLANE_TRIGGER + ", " + (String) playerList.getSelectedItem() + ")");
            }
        });
        quickKeyPanel2.add(playerLabel);
        quickKeyPanel2.add(playerList);
        quickKeyPanel2.add(worldModelButton);
        quickKeyPanel2.add(cplaneButton);
        quickKeyPanel2.add(burstButton);
        quickKeyPanel2.add(yuvplaneButton);
        quickKeyPanel2.add(oplaneButton);

        quickKeyBorder.setTitleFont(new Font("Helvetica", Font.BOLD, 14));
        quickKeyPanel.setBorder(quickKeyBorder);
        quickKeyPanel.setLayout(new BorderLayout());
        quickKeyPanel.add(quickKeyPanel1, BorderLayout.CENTER);
        quickKeyPanel.add(quickKeyPanel2, BorderLayout.SOUTH);
        
        oplaneLogFileCheckBox.setToolTipText("If selected, the oplanes will get saved in the oplane log file specified");
        oplaneLogFileCheckBox.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (oplaneLogFileCheckBox.isSelected()) {
                    if (oplaneLogFileField.getText().length()==0) {
                        oplaneLogFileField.setText(RoboConstant.DEFAULT_OPLANE_LOG_FILE);
                    }
                    oplaneLogFileField.requestFocus();
                }
            }
        });
        oplaneLogFileField.addFocusListener(new FocusAdapter() {
            public void focusGained(FocusEvent e) {
                oplaneLogFileField.selectAll();
            }
        });

        oplaneChooseButton.setToolTipText("Choose a file to log oplane");
        oplaneChooseButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                JFileChooser chooser;
                String currentDir = "";
                int slashPos, dotPos;
                
                // start the file chooser in the directory of the specified file
                currentDir = oplaneLogFileField.getText().trim();
                slashPos   = currentDir.lastIndexOf(File.separator);
                if (slashPos==-1) {
                    chooser = new JFileChooser();
                } else {
                    chooser = new JFileChooser(currentDir.substring(0, slashPos));
                }
                int returnVal = chooser.showOpenDialog(null);
                if(returnVal == JFileChooser.APPROVE_OPTION) {
                    oplaneLogFileField.setText(chooser.getSelectedFile().getPath());
                }
            }
        });
        oplaneLogPanel.add(oplaneLogFileCheckBox);
        oplaneLogPanel.add(new JPanel());
        oplaneLogPanel.add(oplaneLogFileLabel);
        oplaneLogPanel.add(oplaneLogFileField);
        oplaneLogPanel.add(oplaneChooseButton);
        
        logBorder.setTitleFont(new Font("Helvetica", Font.BOLD, 14));
        logPanel.setBorder(logBorder);
        logPanel.add(oplaneLogPanel);

        userPanel.setLayout(new BorderLayout());
        userPanel.add(commandPanel, BorderLayout.NORTH);
        userPanel.add(quickKeyPanel, BorderLayout.CENTER);
        userPanel.add(logPanel, BorderLayout.SOUTH);
                
        this.setDefaultCloseOperation(3);
        if (showInfoPanel) {
            this.getContentPane().add(menuBar, BorderLayout.NORTH);
            this.getContentPane().add(scrollPane, BorderLayout.CENTER);
            this.getContentPane().add(userPanel, BorderLayout.SOUTH);
        } else {
            this.getContentPane().add(menuBar, BorderLayout.NORTH);
            this.getContentPane().add(userPanel, BorderLayout.CENTER);
        }
    }
    
    private void switchVisiblePlayer() {
        if (playerList.getSelectedIndex()==0) {
            playerList.setSelectedItem(DEFAULT_PLAYER);
        } else {
            playerList.setSelectedIndex(0);
        }
    }
    
    private void showHumanControl() {
        parent.getHumanControl().setVisible(true);
    }

    private void showCPlaneControl() {
        parent.getCPlaneDisplay().setVisible(true);
    }

    public void printValues() {
        for (int i = 0; i < speeds.length; i++) {
            System.out.print(speeds[i] + " ");
        }
        System.out.println();
    }

    public void printToConsole(String line) {
        if (showInfoPanel) {
            dataInfoArea.append(line + "\n");
            dataInfoArea.setCaretPosition(dataInfoArea.getDocument().getLength());
        }
    }
    
    
    // utility methods to be used in other java files
    public void send(String name) {
        commandSender.send(name, "");
    }
    public void send(String name, String value) {
        commandSender.send(name, value);
    }

    public void setOPlaneLogFile(PrintWriter file) {
        this.oplaneLogFile = file;
    }
    
    public PrintWriter getOPlaneLogFile() {
        return oplaneLogFile;
    }
    public JCheckBox getOPlaneLogFileCheckBox() {
        return oplaneLogFileCheckBox;
    }
    public void setCommands(String name, String value) {
        nameField.setText(name);
        valueField.setText(value);
    }
    public void moveDog(KeyEvent e) {
        int speed, mult;
        int kc = e.getKeyCode();

        if (kc==KeyEvent.VK_PLUS || kc==KeyEvent.VK_EQUALS) {
            // increase speed
            if ((RoboConstant.KEYS[currentCommand]==RoboConstant.K_FORWARD || 
                 RoboConstant.KEYS[currentCommand]==RoboConstant.K_LEFT) &&
                speeds[currentCommand/2]<RoboConstant.MAX_FORWARD_SPEED) {
                speeds[currentCommand/2] += RoboConstant.FORWARD_STEP;
            } else if ((RoboConstant.KEYS[currentCommand]==RoboConstant.K_BACKWARD ||
                        RoboConstant.KEYS[currentCommand]==RoboConstant.K_RIGHT) &&
                       speeds[currentCommand/2]>-RoboConstant.MAX_FORWARD_SPEED) {
                speeds[currentCommand/2] -= RoboConstant.FORWARD_STEP;
            } else if (RoboConstant.KEYS[currentCommand]==RoboConstant.K_TURNCCW && 
                       speeds[currentCommand/2]<RoboConstant.MAX_TURN_SPEED) {
                speeds[currentCommand/2] += RoboConstant.TURN_STEP;
            } else if (RoboConstant.KEYS[currentCommand]==RoboConstant.K_TURNCW &&
                       speeds[currentCommand/2]>-RoboConstant.MAX_TURN_SPEED) {
                speeds[currentCommand/2] -= RoboConstant.TURN_STEP;
            }
            commandSender.send(RoboConstant.COMMANDS[currentCommand], speeds[currentCommand/2] + "");
            return;
        } else if (kc==KeyEvent.VK_MINUS) {
            // reduce speed
            if ((RoboConstant.KEYS[currentCommand]==RoboConstant.K_FORWARD || 
                 RoboConstant.KEYS[currentCommand]==RoboConstant.K_LEFT) &&
                speeds[currentCommand/2]>-RoboConstant.MAX_FORWARD_SPEED) {
                speeds[currentCommand/2] -= RoboConstant.FORWARD_STEP;
            } else if ((RoboConstant.KEYS[currentCommand]==RoboConstant.K_BACKWARD ||
                        RoboConstant.KEYS[currentCommand]==RoboConstant.K_RIGHT) &&
                       speeds[currentCommand/2]<RoboConstant.MAX_FORWARD_SPEED) {
                speeds[currentCommand/2] += RoboConstant.FORWARD_STEP;
            } else if (RoboConstant.KEYS[currentCommand]==RoboConstant.K_TURNCCW && 
                       speeds[currentCommand/2]>-RoboConstant.MAX_TURN_SPEED) {
                speeds[currentCommand/2] -= RoboConstant.TURN_STEP;
            } else if (RoboConstant.KEYS[currentCommand]==RoboConstant.K_TURNCW &&
                       speeds[currentCommand/2]<RoboConstant.MAX_TURN_SPEED) {
                speeds[currentCommand/2] += RoboConstant.TURN_STEP;
            }
            commandSender.send(RoboConstant.COMMANDS[currentCommand], speeds[currentCommand/2] + "");
            return;
        }
        // test which key is pressed
        for (int i = 0; i<RoboConstant.KEYS.length; i++) {
            if (kc==RoboConstant.KEYS[i]) {
                if (i>=RoboConstant.STANCE_KEYS_INDEX) {
                    if (e.isControlDown() && RoboConstant.COMMANDS[i].equals(RoboConstant.C_NORMAL)) {
                        commandSender.send(RoboConstant.COMMANDS[i].toLowerCase(), RoboConstant.KICK_VALUES[i-RoboConstant.STANCE_KEYS_INDEX]+"");
                    } else {
                        commandSender.send(RoboConstant.COMMANDS[i], RoboConstant.KICK_VALUES[i-RoboConstant.STANCE_KEYS_INDEX]+"");
                    }
                } else {

                    currentCommand = i;

                    // otherwise need to take care of reverse movement and speed
                    speed = i/2 * 2;
                    mult  = 1 - i%2*2;
                    if (e.isShiftDown()) {
                        // slow
                        speed++;
                    }

                    // if diff motion then send command
                    if (speeds[i/2] != mult*RoboConstant.SPEEDS[speed]) {
                        if ((speeds[i/2]>0 && mult*RoboConstant.SPEEDS[speed]<0) ||
                            (speeds[i/2]<0 && mult*RoboConstant.SPEEDS[speed]>0)) {
                            speeds[i/2] = 0;
                        } else {
                            speeds[i/2] = mult*RoboConstant.SPEEDS[speed];
                        }

                        commandSender.send(RoboConstant.COMMANDS[i], speeds[i/2] + "");
                    }
                }
                break;
            }

        }

        if (debugMsg) {
            System.out.println("pressed = " + kc );
        }
    }


    // for sending commands
    private class CommandSender extends Thread {
        private String filename;
        OutputStream out;
        boolean startOPlaneLog = true;

        public CommandSender(String filename) {
            this.filename = filename;
        }
        
        public void run() {
            try {
                ServerSocket ss = new ServerSocket(sendPort);
                System.out.println(filename + " (send," + sendPort + "): ready");
                Socket s = ss.accept();
                out = s.getOutputStream();
                //ss.close();
            } catch(Exception e) {
                e.printStackTrace();
                System.exit(1);
            }
        }

        // sends what ever is in the text fields
        public void send() {
            // something to send
            if (!nameField.getText().equals("") && !valueField.getText().equals("")) {
                send(nameField.getText(), valueField.getText());
            }
        }

        // name and value must be shorter than 255
        public void send(String name, String value) {
            try {
                if (out==null) {
                    System.out.println(filename + " (send," + sendPort + "): not ready yet");
                    return;
                }
                
                // save commands into the oplane logFile
                if (startOPlaneLog && oplaneLogFileCheckBox.isSelected()) {
                    try {
                        if (oplaneLogFile==null) {
                            oplaneLogFile = new PrintWriter(new BufferedWriter(new FileWriter(RoboConstant.DEFAULT_OPLANE_LOG_FILE)));
                        }
                        oplaneLogFile.println(name + "," + value);
                        oplaneLogFile.flush();
                    } catch (IOException e) {
                        JOptionPane.showMessageDialog(null, "Error: Cannot write to file " + RoboConstant.DEFAULT_OPLANE_LOG_FILE, "Error", JOptionPane.ERROR_MESSAGE);
                    }
                }
                
                checkStartCPlaneLog(name, value);
                checkStartOPlaneLog(name, value);
                
                // replace commands with something shorter
                if (name.length()>=RoboConstant.NAME_SIZE) {
                    if (replaceLongCommandDebugMsg) {
                        System.out.println("RoboCommander.java: replaced " + name + " with " + name.substring(0, RoboConstant.NAME_SIZE-1));
                    }
                    name = name.substring(0, RoboConstant.NAME_SIZE-1);
                }
                
                // replace values with something shorter
                if (value.length()>=RoboConstant.VALUE_SIZE) {
                    if (debugMsg) {
                        System.out.println("RoboCommander.java: replaced " + value + " with " + value.substring(0, RoboConstant.VALUE_SIZE-1));
                    }
                    value = value.substring(0, RoboConstant.VALUE_SIZE-1);
                }

                byte[] len1 = new byte[1];
                byte[] len2 = new byte[1];

                len1[0] = (byte) name.length();
                len2[0] = (byte) value.length();

                if (debugMsg) {
                    System.out.println("len12 " + len1 + " " + len2);
                }

                if (LATENCY_TEST3) {
                   long time = System.currentTimeMillis();
                   System.out.println("t3start, trial #"+trial3+"="+time);
                   trial3++;
                }

                out.write(len1);
                out.write(name.getBytes(),0,name.length());
                out.write(len2);
                out.write(value.getBytes(),0,value.length());
                
                if (delayDebugMsg) {
                    System.out.println("RoboCommander: (" + name + ", " + value + ")");
                }

            } catch (IOException e) {
                e.printStackTrace();
                System.exit(1);
            }
        }
        
        private void checkStartCPlaneLog(String name, String value) {
            if (name.trim().equals(RoboConstant.CPLANE_TRIGGER) && RoboUtility.valueType(value)==RoboConstant.START_TYPE) {
                parent.getCPlaneDisplay().getPlayBackFileField().setEnabled(false);
                parent.getCPlaneDisplay().getLoadButton().setEnabled(false);
            } else if (name.trim().equals(RoboConstant.CPLANE_TRIGGER) && RoboUtility.valueType(value)==RoboConstant.STOP_TYPE) {
                parent.getCPlaneDisplay().getPlayBackFileField().setEnabled(true);
                parent.getCPlaneDisplay().getLoadButton().setEnabled(true);
            }
        }


        private void checkStartOPlaneLog(String name, String value) {
            if (name.trim().equals(RoboConstant.OPLANE_TRIGGER) && RoboUtility.valueType(value)==RoboConstant.START_TYPE) {
                startOPlaneLog = true;
                parent.getHumanControl().getPlayBackFileField().setEnabled(false);
                parent.getHumanControl().getLoadButton().setEnabled(false);
            } else if (name.trim().equals(RoboConstant.OPLANE_TRIGGER) && RoboUtility.valueType(value)==RoboConstant.STOP_TYPE) {
                startOPlaneLog = false;
                parent.getHumanControl().getPlayBackFileField().setEnabled(true);
                parent.getHumanControl().getLoadButton().setEnabled(true);
            }
        }
    }

    // receives frame rate info from client
    private class FrameRateDisplay extends Thread {

        public void run() {
            try {

                ServerSocket ss = new ServerSocket(receivePort);
                System.out.println("FrameRateDisplay (receive," + receivePort + "): ready");
                Socket s = ss.accept();
                BufferedReader in = new BufferedReader(new InputStreamReader(s.getInputStream()));
                ss.close();                
        
                double[] data = new double[200]; // temporary for now
                int loc = 0; // current loc in data array
                boolean inData = false;

                for (String line = in.readLine();line != null;line = in.readLine()) {                    
                    dataInfoArea.append(line + "\n");
                    dataInfoArea.setCaretPosition(dataInfoArea.getDocument().getLength());
                }

                System.out.println("Connection Closed");
                in.close();
                s.close();
            } catch (IOException e) {
                System.err.println("GLDisplay.Reader.run: IOException");
                e.printStackTrace();
            }
        }
    }

    // receives world model display data from client
    private class WorldModelDisplay extends Thread {
        private int portNumber;
        private GLDisplay d;

        public WorldModelDisplay(String port) {
            try {
                portNumber = Integer.parseInt(port);
            } catch (NumberFormatException e) {
                System.err.println(e);
            }
        }

        public void run() {
            try {

                ServerSocket ss = new ServerSocket(portNumber);
                System.out.println("WorldModelDisplay (receive," + portNumber + "): ready");
                Socket s = ss.accept();
                BufferedReader in = new BufferedReader(new InputStreamReader(s.getInputStream()));
                ss.close();                
        
                double[] data = new double[200]; // temporary for now
                int loc = 0; // current loc in data array
                boolean inData = false;
                
                initialiseLogFile(oplaneLogFileField, oplaneLogFileCheckBox, oplaneLogFile);

                for (String line = in.readLine();line != null;line = in.readLine()) {                    
			
                    //line = line.trim();
                    if (!inData && line.startsWith(RoboConstant.RINFO_DELIMA_S)) { // start of data
                        if (debugMsg) {
                            System.out.println("Start of info");
                        }
                        inData = true;
                    } else if (inData && line.startsWith(RoboConstant.RINFO_DELIMA_E)) { // end of data
                        if (debugMsg) {
                            System.out.println("end of info");
                        }
                        inData = false;
                        processData(data,loc);
                        loc = 0;
                    } else if (inData) { // robot data
                        try {
                            data[loc++] = Double.parseDouble(line);
                        } catch (NumberFormatException e) {
                            System.err.println("GLDisplay.Reader.run: failed for parse info: " + line);
                            System.err.println(e);
                        }
                    } else {
                        // execution should not reach here
                        // frame count data is sent through another port
                        System.err.println("RoboCommander: This should not get printed");
                    }
                }

                // ignore the rest
                System.out.println("Connection Closed");
                in.close();
                s.close();
                if (oplaneLogFile!=null) {
                    oplaneLogFile.close();
                }
                oplaneLogFileField.setEnabled(true);
            } catch (IOException e) {
                System.err.println("GLDisplay.Reader.run: IOException");
                e.printStackTrace();
            }

            // close world model display
            if (d!=null) {
                d.dispose();
            }
        }
        
        private void initialiseLogFile(JTextField logFileField, JCheckBox logFileCheckBox, PrintWriter logFile) throws IOException {
            String filename = logFileField.getText().trim();
            if (logFileCheckBox.isSelected() && !filename.equals("")) {
                logFile = new PrintWriter(new BufferedWriter(new FileWriter(filename)));
            }
        }

        // display the world model
        public void processData(double[] data, int size) {
            Integer playerNum = new Integer((int) data[0]);
            if (d == null) {
                d = new GLDisplay(playerNum.intValue());
            }    
            d.processData(playerNum.intValue(), data, 1, size - 1);
        }
    }
}
