/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: RoboWirelessBase.java 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 *
**/


public class RoboWirelessBase {

    public static final boolean debugMsg = false;

    public static final int PORT1  = 5005; // Robo commander send
    public static final int PORT2  = 5006; // Robo commander receive 
    public static final int PORT3  = 5010; // cplane
    public static final int PORT4  = 5011; // yuv plane
    public static final int PORT5  = 5014;
    public static final int PORT6  = 5015; // debug stream
    public static final int PORT7  = 5102; // oplane
    public static final int PORT8  = 5016; // joystick receive
    
    private RoboCommander roboCommander;
    private CPlaneDisplay cplaneDisplay;
    private RoboYUVPainter roboYUVPainter;
    private RoboYUVLogger roboYUVLogger;
    private Grapher grapher;
    private DebugModule debugModule;
    private HumanControl humanControl;
    
    private static boolean logYUV;
 
    public static void main(String[] args) {
        if (args.length != 0 && args[0].equals("logYUV")) {
            logYUV = true;		
	} else {
	    logYUV = false;
	}
        new RoboWirelessBase();
    }
    
    RoboWirelessBase() {
        humanControl   = new HumanControl(this, PORT7, PORT8);
        roboCommander  = new RoboCommander(this, PORT1,PORT2);
        cplaneDisplay  = new CPlaneDisplay(this, PORT3);
	if (logYUV) {
	    roboYUVLogger = new RoboYUVLogger();
	} else {
            roboYUVPainter = new RoboYUVPainter(PORT4, roboCommander);
	}
        grapher        = new Grapher(PORT5, roboCommander);
        debugModule    = new DebugModule(PORT6, roboCommander, DebugModule.MODE_POSITION);
    }
    
    public RoboCommander getRoboCommander() {
        return roboCommander;
    }
    
    public HumanControl getHumanControl() {
        return humanControl;
    }

    public CPlaneDisplay getCPlaneDisplay() {
        return cplaneDisplay;
    }
}
