/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/**
 * Title: readJoystick
 * Description: reads joystick input, scales it down to acceptable RoboCommander inputs
 *              and sends it to RoboWirelessBase
 * Copyright: Copyright (c) 2003
 * Company:
 * @author UNSW 2003 Robocup (Terry Tam)
 * @version 1.0
 */

#ifndef NO_JOYSTICK_CONTROL_D

// for joystick
#include "stdio.h"
#include "fcntl.h"
#include "unistd.h"
#include "sys/ioctl.h"
#include "linux/joystick.h"
#include "stdlib.h"
#define JOY_DEV "/dev/js0"

// for udp client
#include "sys/types.h"
#include "sys/socket.h"
#include "netinet/in.h"
#include "arpa/inet.h"
#include "netdb.h"
#include "stdio.h"
#include "unistd.h"

#include "string.h"
#include "sys/time.h"
#include "sys/timeb.h"
#include "time.h"
#include "pthread.h"

#define REMOTE_SERVER_PORT  5016
#define MAX_MSG             20
#define HOST                "LOCALHOST"
#define DEFAULT_AXIS        0
#define NUM_BUTTONS         15
#define NUM_SLIDERS         3
#define LATENCY_TEST1       0

#define JOYSTICK_KEY        "j"
#define commandDebugMsg     0

int abs(int);
int checkPattern(int, int);
extern void sendRobotCommand(char [], char []);

void* readJoystick(void *) {

    clock_t start_time;

    // for joystick
    int joy_fd, *axis = NULL, num_of_axis = 0, num_of_buttons = 0, x = 0;
    int type = 0, action = 0, button_pressed = 0;
    int old_type = 0, old_action = NUM_BUTTONS;
    char *button = NULL;
    struct js_event js;
    int old_x = DEFAULT_AXIS, old_y = DEFAULT_AXIS, old_z = DEFAULT_AXIS;
    int new_x = DEFAULT_AXIS, new_y = DEFAULT_AXIS, new_z = DEFAULT_AXIS;
    int debugMsg = 0;
    int buttonDebugMsg = 0;
    int trial1 = 0;

    // for udp client
    int sd, rc, i;
    struct sockaddr_in cliAddr, remoteServAddr;
    struct hostent *h;
    char msg[20] = "";

    // get server IP address (no check if input is IP address or DNS name)
    h = gethostbyname(HOST);
    if(h==NULL) {
        printf("%s: unknown host\n", HOST);
        pthread_exit(0);
    }

    if (debugMsg) {
    printf("%s: sending data to '%s' (IP : ) \n", h->h_name,
        inet_ntoa(*(struct in_addr *)h->h_addr_list[0]));
    }
    remoteServAddr.sin_family = h->h_addrtype;
    memcpy((char *) &remoteServAddr.sin_addr.s_addr, 
    h->h_addr_list[0], h->h_length);
    remoteServAddr.sin_port = htons(REMOTE_SERVER_PORT);

    // socket creation
//    sd = socket(AF_INET,SOCK_DGRAM,0);
    sd = socket(AF_INET,SOCK_STREAM,0); //!! added for TCP
    if(sd<0) {
        printf("cannot open socket \n");
        pthread_exit(0);
    }
  
    // bind any port
    cliAddr.sin_family = AF_INET;
    cliAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    cliAddr.sin_port = htons(0);

    rc = bind(sd, (struct sockaddr *) &cliAddr, sizeof(cliAddr));
    if(rc<0) {
        printf("cannot bind port\n");
        pthread_exit(0);
    }

//!! added for TCP

    /* connect to server */
/*
    rc = connect(sd, (struct sockaddr *) &remoteServAddr, sizeof(remoteServAddr));
    if(rc<0) {
        printf( "Cannot connect\n" );
        pthread_exit(0);
    }
*/
/*
    if((joy_fd = open(JOY_DEV, O_RDONLY)) == -1) {
        printf( "Couldn't open joystick\n" );
        pthread_exit(0);
    }
*/
    // read number of joystick axes
    ioctl( joy_fd, JSIOCGAXES, &num_of_axis );
    axis = (int *) calloc( num_of_axis, sizeof( int ) );

    // read number of buttons
    ioctl( joy_fd, JSIOCGBUTTONS, &num_of_buttons);
    button = (char *) calloc( num_of_buttons, sizeof( char ) );

    fcntl( joy_fd, F_SETFL, O_NONBLOCK );// use non-blocking mode

    // loop to continuously read from joystick
    while(1) {
        button_pressed = 0;
        type = 0;
        action = NUM_BUTTONS;

        read(joy_fd, &js, sizeof(struct js_event)); // read joystick state

        // see what to do with the event
        switch (js.type & ~JS_EVENT_INIT) {
            case JS_EVENT_AXIS:
                axis [ js.number ] = js.value;
                break;
            case JS_EVENT_BUTTON:
                button [ js.number ] = js.value;
                break;
        }

        // check all buttons (except sliders) to see if they have been pressed
        for( x=0 ; x < NUM_BUTTONS - NUM_SLIDERS  ; ++x ) {
        if (buttonDebugMsg) {
            printf("B%d: %d ", x, button[x] );
        }
            if (button[x]) {
                if (buttonDebugMsg) {
                    printf("PRESSED: button num = %d\n", x);
                }
                button_pressed = 1;
                action = x;

                // check to see what setting slider is set to
                if (button[14]) {
                    type = 1;
                } else if (button[13]) {
                    type = 2;
                } else if (button[12]) {
                    type = 3;
                } 
                fflush(stdout);
                break;
            }
        }

/*
        printf("----------\n");
        printf("button_pressed = %d\n", button_pressed);
        printf("old_type = %d\n", old_type);
        printf("old_action = %d\n", old_action);
        printf("type = %d\n", type);
        printf("action = %d\n", action);
        printf("==========\n");
*/
        // send button data if a button has been pressed
        // and current button press is not same as previous one

        if (button_pressed && (old_type != type || old_action != action)) {

            if (buttonDebugMsg) {
                printf( "Sending... T: %d B: %d\n" ,type , action);
            }
            fflush(stdout);

            // create message of format "B:t,b," button data to send
            // need extra comma at end to delimit end of data with previous buffered data
            // B: signals that the msg is button data, not axis data
            // t  shows the type of button (e.g. type 1, 2 or 3, manipulated by slider on joystick)
            // b  shows which button number was pressed
            sprintf(msg, "B:%d %d\0", type, action);

            // send message//!! change this?
            rc = send(sd, msg, strlen(msg), 0); // added for TCP
//            rc = sendto(sd, msg, strlen(msg)+1, 0, (struct sockaddr *) &remoteServAddr, 
//            sizeof(remoteServAddr));

            if(rc<0) {
                printf("cannot send data %d \n",i-1);
                close(sd);
                pthread_exit(0);
            }
        }
        // update button presses
        old_type = type;
        old_action = action;

        // scale joystick data down to acceptable robocommander inputs
        // from [-512,511] to [-8,8] for x and y-axes
        // from [-512,511] to [-45,45] for z-axis

        // xy-axis approach: 
        // if positive, add 1 to change range from [0,511] to [1,512],
        //              then divide by 64 to get range from [0,8]
        //              change sign, as joystick data is inverted by default
        // if negative, same as positive, but no need to add 1

        if (axis[0] > 0) {
            new_x = -((axis[0]+1) / 64);
        } else {
            new_x = -(axis[0] / 64);
        }

        if (axis[1] > 0) {
            new_y = -((axis[1]+1) / 64);
        } else {
            new_y = -(axis[1] / 64);
        }

        // z-axis approach:
        // same as xy-axis, but scaled to [-21,21] range instead of [-45,45]
        // as it is too sensitive. the new range was calculated via experimentation
        if (axis[2] > 0) {
            new_z = -((axis[2]+1) / 24);
        } else {
            new_z = -(axis[2] / 24); 
        }


// get rid of?

        // throttle joystick data sending rate, so that wireless doesn't drop commands
        // new data is reset back to old data, based on the test condition
        if (checkPattern(new_x,2)) {
            new_x = old_x; // executed if number is NOT in set {-8,-6,-4,-2,0,2,4,6,8}
        }
        
        if (checkPattern(new_y,2)) {
            new_y = old_y; // executed if number is NOT in set {-8,-6,-4,-2,0,2,4,6,8}
        }

        if (new_z!=0 && !checkPattern(new_z,5)) {
            new_z = old_z; // executed if number NOT in set {-21,-16,-11,-6,-1,0,1,6,11,16,21}
        }

// get rid of?



// added
        // if turning and moving at same time, decrease turn sensitivity
        if (new_z != 0 && (new_x != 0) || (new_y != 0)) {
            // ignore turning if sensitivity is less than 2
            if (abs(new_z) < 10) {
                new_z = 0;
            }
        }

/*
        // set values to full speed whenever joystick is moved
        if (new_x > 0) {
            new_x = 8;
        } else if (new_x < 0) {
            new_x = -8;
        }
        if (new_y > 0) {
            new_y = 8;
        } else if (new_y < 0) {
            new_y = -8;
        }
        if (new_z > 0) {
            new_z = 20;
        } else if (new_z < 0) {
            new_z = -20;
        }
*/
        // if strafing and moving forward/backwards, slow down
        if (new_x != 0 && new_y != 0) {
            if (new_x > 0) {
                new_x = 4;
            } else if (new_x < 0) {
                new_x = -4;
            }
            if (new_y > 0) {
                new_y = 4;
            } else if (new_y < 0) {
                new_y = -4;
            }
        }

        // if strafing and turning, slow down
        if (new_x != 0 && new_z != 0) {
            if (new_x > 0) {
                new_x = 4;
            } else if (new_x < 0) {
                new_x = -4;
            }
            if (new_z > 0) {
                new_z = 10;
            } else if (new_z < 0) {
                new_z = -10;
            }
        }

        // if moving forward/backwards and turning, slow down
        if (new_y != 0 && new_z != 0) {
            if (new_y > 0) {
                new_y = 4;
            } else if (new_y < 0) {
                new_y = -4;
            }
            if (new_z > 0) {
                new_z = 10;
            } else if (new_z < 0) {
                new_z = -10;
            }
        }
// added

        // if the new command is same as old, don't send anything
        if ((old_x == new_x) && (old_y == new_y) && (old_z == new_z)) {
            continue;
        }

        // keep copy of new data
        old_x = new_x;
        old_y = new_y;
        old_z = new_z;


        // create message of format "x y z" joystick data to send
        // need extra comma at end to delimit end of data with previous buffered data
//        sprintf(msg, "%d %d %d\0", new_x, new_y, new_z);
/*
        unsigned char *x_value = (unsigned char *) malloc(sizeof(unsigned char) * 4);
        memcpy(x_value, &new_x, sizeof(unsigned char) * 4);
        unsigned char *y_value = (unsigned char *) malloc(sizeof(unsigned char) * 4);
        memcpy(y_value, &new_y, sizeof(unsigned char) * 4);
        unsigned char *z_value = (unsigned char *) malloc(sizeof(unsigned char) * 4);
        memcpy(z_value, &new_z, sizeof(unsigned char) * 4);
*/
/*
        unsigned char *values = (unsigned char *) malloc(sizeof(unsigned char) * 12);
        memcpy(values, &new_x, sizeof(unsigned char) * 4);
        memcpy(values+sizeof(unsigned char)*4, &new_y, sizeof(unsigned char) * 4);
        memcpy(values+2*(sizeof(unsigned char)*4), &new_z, sizeof(unsigned char) * 4);
*/
/*
    if (1) {
        cout << "x_value= ";
        cout << ((int)(x_value[0] & 0xff) + 
                 (int)((x_value[1] & 0xff) << 8) + 
                 (int)((x_value[2] & 0xff) << 16) + 
                 (int)((x_value[3] & 0xff) << 24)) << " | ";
        cout << "y_value= ";
        cout << ((int)(y_value[0] & 0xff) + 
                 (int)((y_value[1] & 0xff) << 8) + 
                 (int)((y_value[2] & 0xff) << 16) + 
                 (int)((y_value[3] & 0xff) << 24)) << " | ";
        cout << "z_value= ";
        cout << ((int)(z_value[0] & 0xff) + 
                 (int)((z_value[1] & 0xff) << 8) + 
                 (int)((z_value[2] & 0xff) << 16) + 
                 (int)((z_value[3] & 0xff) << 24)) << "\n";
    }
*/
        if (LATENCY_TEST1) {
            struct timeb tp;
            ftime(&tp);
            if (tp.millitm < 10) {
                printf("t1start, trial #%d=%ld00%d\n",trial1,tp.time,tp.millitm);
            } else if (tp.millitm < 100) {
                printf("t1start, trial #%d=%ld0%d\n",trial1,tp.time,tp.millitm);
            } else {
                printf("t1start, trial #%d=%ld%d\n",trial1,tp.time,tp.millitm);
            }
            fflush(stdout);
            trial1++;
        }

        // send message // added for TCP
//        rc = send(sd, values, sizeof(const unsigned char*)*3,0);

        char name[3];
        char value[10];	
        sprintf(name, JOYSTICK_KEY);
        sprintf(value, "%d %d %d", new_x, new_y, new_z);
        
        if (commandDebugMsg) {
            printf("(%s, %d %d %d)", JOYSTICK_KEY, new_x, new_y, new_z);
        }
        sendRobotCommand(name, value);
/*
        rc = send(sd, x_value, sizeof(const unsigned char*),0);
        rc = send(sd, y_value, sizeof(const unsigned char*),0);
        rc = send(sd, z_value, sizeof(const unsigned char*),0);
*/
//	rc = send(sd, msg, strlen(msg), 0);
/*
	if(rc<0) {
            perror("cannot send data ");
            close(sd);
            pthread_exit(0);
	}
*/
/*
        rc = sendto(sd, msg, strlen(msg)+1, 0, (struct sockaddr *) &remoteServAddr, 
        sizeof(remoteServAddr));

        if(rc<0) {
            printf("cannot send data %d \n",i-1);
            close(sd);
            pthread_exit(0);
        }
*/

        if (debugMsg) {
            printf( "Sent: %d %d %d$\n", new_x, new_y, new_z);
            fflush(stdout);
        }

        // wait before reading joystick data again, in order to throttle data sending
/*
        start_time = clock();
        while((clock() - start_time) < (0.1 * CLOCKS_PER_SEC)) {
        }
*/
    } // end of while

    close( joy_fd );
    pthread_exit(0);
}

// checks if a given number n has a remainder of 1, when divided by modNum
// written to handle both positive and negative numbers
int checkPattern(int n, int modNum) {
    if (n >= 0) {
        return (n%modNum==1);
    } else {
        int m = -n;
        return (m%modNum==1);
    }
}

int abs(int n) {
    if (n >= 0) {
        return n;
    } else {
        return (-n);
    }
}

#endif // NO_JOYSTICK_CONTROL_D
