/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/*
 * Last modification background information
 * $Id: ActionPro.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#include "ActionPro.h"

ActionPro::ActionPro():eCommander() {
    //actionSuccess = true;
	takenFirstStep = false;
	leading_leg = NULL;
    currentWalkType = NULL;
	shareStep_ = NULL;
	currentPG = NULL;
    desiredPan = NULL;
    desiredTilt = NULL;
    stepComplete = NULL;

}

void ActionPro::initMem() {
    int sizeOfNeededRegion = sizeof(ActuatorSharedMemory);     //for stepComplete only needed in push interface
						   
    shmPWalk = new RCRegion(sizeOfNeededRegion);
    
    ActuatorSharedMemory *ptr = (ActuatorSharedMemory *)shmPWalk->Base();

	if (ptr == NULL) {
		*(char *)NULL  = 5;	// crash
	}

	cout << "ActuatorControl shared memory: " << (int)ptr << endl;
	
	leading_leg = &(ptr->leading_leg);
    currentWalkType = &(ptr->currentWalkType);
    shareStep_ = &(ptr->shareStep_);
	currentPG = &(ptr->currentPG);
	desiredPan = &(ptr->desiredPan);
	desiredTilt = &(ptr->desiredTilt);
	stepComplete = &(ptr->stepComplete);

 	//initialise to avoid crashing caused by accessing random memory
	*leading_leg     = 0;
	*currentWalkType = false;
	*desiredPan       = 0;
	*desiredTilt      = 0;
	*stepComplete    = true;


	//indicatorCommander = new Indicators(sIndicators, eCommander);
    headAgent = new HeadMotion(sHead, eCommander, desiredPan, desiredTilt);
    pAgent    = new PWalk(sEffector, eCommander  /*, forLeadingLeg, forStepComplete*/ );
    eCommander.start();

	*shareStep_ 	 = *(pAgent->ptr_step_);
	*currentPG 	 	 = *(pAgent->ptr_PG);

}


void ActionPro::SetPAgent(OSubject *sbjShmPWalk) {

    MemoryRegionID mem = shmPWalk->MemID();
    size_t offset = shmPWalk->Offset();
    char* base = shmPWalk->Base();
    size_t size = shmPWalk->Size(); 

    sbjShmPWalk->SetData(&mem,    sizeof(mem));
    sbjShmPWalk->SetData(&offset, sizeof(offset));
    sbjShmPWalk->SetData(&base,   sizeof(base));
    sbjShmPWalk->SetData(&size,   sizeof(size));
    sbjShmPWalk->NotifyObservers();
}



void ActionPro::execute (const ONotifyEvent& event) {

    *shareStep_ = *(pAgent->ptr_step_);
	*currentPG 	= *(pAgent->ptr_PG);
	
	next = (AtomicAction)*(const AtomicAction *)event.Data(0);
	
	#ifdef PSTEP_PULL_INTERFACE
	
	if (!takenFirstStep) {
        pAgent->pStep (next.walkType, next.Forward, next.Left, 
                       next.turnCCW, next.speed, next.hF, next.hB, next.hdF, 
                       next.hdB, next.ffO, next.fsO, next.bfO, next.bsO);  
		takenFirstStep = true;
	}		
	
    if (pAgent->pwalkNotUsingHead && sHead->IsReady () && takenFirstStep) {
        headAgent->SendHeadCommand (next.headType, next.panx, next.tilty, next.nullz);
        //actionSuccess |= execHEAD;
    }	
	
	#endif //PSTEP_PULL_INTERFACE
	
	
	#ifdef PSTEP_PUSH_INTERFACE
	
    if (sEffector->IsReady () && *stepComplete) {
        pAgent->pStep (next.walkType, next.Forward, next.Left, 
                       next.turnCCW, next.speed, next.hF, next.hB, next.hdF, 
                       next.hdB, next.ffO, next.fsO, next.bfO, next.bsO);    

        *stepComplete = false;
		*currentWalkType = next.walkType;
        //actionSuccess |= execLEGS;
    }
    if (pAgent->pwalkNotUsingHead && sHead->IsReady ()) {
        headAgent->SendHeadCommand (next.headType, next.tilty, next.panx, next.nullz);
        //actionSuccess |= execHEAD;
    }
	
	#endif //PSTEP_PUSH_INTERFACE
}

void ActionPro::ContinueStep () {

	*shareStep_ = *(pAgent->ptr_step_);
	*currentPG 	= *(pAgent->ptr_PG);

	//extract some info from pAgent through these const pointers.
	int walktype = *(pAgent->ptr_walktype);
	int step_    = *(pAgent->ptr_step_);
	int PG       = *(pAgent->ptr_PG);

    //if haven't finished the kick keep going until i have finished the kick
	if ((walktype == ChestPushWT  && pAgent->chestpush->step < pAgent->chestpush->duration) ||
        (walktype == FwdKickWT    && pAgent->fwdkick->step   < pAgent->fwdkick->duration  ) ||
        (walktype == FastKickWT   && pAgent->fastkick->step  < pAgent->fastkick->duration ) ||
		(walktype == GetupFrontWT && pAgent->getupf->step    < pAgent->getupf->duration   ) ||
		(walktype == GetupSideWT  && pAgent->getups->step    < pAgent->getups->duration   ) ||
		(walktype == GetupRearWT  && pAgent->getupr->step    < pAgent->getupr->duration   ) ||
		(walktype == BackKickWT   && pAgent->backKick->step  < pAgent->backKick->duration )	||
		(walktype == CheerWT      && pAgent->cheer->step     < pAgent->cheer->duration )	
        ) {
        pAgent->pStep (walktype, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
		*leading_leg = 0;
        *stepComplete = false;
    } 
	//if haven't finished half step keep going until i have finished the half step
	else if (step_ != PG && step_ < 2 * PG) {
        pAgent->pStep (walktype, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
        *stepComplete = false;
    } 
	//else read the next action and do that
	else {
        if (step_ == PG) {
            *leading_leg = 1; //right
        } else if (step_ == 2 * PG) {
            *leading_leg = -1; //left
        } else {
            *leading_leg = 0;
        }
        *stepComplete = true;
		
		#ifdef PSTEP_PULL_INTERFACE

		if (sEffector->IsReady () && takenFirstStep) {
        	pAgent->pStep (next.walkType, next.Forward, next.Left, 
                    	   next.turnCCW, next.speed, next.hF, next.hB, next.hdF, 
                    	   next.hdB, next.ffO, next.fsO, next.bfO, next.bsO);    

        	*currentWalkType = next.walkType;
			*stepComplete = false;
        	//actionSuccess |= execLEGS;
    	}
		
		#endif //PSTEP_PULL_INTERFACE
    }

	//cout << "from actuatorcontrol ll, wt, pan, tilt " << *leading_leg << "," << *currentWalkType << "," << *hPanShare << "," << *hTiltShare << endl;

}




