/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * UNSW 2003 Robocup (Eric Chung)
 *
 * Last modification background information
 * $Id: HeadMotion.h 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
 * Head Controls
 *
**/

#ifndef _HeadMotion_h_DEFINED
#define _HeadMotion_h_DEFINED

#include <OPENR/OSubject.h>
#include <OPENR/OObserverVector.h>
#include <OPENR/OPENREvent.h>
#include <OPENR/OPENRAPI.h>

#include "EffectorCommander.h"
#include "../share/Common.h"
#include "../share/ActionProDef.h"
#include "../share/PWalkDef.h"
#include "primitives.h"

/* the magnitude (not sign) should agree with base/work/RoboCommander/RoboConstant.java */
static const double MIN_TILT = -88.5;
static const double MAX_TILT = 43;
static const double MIN_PAN  = -89.6;
static const double MAX_PAN  = 89.6;

class HeadMotion {

public:

    HeadMotion(OSubject * sh, EffectorCommander & eCommander, double *dPan, double *dTilt);
    ~HeadMotion() {}

    /*
     * @param headType is the type of head movement, either ABS_H (absolute), REL_H (relative), or NUL_H, (null)
     * @param tilt is the vertical angle of head in degrees. Up is positive
     * @param pan is the horizontal angle of head in degrees. Left is positive
    **/
    void SendHeadCommand (int headType, double panx, double tilty, double nullz);

    void SetUpHeadBuffer (EffectorCommander & eCommander);

private:
    
    double destTilt, destPan;
    double *desiredTilt;
    double *desiredPan;        

    OCommandVectorData *headVec_;
    MemoryRegionID headVecID_;
    RCRegion *headRegion;
    
    OSubject *sHead;
    
    /* determine tilt and pan from a 3D vector from base of the neck */
    void setDestVarsXYZ(double x, double y, double z);

};

#endif //_HeadMotion_h_DEFINED
