/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id:
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#include "Behaviours.h"
#include "beckham.h"

static int spinDribbleTimer = 0;
static int spinDribbleLastHeadtype = ABS_XYZ;
static double spinDribbleLastHeadpanx = 0;
static double spinDribbleLastHeadtilty = 0;
static double spinDribbleLastHeadnullz = 0;

void Behaviours::aaSpinDribble() {
    Beckham::doBeckhamTrackVisualBall();

    spinDribbleTimer++;
    if (spinDribbleTimer == 1) {
        smartSetBeacon();
        spinDribbleLastHeadtype = headtype;
        spinDribbleLastHeadpanx = panx;
        spinDribbleLastHeadtilty = tilty;
        spinDribbleLastHeadnullz = nullz;
    }    
    if (spinDribbleTimer < 12) {
        setHoldParams();
        headtype = spinDribbleLastHeadtype;            
        panx = spinDribbleLastHeadpanx;
        tilty = spinDribbleLastHeadtilty;
        nullz = spinDribbleLastHeadnullz;
        //Beckham::doBeckhamTrackVisualBall();
        return;
    } else if (spinDribbleTimer < 20) {
        setHoldParams();
        smartActiveGpsLocalise(); 
        return;
    }
        
    if (!gps->canSee(vobBall)) {
        headtype = spinDribbleLastHeadtype;            
        panx = spinDribbleLastHeadpanx;
        tilty = spinDribbleLastHeadtilty;
        nullz = spinDribbleLastHeadnullz;
    }

    double reqAccuracy = requiredAccuracy(20);
    reqAccuracy = MAX(reqAccuracy, 5);

    double minGoalHead = gps->tGoal().pos.head - reqAccuracy;
    double maxGoalHead = gps->tGoal().pos.head + reqAccuracy;

    double min = MIN(abs(minGoalHead), abs(maxGoalHead));
    double max = MAX(abs(minGoalHead), abs(maxGoalHead));

    // if ball aligned correctly, dribble ball towards target goal
    /*if (ABS (gps->tGoal().pos.getHead()) <= accuracyRequirement &&
        !vision->facingFieldEdge) {
    */    
    if (maxGoalHead >= 0 && minGoalHead <= 0 && !vision->facingFieldEdge) {

		gps->currentlySpinChestKicking = false;

        lockMode = Dribble;
        resetDribbleTimer();
        aaDribble();

    } else {
            
        // spin with ball under snout
        walkType = CanterWalkWT;
		gps->currentlySpinChestKicking = true;
        fsO = 25;
        ffO = 80;
		forward = 0;
        left = (gps->tGoal().pos.getHead() > 0) ? -3 : 3;
        turnCCW = CLIP (gps->tGoal().pos.getHead(), 35.0);

        headtype = ABS_H;
        panx = 0;
        tilty = -50;        
    }
}

static int dribbleTimer = 0;

void Behaviours::resetDribbleTimer() {
    dribbleTimer = 0;
    spinDribbleTimer = 0;
}

void Behaviours::aaDribble() {
    dribbleTimer++;

    /* this needs to be carefully considered ...
    if (vtGoal.cf > 0 && Beckham::isBeckhamBallUnderChin()) {
        lockMode = LightningKick;        
        aaLightningKick(); 
        return;
    }    
    */
    
    /*
    if (dribbleTimer < 25 ||
        (gps->canSee(vobBall) && vBall.d < 20 && abs(vBall.h) < 15)) {
    */    
    if ((dribbleTimer < 25) && ((!gps->canSee(vobBall)) || (gps->canSee(vobBall) && vBall.d < 20))) {
    //if (!gps->canSee(vobBall) || (gps->canSee(vobBall) && vBall.d < 20 && abs(vBall.h) < 15)) {
/*
        if (dribbleTimer < 50) {

            if (gps->canSee(vobBall)) {

                Beckham::doBeckhamTrackVisualBall();    
                nullz = 22;

            } else {

                lockMode = Dribble;
                chooseStrategy();
                return;
            }    

        } else {     

            Beckham::doBeckhamTrackVisualBall();    
            nullz = 22;
        }    
*/            
       
        /*
        Beckham::doBeckhamTrackVisualBall();    
        nullz = 22;
        */

        headtype = ABS_H;
        panx = 0;
        tilty = 0;

        if (vtGoal.cf > 0) {
            Beckham::doBeckhamHoverToSpot(vtGoal.d, vtGoal.h);
            leds(2, 2, 2, 2, 2, 2);
        } else {
            double targetx = FIELD_WIDTH/2;
            double targety;
            //double targety = FIELD_LENGTH + WALL_THICKNESS;
            if (gps->self().pos.x <= FIELD_WIDTH/2) {
                targety = (4*WALL_THICKNESS/FIELD_WIDTH)*gps->self().pos.x + 
                    (FIELD_LENGTH - WALL_THICKNESS);
            } else {
                targety = (-4*WALL_THICKNESS/FIELD_WIDTH)*(gps->self().pos.x - FIELD_WIDTH/2) +
                    (FIELD_LENGTH + WALL_THICKNESS);
            }

            Vector localTarget;
            double dx = targetx - gps->self().pos.x;
            double dy = targety - gps->self().pos.y;
            localTarget.setVector(vCART, dx, dy);
            localTarget.rotate(90 - gps->self().h);

            double targetd = localTarget.d;
            double targeth = localTarget.head;

            Beckham::doBeckhamHoverToSpot(targetd, targeth);
            
            /* 
            Beckham::doBeckhamHoverToSpot(gps->tGoal().pos.d, gps->tGoal().pos.getHead());
            */
        }
    } else {

        lockMode = NoMode;
        chooseStrategy();

    }    
}
