/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "Behaviours.h"
#include "../share/BallDef.h"

static const double FIRE_PAWKICK_YOFFSET    = BALL_RADIUS;
static const double FIRE_PAWKICK_APPROACH   = 7.5;

static int firePaw;

// approaches ball and paw kicks with which ever paw nearer the ball
// try to align ball with designated paw
// switch paw if one rolls past opposite paw
// note this kicks straight
void Behaviours::firePawKick(int paw, double offset) {
    double ballx = sin(DEG2RAD(vBall.vob->h)) * vBall.vob->d;
    double bally = cos(DEG2RAD(vBall.vob->h)) * vBall.vob->d;
    
    // decide on a paw. once decided stick with it until ball rolls past the other paw
    // try to avoid problem with switching btwn left and right paws
    if (paw == FIRE_PAWKICK_AUTO) {
        if (firePaw == FIRE_PAWKICK_LEFT && ballx < -offset/2)
            firePaw = FIRE_PAWKICK_RIGHT;
        else if (firePaw == FIRE_PAWKICK_RIGHT && ballx > offset/2) 
            firePaw = FIRE_PAWKICK_LEFT;
    } else {
        firePaw = paw;
    }
    
    double target_x = (firePaw == FIRE_PAWKICK_RIGHT) ? ballx + offset : ballx - offset;
    double target_y = bally - FIRE_PAWKICK_YOFFSET;
    double target_d = sqrt(target_x * target_x + target_y * target_y);
    double target_h = RAD2DEG(asin(target_x / target_d));

    // run at ball
    turnCCW = CLIP(target_h/3, 30.0);
	left = 0;
    if (abs(target_x) <= 1.0 && target_d < BALL_RADIUS * 1.5) {
        walkType = ZoidalWalkWT;
        turnCCW = CLIP(turnCCW, 10.0);
		if (abs(turnCCW) <= 8) {  
			forward = MAX_OFFSET_FORWARD_SMALL_TURN;
		} else {
			forward = MAX_OFFSET_FORWARD_LARGE_TURN;
		}
    } else if (abs(turnCCW) > 13.75) {
        walkType = CanterWalkWT;
        forward = (abs(turnCCW) < 20) ? FIRE_PAWKICK_APPROACH : 0.5;
    } else {
        walkType = ZoidalWalkWT;
        forward = MAX_OFFSET_FORWARD;
    }
}
