

/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/*
 * UNSW 2003 Robocup
 *
 * Last modification background information
 * $$
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#include <iostream>
#include <string.h>
#include "CommandData.h"


// nameMatch represents a string representation of a command and it's
// Instruction equivalent
struct nameMatch {
    const char* name;
    Instruction inst;
};


// Add suitable members to this array if you are adding new robocommander commands
static const struct nameMatch matchArray[] = {
    {"mode", MODE_CHANGE},
    {"ko", UTHER_KICKING_OFF},
    {"p", PLAYER_CHANGE},
    {"f", FORWARD_MOVE},
    {"l", LEFT_MOVE},
    {"t", TURN_MOVE},
    {"v", SET_TILT},
    {"h", SET_PAN},
    {"B", DO_BLOCK_STANCE},
    {"H", DO_HOLD_STANCE},
    {"N", NORMAL_STANCE_ABSH},
    {"n", NORMAL_STANCE_RELH},
    {"K", SET_STANCE},
    {"T", DO_TURN_KICK},
    {"u", PAN_TILT},
    {"k", TILT_PAN},
    {"j", MOVE_COMMAND},
    {"c", HUMAN_CONTROL},
    {"r", SET_COLOUR},
    {"I", SET_INDICATOR},
    {"g", EDGE_DETECT_DEBUG},
#ifdef SENDING_DEBUG
#ifdef SENDING_CPLANE
    {"d", CPLANE_SEND},
    {"burst", BURST_MODE},
#endif // SENDING_CPLANE
#ifdef SENDING_WORLD_MODEL_DEBUG
    {"w", WM_SEND},
#endif // SENDING_WORLD_MODEL_DEBUG
#ifdef SENDING_YUVPLANE
    {"y", YUVPLANE_SEND},
#endif // SENDING_YUVPLANE
#ifdef SENDING_OPLANE
    {"z", OPLANE_SEND},
#endif // SENDING_OPLANE
#endif // SENDING_DEBUG
    {"e", BEACON_DEBUG},
    {"b", SET_BEACON_DIST_CONST},
    {"E", TEST_TURN_KICK},
    {"Z", SET_CURRENTLY_CARRYING},
    {"lat", DO_WM_LAT_TEST},
	{"ctk", CAL_TURN_KICK},
    {"none", NO_COMMAND} // this must be at the end
};


CommandData::CommandData() {
    
    type = ROBOCOMMANDER_DATA;
    RC_COMMAND = NO_COMMAND;
    value = 0;
    
}


CommandData::CommandData(const PackageDef& p) {
    
    type = ROBOCOMMANDER_DATA;
    
    bool set = false;
    for(int i=0 ; strcmp(matchArray[i].name, "none") != 0 ; i++) {
        
        if(strcmp(matchArray[i].name, p.name) == 0) {
            RC_COMMAND = matchArray[i].inst;
            set = true;
            break;
        }
        
    }
    
    if(!set) {
        RC_COMMAND = NO_COMMAND;
    }
    
    int n = sscanf(p.value, "%lf %lf %lf", &value, &value2, &value3);    
    //value = atof(p.value);
    if(n < 3) {
        value3 = 0.0;
        if(n < 2) {
            value2 = 0.0;
            if(n < 1) {
                value = 0.0;
            }
        }
    } 
    printRCCommand();
    std::cout << " - {" << value << "," << value2 << "," << value3 << "}" << std::endl;
}

CommandData::CommandData(const RoboCupGameControlData& gc) {

    type = GAMECONTROLLER_DATA;
    GC_DATA = gc;

}

void CommandData::printRCCommand() const {
    
    switch(RC_COMMAND) {
        case NO_COMMAND:
            std::cout << "NO_COMMAND";
            break;
        case MODE_CHANGE:
            std::cout << "MODE_CHANGE";
            break;
        case UTHER_KICKING_OFF:
            std::cout << "UTHER_KICKING_OFF";
            break;
        case PLAYER_CHANGE:
            std::cout << "PLAYER_CHANGE";
            break;
        case FORWARD_MOVE:
            std::cout << "FORWARD_MOVE";
            break;
        case LEFT_MOVE:
            std::cout << "LEFT_MOVE";
            break;
        case TURN_MOVE:
            std::cout << "TURN_MOVE";
            break;
        case SET_TILT:
            std::cout << "SET_TILT";
            break;
        case SET_PAN:
            std::cout << "SET_PAN";
            break;
        case DO_BLOCK_STANCE:
            std::cout << "DO_BLOCK_STANCE";
            break;
        case DO_HOLD_STANCE:
            std::cout << "DO_HOLD_STANCE";
            break;
        case NORMAL_STANCE_ABSH:
            std::cout << "NORMAL_STANCE_ABSH";
            break;
        case NORMAL_STANCE_RELH:
            std::cout << "NORMAL_STANCE_RELH";
            break;
        case SET_STANCE:
            std::cout << "SET_STANCE";
            break;
        case DO_TURN_KICK:
            std::cout << "DO_TURN_KICK";
            break;
        case PAN_TILT:
            std::cout << "PAN_TILT";
            break;
        case TILT_PAN:
            std::cout << "TILT_PAN";
            break;
        case MOVE_COMMAND:
            std::cout << "MOVE_COMMAND";
            break;
        case HUMAN_CONTROL:
            std::cout << "HUMAN_CONTROL";
            break;
        case SET_COLOUR:
            std::cout << "SET_COLOUR";
            break;
        case SET_INDICATOR:
            std::cout << "SET_INDICATOR";
            break;
        case EDGE_DETECT_DEBUG:
            std::cout << "EDGE_DETECT_DEBUG";
            break;
#ifdef SENDING_DEBUG
#ifdef SENDING_CPLANE
        case CPLANE_SEND:
            std::cout << "CPLANE_SEND";
            break;
        case BURST_MODE:
            std::cout << "BURST_MODE";
            break;
#endif // SENDING_CPLANE
#ifdef SENDING_WORLD_MODEL_DEBUG
        case WM_SEND:
            std::cout << "WM_SEND";
            break;
#endif // SENDING_WORLD_MODEL_DEBUG
#ifdef SENDING_YUVPLANE
        case YUVPLANE_SEND:
            std::cout << "YUVPLANE_SEND";
            break;
#endif // SENDING_YUVPLANE
#ifdef SENDING_OPLANE
        case OPLANE_SEND:
            std::cout << "OPLANE_SEND";
            break;
#endif // SENDING_OPLANE
#endif // SENDING_DEBUG
        case BEACON_DEBUG:
            std::cout << "BEACON_DEBUG";
            break;
        case SET_BEACON_DIST_CONST:
            std::cout << "SET_BEACON_DIST_CONST";
            break;
        case TEST_TURN_KICK:
            std::cout << "TEST_TURN_KICK";
            break;
        case SET_CURRENTLY_CARRYING:
            std::cout << "SET_CURRENTLY_CARRYING";
            break;
        case DO_WM_LAT_TEST:
            std::cout << "DO_WM_LAT_TEST";
            break;
        case CAL_TURN_KICK:
            std::cout << "DO_CAL_TURN_KICK";
            break;			
			
    }
    
}


void CommandData::clear() {
    
    type = ROBOCOMMANDER_DATA;
    RC_COMMAND = NO_COMMAND;

}
    


