/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: Blob.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

//#include <cstdio>
//#include <iostream>
//#include <cstring>
//#include <cfloat>
//#include <vector>

//#include <OPENR/OSubject.h>
//#include <OPENR/OObserverVector.h>
//#include <OPENR/OPENREvent.h>
//#include <OPENR/OPENRAPI.h>

#include "Blob.h"

// ricky: uncomment to run maybe colors
//#define MAYBE 10
// ricky: uncomment to run maybe debug
//#define MAYBE_DEBUG

Blobber::Blobber (GeneralisedColourDetect & gcd) {
    // sets the cplane pointer
    cplane = gcd.cplane;
    nn_multi_colour = gcd.nn_multi_colour;

    for (int i=0;i < NUM_USABLE_COLORS;i++)
        blobs[i].init(i);

    for (int i = 0; i < NUM_USABLE_COLORS; i++)
        barray[i].init(i);

    usingHeap = 0;
    usingAlt = 1;
    multi_colour = 0;
    
    // reset the number of blobs for each color
    for (int k = cBALL; k < NUM_USABLE_COLORS; k++)
        nbob[k] = 0;
}

void
Blobber::updateBlobNums (RunLengthInfo segment[], int min, int max, int oldnum, int newnum, int colorMatch, int newcolor) {
    for (int i = min; i <= max; i++) {
        if (segment[i].blob_number == oldnum && segment[i].color == colorMatch) {
            segment[i].blob_number = newnum;
	        segment[i].color = newcolor;
/*            
#ifdef OFFLINE
            if (newcolor == 0) {
                cout << "segment " << i << " updated from " << oldnum << " to " << newnum << endl;
            }
#endif
*/
        }
	}
}

void 
Blobber::createBlob (RunLengthInfo * segment, int y) {
    int i;
    if (nbob[segment->color] >= MAX_NUM_BLOBS && top_bn[segment->color] >= 0)
        i = available_bn[segment->color][top_bn[segment->color]--];
    else
        i = nbob[segment->color]++;

    if (i < MAX_NUM_BLOBS) {
        segment->blob_number = i;
        blobinfo[segment->color][i].Start (segment->startIndex,
                                           segment->endIndex, y,
                                           segment->length, i);
	blobinfo[segment->color][i].xsum = segment->xsum;
	blobinfo[segment->color][i].ysum = segment->ysum;
        openBlobs[segment->color].push_back(i);
  }
}

void 
Blobber::updateArray (int y) {
    for (int j = 0; j < NUM_USABLE_COLORS; j++) {
        for (vector<int>::reverse_iterator i = openBlobs[j].rbegin(); 
            i != openBlobs[j].rend(); i++) {
            HalfBeacon & blob = blobinfo[j][*i];

            // blob is marked for deletion
            if (blob.area == 0) {
                available_bn[j][++top_bn[j]] = *i;
                *i = openBlobs[j][openBlobs[j].size() - 1];
                openBlobs[j].pop_back();
            } else if (blob.getYMax () != y) {
                /* blob hasn't been altered so we store it if it meets all thresholds.  
                 * always erase it from the openBlobs list because 
                 * it can no longer connect to anything
                **/
		// ricky: only store core blobs, no point storing maybe blobs that 
		// cant connect to anything
		if (blob.color < GREEN_FIELD || blob.color == MAYBE_DOG) {
                    if (!barray[j].add (blob)) {
                        available_bn[j][++top_bn[j]] = *i;
                        blob.area = 0;
                    }
		}
                *i = openBlobs[j][openBlobs[j].size() - 1];
                openBlobs[j].pop_back();
            }
        }
    }
}

#ifndef MAYBE

void 
Blobber::runLengthEncode () {
    int ci = 0;
    int *segmentCount = splane.segmentCount;
    for (int y = 0; y < USABLE_HEIGHT; y++) {
        for (segmentCount[y] = 0;;) {
            int start = 0, end = 0;

            // skips unused colors
            for (start = ci % WIDTH;
                 cplane[ci] > LAST_USABLE_COLOR && cplane[ci] != MULTI
                 && start < WIDTH; ci++, start++);

            if (start < WIDTH) {
                end = start;
                uchar color = cplane[ci];
                if (color > BACKGROUND && multi_colour != 0)
                {
                    color = (lookup_mc (BLUE_DOG, y, end) ? BLUE_DOG : BACKGROUND);
                }
                // skip to end of segment
                if (color <= LAST_USABLE_COLOR)
                {
                    for (++end, ++ci;
                         end < WIDTH && (cplane[ci] == color
                         || (multi_colour != 0
                         && color == BLUE_DOG
                         && cplane[ci] > BACKGROUND
                         && lookup_mc (BLUE_DOG, y, end)));
                         ++end, ++ci);

                    // storing sum of yvalue into segment
                    splane.segment[y][segmentCount[y]].ysum = y * (end - start);
                    splane.segment[y][segmentCount[y]].Set (start, end - 1, color);

                    segmentCount[y]++;
                } else {
                    ++ci;
                }    

            }
            if (start >= WIDTH || end >= WIDTH)
                break;
        }
    }
}

/* NOTE to future generations
 * This algorithm is very very bad
 * it is slow.
 * 
 * A better alternative, was thought of a month before the 2003 comp, but not implemented
 * use disjoint set union thing. 
 * basically, structure blobs as trees of segment. when a new segment is added, have it point
 * to the root of the segment tree. and when two blobs join, join the root of the smaller tree
 * to the root of the larger tree.
 *
 * Also, use path compression when ever the root is accessed.
 *
 * So.. you guys should use that new algorithm. Talk to will about it.
 */
void
Blobber::FormBlobs () {
    int *segmentCount = splane.segmentCount;
  
    //runLengthEncode ();

    for (int x = 0; x < segmentCount[0]; x++) {
        RunLengthInfo *cs = &splane.segment[0][x];

        if (cs->color < 7)
            createBlob (cs, 0);
    }

    // Merge blobs every 2 lines
    int prev_y = 0;
    for (int y = 1; y < USABLE_HEIGHT; y++, prev_y++) {
/*
#ifdef OFFLINE
        cout << "at height " << y << endl;
#endif                        
*/
        int w = 0;
        for (int x = 0; x < segmentCount[y]; x++) {
            RunLengthInfo *cs = &splane.segment[y][x];
            RunLengthInfo *ps = 0;
            for (; w < segmentCount[prev_y]; w++) {
                ps = &splane.segment[prev_y][w];
                if (cs->color == ps->color && ps->overlaps (cs)
                     && ps->blob_number != cs->blob_number)
                {
                    if (cs->blob_number == -1)
                    {
                        cs->blob_number = ps->blob_number;

			            // also update the sum of x, y and pixles
                        blobinfo[ps->color][ps->blob_number].Update(cs->startIndex,
                                                                    cs->endIndex, y,
                                                                    cs->length, cs->xsum,
								    cs->ysum);
/*
#ifdef OFFLINE
                        cout << "new segment joined with segment " << w << " number " << cs->blob_number << " (" << blobinfo[cs->color][cs->blob_number].getXMin();
                        cout << "," << blobinfo[cs->color][cs->blob_number].getYMin() << ") (";
                        cout << blobinfo[cs->color][cs->blob_number].getXMax() << ",";
                        cout << blobinfo[cs->color][cs->blob_number].getYMax() << ")" << endl;
#endif                        
*/
                    } else {
                        BlobInfo *pblob = &blobinfo[ps->color][ps->blob_number];
                        BlobInfo *cblob = &blobinfo[cs->color][cs->blob_number];
                        int oldnum = cs->blob_number;
                        pblob->Update (cblob);
/*
#ifdef OFFLINE
                        cout << ps->blob_number << " took over " << cs->blob_number << ": (" << pblob->getXMin() << "," << pblob->getYMin();
                        cout << ") (" << pblob->getXMax() << "," << pblob->getYMax() << ")" << endl;
                        //cout << "update current height - 1 from " << w << " to " << segmentCount[prev_y]-1 << endl;
#endif                        
*/
                        updateBlobNums (splane.segment[prev_y], w, segmentCount[prev_y]-1, cs->blob_number, ps->blob_number, ps->color, ps->color);
/*
#ifdef OFFLINE
                        //cout << "update current height from 0 to " << x << endl;
#endif                        
*/
                        updateBlobNums (splane.segment[y], 0, x, cs->blob_number, ps->blob_number, ps->color, ps->color);
                        cblob->area = 0;
                    }
                }

                if (ps->endIndex >= cs->endIndex)
                    break;
            }

            // apply these actions no matter the exit condition above
            if (cs->blob_number == -1) {
                createBlob (cs, y);
/*
#ifdef OFFLINE
                cout << "new " << cs->blob_number << " created " << cs->blob_number << " (";
                cout << blobinfo[cs->color][cs->blob_number].getXMin();
                cout << "," << blobinfo[cs->color][cs->blob_number].getYMin() << ") (";
                cout << blobinfo[cs->color][cs->blob_number].getXMax() << ",";
                cout << blobinfo[cs->color][cs->blob_number].getYMax() << ")" << endl;
#endif                        
*/
            }
            if (ps != 0 && ps->endIndex == cs->endIndex)
                w++;
         }
        updateArray (y);
    }
    updateArray (USABLE_HEIGHT);
}

#endif
#ifdef MAYBE
// ricky: 
// initial approach was to merge maybe colours with core colours horizontally
// but that seems to slow down stuff, so instead, it is done after blobbing
// we end up with more segments, but less calculations whilst doing run length
// encoding
// this does not affect normal segmentaion
void 
Blobber::runLengthEncode () {
    int ci = 0;
    int *segmentCount = splane.segmentCount;
    for (int y = 0; y < USABLE_HEIGHT; y++) {
        for (segmentCount[y] = 0;;) {
            int start = 0, end = 0;
            // skips unused colors
            for (start = ci % WIDTH; (cplane[ci] > RED_DOG && cplane[ci] < MAYBE_ORANGE && start < WIDTH); ci++, start++);

            if (start < WIDTH) {
                end = start;
                uchar color = cplane[ci];
		
		// green field ignore
				
                // skip to end of segment
                for (++end, ++ci; end < WIDTH && (cplane[ci] == color); ++end, ++ci);

		if (color != FIELD_OR_GREEN) {
                  // storing sum of yvalue into segment
                  splane.segment[y][segmentCount[y]].ysum = y * (end - start);
                  splane.segment[y][segmentCount[y]].Set (start, end - 1, color);
                  segmentCount[y]++;
		}
            }
            if (start >= WIDTH || end >= WIDTH)
                break;
        }
    }
}

/* ricky: very very twitchy piece of code */
void
Blobber::FormBlobs () {
    int *segmentCount = splane.segmentCount;
  
    //runLengthEncode ();

    for (int x = 0; x < segmentCount[0]; x++) {
        RunLengthInfo *cs = &splane.segment[0][x];
        createBlob (cs, 0);
    }

    // Merge blobs every 2 lines
    int prev_y = 0;
    int alt;
    for (int y = 1; y < USABLE_HEIGHT; y++, prev_y++) {
        int w = 0;
        for (int x = 0; x < segmentCount[y]; x++) {
            RunLengthInfo *cs = &splane.segment[y][x];
            RunLengthInfo *ps = 0;
            for (; w < segmentCount[prev_y]; w++) {
                ps = &splane.segment[prev_y][w];
		// core
		if (cs->color < GREEN_FIELD) {
		    alt = cs->color + MAYBE;
		    if (cs->color == ps->color && ps->overlaps(cs) && ps->blob_number != cs->blob_number) {
		        if (cs->blob_number == -1) {
			    cs->blob_number = ps->blob_number;
			    blobinfo[ps->color][ps->blob_number].Update(cs->startIndex, cs->endIndex, y, cs->length, cs->xsum, cs->ysum);
			} else {
                            BlobInfo *pblob = &blobinfo[ps->color][ps->blob_number];
                            BlobInfo *cblob = &blobinfo[cs->color][cs->blob_number];
                            pblob->Update (cblob);
                            updateBlobNums (splane.segment[y], 0, x, cs->blob_number, ps->blob_number, ps->color, ps->color);
                            updateBlobNums (splane.segment[prev_y], w, segmentCount[prev_y], cs->blob_number, ps->blob_number, ps->color, ps->color);
                            cblob->area = 0;
			}
		    } else if (alt == ps->color && ps->overlaps(cs) && ps->blob_number != cs->blob_number) {
			// set segment to core colour, then create new core blob if dont already have one, merge with maybe blob and
			// mark all segments to used, finally mark maybe blob for deletion
		        if (cs->blob_number == -1) {
			    createBlob(cs, y);
			}
			BlobInfo *pblob = &blobinfo[ps->color][ps->blob_number];
			BlobInfo *cblob = &blobinfo[cs->color][cs->blob_number];
#ifdef MAYBE_DEBUG
			fill_up(y, ps->blob_number, ps->color, cs->color);
#endif
			cblob->Update(pblob);
			updateBlobNums (splane.segment[y], x, ps->blob_number, cs->blob_number, ps->color, cs->color);
			updateBlobNums (splane.segment[y-1], x, ps->blob_number, cs->blob_number, ps->color, cs->color);
			pblob->area = 0;
		    }
		} 
		else {
		    alt = cs->color - MAYBE;
		    if (alt == ps->color && ps->overlaps(cs) && ps->blob_number != cs->blob_number) {
		        if (cs->blob_number == -1) {
			    cs->blob_number = ps->blob_number;
			    cs->color = ps->color;
#ifdef MAYBE_DEBUG
			    for (int tq = cs->startIndex + y * WIDTH; tq < cs->endIndex + y * WIDTH; tq++) {
			        cplane[tq] = ps->color;
			    }
#endif
			    blobinfo[ps->color][ps->blob_number].Update(cs->startIndex, cs->endIndex, y, cs->length, cs->xsum, cs->ysum);
			} else {
			    BlobInfo *pblob = &blobinfo[ps->color][ps->blob_number];
                            BlobInfo *cblob = &blobinfo[cs->color][cs->blob_number];
#ifdef MAYBE_DEBUG
			    fill_up(y, cs->blob_number, cs->color, ps->color);
#endif
			    pblob->Update(cblob);
			    updateBlobNums(splane.segment[y], x, cs->blob_number, ps->blob_number, cs->color, ps->color);
			    updateBlobNums(splane.segment[y-1], x, cs->blob_number, ps->blob_number, cs->color, ps->color);
			    cblob->area = 0;
			    cblob->blob_number = pblob->blob_number;
			    cblob->color = pblob->color;
			}
	  	    } else if (cs->color == ps->color && ps->overlaps(cs) && ps->blob_number != cs->blob_number) {
		        if (cs->blob_number == -1) {
			    cs->blob_number = ps->blob_number;
			    blobinfo[ps->color][ps->blob_number].Update(cs->startIndex, cs->endIndex, y, cs->length, cs->xsum, cs->ysum);
			} else {
                            BlobInfo *pblob = &blobinfo[ps->color][ps->blob_number];
                            BlobInfo *cblob = &blobinfo[cs->color][cs->blob_number];
                            pblob->Update (cblob);
                            updateBlobNums (splane.segment[y], x, cs->blob_number, ps->blob_number, ps->color, ps->color);
                            cblob->area = 0;
			}
                    }
		}
                if (ps->endIndex >= cs->endIndex)
                    break;
            }

            // apply these actions no matter the exit condition above
            if (cs->blob_number == -1)
                createBlob (cs, y);

            if (ps != 0 && ps->endIndex == cs->endIndex)
                w++;
         }
        updateArray (y);
    }
    updateArray (USABLE_HEIGHT);
}
#endif

int
Blobber::findHalfBeacons (int colorIndex, HalfBeacon ** hbArray) {

    int i, cur;
    int biggest, bigIndex;
    BlobArray binfo;
    binfo = barray[colorIndex];

    for (i = 0; i < binfo.getCount (); i++) {
        biggest = -1;
        bigIndex = -1;
        for (cur = 0; cur < binfo.numCandidates (); cur++) {
            if (!binfo[cur].blobProcessed && binfo[cur].area > biggest) {
                biggest = binfo[cur].area;
                bigIndex = cur;
            }
        }
        if (bigIndex == -1) {
            // should never get here
            return 0;
        }
        // now bigIndex is set to the index of the next biggest blob
        hbArray[i] = &binfo[bigIndex];
        // now set the size of the current biggest blob to 0, so we can get
        // the next biggest
        binfo[bigIndex].formationComplete ();

        // hbArray[i]->Print();
    }

    // set the variable so the caller knows how many halfbeacons of this 
    // type we found
    return binfo.getCount ();
}

bool 
Blobber::lookup_mc (int prev_colour, int height, int width) {
    int colour = MULTI;

    if (multi_colour != 0) {
        int numBlue = 0;
        int numBackground = 0;
        int index;

        index = (height - 1) * WIDTH + (width - 1);
        if (index < HEIGHT * WIDTH && index >= 0) {
            if (cplane[index] == BLUE_DOG)
                numBlue++;
            if (cplane[index] >= BACKGROUND)
                numBackground++;
        }

        index = (height - 1) * WIDTH + width;
        if (index < HEIGHT * WIDTH && index >= 0) {
            if (cplane[index] == BLUE_DOG)
                numBlue++;
            if (cplane[index] >= BACKGROUND)
                numBackground++;
        }

        index = (height - 1) * WIDTH + (width + 1);
        if (index < HEIGHT * WIDTH && index >= 0) {
            if (cplane[index] == BLUE_DOG)
                numBlue++;
            if (cplane[index] >= BACKGROUND)
                numBackground++;
        }

        index = height * WIDTH + (width - 1);
        if (index < HEIGHT * WIDTH && index >= 0) {
            if (cplane[index] == BLUE_DOG)
                numBlue++;
            if (cplane[index] >= BACKGROUND)
                numBackground++;
        }

        index = (height + 1) * WIDTH + (width + 1);
        if (index < HEIGHT * WIDTH && index >= 0) {
            if (cplane[index] == BLUE_DOG)
                numBlue++;
            if (cplane[index] >= BACKGROUND)
                numBackground++;
        }

        index = (height + 1) * WIDTH + (width - 1);
        if (index < HEIGHT * WIDTH && index >= 0) {
            if (cplane[index] == BLUE_DOG)
                numBlue++;
            if (cplane[index] >= BACKGROUND)
                numBackground++;
        }

        index = (height + 1) * WIDTH + width;
        if (index < HEIGHT * WIDTH && index >= 0) {
            if (cplane[index] == BLUE_DOG)
                numBlue++;
            if (cplane[index] >= BACKGROUND)
                numBackground++;
        }

        index = (height + 1) * WIDTH + (width + 1);
        if (index < HEIGHT * WIDTH && index >= 0) {
            if (cplane[index] == BLUE_DOG)
                numBlue++;
            if (cplane[index] >= BACKGROUND)
                numBackground++;
        }

        if (numBlue > numBackground)
            colour = BLUE_DOG;
        else
            colour = BACKGROUND;
    }

    if (colour != prev_colour)
        return false;
    else
        return true;
}

void
Blobber::orderMin (int color, int lo0, int hi0) {

    int lo = lo0;
    int hi = hi0;
    int mid;

    if (hi0 > lo0) {
        //mid = blobinfo[color][( lo0 + hi0 ) / 2].getXMin()*cos_roll + 
        //        blobinfo[color][( lo0 + hi0 ) / 2].getYMin()*sin_roll;
        //mid = blobinfo[color][( lo0 + hi0 ) / 2].boundary[SOUTH].midPoint()*
        //        cos_roll + blobinfo[color][( lo0 + hi0 ) / 2].getYMid()*
        //        sin_roll;
        //mid = blobinfo[color][( lo0 + hi0 ) / 2].getXMin();
        mid = barray[color][(lo0 + hi0) / 2].xRmin;

        while (lo <= hi) {
            /* find the first element that is greater than or equal to
             * the partition element starting from the left Index.
             */
            while ((lo < hi0) && (barray[color][lo].xRmin < mid))
              ++lo;

            /* find an element that is smaller than or equal to
             * the partition element starting from the right Index.
             */
            while ((hi > lo0) && (barray[color][hi].xRmin > mid))
              --hi;

            /* if the indexes have not crossed, swap
                 */
            if (lo <= hi) {
                HalfBeacon info = barray[color][lo];
                barray[color][lo] = barray[color][hi];
                barray[color][hi] = info;
                ++lo;
                --hi;
            }
        }

        /* If the right index has not reached the left side of array
         * must now sort the left partition.
         */
        if (lo0 < hi)
            orderMin (color, lo0, hi);

        /* If the left index has not reached the right side of array
         * must now sort the right partition.
         */
        if (lo < hi0)
            orderMin (color, lo, hi0);

    }
}

/* Perform rotation on blobs about the center of the image
 */
void
Blobber::rotate (int color, double sin_eroll, double cos_eroll) {
    int max = barray[color].numCandidates ();
    int xOrigin = WIDTH / 2, yOrigin = HEIGHT / 2;

    for (int i = 0; i < max; i++) {
        int nX, nY, eX, eY, sX, sY, wX, wY;
        int rnX, rnY, reX, reY, rsX, rsY, rwX, rwY;

        nX = (xOrigin - barray[color][i].boundary[NORTH].midPoint ());
        nY = (yOrigin - barray[color][i].getYMin ());

        rnX = (int) (cos_eroll *nX - sin_eroll * nY);
        rnY = (int) (sin_eroll *nX + cos_eroll * nY);

        sX = (xOrigin - barray[color][i].boundary[SOUTH].midPoint ());
        sY = (yOrigin - barray[color][i].getYMax ());

        rsX = (int) (cos_eroll *sX - sin_eroll * sY);
        rsY = (int) (sin_eroll *sX + cos_eroll * sY);

        eX = (xOrigin - barray[color][i].getXMax ());
        eY = (yOrigin - barray[color][i].boundary[EAST].midPoint ());

        reX = (int) (cos_eroll *eX - sin_eroll * eY);
        reY = (int) (sin_eroll *eX + cos_eroll * eY);

        wX = (xOrigin - barray[color][i].getXMin ());
        wY = (yOrigin - barray[color][i].boundary[WEST].midPoint ());

        rwX = (int) (cos_eroll *wX - sin_eroll * wY);
        rwY = (int) (sin_eroll *wX + cos_eroll * wY);

        barray[color][i].xRmin = xOrigin - MAX (MAX (MAX (rnX, rsX), reX), rwX);
        barray[color][i].xRmax = xOrigin - MIN (MIN (MIN (rnX, rsX), reX), rwX);

        barray[color][i].yRmin = yOrigin - MAX (MAX (MAX (rnY, rsY), reY), rwY);
        barray[color][i].yRmax = yOrigin - MIN (MIN (MIN (rnY, rsY), reY), rwY);

        barray[color][i].cRx = (barray[color][i].xRmin + barray[color][i].xRmax) / 2;
        barray[color][i].cRy = (barray[color][i].yRmin + barray[color][i].yRmax) / 2;
    }
}

// ricky debug
void
Blobber::fill_up(int y, int bnum, int color, int fill) {
  int *segmentCount = splane.segmentCount;
  for (int i = y; y >= 0; y--) {
    bool found = false;
    for (int j = 0; j < segmentCount[i]; j++) {
        RunLengthInfo *cs = &splane.segment[i][j];
	if (cs->blob_number == bnum && cs->color == color) {
	    found = true;
	    for (int k = cs->startIndex + (WIDTH * i); k <= cs->endIndex + (WIDTH * i); k++) {
	        cplane[k] = fill;
            }
	}
    }
    if (!found && i < (y - 1)) break;
  }
}

// ricky do colour segmentation and run length encoding at the same time
void
Blobber::replaceCplane (uchar * yp, uchar * up, uchar * vp, long skip) {
    unsigned char c, cc;
    uchar *cp = cplane;
    int *segmentCount = splane.segmentCount;
    int start, end;

    for (int y = 0; y < HEIGHT; ++y) {
        segmentCount[y] = 0;
	    start = end = 0;
        for (int x = 0; x < WIDTH; ++x) {
            c = nn_multi_colour[(*yp / FACTOR) * MAXU * MAXV + (*up / FACTOR) * MAXV + (*vp / FACTOR)];
	        *cp = c;
	        start = x;
    	    for (end = start + 1; end < WIDTH; end++) {
		        cc = nn_multi_colour[(*(++yp) / FACTOR) * MAXU * MAXV + (*(++up) / FACTOR) * MAXV + (*(++vp) / FACTOR)];
		        *(++cp) = cc;
		        if (cc != c) break;
	        }
	        if (c < NUM_USABLE_COLORS) {
                splane.segment[y][segmentCount[y]].ysum = y * (end - start);
                splane.segment[y][segmentCount[y]].Set (start, end - 1, c);
		        segmentCount[y] += 1;
	        }	    
	        x = end-1;
        }
        (++yp) += skip;
        (++up) += skip;
        (++vp) += skip;
	    (++cp);
    }
}


