/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: GeneralisedColourDetect.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#include <cstdio>
#include <iostream>
#include <cstring>
#include <cfloat>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "GeneralisedColourDetect.h"

#ifdef OFFLINE
static const char* calFile = "../../robot/cfg/nnmc.cal";
#else // OFFLINE
static const char* calFile = "/MS/NNMC.CAL";
#endif // OFFLINE

using namespace std;

GeneralisedColourDetect::GeneralisedColourDetect () {

    int fd, rr;
    nn_multi_colour = new unsigned char[CUBE_VOLUME];

    if (nn_multi_colour == NULL)
        cerr << "eek!" << endl;

    // Default method.
    method = NN_MULTI_COLOUR;


    // New Calibration Technique Tim and James 7/2/02

    fd = open (calFile, O_RDONLY);

    if (fd < 0)
        cerr << "Error opening file\n" << endl;
    else
        rr = read (fd, nn_multi_colour, CUBE_VOLUME);
}

void
GeneralisedColourDetect::wlLoadNNMC (const uchar * data, int size) {
    // copy in the new NNMC into memory
    for (int i = 0; i < size; i++)
    {
        //if (nn_multi_colour[i] != data[i]) {

        nn_multi_colour[i] = data[i];
    }
}

void
GeneralisedColourDetect::replaceCplane (uchar * yp, uchar * up,
                    uchar * vp, long skip) {

    uchar *cp = cplane;
    uchar c, cc, mc;

    for (int y = 0; y < HEIGHT; y++) {
        for (int x = 0; x < WIDTH; x++) {

            if (method == GCD) {
                *(cp++) = seachart[plane[*(yp++)]][((*(up++)) << 8) | (*(vp++))];
            } else if (method == DECISION_TREE) {
                // *(cp++) = tree (*yp++, *up++, *vp++);
            } else if (method == NN_MULTI_COLOUR) {
                *(cp++) = nn_multi_colour[(*(yp++) / FACTOR) * MAXU * MAXV +
                    (*(up++) / FACTOR) * MAXV +
                    (*(vp++) / FACTOR)];
            } else {
                cout << "Unknown method: " << method << endl;
            }

        }
        yp += skip;
        up += skip;
        vp += skip;
    }
}

#ifdef VISION_DEBUG
void
GeneralisedColourDetect::changeCalibration (int type) {
    if (type == GCD) {
        method = GCD;
        cout << "Calibration method now GCD" << endl;
    } else if (type == DECISION_TREE) {
        method = DECISION_TREE;
        cout << "Calibration method now Decision Tree" << endl;
    } else if (type == NN_MULTI_COLOUR) {
        method = NN_MULTI_COLOUR;
        cout << "Calibration method now Nearest Neighbour MultiColour Pixels" << endl;
    } else {
        cout << "Unknown calibration type: " << type << endl;
    }
}

/*
 * Returns the calibration method.
 */
int
GeneralisedColourDetect::getCalibration () {
    return method;
}

void
GeneralisedColourDetect::switchFile (int suff) {
    int fd;
    char fileName[80];
    char tmp[10];
    switch (method) {
        case GCD:
        strcpy (fileName, "/MS/NewMAuto");
        if (suff != 0) {
            sprintf (tmp, "%d", suff);
            strcat (fileName, tmp);
        }
        strcat (fileName, ".CAL");
        fd = open (fileName, O_RDONLY);

        if (fd >= 0) {
            for (int i = 0; i < NUM_PLANES; i++) {
                seachart[i] = new uchar[256 * 256];
                if (seachart[i] == NULL)
                    cout << "    NULL ERROR in GCD******************************************\n";
            }

            for (int i = 0; i < NUM_PLANES; i++) {
                read (fd, seachart[i], 256 * 256);
            }

            // initialize the plane array
            for (int i = 0; i < 256; i++) {
                plane[i] = determinePlane (i);
            }

            close (fd);
            cout << "Parsed GCD file\n";
        } else {
            cout << "Unable to Parse GCD file.\n";
        }
        break;

        case NN_MULTI_COLOUR:
        cout << "parsing nnmc in switchfile" << endl;

        strcpy (fileName, "/MS/NNMC");
        if (suff != 0) {
            sprintf (tmp, "%d", suff);
            strcat (fileName, tmp);
        }
        strcat (fileName, ".CAL");

        fd = open (fileName, O_RDONLY);
        if (fd < 0) {
            cout << "Error opening file\n" << endl;
        } else {
            cout << "why the heck are we here" << endl;
            read (fd, nn_multi_colour, CUBE_VOLUME);

            close (fd);
        }
        cout << "****** FINISHED PARSING THE NNMC FILE **************" << endl;
        break;
    }

}

/*
 * Updates the calibration dynamically.
 */
void
GeneralisedColourDetect::updateCalibration (int col, int row,
                        int newcolour, uchar * yplane) {
    int y, u, v;

    y = yplane[row * WIDTH * 3 + col];
    u = yplane[row * WIDTH * 3 + col + WIDTH];
    v = yplane[row * WIDTH * 3 + col + WIDTH * 2];

    if (method == NN_MULTI_COLOUR) {
        nn_multi_colour[(y / FACTOR) * MAXU * MAXV + (u / FACTOR) * MAXV + (v / FACTOR)] = newcolour;
    } else if (method == GCD) {
        seachart[plane[y]][(u << 8) | v] = newcolour;
    }
}
#endif

int
GeneralisedColourDetect::determinePlane (int val) {
    int planeValue;

    if (val < 40)
        planeValue = 0;
    else if (val < 45)
        planeValue = 1;
    else if (val < 50)
        planeValue = 2;
    else if (val < 55)
        planeValue = 3;
    else if (val < 60)
        planeValue = 4;
    else if (val < 65)
        planeValue = 5;
    else if (val < 70)
        planeValue = 6;
    else if (val < 75)
        planeValue = 7;
    else if (val < 80)
        planeValue = 8;
    else if (val < 85)
        planeValue = 9;
    else if (val < 90)
        planeValue = 10;
    else if (val < 100)
        planeValue = 11;
    else if (val < 110)
        planeValue = 12;
    else if (val < 120)
        planeValue = 13;
    else if (val < 140)
        planeValue = 14;
    else if (val < 160)
        planeValue = 15;
    else if (val < 190)
        planeValue = 16;
    else
        planeValue = 17;

    return planeValue;
}

