/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: Vision.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 *
**/

#define OPENR_STUBGEN

#include <OPENR/core_macro.h>

#include <MCOOP.h>
#include <OPENR/OPENR.h>
#include <OPENR/OPENRMessages.h>
#include <OPENR/core_macro.h>

#include "Vision.h"
#include "VisualCortex-priv.h"
#include "../share/Common.h"
#include "../share/PlaneTypeDef.h"
#include "../share/VisionDef.h"
#include "../share/SharedMemoryDef.h"
#include "../share/WirelessTypeDef.h"
#include "../behaviours/Behaviours.h"

/*
#ifdef DELIBERATE
#include "../prolog/PrologData.h"
#endif
*/

#define ALL_WIRELESS_SHARED_ON 1

static const bool delayDebugMsg            = false;
static const bool yuvDebugMsg              = false;
static const bool newOplaneDebugMsg        = false;
static const bool sensorValueDebugMsg      = false;
static const bool sanityValuesDebugMsg     = false;
static const bool compressedCPlaneDebugMsg = false;

/*
 * The roll is multiplied by this factor so that only 4 bytes is used in the transmission, instead of 8.
 * has to agree among the following files: robot/vision/Vision.cc, client.h and RoboConstant.java
**/
static const int SEND_SCALE = 1000;

// this is sent when the robot recognise any object in the colour plane
static const unsigned char zeroBytes[INT_BYTES] = {0, 0, 0, 0};

static const int objectOrder[NUM_OBJ] = {vobBlueGoal,
                                         vobYellowGoal,
                                         vobBlueLeftBeacon,
                                         vobBlueRightBeacon,
                                         vobGreenLeftBeacon,
                                         vobGreenRightBeacon,
                                         vobYellowLeftBeacon,
                                         vobYellowRightBeacon,
                                         vobRedDog,
                                         vobRedDog2,
                                         vobRedDog3,
                                         vobRedDog4,
                                         vobBlueDog,
                                         vobBlueDog2,
                                         vobBlueDog3,
                                         vobBlueDog4,
                                         vobBall};

Vision::Vision () {
    OSYSDEBUG(("Vision::Vision()\n"));
    
    checkPrimitiveSize();
    
    Behaviours::InitBehaviours();
    powerIsOn = false;
	sharedMemIsSetup = false;
    camNotSet = true;
    wait4settle = 30;
#ifdef SENDING_DEBUG
    sharedrc = NULL;
    sharedDebugMemory = NULL;
#endif // SENDING_DEBUG
  
#ifdef FRAMECOUNT
    GetSystemTime(&lastTime);
    GetSystemTime(&curTime);

    frameCounter = 0;
    LastframePt = 0;
#endif // FRAMECOUNT

#ifdef WORLD_MODEL_DEBUG
    Last_WM_Pt = 10 * 125;
#endif // WORLD_MODEL_DEBUG

#ifdef EDGE_DETECT
    sendDebugNow        = false;
    vision.sendDebugNow = false;
    beaconsToGPS = true;
    edgesToGPS = false;
#endif //EDGE_DETECT

    Behaviours::sensors      = &sensors;
    Behaviours::vision       = &vision;
    Behaviours::gps          = &gps;
    Behaviours::transmission = this;

    cplaneCounter = 0;
    oplaneCounter = 0;

    displayEnvironmentCounter = 0;
    shareEnvironmentCounter = 0;

    delayTestingCounter = 0;
	
	gotTMWM = 0;
}

OStatus Vision::DoInit (const OSystemEvent & event) {
    OSYSDEBUG(("Vision::DoInit()\n"));

    // NewComData  /* generates the necessary number of subjects and observers */
    // SetComData  /* registers the communication service with other objects */
    NEW_ALL_SUBJECT_AND_OBSERVER;
    REGISTER_ALL_ENTRY;
    SET_ALL_READY_AND_NOTIFY_ENTRY;


    // wireless
    objectOut       = subject[sbjSPEAKER];
    dogCommOut      = subject[sbjDogComm];
    dogPhotoOut     = subject[sbjDogPhoto];
    playerNumberOut = subject[sbjPlayerNumber];

    sSetIndication = subject[sbjSetIndication];
    sExecuteAA = subject[sbjExecuteAA];
    sensors.oSensor           = observer[obsSensor];

    //wireless
    objectIn  = observer[obsExternalCommand];
    dogCommIn = observer[obsDogComm];
    //dogCommLTIn = observer[obsDogCommLT];
    
#ifdef SENDING_DEBUG
    sharedrc = new RCRegion(sizeof(RC_DEBUG_DATA));
    sharedDebugMemory = (RC_DEBUG_DATA *)sharedrc->Base();
#endif // SENDING_DEBUG

    return oSUCCESS;
}

OStatus Vision::DoStart (const OSystemEvent & event) {
    OSYSDEBUG(("Vision::DoStart()\n"));

    ENABLE_ALL_SUBJECT;
    ASSERT_READY_TO_ALL_OBSERVER;
    /*
    for (int i = 0; i < numOfObserver; ++i)
      if (observer[i]->AssertReady () != oSUCCESS)
      OSYSDEBUG(("AssertReady() failed. observer[%d]\n",i));

    */
	// send player number to wireless
	sendPlayerNumber(Behaviours::PLAYER_NUMBER);

#ifdef SENDING_DEBUG
    int msgType = 0;
    MemoryRegionID mem = sharedrc->MemID();
    size_t offset = sharedrc->Offset();
    size_t size = sharedrc->Size();

    objectOut->ClearBuffer();
    objectOut->SetData(&msgType, sizeof(int));
    objectOut->SetData(&mem, sizeof(mem));
    objectOut->SetData(&offset, sizeof(offset));
    objectOut->SetData(&sharedDebugMemory, sizeof(sharedDebugMemory));
    objectOut->SetData(&size, sizeof(size));
    objectOut->NotifyObservers();
#endif // SENDING_DEBUG

    return oSUCCESS;

}

OStatus Vision::DoStop (const OSystemEvent & event) {
    //eCommander.stop();

    DISABLE_ALL_SUBJECT;
    DEASSERT_READY_TO_ALL_OBSERVER;

    //for (int i = 0; i < numOfObserver; ++i)
    //  observer[i]->DeassertReady();

    return oSUCCESS;
}

OStatus Vision::DoDestroy (const OSystemEvent & event) {
    DELETE_ALL_SUBJECT_AND_OBSERVER;

    /* deletes all observers and subjects spawned in DoInit() */
    DeleteComData         

    return oSUCCESS;
}

/*
void Vision::SetWirelessShmem(OSubject *sbjShmWireless) {
#ifdef SENDING_DEBUG
    int msgType = 0;
    MemoryRegionID mem = sharedrc->MemID();
    size_t offset = sharedrc->Offset();
    size_t size = sharedrc->Size();

    objectOut->ClearBuffer();
    objectOut->SetData(&msgType, sizeof(int));
    objectOut->SetData(&mem, sizeof(mem));
    objectOut->SetData(&offset, sizeof(offset));
    objectOut->SetData(&sharedDebugMemory, sizeof(sharedDebugMemory));
    objectOut->SetData(&size, sizeof(size));
    objectOut->NotifyObservers();
#endif // SENDING_DEBUG
}
*/

void Vision::ResultCamera (const ONotifyEvent & event) {

    // Setting the camera
    // not sure if this is mandatory
    // camNotSet = false;
    if (camNotSet) {
        SetCamera();
        observer[obsYUV]->AssertReady();
        return;
    }

    if (!sharedMemIsSetup) {
        observer[obsYUV]->AssertReady();
        return;
    }

   // Wait for everything to settle
    if (wait4settle > 0) {
        wait4settle--;
        observer[obsYUV]->AssertReady (event.SenderID ());
        return;
    }
    // Setting goal
    if (gps.goalNotSet) {
        gps.SetGoals (sensors.sensorVal[ssMOUTH]);
    //      if (targetGoal == vobYellowGoal)
    //          Produce_Sound(5);
    }

#ifdef FRAMECOUNT 
 
    SystemTime icur;
    GetSystemTime(&icur);

#endif // FRAMECOUNT
#ifdef EDGE_DETECT
	const WMObj& self = gps.self();
//    vision.setSelfPos((int)(self.pos.x), (int)(self.pos.y), self.h, self.posVar, self.posVar, self.hVar);
    vision.ProcessImage (event, sensors.sensorVal[ssHEAD_PITCH], sensors.sensorVal[ssHEAD_YAW],
                  sensors.sensorVal[ssHEAD_ROLL], sensors.range2obstacle, self);
    if (vision.VISUALCORTEX_TIME_PRINT == 1)
    {
      //if the blobber has taken more than 30000us send a YUV plane of the incident. 
      if (vision.peakBlobTime > 30000)
      {
        sendYUVPlane();
      }
    }
#else //EDGE_DETECT
    vision.ProcessImage (event, sensors.sensorVal[ssHEAD_PITCH], sensors.sensorVal[ssHEAD_YAW],
                  sensors.sensorVal[ssHEAD_ROLL], sensors.range2obstacle);
#endif //EDGE_DETECT
#ifdef FRAMECOUNT  

    SystemTime icur2;
    GetSystemTime(&icur2);
    SystemTime idiff = icur2 - icur;
    imgProcessTime = imgProcessTime + idiff;
#endif // FRAMECOUNT


  if (!beaconsToGPS)
  {
    for (int i = vobBlueLeftBeacon; i <= vobYellowRightBeacon; i ++)
    {
      vision.vob[i].cf = 0;
      vision.vob[i].var = HUGE_VAL;
    }
  }
  gps.GPSVisionUpdate(vision.vob, RAD2DEG(vision.effective_pan),vision.ballDistMethod);

#ifdef EDGE_DETECT
    //{{{
    //Stuff to run when debug is called for edge detection (note this runs AFTER process image so 
    //it's safe to use things). 
      //Now check to see if we need to send a debug plane.
  if (!edgesToGPS)
  {
    for (int i = 0; i < HOUGH_NUM_LINES; i ++)
    {
      vision.edges[i].distVar = HUGE_VAL;
      vision.edges[i].angleVar = HUGE_VAL;
    }
  }
  if (sendDebugNow)
  {
    sendDebugPlane();
    //Do this last.
    sendDebugNow = false;
    vision.sendDebugNow = false;
  }
  /*
    //Now check to see if we need to send a debug plane.
    vision.setLineType(vision.maxVals, HOUGH_NUM_LINES, vision.lineType);
    int * tempFieldLineDefs = (int*)malloc(HOUGH_NUM_LINES*2*sizeof(int));
    int * tempFieldLineID = (int*)malloc(HOUGH_NUM_LINES*2*sizeof(int));
    vision.defineFieldLine(vision.fieldEnds, HOUGH_NUM_LINES, tempFieldLineDefs);
    vision.identifyLine(tempFieldLineDefs, HOUGH_NUM_LINES, tempFieldLineID);
    free(tempFieldLineDefs);
    free(tempFieldLineID);
  */
#ifdef FRAMECOUNT  
    SystemTime gcur;
    GetSystemTime(&gcur);
#endif // FRAMECOUNT
    vision.edgesToGPS = false;
    if (vision.sendEdges && EDGE_DETECT_TYPE == EDGE_PATTERN && vision.currentXint >= 0 && vision.currentYint >= 0 && vision.currentTint >= 0 && vision.edgesValid && vision.origNumEdgePixels > 20 && vision.edgeMatchQuality > 150)// && !(vision.abortBorderSet))
    {
      //We have the gradients in x, y and t in the form of vision.patternDx, Dy and Dt. At this stage, if any of these are less than 50, set the variance in 
      //that direction to infinity (in reality it is either over a hundred or under 10 anyway). 
      double varX, varY, varT;
      bool matchX, matchY, matchT;
      if ((matchX = (vision.patternDx < vision.matchCutoffX)))
      {
        varX = get95CF(vision.cfNoMatchX);
      }
      else
      {
        varX = get95CF(vision.cfMatchedX);
      }
      if ((matchY = (vision.patternDy < vision.matchCutoffY)))
      {
        varY = get95CF(vision.cfNoMatchY);
      }
      else
      {
        varY = get95CF(vision.cfMatchedY);
      }
      if ((matchT = (vision.patternDt < vision.matchCutoffT)))
      {
        varT = get95CF(vision.cfNoMatchT);
      }
      else
      {
        varT = get95CF(vision.cfMatchedT);
      }
      if (
          (abs(vision.currentXint - vision.currentX) < vision.maxDiffX && !matchX)
          && 
          (abs(vision.currentYint - vision.currentY) < vision.maxDiffY && !matchY)
          && 
          (abs(vision.currentTint - vision.currentT) < vision.maxDiffT && !matchT)
          )
      {
        vision.edgesToGPS = true;
        gps.GPSDirectPositionUpdate((double)(vision.currentXint), (double)(vision.currentYint), (double)(vision.currentTint),varX,varY,varT);
      }
      else
      {
//        cout << "xdiff = " << abs(vision.currentXint - vision.currentX) << " matchX = " << matchX << " ydiff = " << abs(vision.currentYint - vision.currentY) << " matchY = " << matchY << " diffT = " << abs(vision.currentTint - vision.currentT) << " matchT = " << matchT << endl;
      }
//      cout << "Updating GPS position with " << vision.currentXint << "/" << vision.currentYint << "/" << vision.currentTint << endl;
    }
    else
    {
//      cout << "vision.edgesValid = " << vision.edgesValid << " vision->origNumEdgePixels = " << vision.origNumEdgePixels << " vision.edgeMatchQualit = " << vision.edgeMatchQuality << endl;
    }
//    gps.GPSVisionUpdate(vision.vob, vision.edges, RAD2DEG(vision.effective_pan));//, vision.edges);
#ifdef FRAMECOUNT  
    SystemTime gcur2;
    GetSystemTime(&gcur2);
    SystemTime gdiff = gcur2 - gcur;
    gpsVisTime = gpsVisTime + gdiff;
#endif // FRAMECOUNT
    //}}}
#else //EDGE_DETECT
 
#ifdef FRAMECOUNT  
    SystemTime gcur;
    GetSystemTime(&gcur);
#endif // FRAMECOUNT
    //gps.GPSVisionUpdate(vision.vob, RAD2DEG(vision.effective_pan), vision.ballDistMethod);
#ifdef FRAMECOUNT  
    SystemTime gcur2;
    GetSystemTime(&gcur2);
    SystemTime gdiff = gcur2 - gcur;
    gpsVisTime = gpsVisTime + gdiff;
#endif // FRAMECOUNT
#endif //EDGE_DETECT
    
// Counting the number of frames
#ifdef FRAMECOUNT

	frameCounter++;
    if (vision.cameraFrame > LastframePt + 5 * 125) {
        LastframePt += 5 * 125;
        //cout << "Last 5 seconds averaged " << frameCounter/5.0 << " camera frames per second." << endl;
        //wout << "Last 5 seconds averaged " << frameCounter/5.0 << " camera frames per second.\n";
        GetSystemTime(&curTime);
        SystemTime diff = curTime - lastTime;
        double frames = frameCounter/(diff.seconds + (diff.useconds/1000000.0));
        int piSpeed = (int)((imgProcessTime.seconds*1000000.0 + imgProcessTime.useconds)/frameCounter);
        int gpsSpeed = (int)((gpsVisTime.seconds*1000000.0 + gpsVisTime.useconds)/frameCounter);
        int bSpeed = (int)((behavioursTime.seconds*1000000.0 + behavioursTime.useconds)/frameCounter);
        sprintf (dataResultCamera, 
            "(%d) - %d.%.6d s - %f cps\n   Aves : PI - %dus, gps - %dus, b - %dus\n", Behaviours::PLAYER_NUMBER,
            (int)diff.seconds,(int)diff.useconds,frames,piSpeed,gpsSpeed,bSpeed);
        //sprintf (dataResultCamera, 
        //    "%d.%.6d seconds -  %f Camera Frames Per Second \n",(int)diff.seconds,(int)diff.useconds,frames);
        //cout << dataResultCamera;
        // send data

        int dataType = 0;
        int arrayLength = strlen (dataResultCamera);
        //cout << "ArrayLength : " << arrayLength << endl;
#ifdef SENDING_DEBUG
#ifdef ALL_WIRELESS_SHARED_ON
        memcpy(sharedDebugMemory->message, dataResultCamera, arrayLength);
		
        int msgType = 1;
        objectOut->ClearBuffer();
        objectOut->SetData(&msgType, sizeof(int));
        objectOut->SetData(&dataType, sizeof(int));
        objectOut->SetData(&arrayLength, sizeof(int));
        objectOut->NotifyObservers();
#endif 
#endif // SENDING_DEBUG

        frameCounter            = 0;
        lastTime                = curTime;
        imgProcessTime.seconds  = 0;
        imgProcessTime.useconds = 0;
        gpsVisTime.seconds      = 0;
        gpsVisTime.useconds     = 0;
        behavioursTime.seconds  = 0;
        behavioursTime.useconds = 0;
    }

#endif // FRAMECOUNT

// Debug world model
#ifdef WORLD_MODEL_DEBUG

    if (vision.cameraFrame > Last_WM_Pt + 1 * 125) {
        Last_WM_Pt += 1 * 125;

        /*
        cout << "Start World Model here ";
        cout << (long) rx << " " << (long) ry << " " << (long) rh << " " <<
        (long) rcf;
        cout << " " << (long) wmobj[0].d << " " << (long) wmobj[0].
        h << " " << (long) wmobj[0].cf;
        cout << " " << (long) wmobj[vobRedDog].
        d << " " << (long) wmobj[vobRedDog].
        h << " " << (long) wmobj[vobRedDog].cf;
        cout << " " << (long) wmobj[vobBlueDog].
        d << " " << (long) wmobj[vobBlueDog].
        h << " " << (long) wmobj[vobBlueDog].cf;
        cout << " End World Model here";
        cout << endl;
        */
    }
#endif // WORLD_MODEL_DEBUG

#ifdef FRAMECOUNT  
    SystemTime bcur;
    GetSystemTime(&bcur);
#endif // FRAMECOUNT
    Behaviours::NextAction();
#ifdef FRAMECOUNT  
    SystemTime bcur2;
    GetSystemTime(&bcur2);
    SystemTime bdiff = bcur2 - bcur;
    behavioursTime = behavioursTime + bdiff;
    
    SystemTime tdiff = bcur2 - icur;
    Behaviours::cpuUsage = (tdiff.seconds*1000000.0 + tdiff.useconds);
    Behaviours::currentFrameID = vision.cameraFrame;
#endif // FRAMECOUNT
    observer[obsYUV]->AssertReady (event.SenderID ());

    // send across visual data to all other robots

}

void Vision::SetCamera () {

//Parameter:
//  White Balance:
//      const longword ocamparamWB_INDOOR_MODE  = 1;
//      const longword ocamparamWB_OUTDOOR_MODE = 2;
//      const longword ocamparamWB_FL_MODE      = 3; // Fluorescent Lamp
//
//  Gain:
//      const longword ocamparamGAIN_LOW        = 1;
//      const longword ocamparamGAIN_MID        = 2;
//      const longword ocamparamGAIN_HIGH       = 3;
//
//  Shutter Speed:
//      const longword ocamparamSHUTTER_SLOW    = 1;
//      const longword ocamparamSHUTTER_MID     = 2;
//      const longword ocamparamSHUTTER_FAST    = 3;

    OPrimitiveID fbkID;
    OPENR::OpenPrimitive("PRM:/r1/c1/c2/c3/i1-FbkImageSensor:F1", &fbkID);

    /* White Balance */
    OPrimitiveControl_CameraParam wb (ocamparamWB_OUTDOOR_MODE);
    if (OPENR::ControlPrimitive (fbkID, oprmreqCAM_SET_WHITE_BALANCE, &wb, sizeof(wb), 0, 0) == oSUCCESS)
        OSYSDEBUG(("CAM_SET_WHITE_BALANCE : Success\n"));
    else
        OSYSDEBUG(("CAM_SET_WHITE_BALANCE : Failed\n"));

    /* Gain */
    //OPrimitiveControl_CameraParam gain (ocamparamGAIN_MID);
    OPrimitiveControl_CameraParam gain (ocamparamGAIN_MID);
    if (OPENR::ControlPrimitive (fbkID, oprmreqCAM_SET_GAIN, &gain, sizeof(gain), 0, 0) == oSUCCESS)
        OSYSDEBUG(("CAM_SET_GAIN : Success\n"));
    else
        OSYSDEBUG(("CAM_SET_GAIN : Failed\n"));

    /* Shutter Speed */
    //OPrimitiveControl_CameraParam shutter (ocamparamSHUTTER_MID);
    OPrimitiveControl_CameraParam shutter (ocamparamSHUTTER_FAST);
    if (OPENR::ControlPrimitive (fbkID, oprmreqCAM_SET_SHUTTER_SPEED, &shutter,sizeof(shutter), 0, 0) == oSUCCESS) {
        OSYSDEBUG(("CAM_SET_SHUTTER_SPEED : Success\n"));
    } else {
        OSYSDEBUG(("CAM_SET_SHUTTER_SPEED : Failed\n"));
    }

    camNotSet = false;
}

//the purpose of this fuction is to call GPSMotionUpdate
//and then assert ready.
//needs to be done here because gps doesn't have access to the observer array
void Vision::GPSMotionUpdateAssert(const ONotifyEvent & event) {
#ifdef FRAMECOUNT  
    SystemTime gcur;
    GetSystemTime(&gcur);
#endif // FRAMECOUNT   
    const double dForward = (const double)*(const double *)event.Data(0);
    const double dLeft = (const double)*(const double *)event.Data(1);
    const double dTurn = (const double)*(const double *)event.Data(2);

    gps.oForward = dForward;
    gps.oLeft = dLeft;
    gps.oTurn = dTurn;

    gps.GPSMotionUpdate(dForward, dLeft, dTurn);
    observer[event.ObsIndex()]->AssertReady();
#ifdef FRAMECOUNT  
    SystemTime gcur2;
    GetSystemTime(&gcur2);
    SystemTime gdiff = gcur2 - gcur;
    gpsVisTime = gpsVisTime + gdiff;
#endif // FRAMECOUNT
}

/*void Vision::ResultWireless(const ONotifyEvent& event) {

    //const char* text = (const char *)event.Data(0);
    //OSYSPRINT(("Vision::ResultWireless() %s\n", text));
    //++msgReceiveCounter;

    processWireless(event);

    objectIn->AssertReady();
}*/

void Vision::ExternalCommand(const ONotifyEvent& event) {

    const CommandData& cmd = *(const CommandData *)event.Data(0);
    Behaviours::processCommand(cmd);
    
    objectIn->AssertReady();

}

void Vision::TeammateWM(const ONotifyEvent& event) {
    
	gotTMWM = 1;
    const WMShareInfo *wm = (const WMShareInfo *)event.Data(0);
    gps.processShareInfo (wm->playerNum, wm->obj);
    
    dogCommIn->AssertReady();
    
}

// Taken out by Ross due to change in shared memory use
/*void Vision::LatencyTesting(const ONotifyEvent& event) {

    int msgType = *(const int *)event.Data(0);
    int offset = 0;
    
    //Wireless Latency
     offset = REGION_H;
     WDataInfo info = *(WDataInfo *) (sharedBase+offset);
     int arrayLength = info.size;

     // Test for array of 'a'
     char world[arrayLength];
     memcpy(world, sharedBase+offset+sizeof(WDataInfo), arrayLength);


     //byte *world = (byte *) sharedBase+offset+sizeof(WDataInfo);
     //Test World model Data
     for (int a = 0; a < arrayLength; a++){
         world [a] = 'b';
         //cout << a << (char) world[a] << "  ";
     }
     //cout << "Received correctly.. " << endl;

     //Behaviours::wreceive(info.robot, info.type, info.size, obj);


    int dataType = 500;
    //cout << "ArrayLength : " << arrayLength << endl;
    memcpy(sharedBase, &dataType, sizeof(int));
    memcpy(sharedBase+sizeof(int), &arrayLength, sizeof(int));
    memcpy(sharedBase+REGION_F, world, arrayLength);
    int msgType2 = 1;
    objectOut->ClearBuffer();
    objectOut->SetData(&msgType2, sizeof(int));
    objectOut->NotifyObservers();
    
    dogCommLTIn->AssertReady();
}*/

void Vision::ResultSensor (const ONotifyEvent &event) {
    sensors.ResultSensor(event);
}

void Vision::SetPAgent(const ONotifyEvent& event) {

    const MemoryRegionID mem = *(const MemoryRegionID *)event.Data(0);
    const size_t offset      = *(const size_t *)event.Data(1);
    const voidPtr base       = *(const voidPtr *)event.Data(2);
    const size_t size        = *(const size_t *)event.Data(3);

	if (size != sizeof(ActuatorSharedMemory)) {
		cout << "ERROR: Actuator Shared Memory definition mismatch!" << endl;
	}

    shmPWalk = new RCRegion(mem, offset, base, size);

    ActuatorSharedMemory *ptr = (ActuatorSharedMemory *)shmPWalk->Base();

	if (ptr == NULL) {
		cout << "Bad Actuator Control Shared Memory Base Pointer" << endl;
	} else {
		cout << "Vision has received ActuatorControl shared memory: " << (int)ptr << endl;
		Behaviours::leading_leg = &(ptr->leading_leg);
    	Behaviours::currentWalkType = &(ptr->currentWalkType);
    	Behaviours::shareStep_ = &(ptr->shareStep_);
    	Behaviours::currentPG = &(ptr->currentPG);
		Behaviours::desiredPan = &(ptr->desiredPan);
		Behaviours::desiredTilt = &(ptr->desiredTilt);
    	#ifdef PSTEP_PUSH_INTERFACE
		Behaviours::stepComplete = &(ptr->stepComplete);
    	#endif //PSTEP_PUSH_INTERFACE
		
		vision.currentWalkType = &(ptr->currentWalkType);
		
		sharedMemIsSetup = true;
	}
    //I dont assert ready at the end
    //because i dont expect to be getting a message to set up the shared mem region
    //thus activating this function more than once.
	
}




void Vision::shareEnvironment(bool kicking) {

    if (shareEnvironmentCounter == 0 && 
	    Behaviours::gameState.state != ROBOCUP_STATE_PENALIZED) {
        WMShareObj obj = gps.getShareInfo();

        #ifdef ALL_WIRELESS_SHARED_ON
        int msgType = 1;
        dogCommOut->ClearBuffer();
        dogCommOut->SetData(&obj, sizeof(obj));
        dogCommOut->NotifyObservers();
		#endif 
        
    }
    shareEnvironmentCounter++;
    shareEnvironmentCounter %= SHARE_ENVIRONMENT_INTERVAL;

}


void Vision::sendPlayerNumber(int playerNumber) {
    playerNumberOut->SetData(&playerNumber, sizeof(int));
    playerNumberOut->NotifyObservers();
}

#ifdef SENDING_DEBUG
#ifdef SENDING_WORLD_MODEL_DEBUG
void Vision::displayEnvironment () {

    if (displayEnvironmentCounter == 0) {

        static const int TEAMMEMBERS_BEING_SENT = (NUM_TEAM_MEMBER-1);
        static const int NUM_ATTR_SENT_ROB = 9;
		static const int NUM_ATTR_VIS_PULL = 3;
        static const int NUM_ATTR_SENT_TEAMMATE = 5;
        static const int NUM_ATTR_SENT_BALL = 9;    //for now x y vx vy varx vary varvx varvy
        static const int NUM_ATTR_SENT_WBALL = 3;	// source robot num, x & y of wireless ball position
		static const int NUM_ATTR_SENT_OPPONENTS = 5;

        const int  arraySize =
            2 +                     // for the player number and the team number
            
            1 +                     // for the view variable describing things we can see
            
            3 +                     // for the last direct update values

            1 +                     // for the size of the remaining vision model data
            NUM_ATTR_SENT_ROB +     // the new world model data (x y h varx vary varh varxy varxh varyh)
            NUM_ATTR_VIS_PULL + 	// the exponentially decaying visual pull
			NUM_ATTR_SENT_BALL +
			NUM_ATTR_SENT_WBALL +   // for the wireless ball
            TEAMMEMBERS_BEING_SENT * NUM_ATTR_SENT_TEAMMATE +

           // 1 +                   // for the size of the remaining world model data
           // NUM_ATTR_SENT_ROB +   // the vision data (x y h varx vary varh varxy varxh varyh)
           // NUM_ATTR_SENT_BALL +

            1 +                     // for the size of the remaining wireless model data
            NUM_ATTR_SENT_ROB +     // the old world model data (x y h varx vary varh varxy varxh varyh)
        
            1 +                     // for size of opponents data
            NUM_OPPONENTS*NUM_ATTR_SENT_OPPONENTS; // for opponents


        float data[arraySize];
        int ptr = 0;

        //set player number
        data[ptr] = Behaviours::PLAYER_NUMBER;
        ptr++;

        // set direction team color
        if (gps.targetGoal == vobBlueGoal) {
            data[ptr] = P_RED_DOG;
        } else {
          data[ptr] = P_BLUE_DOG;
        }
        ptr++;
    
        // what we can see
        data[ptr++] = (double)gps.view;
        
        data[ptr++] = (double)gps.directx;
        data[ptr++] = (double)gps.directy;
        data[ptr++] = (double)gps.directh;

        /////////////////////
        // NEW WORLD MODEL
        /////////////////////
        //cout << "*************** got to world model sending *********" << endl;


        data[ptr] = NUM_ATTR_SENT_ROB + NUM_ATTR_VIS_PULL + NUM_ATTR_SENT_BALL + NUM_ATTR_SENT_WBALL +
				(TEAMMEMBERS_BEING_SENT * NUM_ATTR_SENT_TEAMMATE);
        ptr++;
		
		const WMObj& self = gps.self();
		const MMatrix3 scov = gps.selfCov();

        data[ptr]     = self.pos.x;
        data[ptr + 1] = self.pos.y;
        data[ptr + 2] = self.h;
        data[ptr + 3] = scov(0,0); //varx
        data[ptr + 4] = scov(1,1); //vary
        data[ptr + 5] = scov(2,2); //varh
        data[ptr + 6] = scov(0,1); //varxy
        data[ptr + 7] = scov(0,2); // varxh
        data[ptr + 8] = scov(1,2); // varyh
        ptr += NUM_ATTR_SENT_ROB;
		
		const MVec3& vp = gps.getVisualPull();
		data[ptr]     = vp(0,0);
		data[ptr + 1] = vp(1,0);
		data[ptr + 2] = vp(2,0);
		ptr += NUM_ATTR_VIS_PULL;

		//WMObj& ball = gps.ball('g');
		const Vector& ball = gps.getBall(GLOBAL);
		//WMObj& vball = gps.vBall();
		const Vector& vball = gps.getVBall(GLOBAL);
		const MMatrix2& bcov = gps.getBallCov();
		const MMatrix2& bvcov = gps.getVelCov();
		//const MMatrix4& bcov = gps.bCov();
        data[ptr]     = ball.x; //ball.pos.x;
        data[ptr + 1] = ball.y; //ball.pos.y;
        data[ptr + 2] = vball.x; //vball.pos.x;
        data[ptr + 3] = vball.y; //vball.pos.y;
        data[ptr + 4] = bcov(0,0); //varx
        data[ptr + 5] = bcov(1,1); //vary
        data[ptr + 6] = bcov(0,1); //varxy
        data[ptr + 7] = bvcov(0,0); //varvx
        data[ptr + 8] = bvcov(1,1); //varvy
        ptr += NUM_ATTR_SENT_BALL;
		
		if(gps.shareBallvar < LARGE_VAL) {
			// wireless ball data is valid
			data[ptr] = gps.sbRobotNum;
			data[ptr+1] = gps.shareBallx;
			data[ptr+2] = gps.shareBally;
		} else {
			// wireless ball data is invalid
			data[ptr] = 0;
			data[ptr+1] = 0;
			data[ptr+2] = 0;
		}
		ptr += NUM_ATTR_SENT_WBALL;
    
        for (int i=0; i<NUM_TEAM_MEMBER; i++) {
            if (i!=(Behaviours::PLAYER_NUMBER-1)) {
				const WMObj& tm = gps.tmObj(i);
                data[ptr]     = tm.pos.x;
                data[ptr + 1] = tm.pos.y;
                data[ptr + 2] = tm.h;
                data[ptr + 3] = tm.posVar;
                data[ptr + 4] = tm.hVar;
                ptr += NUM_ATTR_SENT_TEAMMATE;
            }
        }
        
        /*
        cout << "**************" << endl;
        cout << "my player number is " << Behaviours::PLAYER_NUMBER << endl;
        cout << "my location is " << gps.r.pos.x << "," << gps.r.pos.y << endl;
        for (int i=0; i<NUM_TEAM_MEMBER; i++) {
            if (i!=(Behaviours::PLAYER_NUMBER-1)) {
                cout << "my teammates number " << (i+1) << "'s location is " << gps.wmTeammate[i].pos.x << "," << 
                gps.wmTeammate[i].pos.y << endl;
            }
        }
        cout << "**************" << endl;
        */
    
        /////////////////////
        // VISION MODELS
        /////////////////////

        //data[ptr] = NUM_ATTR_SENT_ROB + NUM_ATTR_SENT_BALL; //number of attributes being sent over
        //ptr++;


        // note - there is no vision state any more as kalman filtering has changed
        // this should be eventually deleted
        /*data[ptr]     = 0;
        data[ptr + 1] = 0;
        data[ptr + 2] = 0;
        data[ptr + 3] = 1; //varx
        data[ptr + 4] = 1; //vary
        data[ptr + 5] = 1; //varh
        data[ptr + 6] = 0; //varxy
        data[ptr + 7] = 0; // varxh
        data[ptr + 8] = 0; // varyh
        ptr += NUM_ATTR_SENT_ROB;*/

        /*data[ptr]     = gps.newBallx;
        data[ptr + 1] = gps.newBally;
        data[ptr + 2] = gps.newBallvx;
        data[ptr + 3] = gps.newBallvy;
        data[ptr + 4] = (*(gps.ballMeasCov))(0,0); //varx
        data[ptr + 5] = (*(gps.ballMeasCov))(1,1); //vary
        data[ptr + 6] = (*(gps.ballMeasCov))(0,1); //varxy
        data[ptr + 7] = (*(gps.ballMeasCov))(2,2); //varvx
        data[ptr + 8] = (*(gps.ballMeasCov))(3,3); //varvy
        ptr += NUM_ATTR_SENT_BALL;*/

        /////////////////////
        // OLD WORLD MODEL
        /////////////////////

        data[ptr] = NUM_ATTR_SENT_ROB; //number of attributes being sent over
        ptr++;

        data[ptr]     = gps.oldx;
        data[ptr + 1] = gps.oldy;
        data[ptr + 2] = gps.oldh;
        data[ptr + 3] = gps.oldvx; //varx
        data[ptr + 4] = gps.oldvy; //vary
        data[ptr + 5] = gps.oldvh; //varh
        data[ptr + 6] = gps.oldcxy; //varxy
        data[ptr + 7] = gps.oldcxh; //varxh
        data[ptr + 8] = gps.oldcyh; //varyh
        ptr += NUM_ATTR_SENT_ROB;
        
        
        /////////////////////
        // OPPONENTS
        /////////////////////
        data[ptr] = NUM_OPPONENTS*NUM_ATTR_SENT_OPPONENTS;
        ptr++;
        for(int i=0 ; i<NUM_OPPONENTS ; i++) {
            const Vector v = gps.getOppPos(i,GLOBAL);
            const MMatrix2& m = gps.getOppCov(i);
            data[ptr] = v.x;
            data[ptr+1] = v.y;
            data[ptr+2] = m(0,0);
            data[ptr+3] = m(1,1);
            data[ptr+4] = m(0,1);
            ptr += NUM_ATTR_SENT_OPPONENTS;
        }


        int dataType = 100; //i assume this is the number that says "world model stuff"
        int arrayLength = sizeof(data);

        
		#ifdef ALL_WIRELESS_SHARED_ON
        memcpy(sharedDebugMemory->world_model_debug_data, &data, arrayLength);

        int msgType = 1;
        objectOut->ClearBuffer();
        objectOut->SetData(&msgType, sizeof(int));
        objectOut->SetData(&dataType, sizeof(int));
        objectOut->SetData(&arrayLength, sizeof(int));
        objectOut->NotifyObservers();
		#endif
    }        
    displayEnvironmentCounter++;
    displayEnvironmentCounter %= DISPLAY_ENVIRONMENT_INTERVAL;

}
#endif // SENDING_WORLD_MODEL_DEBUG


int Vision::storeSanityValues(unsigned char *planeArray, int offset) {
    int i, j, total;
    unsigned char *dataPtr;
    
    for (i=0; i<NUM_SANITY_COORDS; i++) {
        total   = vision.getSanityArray(i);
        dataPtr = (unsigned char *) &total;
        for (j=0; j<INT_BYTES; j++) {
            planeArray[offset++] = dataPtr[j];
        }

        if (sanityValuesDebugMsg) {
            cout << "Vision.cc: i = " << i << "; NUM_SANITY_COORDS = " << NUM_SANITY_COORDS << "; total = " << total << "; calculated ";
            cout << ((int)((planeArray[offset-4] & 0xff))) + 
                    ((int)((planeArray[offset-3] & 0xff) << 8)) + 
                    ((int)((planeArray[offset-2] & 0xff) << 16)) + 
                    ((int)((planeArray[offset-1] & 0xff) << 24)) << endl;
        }
    }
    return offset;
}

int Vision::storeObjects(unsigned char *planeArray, int offset) {
    int i, j, k, value, objectIndex;
    unsigned char *dataPtr;
    bool objectSend = false;

    // store the pan
    value   = (int)(vision.pan*SEND_SCALE);
    dataPtr = (unsigned char *) &value;
    for (k=0; k<INT_BYTES; k++) {
        planeArray[offset++] = dataPtr[k];
    }

    // store the tilt
    value   = (int)(vision.tilt*SEND_SCALE);
    dataPtr = (unsigned char *) &value;
    for (k=0; k<INT_BYTES; k++) {
        planeArray[offset++] = dataPtr[k];
    }

    // send all stationary objects first (beacons + goals), 
    // followed by all non-stationary objects (dogs + ball).
    // i = 1..8  = stationary beacons + goals
    // i = 9..16 = non-stationary dogs
    // i = 0     = non-stationary ball
    for (i=0; i<NUM_OBJ; i++) {
        objectIndex = objectOrder[i];
        if (vision.vob[objectIndex].cf > 0 || vision.vob[objectIndex].var < LARGE_VAL) {
            if (newOplaneDebugMsg) {
                cout << "objectIndex: " << objectIndex << "; x: " << vision.vob[objectIndex].x << "; y: " << vision.vob[objectIndex].y << "; width: " << vision.vob[objectIndex].width << "; height: " << vision.vob[objectIndex].height << "; d: " << vision.vob[objectIndex].d << endl;
            }
            objectSend = true;
            for (j=0; j<NUM_INT_COORDS; j++) {
                switch (j) {
                    case 0:
                        value = objectIndex; // sends the object number
                        break;
                    case 1:
                        value = (int)vision.vob[objectIndex].x;
                        break;
                    case 2:
                        value = (int)vision.vob[objectIndex].y;
                        break;
                    case 3:
                        value = (int)vision.vob[objectIndex].width;
                        break;
                    case 4:
                        value = (int)vision.vob[objectIndex].height;
                        break;
                    case 5:
                        value = (int)vision.vob[objectIndex].d;
                        break;
                    default:
                        // do nothing
                        break;
                }
                dataPtr = (unsigned char *) &value;
                for (k=0; k<INT_BYTES; k++) {
                    planeArray[offset++] = dataPtr[k];
                }
            }
        }
        // don't send empty plane
    }
    if (!objectSend) {
        /* this check is to make sure at least one (filled or empty) plane get sent to RoboCommander, 
         * in which a blank screen will be showing up in the OPlane
         * so that it will be realized in the debug tool
        **/
        for (j=0; j<NUM_INT_COORDS; j++) {
            for (k=0; k<INT_BYTES; k++) {
                planeArray[offset++] = zeroBytes[k];
            }
        }
    }
    return offset;
}


#ifdef SENDING_CPLANE
void Vision::sendCompressedCPlane() {
    // send CPlane per every 'CPLANE_INTERVAL' frame
    cplaneCounter = (cplaneCounter + 1) % CPLANE_INTERVAL;
    if (cplaneCounter && Behaviours::burstCounter == 0) {
        //cout << "dropped cplane" << endl;
        return;
    }
    //cout << "display cplane " << Behaviours::burstCounter << endl;
    
    // variable for coding
    unsigned char *cp = vision.cplane;
    unsigned char *compressedSizePtr;
    unsigned char pixel, prevPix = *(cp);
    unsigned char count = 0;
    int offset = INT_BYTES; // leave space for the compressed cplane size
    int compressedSize;

    // do the run length encoding
    for (int x = 0; x < HEIGHT; x++) {
        for (int y = 0; y < WIDTH; y++) {
            pixel = *(cp++);
            if (pixel > BACKGROUND && vision.get_mc () != 0) {
                pixel = (vision.lookup_mc (BLUE_DOG, x, y) ? BLUE_DOG : BACKGROUND);
            }
            // byte to store number is full or pix different
            if (count < 254 && (pixel == prevPix || (pixel > BACKGROUND && vision.lookup_mc (prevPix, x, y)))) {
                count++;
            } else {
                // assign into the array
                compressedCPlane[offset++] = count;
                compressedCPlane[offset++] = prevPix;

                prevPix = pixel;
                count = 1;
            }
        }
    }
    compressedCPlane[offset++] = count;
    compressedCPlane[offset++] = prevPix;

    // store the size of compressed cplane at the begining of the array
    if (compressedCPlaneDebugMsg) {
        cout << "Vision.cc: ";
    }
    compressedSize    = offset - INT_BYTES;
    compressedSizePtr = (unsigned char *) &compressedSize;
    for (int j=0; j<INT_BYTES; j++) {
        compressedCPlane[j] = compressedSizePtr[j];
    }
    if (compressedCPlaneDebugMsg) {
        cout << "compressedCPlane: ";
        cout << (int)(compressedCPlane[0] & 0xff) << " ";
        cout << (int)((compressedCPlane[1] & 0xff) << 8) << " ";
        cout << (int)((compressedCPlane[2] & 0xff) << 16) << " ";
        cout << (int)((compressedCPlane[3] & 0xff) << 24) << " = ";
        cout << ((int)(compressedCPlane[0] & 0xff) + 
                 (int)((compressedCPlane[1] & 0xff) << 8) + 
                 (int)((compressedCPlane[2] & 0xff) << 16) + 
                 (int)((compressedCPlane[3] & 0xff) << 24)) << endl;
    }

    if (compressedCPlaneDebugMsg) {
        cout << "cplane: " << offset << " ";
    }
    offset = storeSensors(compressedCPlane, offset);
    if (compressedCPlaneDebugMsg) {
        cout << "sensor: " << offset << " ";
    }
    offset = storeSanityValues(compressedCPlane, offset);
    if (compressedCPlaneDebugMsg) {
        cout << "sanity: " << offset << " ";
    }
    offset = storeObjects(compressedCPlane, offset);
    if (compressedCPlaneDebugMsg) {
        cout << "object: " << offset << " ";
    }
    offset = gps.storeGps(compressedCPlane, offset);
    if (compressedCPlaneDebugMsg) {
        cout << "gps   : " << offset << " ";
    }

    if (isDelayTesting) {
        compressedCPlane[offset++] = (byte) (delayTestingCounter >> 24);
        compressedCPlane[offset++] = (byte) (delayTestingCounter >> 16);
        compressedCPlane[offset++] = (byte) (delayTestingCounter >> 8);
        compressedCPlane[offset++] = (byte) delayTestingCounter;

        cout << "UNSWSkills.cc ---------- delayTestingCounter: " << delayTestingCounter << " ----------" << endl;

        delayTestingCounter++;
    }

	#ifdef ALL_WIRELESS_SHARED_ON
    memcpy(sharedDebugMemory->cplane_data, compressedCPlane, offset);

    if (compressedCPlaneDebugMsg) {
        cout << "total = " << offset << endl;
    }
    
    int msgType  = 1;
    int dataType = CPLANE_DATA_TYPE;
    objectOut->ClearBuffer();
    objectOut->SetData(&msgType, sizeof(int));
    objectOut->SetData(&dataType, sizeof(int));
    objectOut->SetData(&offset, sizeof(int));
    objectOut->NotifyObservers();
	#endif
}

int Vision::storeSensors(unsigned char *planeArray, int offset) {
    int value, i, j;
    unsigned char *dataPtr;
    
    if (sensorValueDebugMsg) {
        cout << "Sensor: ";
    }

    for (i=0; i<NUM_SENSOR_COORDS; i++) {
        value   = vision.getSensor(i);
        dataPtr = (unsigned char *) &value;
//        for (j=0; j<INT_BYTES; j++) {
        for (j=0; j<LONG_BYTES; j++) {
            planeArray[offset++] = dataPtr[j];
        }
        if (sensorValueDebugMsg) {
            cout << value << " ";
        }
    }
    if (sensorValueDebugMsg) {
        cout << value << endl;
    }
    
    return offset;
}
#endif // SENDING_CPLANE

#ifdef SENDING_OPLANE
void Vision::sendOPlane() {
    // send OPlane per every 'OPLANE_INTERVAL' frame
    oplaneCounter = (oplaneCounter + 1) % OPLANE_INTERVAL;
    if (oplaneCounter) {
        return;
    }
    
    int offset = storeObjects(oplane, 0);

    int dataType = OPLANE_DATA_TYPE;
	#ifdef ALL_WIRELESS_SHARED_ON
    memcpy(sharedDebugMemory->oplane_data, oplane, offset);
    
    int msgType = 1;
    objectOut->ClearBuffer();
    objectOut->SetData(&msgType, sizeof(int));  
    objectOut->SetData(&dataType, sizeof(int));  
    objectOut->SetData(&offset, sizeof(int));                  
    objectOut->NotifyObservers(); 
	#endif          
}
#endif // SENDING_OPLANE

#ifdef SENDING_YUVPLANE
void Vision::sendYUVPlane() {
    int msgType     = 1;
    int dataType    = YUVPLANE_CPLANE_DATA_TYPE;
    int arrayLength = CPLANE_SIZE;

    #ifdef ALL_WIRELESS_SHARED_ON
    memcpy(sharedDebugMemory->yuvCplane_data, vision.cplane, arrayLength);

    objectOut->ClearBuffer();
    objectOut->SetData(&msgType, sizeof(int));
    objectOut->SetData(&dataType, sizeof(int));
    objectOut->SetData(&arrayLength, sizeof(int));
    objectOut->NotifyObservers();
	#endif

    if (yuvDebugMsg) {
        cout << "Vision.cc: sendYUVPlane - CPlane Size : " << arrayLength << endl;
    }

    dataType    = YUVPLANE_DATA_TYPE;
    arrayLength = YUVPLANE_SIZE;

    #ifdef ALL_WIRELESS_SHARED_ON
    memcpy(sharedDebugMemory->yuvplane_data, vision.yplane, arrayLength);

    objectOut->ClearBuffer();
    objectOut->SetData(&msgType, sizeof(int));
    objectOut->SetData(&dataType, sizeof(int));
    objectOut->SetData(&arrayLength, sizeof(int));
    objectOut->NotifyObservers();
	#endif 

    if (yuvDebugMsg) {
        cout << "Vision.cc: sendYUVPlane - YPlane Size : " << arrayLength << endl;
    }

}
#endif // SENDING_YUVPLANE
#endif // SENDING_DEBUG


void Vision::sendAA(AtomicAction &nextAA) {
    //sExecuteAA->ClearBuffer(obsExecuteAA); //obsExecuteAA is the observer index of the observer reveiving a msg from this transmission.. number is generated automatically by stub.cfg
    sExecuteAA->ClearBuffer(); //obsExecuteAA is the observer index of the observer reveiving a msg from this transmission.. number is generated automatically by stub.cfg
    sExecuteAA->SetData(&nextAA,sizeof(nextAA));
    sExecuteAA->NotifyObservers ();
}

void Vision::sendIndicators(int *inds, size_t sz) {
    //sSetIndication->ClearBuffer(obsSetIndication); //obsSetIndication is the observer index of the observer reveiving a msg from this transmission.. number is generated automatically by stub.cfg
    //probably dont want to clear the buffer for indicators
    //i dont think it will make much of a difference though
    //sSetIndication->ClearBuffer(); //obsSetIndication is the observer index of the observer reveiving a msg from this transmission.. number is generated automatically by stub.cfg
    sSetIndication->SetData(inds,sz);
    sSetIndication->NotifyObservers ();
}


#ifndef EDGE_DETECT

void Vision::sendDebugPlane() {
}

void Vision::setSendDebug(char * inCmd) {
}

void Vision::setSendDebug(double val) {
}

#else // EDGE_DETECT
//EDGE_DETECT start 
//{{{

// void Vision::sendDebugPlane()
// 
// Prepares AND sends the debug plane (this differs from the 
// previous implementation where all this happened in VisualCortex, 
// now VisualCortex handles the stuff that triggers on every frame whilst
// this section now only handles the stuff that triggers when we need to 
// send a debug plane). 
void Vision::sendDebugPlane() {
#ifdef SENDING_DEBUG
#ifdef SENDING_LINE_DEBUG
    int i, j, k;
    //Houghspace transmission blah
    uchar * sendPlane = (uchar*)malloc(DEBUG_SIZE * sizeof(uchar));
    if (EDGE_DETECT_TYPE == EDGE_HOUGH)
    {//{{{
      uchar * houghSpaceTransmit;
      houghSpaceTransmit = (uchar*)malloc(WIDTH*HEIGHT);
      memcpy(sendPlane, vision.fieldEdges, WIDTH*HEIGHT);
      memcpy(sendPlane + (WIDTH*HEIGHT), vision.cplane, WIDTH*HEIGHT);

    cout << "about to send debug plane\n";
        //Normalize hough space to send
        k = 0;
        for (i = 0; i < HOUGH_MAX_R; i++) {
            for (j = 0; j < 360/HOUGH_THETA; j++) {
                if (vision.houghSpace[i][j][0] > k) {
                    k = vision.houghSpace[i][j][0];
                }
            }
        }
        if (k == 0) {
            cerr << "No values in hough space!\n";
            k = 1;
        } else {
            cout << "max hough = " << k << endl;
        }


        if (HOUGH_THETA == 1) {
            for (i = 0; i < HEIGHT; i ++) {
                for (j = 0; j < 120; j ++) {
                    houghSpaceTransmit[i*WIDTH+j] = (uchar)((double)((vision.houghSpace[i][j*3][0] + vision.houghSpace[i][j*3+1][0] + vision.houghSpace[i][j*3+2][0])/3)/(double)k*255.0);
                }
                for (j = 120; j < WIDTH; j ++) {
                    houghSpaceTransmit[i*WIDTH+j] = 254;
                }
            }
        } else {  //it can only otherwise be equal to 2 at this stage
            for (i = 0; i < HEIGHT/2; i ++) {
                for (j = 0; j < WIDTH; j ++) {
                    houghSpaceTransmit[i*WIDTH+j] = (uchar)((double)((vision.houghSpace[i][j][0]))/(double)k*255);
                }
            }
            for (j = 0; j < WIDTH; j ++) {
                houghSpaceTransmit[(HEIGHT/2)*WIDTH+j] = (uchar)(255);
            }
            for (i = 0; i < HEIGHT/2-1; i ++) {
                for (j = 0; j < 180/HOUGH_THETA; j ++) {
                    houghSpaceTransmit[(i+1+(HEIGHT/2))*WIDTH+j] = (uchar)((double)((vision.houghSpace[i][j+180/HOUGH_THETA][0]))/(double)k*255);
                }
                for (j = 180/HOUGH_THETA; j < WIDTH; j ++) {
                    houghSpaceTransmit[(i+1+(HEIGHT/2))*WIDTH+j] = (uchar)((double)((vision.houghSpace[i][j-180/HOUGH_THETA][0]))/(double)k*255);
                }
            }
        }
        memcpy(sendPlane + 2*(WIDTH*HEIGHT), houghSpaceTransmit, WIDTH*HEIGHT);
        //test, double up houghspace.
   //     memcpy(sendPlane + 3*(WIDTH*HEIGHT), houghSpaceTransmit, WIDTH*HEIGHT);


        for (i = 0; i < HOUGH_NUM_LINES; i ++) {
            if (vision.maxVals[i*2+0] != -1) {
                cout << "Line " << i << " has Hough weight of " << vision.houghSpace[vision.maxVals[i*2+0]][vision.maxVals[i*2+1]][0] << endl;
            } else {
                cout << "Line " << i << " has been removed.\n";
            }
        }

        //Tack maxVals onto the beginning of the transmission
        memcpy(sendPlane, vision.maxVals, HOUGH_NUM_LINES*2*sizeof(int));
        //Tack field endpoints after this.
        memcpy(sendPlane+HOUGH_NUM_LINES*2*sizeof(int), vision.fieldEnds, HOUGH_NUM_LINES*4*sizeof(int));


        //To test horizon stuff, put a line of red dog pixels 
        //in the 2nd debug plane frame (where the cplane is transmitted). 

        if (vision.horizonPointsGreen[0] == HORIZON_HORIZONTAL) {
            for (i = 0; i < WIDTH; i ++) {
                if (vision.horizonPointsGreen[i+1] != 0 && vision.horizonPointsGreen[i+1] != (HEIGHT-1)) {
                    sendPlane[WIDTH*(HEIGHT+vision.horizonPointsGreen[i+1])+i] = RED_DOG;
                }
            }
        } else if (vision.horizonPointsGreen[0] == HORIZON_VERTICAL) {
            for (i = 0; i < HEIGHT; i ++) {
                if (vision.horizonPointsGreen[i+1] != 0 && vision.horizonPointsGreen[i+1] != (WIDTH-1)) {
                    sendPlane[WIDTH*(HEIGHT+i)+vision.horizonPointsGreen[i+1]] = RED_DOG;
                }
            }
        } else {
            cout << "Unknown horizon type " << vision.horizonPointsGreen[0] << endl;
        }

        //Test distance to ground
        cout << "Distance to ground is " << vision.imageDistanceToGround() << endl;

    
        
      //Test line identification
    //  int * tempLineTypes = (int *)malloc(HOUGH_NUM_LINES*sizeof(int));
    //  vision.setLineType(vision.maxVals, HOUGH_NUM_LINES, tempLineTypes);
    //  vision.setLineType(vision.maxVals, HOUGH_NUM_LINES, vision.lineType);


      //Test line matching
    /*
    cout << "About to test matching\n";
      int * tempFieldLineDefs = (int*)malloc(HOUGH_NUM_LINES*2*sizeof(int));
      int * tempFieldLineID = (int*)malloc(HOUGH_NUM_LINES*2*sizeof(int));
      vision.defineFieldLine(vision.fieldEnds, HOUGH_NUM_LINES, tempFieldLineDefs);
      vision.identifyLine(tempFieldLineDefs, HOUGH_NUM_LINES, tempFieldLineID);
      free(tempFieldLineDefs);
      free(tempFieldLineID);
    */
           free(houghSpaceTransmit);

    }//}}}
    else if (EDGE_DETECT_TYPE == EDGE_PATTERN)
    {
     cout << "Sending debug for edge_pattern\n";
    //Note that we have to halve the dimensions of gradField in each direction in order to 
    //get them to fit into the sendPlane memory area!
    //Also note that the field is referenced with zero at the bottom left whereas sendPlane 
    //is referenced with zero at the top left. 
      for (int hs = 0; hs < vision.fieldHeight+2*vision.offsetY; hs = hs + 2)
      {//hs = h reference for sendplane, hf = hf reference for field gradient. 
        int hf = vision.fieldHeight+2*vision.offsetY - hs;
        for (int w = 0; w < vision.fieldWidth+2*vision.offsetX; w = w + 2)
        {
          //we can divide by 4 because since both h and w jump by 2, 
          //it'll all be multiples of 4 anyway. 
          int x = w/2;
          int y = hs/2;
          
          if (vision.sendDirField == 1)
          {
//            sendPlane[(y*(vision.fieldWidth+2*vision.offsetX)/2+x)] = vision.dirField[hf*(vision.fieldWidth+2*vision.offsetX)+w];
          }
          else
          {
            sendPlane[(y*(vision.fieldWidth+2*vision.offsetX)/2+x)] = vision.gradField[hf*(vision.fieldWidth+2*vision.offsetX)+w];
//            sendPlane[(y*(vision.fieldWidth+2*vision.offsetX)/2+x)] = vision.isolationField[hf*(vision.fieldWidth+2*vision.offsetX)+w];
          }
        }
      }
      //Now mark the spots where the robot can see things. 
      
      for (int i = 0; i < vision.numEdgePixels; i ++)
      {
        //Again note that we need to switch the vertical axis. 
        int x = vision.fieldEdgeSpace[i*2+0] + vision.offsetX;
        int y = vision.fieldEdgeSpace[i*2+1] + vision.offsetY;
        y = vision.fieldHeight+2*vision.offsetY - y;
        x = x/2;
        y = y/2;
        if ((int)((y*(vision.fieldWidth+2*vision.offsetX)/2+x)) < (vision.fieldWidth+2*vision.offsetX)*(vision.fieldHeight+2*vision.offsetY)/4 && (int)((y*(vision.fieldWidth+2*vision.offsetX)/2+x)) > 0)
        {
          sendPlane[(int)((y*(vision.fieldWidth+2*vision.offsetX)/2+x))] = 255;
        }
        else 
        {
          cout << "x=" << x << " y=" << y << " out of range\n";
        }
//        cout << "Marking edge at " << vision.fieldEdgeSpace[i*2+0] << "/" << vision.fieldEdgeSpace[i*2+1] << endl;
      }
      //Mark the 'corrected' point positions
      for (int i = 0; i < vision.numEdgePixels; i ++)
      {
        int x = vision.fieldPointsGlobalOrig[i*2+0] + vision.offsetX;
        int y = vision.fieldPointsGlobalOrig[i*2+1] + vision.offsetY;
        y = vision.fieldHeight+2*vision.offsetY - y;
        x = x/2;
        y = y/2;
        if ((int)((y*(vision.fieldWidth+2*vision.offsetX)/2+x)) < (vision.fieldWidth+2*vision.offsetX)*(vision.fieldHeight+2*vision.offsetY)/4 && (int)((y*(vision.fieldWidth+2*vision.offsetX)/2+x)) > 0)
        {
          sendPlane[(int)((y*(vision.fieldWidth+2*vision.offsetX)/2+x))] = 254;
        }
        else 
        {
          cout << "x=" << x << " y=" << y << " out of range (new)\n";
        }
//        cout << "Marking new edge at " << vision.fieldPointsGlobalOrig[i*2+0] << "/" << vision.fieldPointsGlobalOrig[i*2+1] << endl;
      }
      //Mark where the robot is
      int x = vision.currentXint + vision.offsetX;
      int y = vision.currentYint + vision.offsetY;
      y = vision.fieldHeight+2*vision.offsetY-y;
      x = x/2;
      y = y/2;
      sendPlane[(int)((y*(vision.fieldWidth+2*vision.offsetX)/2+x))] = 253;
      cout << "Marking robot position at " << x << "/" << y << endl;
      cout << "GPS recons I'm at " << vision.currentX << "/" << vision.currentY << "/" << vision.currentT << endl;
      cout << "I recon I'm at " << vision.currentXint << "/" << vision.currentYint << "/" << vision.currentTint << endl;
    }

    int dataType = 20;
    int msgType = 1;
    int arrayLength = DEBUG_SIZE;
	#ifdef ALL_WIRELESS_SHARED_ON
    memcpy(sharedDebugMemory->line_debug_data, sendPlane, arrayLength);

    objectOut->ClearBuffer();
    objectOut->SetData(&msgType, sizeof(int));
    objectOut->SetData(&dataType, sizeof(int));
    objectOut->SetData(&arrayLength, sizeof(int));
    objectOut->NotifyObservers();
	#endif
    free(sendPlane);
    
    cout << "\nVision::SendDebugPlane - debug plane size: " << arrayLength << endl;
    //Now send a cplane so we know where we're looking!
    forceCplane = true;
    cout << "vision.tilt = " << vision.tilt << " vision.pan=" << vision.pan << endl;



  int dirAngles[] = {45, 0,- 45, -90, -135, -180, 135, 90};//90 plus this minus currentT gives pan direction. 
  int reqPan[] = {0,0,0,0,0,0,0,0};
  uchar bitMask[] = {1,2,4,8,16,32,64,128};
  int minReqPan = LARGE_VAL;
  cout << "currentT = " << vision.currentT << endl;
  /*
  cout << "Viable pan directions: \n";
  
  for (int i = 0; i < 7; i ++)
  {
    reqPan[i] = dirAngles[i] - ((int)(vision.currentT) - 90);//90 + (dirAngles[i]) - (int)(vision.currentT);
    while (reqPan[i] < 0)
    {
      reqPan[i] = reqPan[i] + 360;
    }
    while (reqPan[i] >= 360)
    {
      reqPan[i] = reqPan[i] - 360;
    }
//    cout << "About to do vision.dirField[(vision.currentY)*(vision.fieldWidth+2*(vision.offsetX))+vision.currentX] & bitMask[i] " << (int)(vision.dirField[(vision.currentY)*(vision.fieldWidth+2*(vision.offsetX))+vision.currentX]) << "/" << (int)(bitMask[i]) << "/" << (int)(vision.dirField[(vision.currentY)*(vision.fieldWidth+2*(vision.offsetX))+vision.currentX] & bitMask[i]) << endl;
    if (
        (vision.dirField[(vision.currentY)*(vision.fieldWidth+2*(vision.offsetX))+vision.currentX] & bitMask[i] != 0)  //this is a viable direction
        )
    {
      cout << reqPan[i] << endl;
    }
  }
  cout << "4 & 8 is " << (int)((uchar)4 & (uchar)8) << " 4 and 7 is " << (int)((uchar)4 & (uchar)7) << endl;
  cout << "patternDx = " << vision.patternDx << " patternDy = " << vision.patternDy << " patternDt = " << vision.patternDt << endl;
*/
    
    sendCompressedCPlane();
#endif // SENDING_LINE_DEBUG
#endif // SENDING_DEBUG
}

//void Vision::setSendDebug(char * inCmd) {
void Vision::setSendDebug(double inVal)
{
    const char * inCmd = "1";
    if (inCmd[0] == '1')
    {
      sendDebugNow        = true;
      vision.sendDebugNow = true;
    }
    if (inCmd[0] == 'e')
    {
      vision.sendEdges = atoi((inCmd+1));
      cout << "Setting sendEdges to " << vision.sendEdges << endl;
    }
    if (inCmd[0] == 'a')
    {
      vision.altProject = atoi((inCmd+1));
      cout << "Setting altProject to " << vision.altProject << endl;
    }
    if (inCmd[0] == 'n')
    {
      vision.numTests = atoi((inCmd+1));
      cout << "Setting numTests to " << vision.numTests << endl;
    }
    if (inCmd[0] == 't')
    {
      vision.evilTiltOffset = (double)atoi((inCmd+1))/1000.0;
      cout << "Evil tilt offset = " << vision.evilTiltOffset << endl;
    }
    if (inCmd[0] == 'h')
    {
      vision.HOUGH_SHIFT_HORIZON = (double)atoi((inCmd+1))/1000.0;
      cout << "HOUGH_SHIFT_HORIZON = " << vision.HOUGH_SHIFT_HORIZON << endl;
    }
    if (inCmd[0] == 'g')
    {
      vision.groundPlane = atoi(inCmd+1);
      cout << "Setting groundPlane to " << vision.groundPlane << endl;
    }
    if (inCmd[0] == 'b')
    {
      vision.enableAbortBorder = atoi(inCmd+1);
      cout << "Setting abortBorder to " << vision.enableAbortBorder << endl;
    }
    if (inCmd[0] == 'd')
    {
      vision.sendDirField = atoi(inCmd+1);
      cout << "Send direction field = " << vision.sendDirField << endl;
    }
    if (inCmd[0] == 'i')
    {
      vision.disableBeacons = atoi(inCmd+1);
      cout << "disableBeacons = " << vision.disableBeacons << endl;
    }
    if (inCmd[0] == 's')
    {
      vision.isolationCutoff = atoi(inCmd+1);
      cout << "isolationCutoff = " << vision.isolationCutoff << endl;
    }
    if (inCmd[0] == 'P')
    {
      cout << "sendEdges = " << vision.sendEdges << endl;
      cout << "altProject = " << vision.altProject << endl;
      cout << "numTests = " << vision.numTests << endl;
      cout << "evilTiltOffset = " << vision.evilTiltOffset << endl;
      cout << "HOUGH_SHIFT_HORIZON = " << vision.HOUGH_SHIFT_HORIZON << endl;
      cout << "groundPlane = " << vision.groundPlane << endl;
      cout << "enableAbortBorder = " << vision.enableAbortBorder << endl;
      cout << "sendDirField = " << vision.sendDirField << endl;
      cout << "disableBeacons = " << vision.disableBeacons  << endl;
      cout << "isolationCutoff = " << vision.isolationCutoff << endl;
    }
    if (inCmd[0] == 'm')
    {
      //change match params
      switch (inCmd[1])
      {
        case 'X':
          vision.cfMatchedX = atoi(inCmd+2);
          break;
        case 'Y':
          vision.cfMatchedY = atoi(inCmd+2);
          break;
        case 'T':
          vision.cfMatchedT = atoi(inCmd+2);
          break;
        case 'x':
          vision.cfNoMatchX = atoi(inCmd+2);
          break;
        case 'y':
          vision.cfNoMatchY = atoi(inCmd+2);
          break;
        case 't':
          vision.cfNoMatchT = atoi(inCmd+2);
          break;
        case 'a':
          vision.matchCutoffX = atoi(inCmd+2);
          break;
        case 'b':
          vision.matchCutoffY = atoi(inCmd+2);
          break;
        case 'c':
          vision.matchCutoffT = atoi(inCmd+2);
          break;
        case 'p':
          cout << "cfMatchedX = " << vision.cfMatchedX << endl;
          cout << "cfMatchedY = " << vision.cfMatchedY << endl;
          cout << "cfMatchedT = " << vision.cfMatchedT << endl;
          cout << "cfNoMatchX = " << vision.cfNoMatchX << endl;
          cout << "cfNoMatchY = " << vision.cfNoMatchY << endl;
          cout << "cfNoMatchT = " << vision.cfNoMatchT << endl;
          cout << "matchCutoffX = " << vision.matchCutoffX << endl;
          cout << "matchCutoffY = " << vision.matchCutoffY << endl;
          cout << "matchCutoffT = " << vision.matchCutoffT << endl;
      }
    }
}
   //}}} 
//EDGE_DETECT end

#endif //EDGE_DETECT
