/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: VisualCortex.h 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 *
 * Object identification from blobs
 * 
**/


#ifndef VisualCortex_h_DEFINED
#define VisualCortex_h_DEFINED


#include <iostream>
#include <fstream>
#include "../share/Common.h"
#include "DistanceCalibrator.h"

#ifdef OFFLINE
#include <cstdio>
using namespace std;
#endif
//#include "Blob.h"

class Blobber;
class HalfBeacon;
class ONotifyEvent;
struct WMObj;

#include "GeneralisedColourDetect.h"
#include "../share/SwitchBoard.h"

#define HORIZON_HORIZONTAL 1 //EDGE_DETECT
#define HORIZON_VERTICAL 2  //EDGE_DETECT
#define EDGE_DETECT_TYPE EDGE_PATTERN

#define ALT_BALL

static const int VOB_COUNT = 17;

// for distance methods
static const int PIXEL_COUNT_DISTANCE = 0;
static const int PIXEL_COUNT_PROJECTION = 1;
static const int FIREBALL_PROJECTION = 2;


// active bar challenge code
//#define BAR_CHALL

// denotes infinity
// this is useful in determining when a variance
// is insignificant to be considered in computation
static const int LARGE_VAL = 10000000;

#ifdef ALT_BALL
static const int VOB_2BALL_COUNT = 18;	// 29/01/02
#endif

static const int kMaxPinkBlobs = 12;	// these are just double the max number we'd expect
static const int kMaxBlueBlobs = 6;		//                  "
static const int kMaxYellowBlobs = 6;	//                  "
static const int kMaxGreenBlobs = 4;	//                  "
static const int kMaxOrangeBlobs = 8;	// randomly chosen magic number
static const int numBeacons = 6;	


// Constants for variances, Tim Tam 13/03/02
//static const double BEACON_M = 0.0077388;
//static const double BEACON_M = 0.0077388;

#ifdef SHAPE_CHALL
/******************************/
// defines for Shape Challenge

typedef enum {
    Triangle,
    Rectangle,
    Square,
    TShape,
    LShape
}
ShapeEnum;

static const int SHAPE_COUNT = 5;
static const int BACKGROUND = 8;	//BACKGROUND
static const int FILL_COLOUR = 6;	//cROBOT_RED
static const int BORDER_COLOUR = 5;	//cROBOT_BLUE

/******************************/
#endif

static const double BEACON_C = 0.0203926;

typedef enum {

#ifdef ALT_BALL
    vobAltBall = 17,    // 29/01/02 new addition
#endif

    vobBall = 0,
    vobBlueGoal = 1,
    vobYellowGoal = 2,
    vobBlueLeftBeacon = 3,
    vobBlueRightBeacon = 4,
    vobGreenLeftBeacon = 5,
    vobGreenRightBeacon = 6,
    vobYellowLeftBeacon = 7,
    vobYellowRightBeacon = 8,
    vobRedDog = 9,
    vobRedDog2 = 10,
    vobRedDog3 = 11,
    vobRedDog4 = 12,
    vobBlueDog = 13,
    vobBlueDog2 = 14,
    vobBlueDog3 = 15,
    vobBlueDog4 = 16
}
VobEnum;

static const char *vobName[VOB_COUNT] = { 
    "ball",
    "blue goal",
    "yellow goal",
    "BlueLeft Beacon",
    "BlueRight Beacon",
    "GreenLeft Beacon",
    "GreenRight Beacon",
    "YellowLeft Beacon",
    "YellowRight Beacon",
    "Red Dog",
    "Blue Dog"
};

/**************Important***********************
// Heading: Left is Positive, Right is Negative
// Elevation: Up is Positive, Down is Negative
***********************************************/

#ifdef EDGE_DETECT
struct EdgeObject {
  double angleVar;
  double distVar;
  double angle;
  double dist;
  double eAngle;
  double eDist;
  int edgeNumber;
  bool isEdge;
  bool direction; //Always true for field edges, true for 
                  //lines where robot is FURTHER FROM ORIGIN than line. 
};

#endif //EDGE_DETECT

/*********************************************************
VOBS: Coming out of VisualCortex, the fields have slightly
different meanings to when they come out of Remembrate
e.g. 'head' is offset from current head position, without
reference to the body's position [when Rememebrate is
done with them, they are relative to the body]
*********************************************************/
class VisualObject {
	public:

    int cf;                   // from 0 to 1000 (0 => not seen)
    double var;               // variance of object determined from equation.
    double angleVar;          // variance of the angle of the object
    double cx, cy;            // centroid
    double x, y, height, width;  // actual object bounding box
    double radius;

    double projected_x;  // projection of bounding box onto image due to occlusion, or limited view
    double projected_y;
    double projected_h;
    double projected_w;
    HalfBeacon *blob;

    /****
    * These idea variables are in a bit of a halfway state because of the confusion between visual objects, world model
    * objects (GPS) and CurObjects (UNSWskills).
    ****/
    int area;        // area of blob(s)
    double imgHead;  // Image relative tilt and elevation
    double imgElev;
    double tilt;     // Internal rotation of vob (XY) [degrees]
    double head;     // Heading rotation of vob (XZ) [degrees]
    double lhead;    // Heading rotation of vob (XZ) [degrees]
    double rhead;    // Heading rotation of vob (XZ) [degrees]
    double h;
    double elev;     // Elevation rotation of vob (YZ) [degrees]
    
	//ERIC!! these are all for centroid testing
    double c2c; // ERIC  TEMP!!!!
    double c2c2; // ERIC  TEMP!!!!
	double c2c3; // ERIC  TEMP!!!!
	double pinkXsum;
	double pinkYsum;
	double pinkarea;
	double closestXsum;
	double closestYsum;
	double closestarea;
	double dist1;
	double dist2;
	double dist3;
	double var1;
	double var2;
	double var3;
	
    // above head and elev are robot relative

    double cam_dist;            // A raw camera distance measurement for ball tracking


    /* Distance from the camera to the centroid of the vob in cm.
     * It is a manipulated value to give accurate distances. It is unsuitable for ball tracking
    **/
    double cam_centroid_dist;   

    // Used in calibration process
    double dist;                // distance from the neck to edge of the vob [centimetres]
    double d;                   /* Two Dimensional Distance from to neck to 
                                   vob edge (parallel to the plane of the field) */
    double misc;                // beacon: c2c, ball: raw y co-ordinate of blob
    int pad;                    // may as well have it...

    double objectPosition[3];   //positiion in three dimensional space relative to the neck,
    double f;                   // the absolute direction the visual object is facing in world model.
    double orient;              // orientation of robots seen 0 - facing away chest, 180 - facing towards chest, -90 left, 90 right

    VisualObject ():var (LARGE_VAL), angleVar (LARGE_VAL) {
        Init ();
    }               // var must be set, else division by 0 occurs
    /*
     * void setDistance(double myX,double myY,double myZ)
     * Sets a number of visual object attributes from the 3 dimensional coordinate space relative to the base of the neck.
     * Attributes are: dist(3d), elev, heading (and h), d(2d)
    **/
    void setDistance (double myX, double myY, double myZ);
    /*
     * @return 3 dimensional distance from robot's neck.
    **/
    inline double getRobotDistance () const {
        return dist;
    }
    /*
     * @return 2 dimensional distance from robot's neck, on a plane parrallel to the field.
    **/
    double getFlatDistance () const;

    /*
     * @return 1 dimensional distance from robot's neck, on a plane parallel to the field and perpendicular to the body's
     * chest.  (the z axis)
    **/
    inline double getPerpDistance () const {
        return objectPosition[2];
    }
    /*
     * @return 1 dimensional distance from robot's neck on a plane parallel to the field and parallel to the body's
     * chest.  (the x axis)
    **/
    inline double getLateralOffset () const {
    return objectPosition[0];
    }
    /*
     * @return 1 dimensional distance from robot's neck on a plane perpendicular to the field plane.  (the y axis)
    **/
    inline double getVerticalOffset () const {
        return objectPosition[1];
    }
    void Print ();
    
    void printPlayData ();
    
    void Init ();

};


class VisualCortex {
public:

    VisualCortex ();
    ~VisualCortex () {
    }
	
	int ballDistMethod;
	
    // ricky: changed
#ifdef OFFLINE
    ofstream *outfile;
    int ballMethod;
    void ProcessImage (uchar *yuv, long s0,
                    long s1, long s2, double r2o, const WMObj& self);
#else // OFFLINE
#ifdef EDGE_DETECT
    void ProcessImage (const ONotifyEvent & event, long s0,
                    long s1, long s2, double r2o, const WMObj& self);
#else //EDGE_DETECT
    void ProcessImage (const ONotifyEvent & event, long s0,
                    long s1, long s2, double r2o);
#endif //EDGE_DETECT
#endif // OFFLINE
    
  void VisualCortex::initGoalDistanceLookup();

#ifdef ALT_BALL
    VisualObject vob[VOB_2BALL_COUNT];
#else
    VisualObject vob[VOB_COUNT];
#endif
#ifdef PRE_SANITY_BLOB
    ofstream *blobOutput;
#endif

    uchar *yplane, *uplane, *vplane, *cplane;
    long cameraFrame;

    double tilt, sin_tilt, cos_tilt;              //rad, +ve in downward direction
    // make consistent with PWalk?
    double pan, sin_pan, cos_pan;                 //rad, positive in left direction
    double roll, sin_roll, cos_roll;              //rad, positive in right direction?
    double effective_roll, sin_eroll, cos_eroll;  // values for unrotating the image
    double usin_eroll, ucos_eroll;                // values for rotating image back to normal

    double effective_tilt, sin_etilt, cos_etilt;  //rad, adjusted for the lift pan rotation produces
    double effective_pan, sin_epan, cos_epan;     //rad, adjusted for the lift pan rotation produces

    double dist_neck2snout;                       // cm, horizontal distance from neck base to snout, where camera is
    GeneralisedColourDetect gcd;

    int goalDistanceLookup[144];                  //Lookup table relating goal height to goal distance.

	// pointer to the current walk type the dog is performing
	const int *currentWalkType;

    void RobotRelative (VisualObject & object, bool dp = false);  // ANdres make this private!
    void set_mc (int);
    int get_mc ();

    // prev_colour is the colour of the previous pixel in run-length encoding
    // and height and width is the height and width location of the pixel.
    bool lookup_mc (unsigned char prev_colour, int height, int width);
    
    // initialise all elements to zero at the begining
    void initialiseSanityArray();

    // fired the given sanity
    void fireSanity(int sanityIndex);

    // get methods for the sanity Array
    int getSanityArray(int i);
    
    // get methods for the sensors (used in method storeObject in Vision.cc)
    long getSensor(int i);

    int countColouredPixels(int colour, int minX, int minY, int maxX, int maxY);

    int player;

#ifdef VISION_DEBUG
    // Keep these information relative to the camera as
    // visual objects get changed in remembrate to robot relative
    // new variables added in VisualObject eliminates the need for these
    //long ballHead, ballElev, bGoalHead, bGoalElev, yGoalHead, yGoalElev;

    void changeCalibration (int type);
    void switchFile (int suff);

    // Update the calibration dynamically via wireless.
    void updateCalibration (int col, int row, int newcolour);

    // Subroutine to determine whether to process the image or not.
    void freezeImageProcessing (int i);
#endif

#ifdef FIREBALL
    /*************************************
     * FIREBALL
     *************************************/

    bool isOutside(int x, int y);
    void findEdge(int start, int end, int fixed, bool outside, bool reverse, bool vertical, int* intersect_x, int* intersect_y, int edgeThreshold);
    void equationEdge(int start, int end, double m, double b, bool reverse, int* intersect_x, int* intersect_y, int edgeThreshold);
    void findCentre(HalfBeacon *preball, int *fbX, int *fbY, double *fbHead, double *fbDist, double *fbRadius, double *fb_obj0, double *fb_obj1, double *fb_obj2);
    void findPointOnCircumferenceOneAndTwo(int bxmin, int bxmax, int bymin, int bymax, bool nl, bool nr, bool sl, bool sr, int *p1x, int *p1y, int *p2x, int *p2y);
    void findPointOnCircumferenceThree(int bxmin, int bxmax, int bymin, int bymax, bool nl, bool nr, bool sl, bool sr, int p1x, int p1y, int p2x, int p2y, int *p3x, int *p3y);
    void findIntersectionOfTwoChords(int p1x, int p1y, int p2x, int p2y, int p3x, int p3y, int *ix, int *iy, double *radius);
    void projectPointsD(int * inPoints, int numPoints, double * outPoints, double projectHeight = 0);
    void projectPointsDAlt(int * inPoints, int numPoints, double * outPoints, double projectHeight = 0);

    // test
    void testFunction(int p1x, int p1y, double p13m, double p13b, double p23m, double p23b, int *ix, int *iy, double *radius);
    // debug
#endif

#ifdef EDGE_DETECT
 
  //{{{
//Defines (change to constants later)
    //{{{
#undef PROJECT_SENTINEL //set this to allow inputPoint[i*2+0] to turn off projectPoints
#define HOUGH_MAX_R 228 //Set to sqrt(PIC_WIDTH^2+PIC_HEIGHT^2)
#define HOUGH_MAX_VALS WIDTH*HEIGHT
#define TILT_DEPRESSION 0.027515989 	//The REAL angle of tilt downwards (in
					//radians when the tilt reading is zero.
#define HOUGH_NUM_LINES 3
//#define HOUGH_CULL 0.6 	//The point at which to cull hough lines (any line 
											//with a value less than the maximum/HOUGH_CULL
											//gets reset to zero). Set to zero to not cull at all.


														//use the upper left corner)
//#define DEBUG_TYPE 20
#define HOUGH_THETA 2 //Reduces Hough angle resolution to this (1 for every degree)

#define HOUGH_INCLUDE_EDGEPLANE	//undefine to speed up processing by not 
				//generating an edge frame (still generates 
				//edge list for Hough transform).
#define HOUGH_BEACON_AS_FIELD	//Treat beacon green as field green for edge detection

//Line ID stuff
#define HOUGH_LINEID_DIST 20 // MAKE THIS VARIABLE!
#define HOUGH_LINEID_NUMPTS 10
#define HOUGH_LINEID_WHITEDIFF 5		//Have a difference in white pixel number between each side of 
																		//at least this much to classify as field edge
#define HOUGH_LINEID_WHITEMIN 5	 //Have at least this many white pixels to class as white
#define HOUGH_LINEID_THRESHOLD3 7	//Min lines in picture to count
#define LINE_UNSET 0		//We've not looked at the line to give it a value yet.
#define LINE_UNINIT 1		//We've looked for the line but it's been reset to -1,-1.
#define LINE_UNIDENTIFIED 2	//We can't identify the line (we may break this up later).
#define LINE_FIELDEDGE 3	//Line is a field edge (checked for 'field in sky')
#define LINE_FIELDLINE 4	//Line is on the field
#define LINE_BEVELBLUE 5	//NOT USED
#define LINE_BEVELYELLOW 6	//NOT USED
#define HOUGH_EDGEDETECT_USEHORIZON
#define HOUGH_FREQUENCY 1
#define HOUGH_MULTILEVEL_DEBUG 0
#define HOUGH_NUM_MULTILEVEL_R 8
#define HOUGH_NUM_MULTILEVEL_T 8
#define MAX_EDGE_PIXELS 300
#define CROSS_SPACING_MASK 0x03  //doesn't work!
#define APERTURE_DISTANCE 161.714647712  //Set to HEIGHT/(2*tan(yField/2));//64.829091184;/
#define NOT_EDGE 10

#define R_WEIGHT 1  //Weightings for r and theta when matching observed lines 
#define T_WEIGHT 1  //to lines on the field

#define numFieldLines 9
#define numEdgeLines 8
#define EDGE_HOUGH 1
#define EDGE_PATTERN 2
bool altProject;
    //}}}

/**
 * Stuff in the 'pipeline' between the cplane coming in to 
 * the identified edges coming out. Higher level stuff mainly. 
 * - setSelfPos <--- REMOVEd
 * - runEdgeDetect
 * - initEdgeDetect
 * - initEdgeTable
 * - findFieldEdges
 * - findHoughSpaceMultilevel <---  REMOVED
 * - setLineEnds  <--- REMOVED
 * - setLineType  <--- REMOVED
 * - projectPoints
 * - defineFieldLine  <--- REMOVED
 * - identifyLine <--- REMOVED
 * - runSanityCheck <--- REMOVED
 */
//{{{
/*
  void VisualCortex::setSelfPos(int inX, int inY, double inT, double inVarX, double inVarY, double inVarT);
  */
  void VisualCortex::runEdgeDetect(const WMObj& self);
  void VisualCortex::initEdgeDetect();
  void VisualCortex::runEdgeDetectHough(const WMObj& self);
  void initEdgeTable();
  int findFieldEdges(uchar *retVal, uchar *edgeListOut);
  void VisualCortex::projectPoints(int * inPoints, int numPoints, int * outPoints, int projectHeight = 0);
  void VisualCortex::projectPointsAlt(int * inPoints, int numPoints, int * outPoints, int projectHeight = 0);

//}}}

/**
 * Pipeline stuff for pattern matching
 */
  
  
/**
 * Various horizon functions
 * - generateHorizonPoints
 * - generateHorizonDefinition
 * - imageDistanceToGround
 * - horizonStartX (dependant on HOUGH_EDGEDETECT_USEHORIZON)
 * - horizonEndX (dependant on HOUGH_EDGEDETECT_USEHORIZON)
 * - horizonCounter (used in the above 2 fns) (dependant on HOUGH_EDGEDETECT_USEHORIZON)
 */
//{{{
  void VisualCortex::generateHorizonPoints(double inAngle, int * retVal);
  double * VisualCortex::generateHorizonDefinition(double inAngle);
  double imageDistanceToGround();
#ifdef HOUGH_EDGEDETECT_USEHORIZON
  int VisualCortex::horizonStartX(int inH);
  int VisualCortex::horizonEndX(int inH);
  int horizonCounter;
#endif //HOUGH_EDGEDETECT_USEHORIZON
//}}}
 

  
/**
 * Reasonably specific matrix functions
 * - generatePicToFieldXfrm
 * - createMatrix4x4
 * - applyMatrix
 */
//{{{
  double * generatePicToFieldXfrm();
  double* VisualCortex::createMatrix4x4(double a1, double b1, double c1, double d1, 
                                      double a2, double b2, double c2, double d2,
                                      double a3, double b3, double c3, double d3,
                                      double a4, double b4, double c4, double d4);
  double * VisualCortex::applyMatrix(double * inMatrix, double * inVector);
//}}}



/**
 * Somewhat generic trig functions, move these into somewhere where they're 
 * accessible to all ASAP.
 * - initDegSinLookup
 * - degSinLookup
 * - sinRAD
 * - sinDEG
 * - degCosLookup
 * - cosRAD
 * - cosDEG
 * - rad
 * - zeroTo360
 */
//{{{
  double rad(int t);
  double degSinLookup(int t);
  double degCosLookup(int t);
  double sinRAD(double t);
  double sinDEG(double t);
  double cosRAD(double t);
  double cosDEG(double t);
  void initDegSinLookup();
//  void zeroTo360(int * inAngle);
  double VisualCortex::zeroTo360(double inAngle);
//}}}



/**
 * NightOwl prototypes
 */
  //{{{
static const int VISUALCORTEX_TIME_PRINT = 0;
uchar * gradField;  //This array holds the field gradient (slopes up to lines) and is referenced:
                    //gradField[(y*(fieldWidth/incX)/incY)+(x/incX)]
uchar * dirField;   //This array holds the direction to look at to get to a line and is referenced in the same way as gradField.
                    //Directions are 0 to 8 where 0 is NorthWest, 1 is North, 2 is NorthEast and so on going clockwise. 
inline uchar const VisualCortex::GF(const int x, const int y) {
    return gradField[y*(fieldWidth+2*offsetX)+x];
}

inline uchar const VisualCortex::DF(const int x, const int y) {
    return dirField[y*(fieldWidth+2*offsetX)+x];
}
uchar * isolationField; //Areas where if the final match has more than isolationCutoff pixels here, we discard the edge data. 
uchar * isolationFieldGoalie; //Like above but without excluding the goalboxes
bool isGoalie;
int * fieldPointsLocal; //Projected points in local co-ords
int * fieldPointsGlobalOrig; //Original set of field points (note, this moves per iteration). 
int * fieldPointsGlobal; //Projected points in global co-ords
int * transformAccs;
bool edgesValid;
static const int fieldWidth = FIELD_WIDTH;  //This is beacon to beacon width (cm).
static const int fieldHeight = FIELD_LENGTH; //This is beacon to beacon height (cm).
static const int incX = 1;  //gradient field increments in X direction (cm)
static const int incY = 1;  //gradient field increments in Y direction (cm)
static const int offsetX = 50;  //Amount by which the gradient field extends beyond the beacons (cm)
static const int offsetY = 50;  //Note, these values must be at least 30cm (the amount the home goal is set back by)
static const double cfX = 1.0/(double)incX; //gradient field coefficients (for generation)
static const double cfY = ((double)(fieldWidth+2*offsetX)/(double)incX)/(double)incY;
static const int gOff = (int)(cfX*offsetX+cfY*offsetY);

int groundPlane;
bool abortBorderSet;
bool enableAbortBorder;
int abortCount;     //The running number of frames aborted in the current 25 frames (NOT running average!). 
int lastAbortCount; //The number of frames aborted in the last 25 frames (updated once per frame). 

//Time stuff
int peakBlobTime, peakFindTime, peakEdgeTime, peakEdgeDetectTime, peakTotalTime;
//Time stuff within runEdgeDetectHough
int peakHorizonT, peakFindT, peakProjectT, peakNightT;
int timeReadingCounter;


//static const int numTests = 5; //we perform 3 iterations only.
int numTests; //must stay less tham maxNumTests in VisualCortex-priv.h
/*
static const int numSamplesX = 5; //Number of samples in each dimension for transforms (must be odd)
static const int numSamplesY = 5;
static const int numSamplesT = 11;
static const int distSamplesX = 2;  //Distance between each sample (cm or degrees)
static const int distSamplesY = 2;
static const int distSamplesT = 2;
*/

/* MOVED INTO VisualCortex-priv.h
//NOTE: all the arrays below must have at least maxNumTests elements in them.
static const int numSamplesX[maxNumTests] = {5,5,5,5,5};
static const int numSamplesY[maxNumTests] = {5,5,5,5,5};
static const int numSamplesT[maxNumTests] = {10,5,5,5,5};
static const int distSamplesX[maxNumTests] = {10,5,3,2,1};
static const int distSamplesY[maxNumTests] = {10,5,3,2,1};
static const int distSamplesT[maxNumTests] = {5,5,4,3,2};
*/
static const int maxNumSamplesX = 5;
static const int maxNumSamplesY = 5;
static const int maxNumSamplesT = 11;

int originTransform[maxNumSamplesX*maxNumSamplesY*maxNumSamplesT*3];//Stores the origin transform that matches the field edgepoint transform of a given index.

void VisualCortex::initCurrentPosition();
void VisualCortex::findCurrentPosition(int * inX, int * inY, int * inT);
void VisualCortex::findCurrentPositionOld(int * inX, int * inY, int * inT);
void VisualCortex::findCurrentPositionNew(int * inX, int * inY, int * inT, int * inFieldEdgeSpace, int inNumEdgePixels);
void VisualCortex::applyTransformNew(int * inOrig, int inX, int inY, int inNumPoints, int * outFinal, int iteration, int inTransformIndex);
          
//Takes inNumPoints points as local co-ordinate pairs in inLocal and transforms them into
//global co-ordinate pairs in outGlobal based on inX, inY and inT. NOTE: outGlobal must be 
//twice inNumPoints in size (as must inLocal) or we'll have problems. 
void VisualCortex::localToGlobal(int inX, int inY, int inT, int * inLocal, int inNumPoints, int * outGlobal);
//Generates a series of transforms based on the static consts at the start then 
//applies them to inOrig to get outFinal. NOTE THAT WE ASSUME OUTFINAL HAS 
//BEEN DEFINED AS BEING LARGE ENOUGH OTHERWISE WE'LL SEGFAULT!
void VisualCortex::transformFieldPoints(int * inOrig, int inX, int inY, int inNumPoints, int * outFinal, int iteration);
//Transforms a xyt (t=theta) triple by the number of the input transform (note that 
//the number of the transform is as defined by the static consts and as used by 
//transformFieldPoints and as in the outFinal array). 
void VisualCortex::transformXYT(int * inX, int * inY, int * inT, int inIndex);
//}}}
/*
 * End NightOwl prototypes
 */




/**
 * Private variables. 
 */
  
  //General book-keeping stuff
  int sendDebugNow;
  int picSerialNo;
  static const int DEGSINLOOKUP_NUMENTRIES = 91;  //we only have 91 entries because we can create all 360 degrees by flipping or offsetting that of the initial 90 (made 91 to make the math easier - we have entries for angles 0 to 90). 
  double degSinLookupTable[DEGSINLOOKUP_NUMENTRIES];

  //Position stuff coming out of GPS
  int currentX, currentY;
  double currentT, varX, varY, varT;
  
  //Edge stuff
  uchar * fieldEdges;
  uchar * edgeList; 
  int numEdgePixels;  //number of pixels after culling (if enabled)
  int origNumEdgePixels; //number of pixels before culling
  static const int NUM_EDGE_TYPES = 256;  //256 edge types because we have one edge type for every combination 
                                          //of 8 surrounding pixels.
  int edgeTypeTable[NUM_EDGE_TYPES];//!!!!!! Remember to populate this
  int avgEdgeX; //average position of edge points in X/Y in the image plane. 
  int avgEdgeY;

  //Hough stuff
  int setVals[HOUGH_MAX_VALS][2];
  int houghSpace[HOUGH_MAX_R][360/HOUGH_THETA][2];
  int houghFrequencyCount;
  int houghSpaceMulti[HOUGH_MAX_R/HOUGH_NUM_MULTILEVEL_R][(359/HOUGH_THETA)/HOUGH_NUM_MULTILEVEL_T + 1][4];
  double HOUGH_SHIFT_HORIZON;
  double HOUGH_CULL;  //Fraction of max that a line must satisfy before being guaranteed inclusion
  int HOUGH_CULL2;  //Minimum a line must satisfy before being guaranteed inclusion
  int HOUGH_CULL3;  //Minimum that the highest line must be
  double HOUGH_GREEN_LOWER_LIMIT;
  int * maxVals;
  //Weightings for the various variances
  double HOUGH_DISTVAR_NUMPIXELS;   //Scaling for 1/(number of pixels matched)
  double HOUGH_ANGVAR_NUMPIXELS;

  //Cutoff for input variances
  double xVarCutoff;
  double yVarCutoff;
  double tVarCutoff;

  //Pattern stuff
  int fieldEdgeSpace[MAX_EDGE_PIXELS*2];
  int fieldEdgeSpaceType[MAX_EDGE_PIXELS];
  int tempFieldEdges[MAX_EDGE_PIXELS*2];
  int currentXint, currentYint, currentTint;
  int edgeMatchQuality; 
  bool sendEdges;
  bool sendDirField;
  int patternDx, patternDy, patternDt;
  int maxDiffX, maxDiffY, maxDiffT;
  int cfMatchedX, cfMatchedY, cfMatchedT, cfNoMatchX, cfNoMatchY, cfNoMatchT; //parameters to get95CF for matched and no match
  int matchCutoffX, matchCutoffY, matchCutoffT; //delta cutoffs before something is not matched
  int isolationCutoff;
  bool edgesToGPS;
  
  
  //Line stuff
  int lineType[HOUGH_NUM_LINES];
  int lineEnds[HOUGH_NUM_LINES*4];

  //Projection stuff
  int fieldEnds[HOUGH_NUM_LINES*4];
  int fieldDefs[HOUGH_NUM_LINES*2];
  int fieldIDs[HOUGH_NUM_LINES*2];//Umm ... why is this 2?
  int * horizonPointsGreen;
  double evilHackFactor1;
  double evilHackFactor2;
  double evilTiltOffset;
  double evilDistanceHack;

  //Final thing to pass to GPS
  int fieldLinesRT[numFieldLines*2*2];
  int edgeLinesRT[numEdgeLines*2];
  EdgeObject edges[HOUGH_NUM_LINES];
  bool disableBeacons;

  //Useful stuff for behaviours
  bool facingFieldEdge; //true if vision sees a lot of fieldwhite (and possibly a lot of ball) but not an awful lot else
  bool facingBlueGoal;
  bool facingYellowGoal;
  int whiteAndBallThreshold;
  int whiteThreshold;
  int blueAndBallThreshold;
  int blueThreshold;
  int yellowAndBallThreshold;
  int yellowThreshold;

    //}}}


#endif //EDGE_DETECT


    bool cmdAdjustDistance;   // distance calibration testing
    // PLEASE remove this after the march
    double beaconDistanceConstant;

    // gets rid of recognizing carpet as beacon
    int beaconSearchPerim;
    double beaconErrorThreshold;
    int filterInnerBeacon;

    double errorScale;   // scale of partially seen obj

    double BEACON_M;

    // the constant to use for variance adjustance of partially seen objects
    double varConst;

    double minGoalDim;

    void wlLoadNNMC (const unsigned char *data, int size);


private:
    // Struct for a half beacon type
    // Struct for a beacon type
    class Beacon {
    	public:
        bool valid;
        double c2c;                // c2c distance of the 2 halfbeacons making up the beacon
		double c2c2; //ricky and ERIC testing
		double c2c3; //ricky and ERIC testing
		
		double pinkXsum;
		double pinkYsum;
		double pinkarea;
	
		double closestXsum;
		double closestYsum;
		double closestarea;
	
        double cx, cy;
        int height, width, area;
        double angle;
        int halfmass;              // area of larger halfbeacon

        void Set (const HalfBeacon & pink, const HalfBeacon & closest);
    };
    
    void LocateLandmarks ();
    void CheckMatchingBeacons (const HalfBeacon * pink, int colorIndex,
                               HalfBeacon ** hbArray, int hbCount,
                               double &minDist, double &maxRatio,
                               HalfBeacon ** closest, int *closestType);
    void TrigInit (const long &micro_tilt, const long &micro_pan,
                   const long &micro_roll);
    void LookForGoal (HalfBeacon ** hbArraye, int hbCount,
                      VisualObject & logoal, VisualObject & rogoal,
                      VisualObject & goal);

    int VobInVob (VisualObject & b1, VisualObject & b2, int ot, int or_, int ob, int ol);

    void FindBall ();
    bool normalFindBall(HalfBeacon *preball, VisualObject &ball);
    bool andresFindBall(HalfBeacon *preball, VisualObject &ball);
    void FindBeacon ();
    void FindGoal ();

    void Reset ();

    bool InYellowNoGreen (VisualObject & ball);

    bool NoGreenNearBy (int xmin, int xmax, int ymin, int ymax);
    bool NoColourXNearBy (int colour, int xmin, int xmax, int ymin,
                          int ymax, int deltaX, int deltaY, int threshHold);

    int FindHalfBeacons (int colorIndex, HalfBeacon ** hbArray, int maxHB, int threshold);
    int FindVisualObjects ();

    double NewBallDistance ();

    bool DarkBackgroundInBetween(int left, int right, int top, int bottom);
    //void FindColorRobot (int vobNum);
    void FindColorRobot (int colour);
    void FindRobot ();
    void RobotNoiseFilter ();

    void PrintBeaconInfo ();

    void convertGoalBeacon ();

    void GoalSanities ();

    void BeaconSanities ();

    void LandmarkSanities ();

    bool BallSanities (VisualObject &, int, int, int, int, double, double, double, double);

    // used to draw lines between centroids of blobs and then returns
    // the number of 'colour's seen on the line
    int drawLine (int x1, int y1, int x2, int y2, uchar colour);

    int numPink, numGreen, numBlue, numYellow;

    Beacon beacon[numBeacons];

    // these are numbered from left to right, top to bottom
    // with the 'top' of the field being the blue goal,
    // and the 'bottom' being the yellow goal.
    // ie #0=blue/pink, and #5=pink/yellow

    HalfBeacon *hbPink[kMaxPinkBlobs];
    HalfBeacon *hbBlue[kMaxBlueBlobs];
    HalfBeacon *hbYellow[kMaxYellowBlobs];
    HalfBeacon *hbGreen[kMaxGreenBlobs];
    HalfBeacon *hbBall[kMaxOrangeBlobs];
    
    int sanityArray[NUM_SANITY_COORDS];
    bool firedSanity[SANITY_COUNT];

    long sensor0;
    long sensor1;
    long sensor2;
    int infaredDistance;

    DistanceCalibrator *ballCalibrator;

//  int CreateRunLengthInfo(int start, RunLengthInfo& info);
    Blobber *blobs;

// private varialbles and functions for Shape Challenge
#ifdef SHAPE_CHALL

    void fillGaps (int height, int width, uchar colour);
    void seedFill (int x, int y, int width, int height, uchar colour, uchar border);
    void runFill (int x, int y, int width, int height, uchar colour, uchar border);
    void calcMoments (uchar fill, uchar border, int height, int width);

    double moment[2][2];                /* zeroth moment and first moment */
    double moment1[4][4];                /* central moment */
    double moment2[4][4];                /*normalization of central moment */

public:
    uchar splane[HEIGHT][WIDTH];
    int findShape (int colour);

    double ptr[7];

#endif

};

double getAng (double dx, double dy);
int imax (int a, int b);
int pi_abs (int a);
int imin (int a, int b);

double PointToHeading (double x);
double PointToElevation (double y);
double PointToHeading (double x, double y, double sin_eroll, double cos_eroll);
double PointToElevation (double x, double y, double sin_eroll, double cos_eroll);
long SizeToDistance (long physicalSize, long apparentSize);

/***
 * Queue class added by Ben
 */
class QueueItem {
private:
    int item;
    QueueItem *next;

public:
    QueueItem () {
    }
    ~QueueItem () {
    }

    QueueItem (int qitem) {
        item = qitem;
        next = NULL;
    }

    // returns the item in this container
    int getItem () {
        return item;
    }

    // returns the next container
    QueueItem *getNext () {
        return next;
    }

    // set the next container
    void setNext (QueueItem * qitem) {
        next = qitem;
    }

}; // class QueueItem

class Queue {
private:
    // this queue is implemented as linked
    QueueItem * first, *last;
    int size;

public:

    Queue () {
        first = NULL;
        last = NULL;
        size = 0;
    }
    ~Queue () {
    }

    bool isEmpty () {
        return first == NULL;
    }

    // queue the object to the end of the queue
    void enqueue (int item) {
        // make a new container for this object
        QueueItem *newItem = new QueueItem (item);
        size++;
        // append the object to the end of the queue
        if (first != NULL) {
            last->setNext (newItem);
            last = newItem;
        } else {
            last = newItem;
            first = last;
        }
    }

    // dequeue the object from the top of the queue return null if
    // the queue is empty
    int dequeue () {
        // dequeue the first item and reset pointers
        if (first != NULL) {
            int result = first->getItem ();
            QueueItem *temp = first->getNext ();

            delete first;

            first = temp;
            size--;

            return result;
        } else {
            return 0;
        }
    }

    int getSize () {
        return size;
    }

};  // class Queue

#endif // VisualCortex_h_DEFINED



