/*

Copyright 2003 The University of New South Wales (UNSW) and National  
ICT Australia (NICTA).

This file is part of the 2003 team rUNSWift RoboCup entry.  You may  
redistribute it and/or modify it under the terms of the GNU General  
Public License as published by the Free Software Foundation; either  
version 2 of the License, or (at your option) any later version as  
modified below.  As the original licensors, we add the following  
conditions to that license:

In paragraph 2.b), the phrase "distribute or publish" should be  
interpreted to include entry into a competition, and hence the source  
of any derived work entered into a competition must be made available  
to all parties involved in that competition under the terms of this  
license.

In addition, if the authors of a derived work publish any conference  
proceedings, journal articles or other academic papers describing that  
derived work, then appropriate academic citations to the original work  
must be included in that publication.

This rUNSWift source is distributed in the hope that it will be useful,  
but WITHOUT ANY WARRANTY; without even the implied warranty of  
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
General Public License for more details.

You should have received a copy of the GNU General Public License along  
with this source code; if not, write to the Free Software Foundation,  
Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 * Last modification background information
 * $Id: VisualCortex.cc 1953 2003-08-21 03:51:39Z eileenm $
 *
 * Copyright (c) 2003 UNSW
 * All Rights Reserved.
 * 
**/

#ifndef OFFLINE
#include <OPENR/OSubject.h>
#include <OPENR/OObserverVector.h>
#include <OPENR/OPENREvent.h>
#include <OPENR/OPENRAPI.h>
#include <OPENR/OFbkImage.h>
#endif // OFFLINE

#include <cstdio>
#include <cstring>
#include <cfloat>

#include "VisualCortex.h"
#include "VisualCortex-priv.h"
#include "../share/Common.h"
#include "../share/PWalkDef.h"
#include "gps.h"
#include "Blob.h"

#include "../share/SwitchBoard.h"
#include "../share/BallDef.h"

#ifdef OFFLINE
#include <valgrind/memcheck.h>
#endif // OFFLINE

using namespace std;
/*
#ifndef OFFLINE // not OFFLINE
static SystemTime idiff;
#endif // OFFLINE
*/
static const int ROBOTS_EACH_COLOUR = 4;
static const int MAX_BLOBS_PER_DOG = 100;

static const bool sensorValueDebugMsg  = false;
static const bool beaconSanityDebugMsg = false;
static const bool sanityBlobDebugMsg   = true;

static const double zoidalVarFactor    = 40.0;
static const double zoidalAngVarFactor = 1.5;

static double lastd = 0;

VisualCortex::VisualCortex () {
    ballCalibrator = new DistanceCalibrator();
	blobs = new Blobber(gcd);
    cplane = gcd.cplane;
    cmdAdjustDistance = 1;
    beaconDistanceConstant = 1900;

    beaconSearchPerim = 2;
    beaconErrorThreshold = 0.6;    // % of green field seen before discarding a blob
    filterInnerBeacon = 3;
    BEACON_M = 0.0077388;
    errorScale = 2;
    varConst = 50;
    minGoalDim = 10;
	
	currentWalkType = NULL;
/*
    //clear timing
    peakBlobTime = 0;
    peakFindTime = 0;
    peakEdgeTime = 0;
    peakEdgeDetectTime = 0;
    
    peakHorizonT = 0;
    peakFindT = 0;
    peakProjectT = 0;
    peakNightT = 0;
    
    timeReadingCounter = 0;
 */   
#ifdef EDGE_DETECT
    initEdgeDetect();
    abortCount = 0;
    lastAbortCount = 0;
    disableBeacons = true;
#else //EDGE_DETECT
    disableBeacons = false;
#endif //EDGE_DETECT
	sendDebugNow = 0;
  initGoalDistanceLookup();

#ifdef OFFLINE
    cout << "VisualCortex.cc: OFFLINE turned on" << endl;
    ballMethod = 0;
#else
    cout << "VisualCortex.cc: OFFLINE turned off" << endl;
#endif

	ballDistMethod = PIXEL_COUNT_DISTANCE;

#ifdef PRE_SANITY_BLOB
    cout << "VisualCortex.cc: PRE_SANITY_BLOB turned on" << endl;
#else
    cout << "VisualCortex.cc: PRE_SANITY_BLOB turned off" << endl;
#endif
}

void
VisualCortex::wlLoadNNMC (const unsigned char *data, int size) {
    gcd.wlLoadNNMC(data, size);
}

// initialise all elements to zero at the begining
void VisualCortex::initialiseSanityArray() {
    for (int i=0; i<NUM_SANITY_COORDS; i++) {
        sanityArray[i] = 0;
    }
    for (int i=0; i<SANITY_COUNT; i++) {
        firedSanity[i] = false;
    }
}

// fired the given sanity
void VisualCortex::fireSanity(int sanityIndex) {
    if (!firedSanity[sanityIndex]) {
        sanityArray[sanityIndex/INT_BITS] += 1 << (sanityIndex%INT_BITS);
        firedSanity[sanityIndex] = true;
    }
}

// get methods for the sanity Array
int VisualCortex::getSanityArray(int i) {
    return sanityArray[i];
}

// get methods for the sensors (used in method storeObject in Vision.cc)
long VisualCortex::getSensor(int i) {
    switch (i) {
        case 0:
            return sensor0;
        case 1:
            return sensor1;
        case 2:
            return sensor2;
    }
    return -1;
}

/*
 *  Resets the variables for a new image feed.
**/
void VisualCortex::Reset () {

  // reset all visual objects
#ifdef ALT_BALL
    for (int x = 0; x < VOB_2BALL_COUNT; x++)
        vob[x].Init ();
#else
    for (int x = 0; x < VOB_COUNT; x++)
        vob[x].Init ();
#endif
    // set all the beacons as invalid
    for (int k = 0; k < numBeacons; k++)
        beacon[k].valid = 0;
    //reset all blobs and halfBeacons
    blobs->reset ();

    // reset the sanity that fired previous
    initialiseSanityArray();
}

void
VisualCortex::TrigInit (const long &micro_tilt, const long &micro_pan, const long &micro_roll) {
    tilt = -MICRO2RAD (micro_tilt) + TILT_OFFSET;
    sin_tilt = sin (tilt);
    cos_tilt = cos (tilt);

    pan = MICRO2RAD (micro_pan);
    sin_pan = sin (pan);
    cos_pan = cos (pan);

    roll = MICRO2RAD (micro_roll);
    sin_roll = sin (roll);
    cos_roll = cos (roll);

    effective_tilt = asin (sin_tilt * cos_pan);
    sin_etilt = sin (effective_tilt);
    cos_etilt = cos (effective_tilt);

    double dist_neck2head = NECK_LENGTH * sin_tilt;
    double dist_head2snout = FACE_LENGTH * cos_etilt;
    dist_neck2snout = sqrt (SQUARE (dist_neck2head) + SQUARE (dist_head2snout) +
                            2 * dist_neck2head * dist_head2snout * cos_pan);

    // can consider using epan = p-4/SQUARE(M_PI)*t*p, error of 3.5 degrees
    effective_pan = asin (dist_head2snout * sin_pan / dist_neck2snout);
    sin_epan = sin (effective_pan);
    cos_epan = cos (effective_pan);

    effective_roll = asin (sin_pan * sin_tilt);
	
	if(tilt > M_PI/2) {
		if(effective_roll < 0)
			effective_roll = -M_PI - effective_roll;
		else
			effective_roll = M_PI - effective_roll;
	}
	
    sin_eroll = sin (effective_roll);
    cos_eroll = cos (effective_roll);

    usin_eroll = sin (-effective_roll);
    ucos_eroll = cos (-effective_roll);

}


void
VisualCortex::initGoalDistanceLookup()
{
  //initializes the int goalDistanceLookup[144] table. 
  //Between goal heights of zero to 80 pixels (infinite to 50 cm)
  //the expression for distance d relative to height h is:
  //
  //d = 6,878.598/(h^1.125)
  //
  //Between goal heights of 80 to 144 pixels (50 to 0 cm) 
  //the expression for distance d relative to height h is:
  //
  //d = 144-1.28*h
  //
  //This is precomputed to reduce processing time for each frame. 
  goalDistanceLookup[0] = (int)LARGE_VAL;
  for (int h = 1; h < 80; h ++)
  {
    goalDistanceLookup[h] = (int)(6878.598/(pow(((double)h),1.125)) * 0.9);
  }
  for (int h = 80; h < 144; h ++)
  {
    goalDistanceLookup[h] = (int)(144.0-1.28*(double)h * 0.9);
  }
}



#ifdef OFFLINE
void
VisualCortex::ProcessImage (uchar *yuv, long s0,
                long s1, long s2, double r2o, const WMObj& self) {
				
	// invalidates the memory holding all the vobs for valgrind
	VALGRIND_MAKE_WRITABLE(vob,VOB_2BALL_COUNT*sizeof(VisualObject));
#else // OFFLINE
#ifdef EDGE_DETECT
void
VisualCortex::ProcessImage (const ONotifyEvent & event, long s0,
                long s1, long s2, double r2o, const WMObj& self) {
#else //EDGE_DETECT
void
VisualCortex::ProcessImage (const ONotifyEvent & event, long s0,
                long s1, long s2, double r2o) {
#endif //EDGE_DETECT
//SystemTime timeBlobDone, timeFindDone, timeEdgeDone;
GetSystemTime(&timeStart);
#endif // OFFLINE
    if (sensorValueDebugMsg) {
        cout << s0 << " " << s1 << " " << s2 << endl;
    }
    
    sensor0 = s0;
    sensor1 = s1;
    sensor2 = s2;
    infaredDistance = (int)r2o;    // stands for range to obstacle

    Reset ();
    
    TrigInit (s0, s1, s2);

#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "effective roll " << RAD2DEG(effective_roll) << endl;
#endif
    yplane = yuv;
#else // OFFLINE
    //save only the latest image
    OFbkImageVectorData *imageVec =
    reinterpret_cast <OFbkImageVectorData*>(event.RCData (event.NumOfData () - 1)->Base ());

    cameraFrame = imageVec->GetInfo (0)->frameNumber;

    // pointer to start of actual image data
    // medium resolution index=1
    // high resolution index=0
    yplane = imageVec->GetData (RES);
#endif // OFFLINE
    uplane = yplane + WIDTH;
    vplane = uplane + WIDTH;

    // ricky: segmentation and run length encoding done in same run
    //    gcd.replaceCplane (yplane, uplane, vplane, SKIP);
    blobs->replaceCplane(yplane, uplane, vplane, SKIP);

    blobs->FormBlobs();
    
#ifndef OFFLINE
//GetSystemTime(&timeBlobDone);
#endif
    FindVisualObjects();

#ifndef OFFLINE
    //GetSystemTime(&timeFindDone);
#endif

#ifdef EDGE_DETECT
#ifdef ENABLE_EDGE_DETECT
    runEdgeDetect(self);
#endif //ENABLE_EDGE_DETECT
#endif //EDGE_DETECT
#ifndef OFFLINE
//GetSystemTime(&timeEdgeDone);
/*
if (peakBlobTime < (int)((timeBlobDone - timeStart).useconds))
{
  peakBlobTime = (int)((timeBlobDone - timeStart).useconds);
}
if (peakFindTime < (int)((timeFindDone - timeBlobDone).useconds))
{
  peakFindTime = (int)((timeFindDone - timeBlobDone).useconds);
}
if (peakEdgeTime < (int)((timeEdgeDone - timeFindDone).useconds))
{
  peakEdgeTime = (int)((timeEdgeDone - timeFindDone).useconds);
}
if (peakTotalTime < (int)((timeEdgeDone - timeStart).useconds))
{
  peakTotalTime = (int)((timeEdgeDone - timeStart).useconds);
}

if (timeReadingCounter > 25)
{
  timeReadingCounter = 0;
  if (VISUALCORTEX_TIME_PRINT)
  {
    cout << "peakBlobTime=" << peakBlobTime << " peakFindTime=" << peakFindTime << " peakEdgeTime=" << peakEdgeTime << " peakHorizonT=" << peakHorizonT << " peakFindT=" << peakFindT << " peakProjectT=" << peakProjectT << " peakNightT=" << peakNightT << " peak total vision time=" << peakTotalTime << " number of frames aborted=" << abortCount << " Percent aborted:" << ((double)abortCount/25.0)*100.0 << endl;
  }
  lastAbortCount = abortCount;
  abortCount = 0;
  peakBlobTime = 0;
  peakFindTime = 0;
  peakEdgeTime = 0;
  peakTotalTime = 0;

  peakHorizonT = 0;
  peakFindT = 0;
  peakProjectT = 0;
  peakNightT = 0;
}
timeReadingCounter ++;
*/
#endif
}


/* Find the blue and yellow goals.  As each goal can be composed of 
 * separate patches, there is sanity
 * checking on each patch to determine whether it is a valid "goal" patch.
**/
void VisualCortex::FindGoal () {
    blobs->rotate (cBEACON_BLUE, sin_eroll, cos_eroll);
    blobs->rotate (cBEACON_YELLOW, sin_eroll, cos_eroll);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
	        (*outfile) << "look for blue goal" << endl;
#endif
#endif	    
    LookForGoal (hbBlue, numBlue, vob[vobBlueLeftBeacon],
           vob[vobBlueRightBeacon], vob[vobBlueGoal]);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
	        (*outfile) << "look for yellow goal" << endl;
#endif
#endif	    
    LookForGoal (hbYellow, numYellow, vob[vobYellowRightBeacon],
           vob[vobYellowLeftBeacon], vob[vobYellowGoal]);
}

inline void
VisualCortex::LandmarkSanities () {
    if (vob[vobYellowGoal].cf > 0 && vob[vobBlueRightBeacon].cf > 0) {
        vob[vobBlueRightBeacon].cf = 0;
        vob[vobBlueRightBeacon].var = LARGE_VAL;
        vob[vobBlueRightBeacon].angleVar = LARGE_VAL;
        fireSanity(YELLOW_GOAL_NOT_BLUE_RIGHT_BEACON);
    }

    if (vob[vobYellowGoal].cf > 0 && vob[vobBlueLeftBeacon].cf > 0) {
        vob[vobBlueLeftBeacon].cf = 0;
        vob[vobBlueLeftBeacon].var = LARGE_VAL;
        vob[vobBlueLeftBeacon].angleVar = LARGE_VAL;
        fireSanity(YELLOW_GOAL_NOT_BLUE_LEFT_BEACON);
    }

    if (vob[vobBlueGoal].cf > 0 && vob[vobYellowRightBeacon].cf > 0) {
        vob[vobYellowRightBeacon].cf = 0;
        vob[vobYellowRightBeacon].var = LARGE_VAL;
        vob[vobYellowRightBeacon].angleVar = LARGE_VAL;
        fireSanity(BLUE_GOAL_NOT_YELLOW_RIGHT_BEACON);
    }

    if (vob[vobBlueGoal].cf > 0 && vob[vobYellowLeftBeacon].cf > 0) {
        vob[vobYellowLeftBeacon].cf = 0;
        vob[vobYellowLeftBeacon].var = LARGE_VAL;
        vob[vobYellowLeftBeacon].angleVar = LARGE_VAL;
        fireSanity(BLUE_GOAL_NOT_YELLOW_LEFT_BEACON);
    }

    if (vob[vobYellowLeftBeacon].cf > 0 && vob[vobBlueRightBeacon].cf > 0) {
        if (PointToElevation(vob[vobYellowLeftBeacon].cx, vob[vobYellowLeftBeacon].cy,
                             sin_eroll, cos_eroll) > 
            PointToElevation(vob[vobBlueRightBeacon].cx, vob[vobBlueRightBeacon].cy,
                             sin_eroll, cos_eroll)) {
            vob[vobYellowLeftBeacon].cf = 0;
            vob[vobYellowLeftBeacon].var = LARGE_VAL;
            vob[vobYellowLeftBeacon].angleVar = LARGE_VAL;
            fireSanity(BLUE_RIGHT_BEACON_NOT_YELLOW_LEFT_BEACON);
        } else {
            vob[vobBlueRightBeacon].cf = 0;
            vob[vobBlueRightBeacon].var = LARGE_VAL;
            vob[vobBlueRightBeacon].angleVar = LARGE_VAL;
            fireSanity(YELLOW_LEFT_BEACON_NOT_BLUE_RIGHT_BEACON);
        }
    }

    if (vob[vobYellowLeftBeacon].cf > 0 && vob[vobBlueLeftBeacon].cf > 0) {
        if (PointToElevation(vob[vobYellowLeftBeacon].cx, vob[vobYellowLeftBeacon].cy,
                             sin_eroll, cos_eroll) > 
            PointToElevation(vob[vobBlueLeftBeacon].cx, vob[vobBlueLeftBeacon].cy,
                             sin_eroll, cos_eroll)) {
            vob[vobYellowLeftBeacon].cf = 0;
            vob[vobYellowLeftBeacon].var = LARGE_VAL;
            vob[vobYellowLeftBeacon].angleVar = LARGE_VAL;
            fireSanity(BLUE_LEFT_BEACON_NOT_YELLOW_LEFT_BEACON);
        } else {
            vob[vobBlueLeftBeacon].cf = 0;
            vob[vobBlueLeftBeacon].var = LARGE_VAL;
            vob[vobBlueLeftBeacon].angleVar = LARGE_VAL;
            fireSanity(YELLOW_LEFT_BEACON_NOT_BLUE_LEFT_BEACON);
        }
    }

    if (vob[vobYellowRightBeacon].cf > 0 && vob[vobBlueRightBeacon].cf > 0) {
        if (PointToElevation(vob[vobYellowRightBeacon].cx, vob[vobYellowRightBeacon].cy,
                             sin_eroll, cos_eroll) > 
            PointToElevation(vob[vobBlueRightBeacon].cx, vob[vobBlueRightBeacon].cy,
                             sin_eroll, cos_eroll)) {
            vob[vobYellowRightBeacon].cf = 0;
            vob[vobYellowRightBeacon].var = LARGE_VAL;
            vob[vobYellowRightBeacon].angleVar = LARGE_VAL;
            fireSanity(BLUE_LEFT_BEACON_NOT_YELLOW_RIGHT_BEACON);
        } else {
            vob[vobBlueLeftBeacon].cf = 0;
            vob[vobBlueLeftBeacon].var = LARGE_VAL;
            vob[vobBlueLeftBeacon].angleVar = LARGE_VAL;
            fireSanity(YELLOW_RIGHT_BEACON_NOT_BLUE_LEFT_BEACON);
        }
    }

    if (vob[vobYellowRightBeacon].cf > 0 && vob[vobBlueLeftBeacon].cf > 0) {
        if (PointToElevation(vob[vobYellowLeftBeacon].cx, vob[vobYellowLeftBeacon].cy,
                             sin_eroll, cos_eroll) > 
            PointToElevation(vob[vobBlueRightBeacon].cx, vob[vobBlueRightBeacon].cy,
                             sin_eroll, cos_eroll)) {
            vob[vobYellowLeftBeacon].cf = 0;
            vob[vobYellowLeftBeacon].var = LARGE_VAL;
            vob[vobYellowLeftBeacon].angleVar = LARGE_VAL;
            fireSanity(BLUE_RIGHT_BEACON_NOT_YELLOW_RIGHT_BEACON);
        } else {
            vob[vobBlueRightBeacon].cf = 0;
            vob[vobBlueRightBeacon].var = LARGE_VAL;
            vob[vobBlueRightBeacon].angleVar = LARGE_VAL;
            fireSanity(YELLOW_RIGHT_BEACON_NOT_BLUE_RIGHT_BEACON);
        }
    }

    if (vob[vobGreenRightBeacon].cf > 0 && vob[vobGreenLeftBeacon].cf > 0) {
        if (PointToElevation(vob[vobGreenRightBeacon].cx, vob[vobGreenRightBeacon].cy,
                             sin_eroll, cos_eroll) > 
            PointToElevation(vob[vobGreenLeftBeacon].cx, vob[vobGreenLeftBeacon].cy,
                             sin_eroll, cos_eroll)) {
            vob[vobGreenRightBeacon].cf = 0;
            vob[vobGreenRightBeacon].var = LARGE_VAL;
            vob[vobGreenRightBeacon].angleVar = LARGE_VAL;
            fireSanity(GREEN_LEFT_BEACON_NOT_GREEN_RIGHT_BEACON);
        } else {
            vob[vobGreenLeftBeacon].cf = 0;
            vob[vobGreenLeftBeacon].var = LARGE_VAL;
            vob[vobGreenLeftBeacon].angleVar = LARGE_VAL;
            fireSanity(GREEN_RIGHT_BEACON_NOT_GREEN_LEFT_BEACON);
        }
    }
}

inline void VisualCortex::BeaconSanities () {
//from 2002 goalie
/*
#ifdef GOALIE
    int border = 10;
    int far = 300;

    if (vob[vobBlueLeftBeacon].cf!=0) {
        wout << "blueLeftBeacon.dist = " << vob[vobBlueLeftBeacon].dist << "\n";
    }

    if (vob[vobBlueRightBeacon].cf!=0) {
        wout << "blueRightBeacon.dist = " << vob[vobBlueRightBeacon].dist << "\n";
    }

     // If the opposite end beacons are at the edge of the screen, then discard the beacon
    if ((vob[vobBlueLeftBeacon].cx < border
       || vob[vobBlueLeftBeacon].cx > WIDTH - border
       || vob[vobBlueLeftBeacon].cy < border
       || vob[vobBlueLeftBeacon].cy > HEIGHT - border)
       && vob[vobBlueLeftBeacon].dist > far) {
//       wout<<"BlueLeftBeacon Discarded!\n";
        vob[vobBlueLeftBeacon].cf = 0;
        vob[vobBlueLeftBeacon].var = LARGE_VAL;
        vob[vobBlueLeftBeacon].angleVar = LARGE_VAL;
        fireSanity(BLUE_LEFT_BEACON_OUTSIDE_EDGE);
    }

  if ((vob[vobBlueRightBeacon].cx < border
       || vob[vobBlueRightBeacon].cx > WIDTH - border
       || vob[vobBlueRightBeacon].cy < border
       || vob[vobBlueRightBeacon].cy > HEIGHT - border)
      && vob[vobBlueRightBeacon].dist > far) {
        vob[vobBlueRightBeacon].cf = 0;
        vob[vobBlueRightBeacon].var = LARGE_VAL;
        vob[vobBlueRightBeacon].angleVar = LARGE_VAL;
        fireSanity(BLUE_RIGHT_BEACON_OUTSIDE_EDGE);
    }

  if ((vob[vobYellowLeftBeacon].cx < border
       || vob[vobYellowLeftBeacon].cx > WIDTH - border
       || vob[vobYellowLeftBeacon].cy < border
       || vob[vobYellowLeftBeacon].cy > HEIGHT - border)
      && vob[vobYellowLeftBeacon].dist > far) {
        vob[vobYellowLeftBeacon].cf = 0;
        vob[vobYellowLeftBeacon].var = LARGE_VAL;
        vob[vobYellowLeftBeacon].angleVar = LARGE_VAL;
        fireSanity(YELLOW_LEFT_BEACON_OUTSIDE_EDGE);
    }

  if ((vob[vobYellowRightBeacon].cx < border
       || vob[vobYellowRightBeacon].cx > WIDTH - border
       || vob[vobYellowRightBeacon].cy < border
       || vob[vobYellowRightBeacon].cy > HEIGHT - border)
      && vob[vobYellowRightBeacon].dist > far) {
        vob[vobYellowRightBeacon].cf = 0;
        vob[vobYellowRightBeacon].var = LARGE_VAL;
        vob[vobYellowRightBeacon].angleVar = LARGE_VAL;
        fireSanity(YELLOW_RIGHT_BEACON_OUTSIDE_EDGE);
    }
#endif
*/
    for (int i = vobBlueLeftBeacon; i <= vobYellowRightBeacon; i++) {
        if (vob[i].cf > 0) {
/*
            // used for logical checks
            // haven't implemented yet

            int leftBeacon, rightBeacon;
            switch (i) {
                case vobBlueLeftBeacon:
                    leftBeacon = vobGreenLeftBeacon;
                    rightBeacon = vobBlueRightBeacon;
                    break;
                case vobBlueRightBeacon:
                    leftBeacon = vobBlueLeftBeacon;
                    rightBeacon = vobGreenRightBeacon;
                    break;
                case vobGreenLeftBeacon:
                    leftBeacon = vobYellowLeftBeacon;
                    rightBeacon = vobBlueLeftBeacon;
                    break;
                case vobGreenRightBeacon:
                    leftBeacon = vobBlueRightBeacon;
                    rightBeacon = vobYellowRightBeacon;
                    break;
                case vobYellowLeftBeacon:
                    leftBeacon = vobGreenLeftBeacon;
                    rightBeacon = vobYellowRightBeacon;
                    break;
                case vobYellowRightBeacon:
                    leftBeacon = vobYellowLeftBeacon;
                    rightBeacon = vobGreenRightBeacon;
                    break;
            }
*/

#ifdef OFFLINE
#ifdef PRE_SANITY_BLOB
            if (sanityBlobDebugMsg) {
                cout << "PRE_SANITY_BLOB: " << i << " " << vob[i].x << " " << vob[i].y << " " << vob[i].width << " " << vob[i].height << endl;
            }
            (*blobOutput) << i << " " << vob[i].x << " " << vob[i].y << " " << vob[i].width << " " << vob[i].height << endl;
#endif
#endif
            // if the height of the beacon is greater than 60cm
            // its usually around 30ish but
            // the error increases as the dist gets larger.
            // 45 is the maximum observed from
            // testing with a walking dog looking at the furtherest
            // beacon
            if (tan (DEG2RAD (vob[i].elev)) * vob[i].dist > 50.0 && vob[i].cf < 300) {
                vob[i].cf = 0;
                vob[i].var = LARGE_VAL;
                vob[i].angleVar = LARGE_VAL;
                fireSanity(BEACON_HEIGHT_TOO_HIGH+i-vobBlueLeftBeacon);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "beacon sanity, beacon too high " << i;
                (*outfile) << " elev " << vob[i].elev << " dist " << vob[i].dist;
                (*outfile) << " height " << tan (DEG2RAD (vob[i].elev)) * vob[i].dist << " cf " << vob[i].cf << endl;
#endif
#endif
            } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "beacon height sanity passed " << i;
                (*outfile) << " elev " << vob[i].elev << " dist " << vob[i].dist;
                (*outfile) << " height " << tan (DEG2RAD (vob[i].elev)) * vob[i].dist << " cf " << vob[i].cf << endl;
#endif
#endif
            }
            
            // can't have negative elevation
            if (vob[i].elev < -15.0) {	 
                vob[i].cf = 0;
                vob[i].var = LARGE_VAL;
                vob[i].angleVar = LARGE_VAL;
                fireSanity(NEGATIVE_BEACON_ELEVATION+i-vobBlueLeftBeacon);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "beacon sanity, beacon " << i << " too low " << vob[i].elev << endl;
#endif
#endif
            }
        }
    }
  // can never see both beacons from opposite sides of the field
}

inline void VisualCortex::GoalSanities () {
// this whole goal too low thing is doing more harm than good
// will look into it a bit more
/*
    if (vob[vobYellowGoal].cf && vob[vobYellowGoal].elev < -10.0
      && vob[vobYellowGoal].area < 1584) {
        vob[vobYellowGoal].cf = 0;
        vob[vobYellowGoal].var = LARGE_VAL;
        vob[vobYellowGoal].angleVar = LARGE_VAL;
        fireSanity(YELLOW_GOAL_TOO_LOW_AREA_1584);
    }

    if (vob[vobBlueGoal].cf && vob[vobBlueGoal].elev < -10.0
      && vob[vobBlueGoal].area < 1584) {
        vob[vobBlueGoal].cf = 0;
        vob[vobBlueGoal].var = LARGE_VAL;
        vob[vobBlueGoal].angleVar = LARGE_VAL;
        fireSanity(BLUE_GOAL_TOO_LOW_AREA_1584);
    }
*/

#ifdef OFFLINE
#ifdef PRE_SANITY_BLOB
    if (vob[vobYellowGoal].cf) {
        if (sanityBlobDebugMsg) {
            cout << "PRE_SANITY_BLOB: " << vobYellowGoal << " " << vob[vobYellowGoal].x << " " << vob[vobYellowGoal].y << " " << vob[vobYellowGoal].width << " " << vob[vobYellowGoal].height << endl;
        }
        (*blobOutput) << vobYellowGoal << " " << vob[vobYellowGoal].x << " " << vob[vobYellowGoal].y << " " << vob[vobYellowGoal].width << " " << vob[vobYellowGoal].height << endl;
    }
    if (vob[vobBlueGoal].cf) {
        if (sanityBlobDebugMsg) {
            cout << "PRE_SANITY_BLOB: " << vobBlueGoal << " " << vob[vobBlueGoal].x << " " << vob[vobBlueGoal].y << " " << vob[vobBlueGoal].width << " " << vob[vobBlueGoal].height << endl;
        }
        (*blobOutput) << vobBlueGoal << " " << vob[vobBlueGoal].x << " " << vob[vobBlueGoal].y << " " << vob[vobBlueGoal].width << " " << vob[vobBlueGoal].height << endl;
    }
#endif
#endif

    if (vob[vobYellowGoal].cf && vob[vobYellowGoal].width < WIDTH - 5 && vob[vobYellowGoal].elev > 25.0) {
        vob[vobYellowGoal].cf = 0;
        vob[vobYellowGoal].var = LARGE_VAL;
        vob[vobYellowGoal].angleVar = LARGE_VAL;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "yellow goal too high, sanitied out" << endl;
#endif
#endif
        fireSanity(YELLOW_GOAL_TOO_HIGH);
    }

    if (vob[vobBlueGoal].cf && vob[vobBlueGoal].width < WIDTH - 5 && vob[vobBlueGoal].elev > 25.0) {
        vob[vobBlueGoal].cf = 0;
        vob[vobBlueGoal].var = LARGE_VAL;
        vob[vobBlueGoal].angleVar = LARGE_VAL;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "blue goal too high, sanitied out" << endl;
#endif
#endif
        fireSanity(BLUE_GOAL_TOO_HIGH);
    }

/*
    if (vob[vobYellowGoal].cf && vob[vobYellowGoal].elev < 0.0)
    {
        vob[vobYellowGoal].cf = 0;
        vob[vobYellowGoal].var = LARGE_VAL;
        vob[vobYellowGoal].angleVar = LARGE_VAL;
        fireSanity(YELLOW_GOAL_TOO_LOW);
    }

    if (vob[vobBlueGoal].cf && vob[vobBlueGoal].elev < 0.0)
    {
        vob[vobBlueGoal].cf = 0;
        vob[vobBlueGoal].var = LARGE_VAL;
        vob[vobBlueGoal].angleVar = LARGE_VAL;
        fireSanity(BLUE_GOAL_TOO_LOW);
    } 
*/

    // we can't see any yellow beacons and the blue goal at the same time
    if ((vob[vobYellowLeftBeacon].cf > 200
       || vob[vobYellowRightBeacon].cf > 200) && vob[vobBlueGoal].cf) {
        //if ((vob[vobYellowLeftBeacon].var < get95CF(40) || vob[vobYellowRightBeacon].var < get95CF(40)) && vob[vobBlueGoal].var < LARGE_VAL) {
        //wout << "sanity 1\n";
        vob[vobBlueGoal].cf = 0;
        vob[vobBlueGoal].var = LARGE_VAL;
        vob[vobBlueGoal].angleVar = LARGE_VAL;
        fireSanity(YELLOW_BEACON_NOT_BLUE_GOAL);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "yellow beacon and blue goal same time, sanitied out" << endl;
#endif
#endif
    }
    // we can't see any blue beacons and the yellow goal at the same time
    if ((vob[vobBlueLeftBeacon].cf > 200
       || vob[vobBlueRightBeacon].cf > 200) && vob[vobYellowGoal].cf) {
        //if ((vob[vobBlueLeftBeacon].var < get95CF(40) || vob[vobBlueRightBeacon].var < get95CF(40)) && vob[vobYellowGoal].cf < LARGE_VAL) {
        //wout << "sanity 2\n";
        vob[vobYellowGoal].cf = 0;
        vob[vobYellowGoal].var = LARGE_VAL;
        vob[vobYellowGoal].angleVar = LARGE_VAL;
        fireSanity(BLUE_BEACON_NOT_YELLOW_GOAL);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "blue beacon and yellow goal same time, sanitied out" << endl;
#endif
#endif
    }

/*

    // if the goal is too far away, forget it
    if (vob[vobBlueGoal].cf && vob[vobBlueGoal].dist > (FIELD_LENGTH * 1.2)) {
        //if (vob[vobBlueGoal].var < LARGE_VAL && vob[vobBlueGoal].dist > (FIELD_LENGTH * 1.2)) {
        //wout << "sanity 3\n";
        vob[vobBlueGoal].cf = 0;
        vob[vobBlueGoal].var = LARGE_VAL;
        fireSanity(BLUE_GOAL_TOO_FAR);
    }

    // if the goal is too far away, forget it
    if (vob[vobYellowGoal].cf && vob[vobYellowGoal].dist > (FIELD_LENGTH * 1.2)) {
        //if (vob[vobYellowGoal].var < LARGE_VAL && vob[vobYellowGoal].dist > (FIELD_LENGTH * 1.2)) {
        //wout << "sanity 4\n";
        vob[vobYellowGoal].cf = 0;
        vob[vobYellowGoal].var = LARGE_VAL;
        fireSanity(YELLOW_GOAL_TOO_FAR);
    }
*/
    // we can't see both goals at the same time
    if (vob[vobBlueGoal].cf && vob[vobYellowGoal].cf) {
        //if (vob[vobBlueGoal].var < LARGE_VAL && vob[vobYellowGoal].var < LARGE_VAL) {
        //wout << "sanity 5\n";
        if (vob[vobBlueGoal].cf >= vob[vobYellowGoal].cf) {
            vob[vobYellowGoal].cf = 0;
            fireSanity(BLUE_GOAL_CONFIDENT_THAN_YELLOW_GOAL);
        } else {
            vob[vobBlueGoal].cf = 0;
            fireSanity(YELLOW_GOAL_CONFIDENT_THAN_BLUE_GOAL);
        }

        if (vob[vobBlueGoal].var <= vob[vobYellowGoal].var) {
            vob[vobYellowGoal].var = LARGE_VAL;
            vob[vobYellowGoal].angleVar = LARGE_VAL;
            fireSanity(BLUE_GOAL_VARIANCE_LESS_THAN_YELLOW_GOAL);
        } else {
            vob[vobBlueGoal].var = LARGE_VAL;
            vob[vobBlueGoal].angleVar = LARGE_VAL;
            fireSanity(YELLOW_GOAL_VAIRANCE_LESS_THAN_BLUE_GOAL);
        }
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "both goal seen, sanitied out" << endl;
#endif
#endif
    }
}

/*
 * Copies all the beacon information from beacon[] into vob[].
 */
void
VisualCortex::FindBeacon () {
  // Assign beacons
    VisualObject *cv = &vob[vobBlueLeftBeacon];
    for (int i = 0; i < 6; i++, cv++) {

        if (!beacon[i].valid) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Invalid beacon " << i << endl;
#endif
#endif
            cv->cf = 0;
            cv->var = LARGE_VAL;
            fireSanity(INVALID_BEACON+i);
            continue;
        } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Valid beacon " << i << endl;
#endif
#endif
        }
		//Will: XXX: Um, doesn't this code mean that the bounding box
		// is not really the bounding box?

        cv->x = beacon[i].cx - beacon[i].width / 2.0;
        cv->y = beacon[i].cy - beacon[i].height / 2.0;
        cv->width = beacon[i].width;
        cv->height = beacon[i].height;

        cv->cx = beacon[i].cx;
        cv->cy = beacon[i].cy;

        cv->projected_x = cv->x;
        cv->projected_y = cv->y;
        cv->projected_w = cv->width;
        cv->projected_h = cv->height;

        cv->area = beacon[i].area;

        //cv->tilt = beacon[i].angle*400/360; not sure why 400/360
        cv->tilt = beacon[i].angle;
        cv->imgHead = PointToHeading (beacon[i].cx);
        cv->imgElev = PointToElevation (beacon[i].cy);

#ifdef HIGH
        //calculate the distance based on centre 2 centre of beacon
        //calibrated by martin and tom (curve is hyperbola)
        //plus dist_neck2snout to make distance relative to neck

/*
        // Removed using area calulation for Osaka game
        if (cv->area < 100) {
            int vobIndex = i + vobBlueLeftBeacon;

            if (vobIndex == vobBlueLeftBeacon || vobIndex == vobBlueRightBeacon) {
                // use area instead of c2c.
                // constants are compute using
                // Microsoft Excel tool ==> power series trendline
                cv->dist = 1369.3 * pow ((double)cv->area, -0.4202);
            } else if (vobIndex == vobYellowLeftBeacon || vobIndex == vobYellowRightBeacon) {
                cv->dist = 1773.8 * pow ((double)cv->area, -0.4739);
            } else {
                cv->dist = 1700.0 / beacon[i].c2c;
            }
        } else
*/

		#ifdef OLD_C2C
				cv->cam_dist = cv->dist = beaconDistanceConstant / beacon[i].c2c;
		#endif

		#ifdef NEWCENTROID_C2C
        		cv->cam_dist = cv->dist = beaconDistanceConstant / beacon[i].c2c2;
		#endif

		#ifdef NEWCENTROID_BEACONEXT_C2C
        		cv->cam_dist = cv->dist = beaconDistanceConstant / beacon[i].c2c3;
		#endif

		cv->dist1 = beaconDistanceConstant / beacon[i].c2c;
		cv->dist2 = beaconDistanceConstant / beacon[i].c2c2;
		cv->dist3 = beaconDistanceConstant / beacon[i].c2c3;				


		//ERIC these printouts are all for centroid testing
		cv->c2c = beacon[i].c2c;
		cv->c2c2 = beacon[i].c2c2;
		cv->c2c3 = beacon[i].c2c3;
		cv->pinkXsum = beacon[i].pinkXsum;
		cv->pinkYsum = beacon[i].pinkYsum;
#include <cstdio>
		cv->pinkarea = beacon[i].pinkarea;
		cv->closestXsum = beacon[i].closestXsum;
		cv->closestYsum = beacon[i].closestYsum;
		cv->closestarea = beacon[i].closestarea;

	
        // the max distance of a  (snout) on the field
        // to any beacon is 320cm. This reduces the error of
        // overestimate the distance to yellow left/right beacon
        cv->cam_dist = cv->dist = MIN (cv->dist, FIELD_DIAGONAL);

#endif

#ifdef MED
        cv->cam_dist = cv->dist = SizeToDistance (20, cv->height);

        // 20.5 rounded DOWN to counteract flattening of aspect ratio in SizeToDistance
        // calculate distance to beacons based on number of pixel seen.
        double threshold = 7.0 * 130.0 / sqrt (double (beacon[i].halfmass));
        if (cv->dist - 50 > threshold) {
            cv->cam_dist = cv->dist = threshold;
        }
#endif

        int halfExpectedArea = (int)SQUARE (beacon[i].c2c);
        if (halfExpectedArea) {
            cv->misc = beacon[i].c2c;
            cv->cf = MIN ((1000 * beacon[i].area) / (halfExpectedArea * 2), 1000);

            // Add the variance for beacon.  This is determined from an equation that
            // was generated by using linear regression in MiniTab.
            //cv->var = exp(2 * BEACON_C + 2 * BEACON_M * cv->dist)/
            //          pow(beacon[i].area/(double)(halfExpectedArea*2),errorScale);
            cv->var = exp (2 * BEACON_C + 2 * BEACON_M * cv->dist) +
                      SQUARE (varConst * (1 - beacon[i].area / (double) (halfExpectedArea * 2)));

            //ERIC DISTANCE CALIBRATION
			cv->var1 = exp (2 * BEACON_C + 2 * BEACON_M * cv->dist1) +
                      SQUARE (varConst * (1 - beacon[i].area / (double) (halfExpectedArea * 2)));
            cv->var2 = exp (2 * BEACON_C + 2 * BEACON_M * cv->dist2) +
                      SQUARE (varConst * (1 - beacon[i].area / (double) (halfExpectedArea * 2)));
            cv->var3 = exp (2 * BEACON_C + 2 * BEACON_M * cv->dist3) +
                      SQUARE (varConst * (1 - beacon[i].area / (double) (halfExpectedArea * 2)));




            // this is the variance of the beacon angle
            cv->angleVar = 25 / (beacon[i].area / (double) (halfExpectedArea * 2));

        } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "beacon sanity: halfExpectedArea too small\n";
#endif
#endif
            cv->cf = 0;
            cv->var = LARGE_VAL;
            cv->angleVar = LARGE_VAL;
            fireSanity(BEACON_EXPECTED_AREA_TOO_SMALL + i);
        }

    }
}

/*
 *  Using the blobs generated by Process(), identify visual objects.
 *
 *  @return     number of visual objects identified.
 */
int VisualCortex::FindVisualObjects () {
    // locate land marks find all patches which is possibly beacons or goals.
    // it will store the number of blue, yellow green pink blobs.
    LocateLandmarks ();

    // Copy all the beacon information from beacon[] into vob[].
    FindBeacon ();

    for (int i = vobBlueLeftBeacon; i <= vobYellowRightBeacon; i++) {
        if (vob[i].cf || vob[i].var != LARGE_VAL) {
            RobotRelative (vob[i]);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "beacon " << i << " still in, with elev " << vob[i].elev << endl;
#endif
#endif
	    }
    }


    BeaconSanities ();

    // Find the blue and yellow goals.
    FindGoal ();

    //wout << "==> goal dist, var: " << vob[vobBlueGoal].dist << ", " << vob[vobBlueGoal].var << "\n";

    // convert all objects robot relative except ball and robot.
    // The ball's and robot's position get transformed in FindBall() and FindRobot() respectively

    // Some extra sanity checking for the goal.
    GoalSanities ();

    //Logical Landmark Sanities

    LandmarkSanities ();

#ifndef SHAPE_CHALL
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "started find ball" << endl;
#endif
#endif
    FindBall ();
    FindRobot ();
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "done all finding" << endl;
#endif
#endif

#endif
    // << "red  h,d,var: " << vob[vobRedDog].head << ", " << vob[vobRedDog].dist << ", " << vob[vobRedDog].var << "\n";
    //wout << "blue h,d,var: " << vob[vobBlueDog].head << ", " << vob[vobBlueDog].dist << ", " << vob[vobBlueDog].var << "\n";
    //wout << "tilt, pan, roll: " << tilt << ", " << pan << ", " << roll << "\n";

    //wout << "==2> goal dist, var: " << vob[vobBlueGoal].dist << ", " << vob[vobBlueGoal].var << "\n";

    /* Now return the number of vobs we found */
    int n = 0;
    for (int i = 0; i < VOB_COUNT; i++) {
        if (vob[i].cf || vob[i].var != LARGE_VAL) {
            n++;
            //cout << "vob" << i << " " << vob[i].var << "\n";
        }
    }

    return n;
}

// return if b1 is completely inside b2
// return if b1 is completely inside b2
// offset grows b2
int VisualCortex::VobInVob (VisualObject & b1, VisualObject & b2, int ot,
                        int or_, int ob, int ol) {
    return (b1.x >= b2.x - ol &&
            b1.y >= b2.y - ot &&
            (b1.x + b1.width) <= (b2.x + b2.width + or_) &&
            (b1.y + b1.height) <= (b2.y + b2.height + ob));
}

#ifdef FIREBALL

// i'm putting them here coz it's faster to change and compile
// until i've settled down with a good set of values
// remind me
static const int NECK_HEIGHT    = 7;
static const int FRAME          = 5;
static const int MAXPOINTS      = 8;
static const int TILT_TRIGGER   = 25;
static const int INFINITE       = 10000;
static const int COVERSCREEN    = 22000;
static const int AREA_TRIGGER   = 250;  // if less than this not really big enough to be a ball
                                        // within the frame box thingie anyways.. well it is but not really

// New project point transformation, with Bernhard's revised formula
void VisualCortex::projectPointsDAlt(int * inPoints, int numPoints, double * outPoints, double projectHeight) {
    double hs = 7;
    double h = NECKBASE_HEIGHT;
    double b = 11;
    double s = 11.9;
    double T = (tilt - TILT_OFFSET + evilTiltOffset);
    double P = -pan;
    double c = FACE_LENGTH;
    double f = (double)APERTURE_DISTANCE;
    double n = NECK_LENGTH;
    double p = (double)projectHeight;
    for (int i = 0; i < numPoints; i ++) {
        double A, R, xc, yc, zc, xi, yi, zi, k, x, y, u, w;

        w = (double)(inPoints[i*2+0]);
        u = (double)(inPoints[i*2+1]);
        // Effective tilt = body tilt + head tilt
        A = asin((b-hs)/s);
        R = A+T;

        // Coordiates of camera point(c)
        xc = c*sin(P);
        yc = n*sin(R)+c*cos(R)*cos(P);
        zc = h-p-c*sin(R)*cos(P)+n*cos(R);

        // Coordinates of image point (i) to be projected
        xi = w*cos(P)+(c+f)*sin(P);
        yi = -w*sin(P)*cos(R)+(c+f)*cos(R)*cos(P)+(n+u)*sin(R);
        zi = h-p+w*sin(P)*sin(R)-(c+f)*sin(R)*cos(P)+(n+u)*cos(R);
//Raymond: Added: Check to make sure that zi-zc is greater than 0 (actually check 
//against 1E-5 just to make sure the FPU doesn't stuff up). If this is the 
//case then the points exist over the horizon so set the outpoints at 
//that point to equal infinity. 
        if (zi-zc > -(1E-5))
        {
          outPoints[i*2+0] = LARGE_VAL;
          outPoints[i*2+1] = LARGE_VAL;
        }
        else
        {

          // Ground plane z=0 intersection of line going through points c and i
          k = zi/(zi-zc);
          x = xi+k*(xc-xi);
          y = yi+k*(yc-yi);


          outPoints[i*2+0] = x;
          outPoints[i*2+1] = y;
      }
    }
}

                                 
// Raymond's project point transformation
void VisualCortex::projectPointsD(int * inPoints, int numPoints, double * outPoints, double projectHeight)
{//{{{
  //Find co-ordinates of the end points. 
  double **endVectors, *origin, *newOrigin, **newEndVectors;
  double  *fieldPoints;
  //Matrices.
  //single letters are individual transform matrix elements. 
  double * xfrmMatrix, a, b, c, d, e, f, g, h, ii, jj, kk, u;// *camToPan, *removePan, *panToTilt, *removeTilt, *tiltToGround;
  int i,j, k;
  endVectors = (double **)malloc(numPoints*sizeof(double*));
  newEndVectors = (double**)malloc(numPoints * sizeof(double*));
  fieldPoints = (double*)malloc(numPoints * 2 * sizeof(double));  //array of numpoints times numbers per point (2)
  for (i = 0; i < numPoints; i ++)
  {//{{{
    endVectors[i] = (double*)malloc(4*sizeof(double));  //4 for homogenous vectors
    //Allocate the last dimension of newEndVectors later 
    //(when applying the matrix)
  }//}}}
  origin = (double*)malloc(4*sizeof(double));  //location of the origin. 
//  newOrigin = (double*)malloc(4*sizeof(double));  //I don't think we need this as the matrix apply function should allocate this for us later. 
  for (i = 0; i < numPoints; i ++)
  {//{{{
    endVectors[i][0] = inPoints[i*2+0];
    endVectors[i][1] = inPoints[i*2+1];
    endVectors[i][2] = -(int) APERTURE_DISTANCE;  
    endVectors[i][3] = (int) 0;
  }//}}}


  //We now have the endpoints of the Hough lines in endVectors[i*4+k]. Form 
  //rays that go through the camera aperture. 
  origin[0] = 0;
  origin[1] = 0;
  origin[2] = 0;
  origin[3] = 1;

  //Initialize transformation matrices.
  a = -FACE_LENGTH;
  b = cosRAD(pan);
  c = sinRAD(pan);
  d = -c;
  e = b;
  f = NECK_LENGTH;
  g = cosRAD((-(tilt+evilTiltOffset)));
  h = -sinRAD((-(tilt+evilTiltOffset)));
  ii = -h;
  jj = g;
  kk = NECKBASE_HEIGHT - projectHeight;
  xfrmMatrix = createMatrix4x4(  b,    0,    c,    c*a,
                                d*h,  g,    h*e,  a*h*e+f*g+kk,
                                d*jj,  ii,    e*jj,  a*e*jj+ii*f,
                                0,    0,    0,    1);
  newOrigin = applyMatrix(xfrmMatrix, origin);
  for (i = 0; i < numPoints; i ++)
  {//{{{
    if (inPoints[i*2+0] != -1 || projectHeight != 0)
    {//{{{
      newEndVectors[i] = applyMatrix(xfrmMatrix, endVectors[i]);
    }//}}}
  }//}}}
  //Now we have the rays through the lens which start at the point newOrigin and 
  //extend in the direction of each newEndVectors element by an amount u.
  //Now intersect it with the plane y=0. 
  for (i = 0; i < numPoints; i ++)
  {//{{{
      //NOTE: These vectors are STILL in the co-ordinate system centred about the spot 
      //on the ground under the robot's neck where X is to the right, Y is straight up 
      //and Z is pointing towards the BACK of the robot, where we have [X][Y][Z].
      if (inPoints[i*2+0] != -1 || projectHeight != 0)
      {//{{{
        u = newOrigin[1]/(newOrigin[1] - newEndVectors[i][1]);
//EVIL HACK:
if (u < 0)
{//{{{
//  u = -u;
}//}}}
      fieldPoints[i*2+0] = (newOrigin[0] + u*(newEndVectors[i][0] - newOrigin[0]));
      fieldPoints[i*2+1] = -(newOrigin[2] + u*(newEndVectors[i][2] - newOrigin[2]));

      }//}}}
      else
      {//{{{
        //Line was deleted ...
        fieldPoints[i*2+0] = -1;
        fieldPoints[i*2+1] = -1; 
      }//}}}
  }//}}}
  //For now, we just copy this straight to the out array (for debugging!). 
  memcpy(outPoints, fieldPoints, numPoints*2*sizeof(double));
  free(newOrigin);
  free(origin);
  free(fieldPoints);
  for (i = 0; i < numPoints; i ++)
  {//{{{
    free(endVectors[i]);
    if (inPoints[i*2+0] != -1 || projectHeight != 0)
    {
      //Remember, we didn't allocate this if inPoints was -1.
      free(newEndVectors[i]);
    }
  }//}}}
  free(endVectors);
  free(newEndVectors);
}//}}}

void VisualCortex::equationEdge(int start, int end, double m, double b, bool reverse, int* intersect_x, int* intersect_y, int edgeThreshold) {
    const int INFINITE = 10000;
    int c;
    if (reverse) {
        for (int i = start; i >= end; i--) {
            if (m >= INFINITE) {
                int c = cplane[(int)b + i * WIDTH];
                if (c == BALL || c == RED_DOG) {
                    int j = i - 1;
                    for (; j >= 0; j--) {
                        int nc = cplane[(int)b + j * WIDTH];
                        if (nc != BALL && nc != RED_DOG) break;
                    }
                    if (j <= 0 || j <= i - edgeThreshold) {
                        *intersect_x = (int)b;
                        *intersect_y = i;
                        return;
                    }
                }                
            } else if (abs(m) <= 1) {
                int c = cplane[i + (int)(i * m + b) * WIDTH];
                if (c == BALL || c == RED_DOG) {
                    int j = i - 1;
                    for (; j >= 0; j--) {
                        int nc = cplane[j + (int)(j * m + b) * WIDTH];
                        if (nc != BALL && nc != RED_DOG) break;
                    }
                    if (j <= 0 || j <= i - edgeThreshold) {
                        *intersect_x = i;
                        *intersect_y = (int)(i * m + b);
                        return;
                    }
                }
            } else {
                int c = cplane[(int)((i - b) / m) + i * WIDTH];
                if (c == BALL || c == RED_DOG) {
                    int j = i - 1;
                    for (; j >= 0; j--) {
                        int nc = cplane[(int)((j - b) / m) + j * WIDTH];
                        if (nc != BALL && nc != RED_DOG) break;
                    }
                    if (j <= 0 || j <= i - edgeThreshold) {
                        *intersect_x = (int)((i - b) / m);
                        *intersect_y = i;
                        return;
                    }
                }
            }
        }
    } else {
        for (int i = start; i <= end; i++) {
            if (m >= INFINITE) {
                int c = cplane[(int)b + i * WIDTH];
                if (c == BALL || c == RED_DOG) {
                    int j = i + 1;
                    for (; j < HEIGHT; j++) {
                        int nc = cplane[(int)b + j * WIDTH];
                        if (nc != BALL && nc != RED_DOG) break;
                    }
                    if (j >= HEIGHT || j >= i + edgeThreshold) {
                        *intersect_x = (int)b;
                        *intersect_y = i;
                        return;
                    }
                }                
            } else if (abs(m) <= 1) {
                int c = cplane[i + (int)(i * m + b) * WIDTH];
                if (c == BALL || c == RED_DOG) {
                    int j = i + 1;
                    for (; j < WIDTH; j++) {
                        int nc = cplane[j + (int)(j * m + b) * WIDTH];
                        if (nc != BALL && nc != RED_DOG) break;
                    }
                    if (j >= WIDTH || j >= i + edgeThreshold) {
                        *intersect_x = i;
                        *intersect_y = (int)(i * m + b);
                        return;
                    }
                }
            } else {
                int c = cplane[(int)((i - b) / m) + i * WIDTH];
                if (c == BALL || c == RED_DOG) {
                    int j = i + 1;
                    for (; j < HEIGHT; j++) {
                        int nc = cplane[(int)((j - b) / m) + j * WIDTH];
                        if (nc != BALL && nc != RED_DOG) break;
                    }
                    if (j >= HEIGHT || j >= i + edgeThreshold) {
                        *intersect_x = (int)((i - b) / m);
                        *intersect_y = i;
                        return;
                    }
                }
            }
        }
    }    
    *intersect_x = *intersect_y = -1;
}


void VisualCortex::findEdge(int start, int end, int fixed, bool outside, bool reverse, bool vertical, int* resultx, int* resulty, int edgeThreshold) {
/*
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "start " << start << " end " << end << " fixed " << fixed << " reverse " << reverse << " vertical " << vertical << endl;
#endif
#endif
*/
    if (!reverse) {
        for (int i = start; i <= end; i++) {
            int c = (vertical ? cplane[fixed + i * WIDTH] : cplane[i + fixed * WIDTH]);
            if (outside && (c == BALL || c == RED_DOG)) {
                // see if the subsequent pixels are ball
                int j = i + 1;
                for (; j < (vertical ? HEIGHT : WIDTH); j++) {
                    int nc = (vertical ? cplane[fixed + j * WIDTH] : cplane[j + fixed * WIDTH]);
                    if (nc != BALL && nc != RED_DOG) break;
                }
                if (j == (vertical ? HEIGHT : WIDTH) || j >= i + edgeThreshold) {
                    *resultx = (vertical ? fixed : i);
                    *resulty = (vertical ? i : fixed);
/*
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "found (" << *resultx << "," << *resulty << ")" << endl;
#endif
#endif
*/
                    return;
                }
                i = j;
            } else if (!outside && (c != BALL && c != RED_DOG)) {
                // see if the subsequent pixels are bg
                int j = i + 1;
                for (; j < (vertical ? HEIGHT : WIDTH); j++) {
                    int nc = (vertical ? cplane[fixed + j * WIDTH] : cplane[j + fixed * WIDTH]);
                    if (nc == BALL || nc == RED_DOG) break;
                }
                if (j == (vertical ? HEIGHT : WIDTH) || j >= i + edgeThreshold) {
                    *resultx = (vertical ? fixed : i - 1);
                    *resulty = (vertical ? i - 1: fixed);
/*
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "found (" << *resultx << "," << *resulty << ")" << endl;
#endif
#endif
*/
                    return;
                }
                i = j;
            }
        }
    } else {
        for (int i = start; i >= end; i--) {
            int c = (vertical ? cplane[fixed + i * WIDTH] : cplane[i + fixed * WIDTH]);
            if (outside && (c == BALL || c == RED_DOG)) {
                // see if the subsequent pixels are ball
                int j = i - 1;
                for (; j >= 0; j--) {
                    int nc = (vertical ? cplane[fixed + j * WIDTH] : cplane[j + fixed * WIDTH]);
                    if (nc != BALL && nc != RED_DOG) break;
                }
                if (j <= 0 || j <= i - edgeThreshold) {
                    *resultx = (vertical ? fixed : i);
                    *resulty = (vertical ? i: fixed);
/*
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "found (" << *resultx << "," << *resulty << ")" << endl;
#endif
#endif
*/
                    return;
                }
                i = j;
            } else if (!outside && (c != BALL && c != RED_DOG)) {
                // see if the subsequent pixels are bg
                int j = i - 1;
                for (; j >= 0; j--) {
                    int nc = (vertical ? cplane[fixed + j * WIDTH] : cplane[j + fixed * WIDTH]);
                    if (nc == BALL || nc == RED_DOG) break;
                }
                if (j <= 0 || j <= i - edgeThreshold) {
                    *resultx = (vertical ? fixed : i + 1);
                    *resulty = (vertical ? i + 1: fixed);
/*
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "found (" << *resultx << "," << *resulty << ")" << endl;
#endif
#endif
*/
                    return;
                }
                i = j;
            }
        }
    }
    *resultx = *resulty = -1;
/*
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "found (" << *resultx << "," << *resulty << ")" << endl;
#endif
#endif
*/
}        

// whether a point is outside or inside the circle
// this is a very strict check because it is only applied to 
// pixels outside of the frame anyway, so it basically only
// needs to tell whether we are directly inside a ball or out
// and if directly in, then this should be predominantly orange
// i expect something like.. 20 / 25 pixels orange
bool VisualCortex::isOutside(int px, int py) {
    int count = 0;
    int total = 0;
    for (int i = py - 2; i <= py + 2; i++) {
        for (int j = px - 2; j <= px + 2; j++) {
            int c = cplane[i * WIDTH + j];
            if (c == BALL || c == RED_DOG)
                count++;
            total++;
        }
    }
    return ((count * 1.0) / total) <= 0.8;
}

// counts the number of specified colour points in the rectangle specified
// by minX, minY, maxX and maxY
int VisualCortex::countColouredPixels(int colour, int minX, int minY, int maxX, int maxY) {
    int count = 0;        
    for (int i = minY; i <= maxY; i++) { 
        for (int j = minX; j <= maxX; j++) {
            int c = cplane[i * WIDTH + j];
            if (c == colour)
                count++;
        }    
    }    
    return count;
}    

// find two points on the frame that intersects circumference of ball
void VisualCortex::findPointOnCircumferenceOneAndTwo(int bxmin, int bxmax, int bymin, int bymax, bool nl, bool nr, bool sl, bool sr, int *p1x, int *p1y, int *p2x, int *p2y) {
    int candidate_x[MAXPOINTS], candidate_y[MAXPOINTS];
    // clear the candidates
    for (int i = 0; i < MAXPOINTS; i++) {
        candidate_x[i] = candidate_y[i] = -1;
    }
    int start, end, fixed;
    int edgeThresholdX = max(FRAME - 1, (bxmax - bxmin)/10);
    int edgeThresholdY = max(FRAME - 1, (bymax - bymin)/10);
    //int edgeThresholdX = 4;
    //int edgeThresholdY = 4;

    // south
    start = max(bxmin, FRAME);
    end = min(bxmax, WIDTH - FRAME);
    fixed = HEIGHT - FRAME;
    findEdge(start, end, fixed, sl, false, false, &(candidate_x[0]), &(candidate_y[0]), edgeThresholdX);
    findEdge(end, start, fixed, sr, true, false, &(candidate_x[1]), &(candidate_y[1]), edgeThresholdX);

    // north
    fixed = FRAME;
    findEdge(start, end, fixed, nl, false, false, &(candidate_x[2]), &(candidate_y[2]), edgeThresholdX);
    findEdge(end, start, fixed, nr, true, false, &(candidate_x[3]), &(candidate_y[3]), edgeThresholdX);

    // left
    start = max(bymin, FRAME);
    end = min(bymax, HEIGHT - FRAME);
    fixed = FRAME;
    findEdge(start, end, fixed, nl, false, true, &(candidate_x[4]), &(candidate_y[4]), edgeThresholdY);
    findEdge(end, start, fixed, sl, true, true, &(candidate_x[5]), &(candidate_y[5]), edgeThresholdY);

    // right
    fixed = WIDTH - FRAME;
    findEdge(start, end, fixed, nr, false, true, &(candidate_x[6]), &(candidate_y[6]), edgeThresholdY);
    findEdge(end, start, fixed, sr, true, true, &(candidate_x[7]), &(candidate_y[7]), edgeThresholdY);

    // try to pick the points intelligently
    // can be done more cleverly if i actually order the candidates in clockwise and
    // anti clockwise direction, REMIND ME to come back
    if (nl && nr && sl && sr) {
        for (int i = 0; i < MAXPOINTS; i++) {
            if (candidate_x[i] > 0 && i%2 == 0) {
                *p1x = candidate_x[i];
                *p1y = candidate_y[i];
                *p2x = candidate_x[i+1];
                *p2y = candidate_y[i+1];
                break;
            }
        }
    } else if (nl && nr && sl) {
        *p1x = candidate_x[0]; *p1y = candidate_y[0];
        *p2x = candidate_x[6]; *p2y = candidate_y[6];
    } else if (nl && nr && sr) {
        *p1x = candidate_x[1]; *p1y = candidate_y[1];
        *p2x = candidate_x[4]; *p2y = candidate_y[4];
    } else if (sl && sr && nr) {
        *p1x = candidate_x[3]; *p1y = candidate_y[3];
        *p2x = candidate_x[5]; *p2y = candidate_y[5];
    } else if (sl && sr && nl) {
        *p1x = candidate_x[2]; *p1y = candidate_y[2];
        *p2x = candidate_x[7]; *p2y = candidate_y[7];
    } else if (nl && nr) {
        *p1x = candidate_x[6]; *p1y = candidate_y[6];
        *p2x = candidate_x[4]; *p2y = candidate_y[4];
    } else if (sl && sr) {
        *p1x = candidate_x[5]; *p1y = candidate_y[5];
        *p2x = candidate_x[7]; *p2y = candidate_y[7];
    } else if (nr && sr) {
        *p1x = candidate_x[3]; *p1y = candidate_y[3];
        *p2x = candidate_x[1]; *p2y = candidate_y[1];
    } else if (nl && sl) {
        *p1x = candidate_x[2]; *p1y = candidate_y[2];
        *p2x = candidate_x[0]; *p2y = candidate_y[0];
    } else if (nr) {
        *p1x = candidate_x[3]; *p1y = candidate_y[3];
        *p2x = candidate_x[6]; *p2y = candidate_y[6];
    } else if (nl) {
        *p1x = candidate_x[2]; *p1y = candidate_y[2];
        *p2x = candidate_x[4]; *p2y = candidate_y[4];
    } else if (sr) {
        *p1x = candidate_x[1]; *p1y = candidate_y[1];
        *p2x = candidate_x[7]; *p2y = candidate_y[7];
    } else if (sl) {
        *p1x = candidate_x[5]; *p1y = candidate_y[5];
        *p2x = candidate_x[0]; *p2y = candidate_y[0];
    }
}

void VisualCortex::findPointOnCircumferenceThree(int bxmin, int bxmax, int bymin, int bymax, bool nl, bool nr, bool sl, bool sr, int p1x, int p1y, int p2x, int p2y, int* p3x, int* p3y) {
    // now find definition of perpendicular bisector of p1 and p2
    int p12x, p12y;
    double normal, tangent, b;
    int edgeThresholdX = max(FRAME - 1, (bxmax - bxmin)/10);
    int edgeThresholdY = max(FRAME - 1, (bymax - bymin)/10);
    //int edgeThresholdX = 4;
    //int edgeThresholdY = 4;
    p12x = (p1x + p2x)/2;
    p12y = (p1y + p2y)/2;
    if (p2x != p1x) {
        tangent = ((p2y - p1y) * 1.0) / (p2x - p1x);
        normal = (tangent == 0) ? INFINITE : (-1 / tangent);
    } else {
        tangent = INFINITE;
        normal = 0;
    }            
    if (normal < INFINITE)
        b = p12y - normal * p12x;
    else
        b = p12x;

    // now id the 3rd edge
    int leftCount, rightCount, topCount, bottomCount;
    leftCount = rightCount = topCount = bottomCount = 0;
    if (sl) { leftCount++; bottomCount++; }
    if (nl) { leftCount++; topCount++; }
    if (sr) { rightCount++; bottomCount++; }
    if (nr) { rightCount++; topCount++; }

    if (abs(normal) <= 1) {
        // the case where more x than y
        if (leftCount > rightCount) {
            //cout << "start from left" << endl;
            equationEdge(bxmin, bxmax, normal, b, false, p3x, p3y, edgeThresholdX);
        } else if (rightCount > leftCount) {
            //cout << "start from right" << endl;
            equationEdge(bxmax, bxmin, normal, b, true, p3x, p3y, edgeThresholdX);
        } else {
            // depends which edge closer
            if ((bxmax - (WIDTH - FRAME)) > (FRAME - bxmin)) {
                //cout << "start from left" << endl;
                equationEdge(bxmin, bxmax, normal, b, false, p3x, p3y, edgeThresholdX);
            } else {
                //cout << "start from right" << endl;
                equationEdge(bxmax, bxmin, normal, b, true, p3x, p3y, edgeThresholdX);
            }
        }
    } else {
        // the case where more y than x
        if (topCount > bottomCount) {
            //cout << "start from up" << endl;
            equationEdge(bymin, bymax, normal, b, false, p3x, p3y, edgeThresholdY);
        } else if (bottomCount > topCount) {
            //cout << "start from down" << endl;
            equationEdge(bymax, bymin, normal, b, true, p3x, p3y, edgeThresholdY);
        } else {
            //depends which edge closer
            if ((bymax - (HEIGHT - FRAME)) > (FRAME - bymin)) {
                //cout << "start from up" << endl;
                equationEdge(bymin, bymax, normal, b, false, p3x, p3y, edgeThresholdY);
            } else {
                //cout << "start from bottom" << endl;
                equationEdge(bymax, bymin, normal, b, true, p3x, p3y, edgeThresholdY);
            }
        }
    }
}

// debugging, basically i put stuff in here so if they crash then i know it's them.. 
// dumb way of finding out for sure what crashed.. when i have absolutely no idea
void VisualCortex::testFunction(int p1x, int p1y, double p13m, double p13b, double p23m, double p23b, int *ix, int *iy, double *radius) {}

// find intersection of the perpendicular bisector between 3 points
void VisualCortex::findIntersectionOfTwoChords(int p1x, int p1y, int p2x, int p2y, int p3x, int p3y, int *ix, int *iy, double *radius) {
    // cout << "found points : (" << p1x << "," << p1y << ") (" << p3x << "," << p3y << ") (" << p2x << "," << p2y << ")" << endl;
    // now that we have the three points, try to find the centre by finding the intersection of 
    // perpendicular bisector of the chords p1 p3 and p1 p2
    int p13x, p13y, p23x, p23y;
    double p13m, p13b, p23m, p23b;

    p13x = (p1x + p3x)/2;
    p13y = (p1y + p3y)/2;
    if (p1y == p3y)
        p13m = INFINITE;
    else
        p13m = (p1x == p3x) ? 0 : -1 / (((p3y - p1y) * 1.0) / (p3x - p1x));
    p13b = (p13m == INFINITE) ? p13x : p13y - p13m * p13x;

    p23x = (p2x + p3x)/2;
    p23y = (p2y + p3y)/2;
    if (p2y == p3y)
        p23m = INFINITE;
    else
        p23m = (p2x == p3x) ? 0 : -1 / (((p2y - p3y) * 1.0) / (p2x - p3x));
    p23b = (p23m == INFINITE) ? p23x : p23y - p23m * p23x;

    if (p23m == p13m) {
        //cout << "ERROR !!!!!!!!!!!!!!!!!!! parallel lines found !!!!!!!!!!!!!!!!!!!!!! ERROR" << endl;
        return;
    }
    
    if (p13m == INFINITE) {
        //cout << "p1 and p3 horizontal" << endl;
        *ix = (int)p13b;
        *iy = (int)(*ix * p23m + p23b);
    } else if (p23m == INFINITE) {
        //cout << "p2 and p3 horizontal" << endl;
        *ix = (int)p23b;
        *iy = (int)(*ix * p13m + p13b);
    } else {
        //cout << "normal simultaneous equation" << endl;
        *ix = (int)((p13b - p23b) / (p23m - p13m));
        *iy = (int)(*ix * p13m + p13b);
    }
    // when the ball is close and distorted, it is possible to detect 3 points that ALMOST
    // lie in a straight line. this would then create a VERY large circle
    // and throw off heading.
    // restrict the image centre of ball to within one screen width from the centre
    *ix = (*ix < WIDTH/2) ? MAX(-WIDTH/2, *ix) : MIN(3 * WIDTH / 2, *ix);
    *iy = (*iy < HEIGHT/2) ? MAX(-HEIGHT/2, *iy) : MIN(3 * HEIGHT / 2, *iy);
    double r1 = sqrt((double)((*ix - p1x) * (*ix - p1x) + (*iy - p1y) * (*iy - p1y)));
    double r2 = sqrt((double)((*ix - p2x) * (*ix - p2x) + (*iy - p2y) * (*iy - p2y)));
    double r3 = sqrt((double)((*ix - p3x) * (*ix - p3x) + (*iy - p3y) * (*iy - p3y)));
    *radius = (r1 + r2 + r3) / 3.0;
}

// finds centre of ball
void VisualCortex::findCentre(HalfBeacon *preball, int *fbX, int *fbY, double *fbHead, double *fbDist, double *fbRadius, double *fb_obj0, double *fb_obj1, double *fb_obj2) {
    int p1x, p1y, p2x, p2y, p3x, p3y;
    int bxmin, bxmax, bymin, bymax;
    bool nl, nr, sl, sr;

    double xmin = preball->getXMin ();
    double xmax = preball->getXMax ();
    double ymin = preball->getYMin ();
    double ymax = preball->getYMax ();
    double raw_y = ymin;
                        
    // initialise
    bxmin = (int)xmin;
    bymin = (int)ymin;
    bxmax = (int)xmax;
    bymax = (int)ymax;
    p1x = p2x = p3x = p1y = p2y = p3y = -1;
    nl = isOutside(FRAME, FRAME);
    nr = isOutside(WIDTH - FRAME, FRAME);
    sl = isOutside(FRAME, HEIGHT - FRAME);
    sr = isOutside(WIDTH - FRAME, HEIGHT - FRAME);

#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "try to find 1st & 2nd points on circumference.. " << endl;
#endif
#endif
    findPointOnCircumferenceOneAndTwo(bxmin, bxmax, bymin, bymax, nl, nr, sl, sr, &p1x, &p1y, &p2x, &p2y);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "found points .. (" << p1x << "," << p1y << ") (" << p2x << "," << p2y << ")" << endl;
#endif
#endif

    if (p1x < 0 || p2x < 0) { *fbRadius = -1; return; }
    // else cout << "pt1 : " << p1x << "," << p1y << "  pt2 : " << p2x << "," << p2y << endl;
    
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "try to find 3rd points on circumference.. " << endl;
#endif
#endif
    findPointOnCircumferenceThree(bxmin, bxmax, bymin, bymax, nl, nr, sl, sr, p1x, p1y, p2x, p2y, &p3x, &p3y);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "found points .. (" << p3x << "," << p3y << ")" << endl;
#endif
#endif

    if (p3x < 0) { *fbRadius = -1; return; }
    // else cout << "pt3 : " << p3x << "," << p3y << endl;

    if (p1x < 0 || p2x < 0|| p3x < 0 || ((p1x == p3x) && (p2x == p3x)) || ((p1y == p3y) && (p2y == p3y))) { *fbRadius = -1; return; }

    int ix, iy;
    double radius = -1;

#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "try to find intersection of chords " << endl;
#endif
#endif
    findIntersectionOfTwoChords(p1x, p1y, p2x, p2y, p3x, p3y, &ix, &iy, &radius);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "done intersection of chords " << endl;
#endif
#endif
    if (radius < 0) { *fbRadius = -1; return; }
    
	// if the calculated ball fails to cover the orange blob, then something is wrong
    // it is possible that the orange blob is not covered because part of the orange is noise
    // ie when next to white boundary,
    // but if obviously missing (contain < half of orange blob) then there is error in 
    // detecting edge of ball
    double xdiff = (ix < (bxmax + bxmin) / 2) ? ix + radius - bxmin : bxmax - (ix - radius);
    double ydiff = (iy < (bymax + bymin) / 2) ? iy + radius - bymin : bymax - (iy - radius);
    xdiff = MIN(xdiff, bxmax - bxmin);
    ydiff = MIN(ydiff, bymax - bymin);
    double intersectArea = xdiff * ydiff;
    double blobArea = (bymax - bymin) * (bxmax - bxmin);
	if ((intersectArea / blobArea) < 0.9) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "fireball sanity, blob intersect.. " << intersectArea << "/" << blobArea << " = " << (intersectArea / blobArea) << endl;;
#endif
#endif
        *fbRadius = -1; 
        return;
    } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "fireball sanity passed blob intersect.. " << intersectArea << "/" << blobArea << endl;;
#endif
#endif
    }
    
    // if fireball calculates the ball to be entirely within frame, then it is either unneccesary, or wrong
	if (ix - radius > 0 && ix + radius < WIDTH && iy + radius < HEIGHT && iy - radius > 0) {
        *fbRadius = -1;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "fireball sanity, ball not occluded" << endl;
#endif
#endif
        return;
    }

    // something is wrong if the calculated centre is higher than the top edge of blob
    double centreElev = PointToElevation(ix, iy, sin_eroll, cos_eroll);

    int ox = (int)xmin;
    int oy = (int)ymin;
    int ow = (int)(xmax - xmin);
    int oh = (int)(ymax - ymin);
    
    int xOrigin = WIDTH/2;
    int yOrigin = HEIGHT/2;

    int nX, nY, eX, eY, sX, sY, wX, wY;
    int rnX, rnY, reX, reY, rsX, rsY, rwX, rwY;

    nX = xOrigin - (int)(ox + ow/2);
    nY = yOrigin - (int)(oy);

    rnX = (int) (cos_eroll * nX - sin_eroll * nY);
    rnY = (int) (sin_eroll * nX + cos_eroll * nY);

    sX = xOrigin - (int)(ox + ow/2);
    sY = yOrigin - (int)(oy + oh);

    rsX = (int) (cos_eroll * sX - sin_eroll * sY);
    rsY = (int) (sin_eroll * sX + cos_eroll * sY);

    eX = xOrigin - (int)(ox);
    eY = yOrigin - (int)(oy + oh/2);

    reX = (int) (cos_eroll * eX - sin_eroll * eY);
    reY = (int) (sin_eroll * eX + cos_eroll * eY);

    wX = xOrigin - (int)(ox + ow);
    wY = yOrigin - (int)(oy + oh/2);

    rwX = (int) (cos_eroll * wX - sin_eroll * wY);
    rwY = (int) (sin_eroll * wX + cos_eroll * wY);

    int yRmin = yOrigin - MAX (MAX (MAX (rnY, rsY), reY), rwY);
    double topElev = PointToElevation(yRmin);
     
    if (centreElev > topElev && bymin > 0) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "fireball sanity, centre elev " << centreElev << " > top Elev " << topElev << endl;;
#endif
#endif
        *fbRadius = -1; 
        return;
    }

    int inpoints[2] = {((ix - WIDTH / 2)), (-(iy - HEIGHT / 2))};
    double outpoints[2] = {-1, -1};
    projectPointsDAlt(inpoints, 1, outpoints, BALL_RADIUS);
    *fbDist = sqrt((double)(outpoints[0] * outpoints[0] + outpoints[1] * outpoints[1]));
    if (*fbDist == 0) { *fbRadius = -1; return; }
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "fireball found, image centre.. (" << ix << "," << iy << ")" << endl;;
#endif
#endif
    *fbHead = -1.0*RAD2DEG(asin((1.0 * outpoints[0]) / *fbDist));
    *fbRadius = radius;
    *fbX = ix;
    *fbY = iy;
    *fb_obj0 = outpoints[0];
    *fb_obj2 = outpoints[1];
    *fb_obj1 = NECK_HEIGHT; 
}    

#endif

// this was the method andres developed in 2002, it was intended for locating 
// close balls accurately. It works for moderately close balls, however, it does 
// not work well for balls that are very close.
// 
bool VisualCortex::andresFindBall(HalfBeacon *preball, VisualObject &ball) {
    double xmin = preball->getXMin ();
    double xmax = preball->getXMax ();
    double ymin = preball->getYMin ();
    double ymax = preball->getYMax ();
    double raw_y = ymin;
    Point p[3];
    int i;
    for (i = 0; i < 3; i++) {
        if (!preball->getNextPoint (p[i]))
            break;
    }
    // If we didn't obtain enough points, then there are not enough points to form a circle
    if (i != 3) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "3pt circle crap " << endl;
#endif
#endif
        fireSanity(NEED_3_POINTS_TO_FORM_BALL);
        return false;
    }

    double centreX = 0, centreY = 0, radius = 0, expected = 1;

    int sqr_y[3];
    sqr_y[0] = SQUARE (p[0].y);
    sqr_y[1] = SQUARE (p[1].y);
    sqr_y[2] = SQUARE (p[2].y);
    int sqr_x[3];
    sqr_x[0] = SQUARE (p[0].x);
    sqr_x[1] = SQUARE (p[1].x);
    sqr_x[2] = SQUARE (p[2].x);

    double a = determinant (p[0].x, p[0].y, 1,
                            p[1].x, p[1].y, 1,
                            p[2].x, p[2].y, 1);

    double d = -determinant (sqr_x[0] + sqr_y[0], p[0].y, 1,
                             sqr_x[1] + sqr_y[1], p[1].y, 1,
                             sqr_x[2] + sqr_y[2], p[2].y, 1);

    double e = determinant (sqr_x[0] + sqr_y[0], p[0].x, 1,
                            sqr_x[1] + sqr_y[1], p[1].x, 1,
                            sqr_x[2] + sqr_y[2], p[2].x, 1);

    double f = -determinant (sqr_x[0] + sqr_y[0], p[0].x, p[0].y,
                             sqr_x[1] + sqr_y[1], p[1].x, p[1].y,
                             sqr_x[2] + sqr_y[2], p[2].x, p[2].y);

    // a= 0 if the points are on a straight line
    if (a == 0) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "points in straight line" << endl;
#endif
#endif
        return false;
    }
        
    centreX = -d / (2 * a);
    centreY = -e / (2 * a);
    radius = sqrt ((SQUARE (d) + SQUARE (e)) / (4 * SQUARE (a)) - f / a);
    if (!radius) {
        // andres, probably don't need this check. I hypothesize that a=0 => radius=0
        return false;
    }

    ball.blob = preball;
    ball.area = preball->area;        //M_PI* SQUARE(radius)
    ball.radius = radius;
    ball.cx = centreX;
    ball.cy = centreY;
    ball.width = radius * 2;
    ball.height = radius * 2;
    ball.x = centreX - radius;
    ball.y = centreY - radius;

    xmin = ball.x;
    xmax = ball.x + ball.width;
    ymin = ball.y;
    ymax = ball.y + ball.height;

    // calculating confidence
    expected = MIN (pi * SQUARE (ball.radius), HEIGHT * WIDTH);        //25/1/02 used to be width/2
    ball.cf = (int) MIN (ball.area * 1000.0 / expected, 1000);

    // apply clippings
    int pxmin = (int) MAX (xmin, 0);
    int pxmax = (int) MIN (xmax, WIDTH);
    int pymin = (int) MAX (ymin, 0);
    int pymax = (int) MIN (ymax, HEIGHT);

    ball.projected_x = pxmin;
    ball.projected_y = pymin;
    ball.projected_w = pxmax - pxmin;
    ball.projected_h = pymax - pymin;

    ball.tilt = 0;
    ball.imgHead = PointToHeading (centreX);
    ball.imgElev = PointToElevation (centreY);
    ball.misc = raw_y;

#ifdef HIGH
    double divisor;
    if (cmdAdjustDistance) {
        divisor = ballCalibrator->getAdjustedDistance (ball.radius);
    } else {
        divisor = 600.0;
    }

    ball.cam_dist = 600.0 / ball.radius;
    ball.dist = divisor / ball.radius;
    ball.var = exp (-0.864452 + 0.0215735 * ball.dist) /
    pow (ball.area / expected, errorScale);
#endif
#ifdef MED
    ball.dist = SizeToDistance (BALL_RADIUS, ball.radius);
#endif
    ball.cam_centroid_dist = ball.dist;
    RobotRelative (ball);
    return true;
}

// this method tries to find ball distance using blob size
// blob size is extended to the largest dimension since ball is round
// ie width = height and extend the smaller to be equal to the greater
// to cater for occlusions.
// this method fails when the ball is in the corner, or in other words
// when both and width and height != diametre of ball.
// at close distances, this error is magnified.
bool VisualCortex::normalFindBall(HalfBeacon *preball, VisualObject &ball) {
    double xmin = preball->getXMin ();
    double xmax = preball->getXMax ();
    double ymin = preball->getYMin ();
    double ymax = preball->getYMax ();
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "ball blob top left (" << xmin << "," << ymin << ") bottom right (" << xmax << "," << ymax << ")" << endl;
#endif
#endif

    double raw_y = ymin;

    // First of all, we expand the rectangle to full size
    // assuming the ball is partially visible.
    // This assumes we can at least see two corners
    // Find the centre of the sides.
    int counter1 = 0;
    int counter2 = 0;
    double cleft = 0;
    double cright = 0;
    double ctop = 0;
    double cbottom = 0;

    for (int y = (int)ymin; y <= (int)ymax; y++) {
        if (cplane[y * WIDTH + (int) xmin] == cBALL) {
            cleft += y;
            counter1++;
        }
        if (cplane[y * WIDTH + (int) xmax] == cBALL) {
            cright += y;
            counter2++;
        }
    }

    // added by Martin 21/06/2001
    // try get rid of the posibilities of divide by zero
    // but not sure what to sure, so set counter to 1.
    // any other suggestion?
    if (!counter1)
        counter1 = 1;

    if (!counter2)
        counter2 = 1;

    cleft /= counter1;
    cright /= counter2;

    counter1 = counter2 = 0;
    for (int x = (int)xmin; x <= (int)xmax; x++) {
        if (cplane[(int) ymin * WIDTH + x] == cBALL) {
            ctop += x;
            counter1++;
        }
        if (cplane[(int) ymax * WIDTH + x] == cBALL) {
            cbottom += x;
            counter2++;
        }
    }

    if (!counter1)
        counter1 = 1;

    if (!counter2)
        counter2 = 1;

    ctop /= counter1;
    cbottom /= counter2;

    // We've now got the centres of the sides.
    // So we expand the box so the centre is in the centre.
    // The basic idea is to make it as big as we can justify

    // Adjust the top or bottom
    //double diff1 = MAX(ymax-cleft, cleft-ymin);
    //double diff2 = MAX(ymax-cright, cright-ymin);
    //double centre_y = (MAX(ymax-cleft, cleft-ymin) > MAX(ymax-cright, cright-ymin)) ? cleft : cright;

    // Adjust the left or right
    //diff1 = MAX(xmax-ctop, ctop-xmin);
    //diff2 = MAX(xmax-cbottom, cbottom-ymin);
    //double centre_x = (MAX(xmax-ctop, ctop-xmin) > MAX(xmax-cbottom, cbottom-ymin)) ? ctop : cbottom;

    // Now finally make it a square
    // This should give us an even better centreing - 'tho not perfect.
    // We should get max(sin(th),cos(th))*rad, where rad is the real
    // radius and th is the angle from the nearest axis of the chord to
    // the edgemost point available. (That sort of idea anyway)
    // Hmm.. so maybe the expansion stuff above is unnecessary?
    // But not, it's good if it's obscured anyway.
    // The justification for some of this is slightly dodgy... :)
    double width = xmax - xmin;
    double height = ymax - ymin;

    double offset;
    int delta = 3;

    if (width > height) {
        if (ymax >= USABLE_HEIGHT - delta)        // object is on the bottom edge of the FOV
            ymax = ymin + width;
        else if (ymin <= delta)        // object is on the top edge of the FOV
            ymin = ymax - width;
        else {
// object's bbox is not rectangular?
            offset = (width - height) / 2.0;
            ymin -= offset;
            ymax += offset;
        }
        height = width;
    } else if (width < height) {
        if (xmax >= WIDTH - delta)        // object is on the right edge of the FOV
            xmax = xmin + height;
        else if (xmin <= delta)        // object is on the left edge of the FOV
            xmin = xmax - height;
        else {                        // object's bbox is not rectangular?
            offset = (height - width) / 2.0;
            xmin -= offset;
            xmax += offset;
        }
        width = height;
    }

    double centre_x = (xmin + xmax) / 2.0;
    double centre_y = (ymax + ymin) / 2.0;

    ball.blob = preball;
    ball.area = preball->area;
    ball.x = xmin;
    ball.y = ymin;
    ball.width = width;
    ball.height = height;
    ball.cx = centre_x;
    ball.cy = centre_y;
	ball.radius = MAX(width, height) /2;

    ball.cf = ball.cf;
    ball.tilt = 0;
    ball.imgHead = PointToHeading (centre_x);
    ball.imgElev = PointToElevation (centre_y);
    ball.misc = raw_y;

    ball.dist = 1200.0 / ball.height;
    if (ball.dist < 25 && ABS (pan) < 0.5)        //0.5rad = 30degree
        ball.dist = MIN (ball.dist, NewBallDistance ());
    ball.cam_dist = ball.dist;
    RobotRelative (ball);

    ball.tilt = 0;

    ball.projected_x = MAX(ball.x, 0);
    ball.projected_y = MAX(ball.y, 0);
    ball.projected_w = ball.width;
    ball.projected_h = ball.height;

    double expected = MIN (pi * SQUARE (ball.radius), HEIGHT * WIDTH);        //25/1/02 used to be width/2
    ball.cf = (int) MIN (ball.area * 1000.0 / expected, 1000);
    ball.var = exp (-0.864452 + 0.0215735 * ball.dist) /
    pow (ball.area / expected, errorScale);
    return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////
// above 40cm, use blob to determine distance
// under 40cm, use projection method
// when using projection, if the ball is occluded
// use fireball, otherwise just project centre of ball.
//////////////////////////////////////////////////////////////////////////////////////////////////////////
void VisualCortex::FindBall () {
    double fbHead, fbDist, fbRadius;
    double fb_obj0, fb_obj1, fb_obj2;
    bool fbFound = false;
    bool cover = false;
    bool nbFound = true;
    int fbX, fbY;
    fbHead = fbDist = fbRadius = fb_obj0 = fb_obj1 = fb_obj2 = fbX = fbY = -1;

    // added by Spencer, selection sort the base blobs in order
    // modified by Martin, added threshold... it will filter out small blobs
    int numBalls = blobs->findHalfBeacons (cBALL, hbBall);
  /*
     cout<< endl << numBalls << " findball:" << endl;
     for (int i=0;i<numBalls;i++) {
     cout << i << ": " <<endl;
     hbBall[i]->Print();
     }
     cout << "end"  << endl;
   */
    int totalRedArea = 0;
    int totalOrangeArea = 0;

    for (int n = 0; n < numBalls; n++) {
        HalfBeacon *pball = hbBall[n];
        totalOrangeArea += pball->area;
    }

    int nb =  blobs->count (cROBOT_RED);
    for (int n = 0; n < nb; n++) {
        totalRedArea += blobs->getBlob (cROBOT_RED, n).area;
    }

    for (int n = 0; n < numBalls; n++) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        //cout << n << "/" << numBalls << endl;
#endif
#endif
        HalfBeacon *preball = hbBall[n];
        preball->initBallBlob (&blobs->splane);

        double xmin = preball->getXMin ();
        double xmax = preball->getXMax ();
        double ymin = preball->getYMin ();
        double ymax = preball->getYMax ();
        double raw_y = ymin;

#ifdef OFFLINE
        (*outfile) << "blob " << xmin << " " << ymin << " " << (xmax-xmin) << " " << (ymax-ymin) << endl;
#ifdef PRE_SANITY_BLOB
        if (sanityBlobDebugMsg) {
            cout << "PRE_SANITY_BLOB: " << vobBall << " " << xmin << " " << ymin << " " << (xmax-xmin) << " " << (ymax-ymin) << endl;
        }
        (*blobOutput) << vobBall << " " << xmin << " " << ymin << " " << (xmax-xmin) << " " << (ymax-ymin) << endl;
#endif
#endif

        // New desparate measure - it's not the ball if it's too thin or too flat
        if (xmax + 1 - xmin <= 2 || ymax + 1 - ymin <= 2) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Sanity L" << endl;
#endif
#endif
            fireSanity(BALL_TOO_THIN_OR_TOO_FLAT);
            continue;
        }

        double elev = PointToElevation (xmin, ymin, sin_eroll, cos_eroll);

        // check the elevation of the ball, if it is higher than the head.
        if (RAD2DEG (effective_tilt) - elev < -15) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "ball too high Sanity I\n";
#endif
#endif
            fireSanity(BALL_ELEVATION_TOO_HIGH);
            continue;
        }
        
#ifdef FIREBALL                
        // REMIND: TRIGGER needs WORK
        //if ((RAD2DEG(effective_tilt) > TILT_TRIGGER && preball->area > AREA_TRIGGER) || (preball->area > AREA_TRIGGER * 8)) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "blob area " << totalOrangeArea << " + red dog area " << totalRedArea << " = " << totalOrangeArea + totalRedArea << "/" << (WIDTH - 10) * (HEIGHT - 10) << endl;
#endif
#endif
            findCentre(preball, &fbX, &fbY, &fbHead, &fbDist, &fbRadius, &fb_obj0, &fb_obj1, &fb_obj2);
            fbFound = (fbRadius > 0); // whether it's found
	        // if the radius found is shorter than the orange blob, then something is wrong, most likely the
	        // edges of ball were detected falsely due to noise around ball
            double maxDimension = MAX(preball->getXMax() - preball->getXMin(), preball->getYMax() - preball->getYMin());
	    if (2 * fbRadius < 0.9 * maxDimension) {
	    	    fbFound = false;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "fireball sanity, diametre " << 2*fbRadius << " smaller than 90% max blob dimension " << 0.9*maxDimension << endl;;
#endif
#endif
    	    }


            // check whether the ball covers the entire screen.
            // however cant check for full orange because when ball close enough there'll be alot of red as well
            // on the other hand, it's possible that you are looking straight into red dog, 
            // so if ball is reasonable size, but rest of screen is covered by red dog, then fire.
            if (totalOrangeArea + totalRedArea >= 21000 && !fbFound) {
                int inpoints[2] = {0, 0};
                double outpoints[2] = {-1, -1};
                projectPointsDAlt(inpoints, 1, outpoints, BALL_RADIUS);
                fbDist = sqrt((double)(outpoints[0] * outpoints[0] + outpoints[1] * outpoints[1]));
                fbHead = -1.0*RAD2DEG(asin((1.0 * outpoints[0]) / fbDist));
                fbRadius = max(WIDTH/2, HEIGHT/2);
                fbX = WIDTH / 2;
                fbY = HEIGHT / 2;
                fb_obj0 = outpoints[0];
                fb_obj2 = outpoints[1];
                fb_obj1 = NECK_HEIGHT; 
                fbFound = true;
                cover = true;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "fireball screen covered with ball" << endl;
#endif
#endif
            } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "not cover screen, area " << totalOrangeArea + totalRedArea << " fbFound " << fbFound << endl;
#endif
#endif
            }
	        // if fireball returns big distance, then dont trust it
	        if (fbDist > 30) {
	    	    fbFound = false;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "fireball sanity, dist > 30" << endl;
#endif
#endif
    	    }

        //}
#endif
        nbFound = normalFindBall(preball, vob[vobBall]);

        const int PROJECTION_TRIGGER = 40;
        const int NECK_HEIGHT = 7;
        int pxmin = (int) MAX (xmin, 0);
        int pxmax = (int) MIN (xmax, WIDTH);
        int pymin = (int) MAX (ymin, 0);
        int pymax = (int) MIN (ymax, HEIGHT);
        VisualObject &ball = vob[vobBall];

#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        if (fbFound || nbFound)
            (*outfile) << "ball found before sanity fb:" << fbFound << " nb:" << nbFound << endl;
#endif
#endif

        // Perform sanity checks on ball, if false is returned then it can't be a ball.
	    

        if (cover || BallSanities(ball, pxmin, pxmax, pymin, pymax, xmin, xmax, ymin, ymax)) {
#ifdef FIREBALL
            double currentHead = ball.h;
            double currentDist = ball.d;

            int inpoints[2] = {((int)(ball.cx) - WIDTH / 2), -1*((int)(ball.cy) - HEIGHT / 2)};
            double outpoints[2] = {-1, -1};
            projectPointsDAlt(inpoints, 1, outpoints, BALL_RADIUS);
            double currentRDist = sqrt((double)(outpoints[0] * outpoints[0] + outpoints[1] * outpoints[1]));
            double currentRHead = -1.0 * RAD2DEG(asin((1.0 * outpoints[0]) / currentRDist));

            // if ball is small but vision says it's very close then likely it is reflection from looking down at other robot
            if (currentRDist < 50 && ball.area < 50 && ball.radius <= 5) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "Sanity Q " << ball.cx << " " << ball.cy << " area " << ball.area << " d " << ball.d;
                (*outfile) << " projected_d " << currentRDist << " radius " << ball.radius << endl;
#endif
#endif            
                ball.cf = 0;
                ball.var = LARGE_VAL;
                continue;
            } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "Sanity Q not fired " << ball.cx << " " << ball.cy << " area " << ball.area << " d " << ball.d << " projected_d " << currentRDist << endl;
#endif
#endif            
            }
	    
            // if blob d < 40 use projection
		    if (currentRDist < PROJECTION_TRIGGER) {
                // if intersect frame then fireball.
                if (fbFound && (ball.x <= 0 || (ball.x + ball.radius * 2) >= WIDTH - 1 || ball.y <= 0 || (ball.y + ball.radius * 2) >= HEIGHT - 1)) {
                    ball.radius = fbRadius;
                    ball.cx = fbX;
                    ball.cy = fbY;
                    ball.width = fbRadius * 2;
                    ball.height = fbRadius * 2;
                    ball.x = fbX - fbRadius;
                    ball.y = fbY - fbRadius;
		            ball.area = preball->area;

                    xmin = ball.x;
                    xmax = ball.x + ball.width;
                    ymin = ball.y;
                    ymax = ball.y + ball.height;

                    // calculating confidence
                    double expected = MIN (pi * SQUARE (ball.radius), HEIGHT * WIDTH);        //25/1/02 used to be width/2
                    //ball.cf = (int) MIN (ball.area * 1000.0 / expected, 1000);
                    ball.cf = 500;

                    // apply clippings
                    int pxmin = (int) MAX (xmin, 0);
                    int pxmax = (int) MIN (xmax, WIDTH);
                    int pymin = (int) MAX (ymin, 0);
                    int pymax = (int) MIN (ymax, HEIGHT);
                    ball.projected_x = pxmin;
                    ball.projected_y = pymin;
                    ball.projected_w = pxmax - pxmin;
                    ball.projected_h = pymax - pymin;

                    ball.tilt = 0;
                    ball.imgHead = PointToHeading (fbX);
                    ball.imgElev = PointToElevation (fbY);

                    ball.cam_dist = sqrt(fbDist * fbDist + NECK_HEIGHT * NECK_HEIGHT) - FACE_LENGTH;
                    ball.dist = sqrt(fbDist * fbDist + NECK_HEIGHT * NECK_HEIGHT);
                    ball.d = fbDist;
                    ball.h = ball.head = fbHead;
                    ball.objectPosition[0] = fb_obj0;
	                ball.objectPosition[1] = fb_obj1;
	                ball.objectPosition[2] = fb_obj2;
                    ball.var = exp (-0.864452 + 0.0215735 * ball.dist) / pow (ball.area / expected, errorScale);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
		            if (cover)
	                    (*outfile) << "fire ball, screen covered" << endl;
		            else
	                    (*outfile) << "fire ball, normal" << endl;
                    ballMethod = 0;
#endif
#endif            
					ballDistMethod = FIREBALL_PROJECTION;
                } else {
                    ball.cam_dist = sqrt(currentRDist * currentRDist + NECK_HEIGHT * NECK_HEIGHT) - FACE_LENGTH;
                    ball.dist = sqrt(currentRDist * currentRDist + NECK_HEIGHT * NECK_HEIGHT);
                    ball.d = currentRDist;
                    ball.h = ball.head = currentRHead;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "projected ball" << endl;
                    ballMethod = 1;
#endif
#endif            
					ballDistMethod = PIXEL_COUNT_PROJECTION;
                }
            } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "old ball" << endl;
                ballMethod = 2;
#endif
#endif           
				ballDistMethod = PIXEL_COUNT_DISTANCE;
            }

#endif

#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            if (ball.cf > 0) {
                (*outfile) << "found ball, distance " << ball.d << " variance " <<
                ball.var << " cf " << ball.cf << " elev " << ball.elev << " area " << ball.area << endl;
            }
#endif
#endif
            ball.misc = raw_y;
	    
            // temporary fix
            ball.cf = (ball.var >= LARGE_VAL) ? 3 : ball.cf;
            ball.var = (ball.var >= LARGE_VAL) ? LARGE_VAL - 1 : ball.var;

            // smallest dist from base of neck to ball is ball_radius
            ball.d = (ball.d <= 0) ? BALL_RADIUS : ball.d;
			vob[cBALL].angleVar = get95CF(PAN_95CF);
			
			if(currentWalkType && (((*currentWalkType) == ZoidalWalkWT) || ((*currentWalkType) == OffsetWalkWT))) {
				vob[cBALL].var *= zoidalVarFactor;
				vob[cBALL].angleVar *= zoidalAngVarFactor;
			}
            return;
        } else {
            continue;
        }
    }
    lastd = 0;
    vob[cBALL].cf = 0;
    vob[cBALL].var = LARGE_VAL; 
    vob[cBALL].angleVar = LARGE_VAL;
}


bool
VisualCortex::BallSanities (VisualObject & ball, int pxmin, int pxmax,
                            int pymin, int pymax, double xmin, double xmax,
                           double ymin, double ymax) {
    int k;
    // note: for elevation - up is +ve, down is -ve
    // too high - can't be the ball
    // This is a dumb sanity check because when the flat distance is relatively small in comparison to the value of
    // y, then this will get called.  The only reason we had this sanity check is because we never measured distance accurately
    // enough. I added ball.getFlatDistance()>15 to the condition
    // Contact andres if you disagree

    // ricky: removed sanities to do with elevtion
    // somehow i think the numbers might be transformed wrong somewhere
    // raymond mentioned that if you times everything by 2 then the numbers are more sensible
    // however the numbers worked last year so something must hav been modified somewhere and 
    // I'm not sure if timing everything by 2 is a good way to go
    // will look in to it more when i get the chance
    // ricky: ok there's no error with readings - although they do screw up if
    // head was in extreme positions ie full pan + full tilt etc
    // anyway sanity A seems redundant.. 
    // Raymond: Well it looks like we need elevation sanities anyway because we're seeing a lot of green 
    // in the background. 
    /*
        if (ball.elev > 5 && ball.getFlatDistance () > 15
             && RAD2DEG (effective_tilt) < 40) {
            //cout << "Sanity A" << endl;
            //wout << "Sanity A\n";
            fireSanity(INCORRECT_BALL_ELEVATION_OR_FLAT_DISTANCE);
            return false;
        }
    */    

#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "start ball sanity" << endl;
#endif
#endif

    double elev = PointToElevation (ball.cx, ball.cy, sin_eroll, cos_eroll);
    k = vobBlueGoal;
    //Raymond: Check to make sure ball is below the mathematical horizon (ie. it is closer than infinity when 
    //projected onto the field)
    // however this is very sensitive, if for whatever reason (stance change, pushed by another robot, leg slipped)
    // the robot's head position is different from that received from the sensors, then projection will be 
    // wrong.
    int ballPoints[2], retBallPoints[2];
    ballPoints[0] = (int)ball.cx - WIDTH/2;
    ballPoints[1] = -((int)ball.cy - HEIGHT/2);
    projectPoints(ballPoints, 1, retBallPoints, 0);
    if (retBallPoints[0] == LARGE_VAL && retBallPoints[1] == LARGE_VAL && ball.elev > 5)
    {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "Horizon sanity, elev " << elev << endl;
#endif
#endif
        fireSanity(BALL_NOT_BELOW_HORIZON);
      return false;
    }
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "passed horizon sanity, elev " << elev << endl;
#endif
#endif

    if (tan(DEG2RAD(ball.elev)) * ball.dist > 10.0) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "Ball height sanity fired " << tan(DEG2RAD(ball.elev)) * ball.dist << endl;
#endif
#endif
      return false;
    }
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "Ball height sanity passed " << tan(DEG2RAD(ball.elev)) * ball.dist << endl;
#endif
#endif
    

    // check elev of goals
    for (; k <= vobYellowGoal; k++) 
    {
        // cy already rotated
        double goalElev = PointToElevation (vob[k].cy);
        //if (vob[k].var < get95CF(60)) {
        if (vob[k].cf > 100) 
        {
            if (elev > goalElev + 5 /*&& !VobInVob (vob[vobYellowGoal], ball, 20, 0, 0, 0)*/) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    	        (*outfile) << "ball elev " << elev << " goal elev " << goalElev << endl;
#endif
#endif
                break;
            }
        }
    }
    
    if (k <= vobYellowGoal) {
        //eout << "Sanity B1" << endl;
        //wout << "Sanity B\n";
        fireSanity(BALL_ELEVATION_GREATER_THAN_GOAL_ELEVATION + k - vobBlueGoal);
        return false;
    }
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "passed goal elev sanity" << endl;
#endif
#endif

  
    // check elev of other beacons
    k = vobBlueLeftBeacon;
    for (; k <= vobYellowRightBeacon; k++) {
        if (vob[k].cf > 100) {
            //if (vob[k].var < get95CF(60))

            double beaconElev = PointToElevation (vob[k].cx, vob[k].cy, sin_eroll,
                                                  cos_eroll);
            //if ( ball.elev > vob[k].elev )
            if (elev > beaconElev + 5) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    	        (*outfile) << "ball elev " << elev << " beacon elev " << beaconElev << endl;
#endif
#endif
                break;
            }
        }
    }
    if (k <= vobYellowRightBeacon) {
        //eout << "Sanity C1" << endl;
        //wout << "Sanity C\n";
        fireSanity(BALL_ELEVATION_GREATER_THAN_BEACON_ELEVATION + k - vobBlueLeftBeacon);
        return false;
    }
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "passed beacon elev sanity" << endl;
#endif
#endif

    // another EBET rule - ball cannot overlap the projection of a beacon
    // back onto the camera. With this perceived ball distance must exceed beacon's
    k = vobBlueLeftBeacon;
    for (; k <= vobYellowRightBeacon; k++) {
        //if (vob[k].var < get95CF(40))
        if (vob[k].cf > 200) {
            // beacons should be further away than ball
            if (ABS (vob[k].head - ball.head) < 5
                    && vob[k].dist < ball.dist - 50) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    	        (*outfile) << "ball overlap beacon " << endl;
#endif
#endif
                    
                fireSanity(BALL_TOO_CLOSE_TO_BEACON + k - vobBlueLeftBeacon);
                break;
            }

            // ball cannot be occluded by beacons
            if (ball.cx > vob[k].x && ball.cx < (vob[k].x + vob[k].width)
                    && ball.cy > vob[k].y && ball.cy < (vob[k].y + vob[k].height)) {
                fireSanity(BALL_OCCLUDED_BY_BEACON + k - vobBlueLeftBeacon);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    	        (*outfile) << "ball overlap beacon 2" << endl;
#endif
#endif
                break;
            }
        }

    }
    
    if (k <= vobYellowRightBeacon) {
        //eout << "Sanity D" << endl;
        //wout << "Sanity D\n";
        return false;
    }
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "passed beacon overlap" << endl;
#endif
#endif

    // number of red pixels near the projected bounding box should be low
    if (ball.area <= 150) {
        int numRed = 0;
        int numYellow = 0;
        int numField = 0;
        int starty = pymin - ((int)(ball.projected_h/2.0) + 1);
        int endy   = pymax + ((int)(ball.projected_h/2.0) + 1);
        int startx = pxmin - ((int)(ball.projected_w/2.0) + 1);
        int endx   = pxmax + ((int)(ball.projected_w/2.0) + 1);

        if (starty < 0)
            starty = 0;
        if (endy >= HEIGHT)
            endy = HEIGHT - 1;
        if (startx < 0)
            startx = 0;
        if (endx >= WIDTH)
            endx = WIDTH - 1;

        int ystarty = starty;
        int yendy   = endy;
        int ystartx = startx;
        int yendx   = endx;

        for (int i = starty; i <= endy; i++) {
            for (int j = startx; j <= endx; j++) {
                if (cplane[i * WIDTH + j] == cROBOT_RED)
                    numRed++;
                if (cplane[i * WIDTH + j] == cBEACON_YELLOW)
                    numYellow++;
                if (cplane[i * WIDTH + j] == GREEN_FIELD)
                    numField++;
            }
        }

        // Too many red pixels
        if (numRed > ball.area + numField) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Sanity K   num red " << numRed << " ball area " << ball.area << " field " << numField << endl;
#endif
#endif
            fireSanity(TOO_MUCH_RED_IN_BALL);
            return false;
        } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Sanity K passed num red " << numRed << " ball area " << ball.area << " field " << numField << endl;
#endif
#endif
	    }

        // both yellow and red pixel is too much
	    // i remember this sanity now.. it's to filter out ball formed btwn
	    // red goalie and yellow goal, toned down slightly to check green field around perimetre
	    // if green field exists, then probably a ball.
        if (numRed + numYellow > numField + ball.area) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Sanity M" << endl;
#endif
#endif
            fireSanity(TOO_MUCH_YELLOW_RED_IN_BALL);
            return false;
        } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Sanity M passed num red " << numRed << " and num yellow " << numYellow << " ball area " << ball.area << " field " << numField << endl;
#endif
#endif
        }
    }

#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "passed red and yellow around ball check" << endl;
#endif
#endif

    // check if the bounding box of ball is completely inside bounding box of red robot
    // if it is then it's likely that it's not the ball but misclassification of red patches
    if (ball.area <= 20) {
        int redxmax = 0, redymax = 0;
        int redxmin = 511, redymin = 511;
        for (int i = 0; i < blobs->count (cROBOT_RED); i++) {
            BlobInfo & info = blobs->getBlob (cROBOT_RED, i);
            if (info.area > 20) {
                redxmax = MAX (redxmax, info.getXMax ());
                redymax = MAX (redymax, info.getYMax ());
                redxmin = MIN (redxmin, info.getXMin ());
                redymin = MIN (redymin, info.getYMin ());
            }
        }
	    int bxmin = (int)MAX(0, ball.x);
	    int bxmax = (int)MIN(WIDTH-1, ball.x+ball.width);
	    int bymin = (int)MAX(0, ball.y);
	    int bymax = (int)MIN(HEIGHT-1, ball.y+ball.height);
        if (xmax <= redxmax && ymax <= redymax && xmin >= redxmin && ymin >= redymin && NoGreenNearBy(bxmin, bxmax, bymin, bymax)) { 
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Sanity E on blob " << ball.cx << " " << ball.cy << endl;
    	    (*outfile) << "red xmax ymax xmin ymin " << redxmax << " " << redymax << " " << redxmin << " " << redymin << endl;
#endif
#endif
            fireSanity(BALL_BOUNDING_BOX_INSIDE_RED_ROBOT_BOUNDING_BOX);
            return false;
        }
    }

    // if ball is small and it touches the red blob -> unlikely it's a ball
    if (ball.area <= 40) {        
        k = 0;
        for (; k < blobs->count (cROBOT_RED); k++) {
            BlobInfo & info = blobs->getBlob (cROBOT_RED, k);
            if (info.area > 20) {
                if (info.getXMax () >= xmax && info.getXMin () <= xmin && info.getYMin() >= ymin - ball.radius)
                    break;
		/*
                if (info.getYMax () >= ymax && info.getYMin () <= ymin)
                    break;
		*/
            }
        }
        if (k < blobs->count (cROBOT_RED)) {
            //eout << "Sanity F" << endl;
			//eout << "that touch red blob sanity says its not a ball" << endl;
            fireSanity(BALL_TOUCHES_RED_BLOB);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "ball " << ball.cx << "," << ball.cy << " touch red blob" << endl;
#endif
#endif
            return false;
        }
    }

#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "passed ball and red dog sanity" << endl;
#endif
#endif

    if (ball.area <= 50 && InYellowNoGreen (ball)) {
        //eout << "Sanity G" << endl;
      //wout << "Sanity G\n";
        fireSanity(BALL_INSIDE_YELLOW_GOAL);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "in yellow no green" << endl;
#endif
#endif
        return false;
    }

    // too far away outside the field, cant be the ball
    if (ball.dist > FIELD_DIAGONAL) {
      //eout << "Sanity H" << endl;
      //wout << "Sanity H\n";
        fireSanity(BALL_OUTSIDE_FIELD);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "ball outside field" << endl;
#endif
#endif
        return false;
    }

    // This will get rid of the problem of mis-recognise the
    // edge of a red robot (inside the yellow goal) as a ball.
    // The number of orange pixels must occupy
    // more that a certain percentage of the area of
    // projected bounding box.
    double pi = 3.14159;
    double occupied_area = (long)(ball.radius * ball.radius * pi);

    if (occupied_area == 0) {
        //eout << "SanityHb" << endl;
        fireSanity(BALL_OCCUPIED_AREA_0);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "ball sanity: huh?" << endl;
#endif
#endif
        return false;
    }
    
    if ((ball.area < 100) && (((double) ball.area / occupied_area) < 0.5)) {
        // try to filter out consecutive runs of orange pixels
        // on the edge of red robot
        //eout << "Sanity I" << endl;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "another atttempt sanity " << ball.area << " " << occupied_area << endl;
#endif
#endif
        return false;
    } else if ((ball.area < 150) && (((double) ball.area / occupied_area) < 0.3)) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "Sanity J" << endl;
#endif
#endif
        fireSanity(BALL_AREA_LT150_NOT_OCCUPY_MT25_PROJECTED_BOUNDING_BOX);
        return false;
    } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "ball passed density check " << ball.area << "/" << occupied_area << " ratio " << ((double) ball.area / occupied_area)
        << endl;
#endif
#endif
    }
        
/*
    Commented out by Ben 11/4 after complaints from james - miss classification of green
    makes some of the fields white .... don't actually see the point of
    using NoGreenNearBy in the first place

    if (ball.area < 300 && NoGreenNearBy (pxmin, pxmax, pymin, pymax)) {
        //          cout << "Sanity 1" << endl;
        fireSanity(BALL_AREA_LT300_NO_GREEN_NEARBY);
        return false;
    }

*/

/*
    //we have passed every ball sanity check
    if(VobInVob(vob[vobYellowGoal], ball, 0, 0, 0, 0)) {
        cout << "goal inside ball no offset" << endl;
    }
*/
    //if yellow goal is inside ball, reject the goal
    int buffer = 20;
	int xOrigin = WIDTH/2;
	int yOrigin = HEIGHT/2;
	int uYGoalX = (int)(xOrigin - (ucos_eroll*(xOrigin - vob[vobYellowGoal].x) - usin_eroll*(yOrigin - vob[vobYellowGoal].x)));
	int uYGoalY = (int)(yOrigin - (usin_eroll*(xOrigin - vob[vobYellowGoal].y) + ucos_eroll*(yOrigin - vob[vobYellowGoal].y)));
    if ((ball.x <= uYGoalX + buffer) && (ball.x + ball.width >= uYGoalX + vob[vobYellowGoal].width - buffer) && 
            (ball.y <= uYGoalY + buffer) && (ball.y + ball.height >= uYGoalY + vob[vobYellowGoal].height - buffer)) {
        vob[vobYellowGoal].cf = 0;
        vob[vobYellowGoal].var = LARGE_VAL;
        fireSanity(YELLOW_GOAL_INSIDE_BALL);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "goal in ball" << endl;
#endif
#endif
    }
            

    /*
    if (VobInVob (vob[vobYellowGoal], ball, 20, 0, 0, 0)) {

        vob[vobYellowGoal].cf = 0;
        vob[vobYellowGoal].var = LARGE_VAL;
//          cout << "goal inside ball with top offset = 5" << endl;
        fireSanity(YELLOW_GOAL_INSIDE_BALL);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        cout << "goal in ball" << endl;
#endif
#endif
    }
    */
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "passed ball sanity" << endl;
#endif
#endif

    return true;
}

bool
VisualCortex::InYellowNoGreen (VisualObject & ball) {
    //if orange blob index is inside yellow goal and there is no green
    //of the field nearby, it is not the ball

    if (ball.x < vob[vobYellowGoal].x || ball.y < vob[vobYellowGoal].y ||
         ball.x + ball.width >
         vob[vobYellowGoal].x + vob[vobYellowGoal].width
         || ball.y + ball.height >
         vob[vobYellowGoal].y + vob[vobYellowGoal].height) {
        return false;
    }
    // make bounding box bigger to check the green color
    return NoGreenNearBy ((int)ball.x, (int)(ball.x + ball.width), (int)ball.y,
                          (int)(ball.y + ball.height));
}

bool VisualCortex::NoColourXNearBy (int colour, int xmin, int xmax,
                                    int ymin, int ymax, int delta_x,
                                    int delta_y, int threshHold) {
    // check if there is any green near this box
    int x, y, num;
    //int min_delta = 7;
    //int delta_x = MAX(min_delta, (xmax-xmin)/3);
    //int delta_y = MAX(min_delta, (ymax-ymin)/3);

    xmin -= delta_x;
    if (xmin < 0)
        xmin = 0;

    ymin -= delta_y;
    if (ymin < 0)
        ymin = 0;

    xmax += delta_x;
    if (xmax > WIDTH)
        xmax = WIDTH;

    ymax += delta_y;
    if (ymax > USABLE_HEIGHT)
        ymax = USABLE_HEIGHT;

    // so what if its at the bottom of the image ?
    /*
     if (ymax==USABLE_HEIGHT)
        return false; // at the bottom of the image
    */

    //threshold = ((xmax-xmin)*(ymax-ymin))/(3*3*5);
    //if (threshold>30) threshold = 30;

    num = 0;
    for (y = ymax; y >= ymin; y--) {
        for (x = xmin; x <= xmax; x++) {
            if (cplane[y * WIDTH + x] == colour) {
                num++;

                if (num >= threshHold)
                    return false;
            }
        }
    }

  return true;
}

void
VisualCortex::CheckMatchingBeacons (const HalfBeacon * pink,
                                    int colorIndex,
                                    HalfBeacon ** hbArray, int hbCount,
                                    double &minDist, double &maxRatio,
                                    HalfBeacon ** closest, int *closestType) {
    for (int i = 0; i < hbCount; i++) {
        if (hbArray[i]->matched)
            continue;

        if (hbArray[i]->area == 0 || pink->area == 0) {
            //cout << "area zero!!!" << endl;
            fireSanity(ZERO_HALF_BEACON_AREA_OR_PINK_AREA);
            continue;
        }
        
        double rdist = sqrt(SQUARE(pink->cx - hbArray[i]->cx) + SQUARE(pink->cy - hbArray[i]->cy));
        //if (beaconSanityDebugMsg && drawLine((int)(pink->cx), (int)(pink->cy), (int)(hbArray[i]->cx), (int)(hbArray[i]->cy), BACKGROUND) > rdist / 2) {
        //    //cout << "beacon sanity: blob too far apart" << endl;
        //    continue;
        //}
        // sanity that checks out beacons with over 50% noise
        if (drawLine((int)(pink->cx), (int)(pink->cy), (int)(hbArray[i]->cx), (int)(hbArray[i]->cy), PINK) +
          drawLine((int)(pink->cx), (int)(pink->cy), (int)(hbArray[i]->cx), (int)(hbArray[i]->cy), colorIndex)
          < rdist * 3.0 / 5.0) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "beacon sanity: blob too far apart (pink, color) .. ";
            (*outfile) << (int)(pink->cx) << "," << (int)(pink->cy) << " " << (int)(hbArray[i]->cx) << "," << (int)(hbArray[i]->cy) << endl;
#endif
#endif
            fireSanity(PINK_TOO_FAR_APART_FROM_COLOUR_BLOB);
            continue;
        }

        //double blobDist = ABS(pink->cx-hbArray[i]->cx) + ABS(pink->cy-hbArray[i]->cy);
        //double blobDist = ABS (pink->cy - hbArray[i]->cy);
        double blobDist = rdist;
        double blobRatio = hbArray[i]->area <
                            pink->area ? (double) hbArray[i]->area /
                            pink->area : (double) pink->area / hbArray[i]->area;
    	double beaconRatio = (pink->area + hbArray[i]->area) * 1.0 / 
			       (
                     (MAX(pink->getYMax(), hbArray[i]->getYMax()) - MIN(pink->getYMin(), hbArray[i]->getYMin())) *
                     (MAX(pink->getXMax(), hbArray[i]->getXMax()) - MIN(pink->getXMin(), hbArray[i]->getXMin()))
			       );
                            
        double dx = pink->cx - hbArray[i]->cx;
        double dy = pink->cy - hbArray[i]->cy;
				
	    //double angle = dx == 0 ? pi / 2 + effective_roll : atan (dy / dx) + effective_roll;
	    double angle = NormalizeAngle_180(RAD2DEG(atan2(dy,dx) + effective_roll));

        // int limit_dist = pink->half_perimeter + hbArray[i]->half_perimeter;
	    // this means beacons can be as far as one whole half beacon 's height part
	    // would cause problems, like joining pink half to other coloured blobs
	    // in the background
        // int limit_dist = (pink->getYMax () - pink->getYMin ()) + (hbArray[i]->getYMax () - hbArray[i]->getYMin ());
	    // so just try to half it ie cant be longer than half a beacon 
	    double limit_dist = ((pink->getYMax() - pink->getYMin() + 1) + (hbArray[i]->getYMax() - hbArray[i]->getYMin() + 1)); // *0.75

        if (blobDist <= limit_dist && blobRatio > 0.1 && abs (angle) > 60.0 && abs (angle) < 120.0 &&        // angle between the 2 blobs is between 60 and 120
            (blobDist * 2.0 / 3.0) < minDist && (blobRatio + beaconRatio) > maxRatio && beaconRatio > 0.3) {



//            wout << "Found match\n";

            // David:
            // check to see that there is not alot of green field on the bottom
            // half of the lower beacon
            //
            // the two half beacons needs be close enough, and have reasonable ratio
            const HalfBeacon *lower = pink;
            bool notFloor = true;

            if (hbArray[i]->cy > pink->cy) {
                lower = hbArray[i];
                //wout << "pink " << pink->cy << " is on top of " << hbArray[i]->cy << " \n";
            }
            
            // don't use the beacon if it is surroundered or contains alot of
            // green field colour
            int width = lower->boundary[EAST].intercept - lower->boundary[WEST].intercept;
            int height = lower->boundary[SOUTH].intercept -
            lower->boundary[NORTH].intercept;
            double threshold = width * height * beaconErrorThreshold + 1;        // +1 so you don't get zero

/*	misfiring too much now that green needs to be saturated for lines to work
            if (!NoColourXNearBy (cFIELD_GREEN,
                                  lower->boundary[WEST].intercept,
                                  lower->boundary[EAST].intercept,
                                  lower->boundary[NORTH].intercept,
                                  lower->boundary[SOUTH].intercept,
                                  beaconSearchPerim,
                                  beaconSearchPerim, (int) threshold)) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "beacon sanity: too much green field nearby" << endl;
#endif
#endif
                notFloor = false;
                //wout << "Lower blob avoided xmin,xmax,ymin,yman \n" << lower->boundary[WEST].intercept << ","
                //     << lower->boundary[EAST].intercept << ","
                //     << lower->boundary[NORTH].intercept << ","
                //     << lower->boundary[SOUTH].intercept << "\n";
                fireSanity(BEACON_SURROUNDERED_BY_GREEN);
            }
*/
            // ricky: remove really unproportioned .. unproportioned? or disproportioned? beacons
            // prevent noise from being recognised as beacons
            double proportion;
            if (max(pink->getXMax(), hbArray[i]->getXMax()) - min(pink->getXMin(), hbArray[i]->getXMin()) == 0) {
                proportion = LARGE_VAL;
                fireSanity(UNPROPROTIONAL_BEACON);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "beacon sanity: overly unproportioned blobs" << endl;
#endif
#endif
            } else {
                proportion = ((double) max(pink->getYMax(), hbArray[i]->getYMax()) - min(pink->getYMin(), hbArray[i]->getYMin())) / 
                    ((double) max(pink->getXMax(), hbArray[i]->getXMax()) - min(pink->getXMin(), hbArray[i]->getXMin()));
            }
			
            if (notFloor && proportion < 5 && proportion > 0.2) {
                maxRatio = beaconRatio + blobRatio;
                //maxRatio = blobRatio;
                minDist = blobDist;
                *closest = hbArray[i];
                *closestType = colorIndex;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "beacon sanity: found beacon " << colorIndex << endl;
#endif
#endif
            } else {
                fireSanity(BEACON_ON_FLOOR_OR_UNPROPROTION);
            }
        } 
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        else {
            (*outfile) << "beacon sanity: strange color " << colorIndex << " (" << hbArray[i]->cx << "," << hbArray[i]->cy << ")";
            (*outfile) << " angle btwn blob " << NormalizeAngle_180(RAD2DEG(atan2(dy, dx)));
            (*outfile) << " effective_roll " << RAD2DEG(effective_roll);
            (*outfile) << " angle " << angle;
            (*outfile) << " limitDist " << limit_dist << " blobDist " << blobDist;
            (*outfile) << " blobRatio " << blobRatio << " minDist " << minDist << " maxRatio " << maxRatio;
            (*outfile) << " beaconRatio " << beaconRatio << endl;
        }
#endif
#endif
    }
}

void
VisualCortex::LocateLandmarks () {
    int checkedBeacons[6] = { 0, 0, 0, 0, 0, 0 };

    // check for any pink in the shot. if there is no pink, then there cannot
    // be any beacons
    if (blobs->count (cBEACON_PINK) > 0) {
        numPink = blobs->findHalfBeacons (cBEACON_PINK, hbPink);
        numGreen = blobs->findHalfBeacons (cBEACON_GREEN, hbGreen);
    } else {
        numPink = 0;
        numGreen = 0;
    }

    // If there are no pink beacons, there can still be blue and yellow, i.e. goals
    // of course, we *hope* that the biggest blue and yellow beacons will
    // be the goals :)
    numBlue = blobs->findHalfBeacons (cBEACON_BLUE, hbBlue);
    numYellow = blobs->findHalfBeacons (cBEACON_YELLOW, hbYellow);

    // Check for matches
    for (int p = 0; p < numPink; p++) {
        // city block distance = mahattan distance = x+y
        double minDist = 255;
        double maxRatio = 0.0;
        int closestType = -1;
        HalfBeacon *closest = NULL;

        /*
         * Find the half beacon that matches the pink half beacon.
         */
        CheckMatchingBeacons (hbPink[p], cBEACON_GREEN, hbGreen, numGreen,
                              minDist, maxRatio, &closest, &closestType);

        CheckMatchingBeacons (hbPink[p], cBEACON_BLUE, hbBlue, numBlue,
                              minDist, maxRatio, &closest, &closestType);

        CheckMatchingBeacons (hbPink[p], cBEACON_YELLOW, hbYellow,
                              numYellow, minDist, maxRatio, &closest,
                              &closestType);

        // couldn't match that pink blob
        if (!closest) {
            fireSanity(CANNOT_MATCH_PINK_BLOB);
          continue;
        }
        // figure out which beacon we are dealing with, defaulting to
        // having pink on the bottom (ie the left side of the field)
        closest->matched = true;
        int beaconID = -1;
        switch (closestType) {
            case cBEACON_BLUE:
            beaconID = 0;
            break;
            case cBEACON_GREEN:
            beaconID = 2;
            break;
            case cBEACON_YELLOW:
            beaconID = 4;
        }

        // using 3dtransform to work out the elevation of the two
        // matching half beacon, and determine whether it's on the right
        // or left side.
        // Previously: it only checks the y-value of the two half-beacon
        // on the image, which doesn't make sense when the image is tilted
        // more than 90 degree. Therefore need to use 3dTransform.
        // Previous implementation will simply ignores the beacon when head
        // tilt is < -60 and ABS(pan) is > 60
        VisualObject closest_obj;
        VisualObject pink_obj;
        // note: the distance to the two halfbeacon doesnt matter here
        // because we only interested in the elevation.
        closest_obj.imgHead = PointToHeading (closest->cx);
        closest_obj.imgElev = PointToElevation (closest->cy);
        closest_obj.dist = 100;        // comparing elevation with uniform distance
        RobotRelative (closest_obj);

        pink_obj.imgHead = PointToHeading (hbPink[p]->cx);
        pink_obj.imgElev = PointToElevation (hbPink[p]->cy);
        pink_obj.dist = 100;        // comparing elevation with uniform distance
        RobotRelative (pink_obj);
        
		// otherwise, if pink is on top, the beacon is on the right hand side
        // of the field
        if (closest_obj.elev < pink_obj.elev) {
            beaconID++;
	}


		double pinkboxheight = hbPink[p]->getYMax() - hbPink[p]->getYMin();
		double closestboxheight = closest->getYMax() - closest->getYMin();
		
		// make it simple, only apply the fix if image is not too much rotated
        if (!(RAD2DEG (tilt) > 45 && ABS (RAD2DEG (pan)) > 60)) {
            if (closest->area > hbPink[p]->area) {
                // pink is much smaller, use closest's height as reference
                double h = closest->getYMax () - closest->getYMin ();

                if (closest_obj.elev < pink_obj.elev) {
                    // pink is on top, extends pink upwards
                    hbPink[p]->setYMin ((int)(hbPink[p]->getYMax () - h));
					hbPink[p]->cy3 = hbPink[p]->cy3 - ((closestboxheight-pinkboxheight) / 2.0);
                } else {
                    // pink is on bottom, extends pink downwards
                    hbPink[p]->setYMax ((int)(hbPink[p]->getYMin () + h));
					hbPink[p]->cy3 = hbPink[p]->cy3 + ((closestboxheight-pinkboxheight) / 2.0);
                }

                // some recalculation
				//ERIC this is the old method of centroid extention
				//still in there cause i still calculate the old centroids so that i can look at the difference
				//
                hbPink[p]->cy = (hbPink[p]->getYMin () + hbPink[p]->getYMax ()) / 2.0;
                hbPink[p]->half_perimeter = hbPink[p]->getXMax () - hbPink[p]->getXMin () +
                                            hbPink[p]->getYMax () - hbPink[p]->getYMin ();
            } else if (hbPink[p]->area > closest->area) {
                // closest is much smaller, use pink's height as reference
                double h = hbPink[p]->getYMax () - hbPink[p]->getYMin ();

                if (closest_obj.elev < pink_obj.elev) {
                    // pink is on top, extends closest downwards
                    closest->setYMax ((int)(closest->getYMin () + h));
					closest->cy3 = closest->cy3 + ((pinkboxheight-closestboxheight) / 2.0);
                } else {
                    // pink is on bottom, extends closest upwards
                    closest->setYMin ((int)(closest->getYMax () - h));
					closest->cy3 = closest->cy3 - ((pinkboxheight-closestboxheight) / 2.0);
                }

                // some recalculation
				//ERIC this is the old method of centroid extention
				//still in there cause i still calculate the old centroids so that i can look at the difference
				//
                closest->cy = (closest->getYMin () + closest->getYMax ()) / 2.0;
                closest->half_perimeter = closest->getXMax () - closest->getXMin () +
                                          closest->getYMax () - closest->getYMin ();
            }
        }

		
        // don't override old good beacon
        // **********************************
        // Form beacon,
        if (!checkedBeacons[beaconID]) {
            checkedBeacons[beaconID] = 1;
            beacon[beaconID].Set (*hbPink[p], *closest);

        } else {
        // skip the later sanity check (more efficient)
            continue;
        }

        // David: go through all the beacons to check that they are not
        // overlapping or too close
        for (int i = 0; i < 6 && filterInnerBeacon; i++) {
            if (beaconID == i)
                continue;
            if (beacon[i].valid &&
                 ((beacon[i].cx - beacon[i].width / 2 - filterInnerBeacon <=
                 beacon[beaconID].cx - beacon[beaconID].width / 2
                 && beacon[beaconID].cx - beacon[beaconID].width / 2 <=
                 beacon[i].cx + beacon[i].width / 2 + filterInnerBeacon)
                 || 
                 (beacon[i].cx - beacon[i].width / 2 -
                 filterInnerBeacon <=
                 beacon[beaconID].cx + beacon[beaconID].width / 2
                 && beacon[beaconID].cx + beacon[beaconID].width / 2 <=
                 beacon[i].cx + beacon[i].width / 2 +
                 filterInnerBeacon))
                 &&
                 ((beacon[i].cy - beacon[i].height / 2 -
                 filterInnerBeacon <=
                 beacon[beaconID].cy - beacon[beaconID].height / 2
                 && beacon[beaconID].cy - beacon[beaconID].height / 2 <=
                 beacon[i].cy + beacon[i].height / 2 + filterInnerBeacon)
                 || 
                 (beacon[i].cy - beacon[i].height / 2 -
                 filterInnerBeacon <=
                 beacon[beaconID].cy + beacon[beaconID].height / 2
                 && beacon[beaconID].cy +
                 beacon[beaconID].height / 2 <=
                 beacon[i].cy + beacon[i].height / 2 + filterInnerBeacon))) {
                beacon[beaconID].valid = false;
                fireSanity(BEACON_EITHER_OVERLAP_OR_TOO_CLOSE + i);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "beacon overlapping" << endl;
#endif
#endif
                break;
            }
        }
    }
}

void
VisualCortex::LookForGoal (HalfBeacon ** hbArray, int hbCount,
                           VisualObject & logoal,
                           VisualObject & rogoal, VisualObject & goal) {
    // Combine all the unmatched half_beacons into 1 big one
    HalfBeacon goal_blob;
    goal_blob.setXMin (100000);
    goal_blob.setYMin (100000);
    goal_blob.setXMax (0);
    goal_blob.setYMax (0);
    goal_blob.area = 0;

    // keep track of unrotated size
    int uxmin = 100000;
    int uymin = 100000;
    int uxmax = 0;
    int uymax = 0;    

    // This loop will look for the remaining un-matched beacon patches and
    // combine them to form a goal
    for (int b = 0; b < hbCount; b++) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
	        (*outfile) << hbArray[b]->cx << "," << hbArray[b]->cy << endl;
#endif
#endif	    

        if (hbArray[b]->matched) {
            //wout << "Sanity C\n";
            fireSanity(BLOB_ALREADY_MATCHED);
            continue;
        }
        
        if (hbArray[b]->area < 30) {
            //wout << "Sanity C\n";
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
	        (*outfile) << "goal sanity: area under 30" << endl;
#endif
#endif	    
            fireSanity(BLOB_AREA_LT30);
            continue;
        }

        VisualObject patch;

        int flag = 0;

        patch.imgHead = PointToHeading (hbArray[b]->cx, hbArray[b]->cy, sin_eroll, cos_eroll);
        patch.imgElev = PointToElevation (hbArray[b]->cx, hbArray[b]->cy, sin_eroll, cos_eroll);
	
	// ricky : throw out really high blobs
        if (RAD2DEG (effective_tilt) - patch.imgElev < -15) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
	        (*outfile) << "goal sanity: blob too high" << endl;
#endif
#endif	    
            fireSanity(BLOB_TOO_HIGH);
            continue;
        }

        for (int k = vobBlueLeftBeacon; k <= vobYellowRightBeacon; k++) {
            if (vob[k].cf > 200) {

                VisualObject beacon;

                beacon.imgHead = PointToHeading (vob[k].cx, vob[k].cy, sin_eroll, cos_eroll);
                beacon.imgElev = PointToElevation (vob[k].cx, vob[k].cy, sin_eroll, cos_eroll);

                if (patch.imgElev > beacon.imgElev + 5) { // +5 leniency due to image distortion
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "Goal blob is too high " << patch.imgElev << " beacon " << k << " " << PointToElevation(vob[k].cx, vob[k].cy, sin_eroll, cos_eroll) << "\n";
                    (*outfile) << "Goal " << hbArray[b]->cx << "," << hbArray[b]->cy << " Beacon " << vob[k].cx << " " << vob[k].cy << "sin_eroll cos_eroll " << sin_eroll << " " << cos_eroll << endl;
#endif
#endif	    
                    flag = 1;
                    fireSanity(BLOB_ELEVATION_GT_BEACON_ELEVATION+k-vobBlueLeftBeacon);
                    break;
                }
            }
        }

        if (flag) {
            continue;
        }
        // it's too small, so forget about it altogether
        if (((hbArray[b]->getXMax () + 1 - hbArray[b]->getXMin ()) < kGoalMinWidth)
             || ((hbArray[b]->getYMax () + 1 - hbArray[b]->getYMin ()) < kGoalMinHeight)) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
	        (*outfile) << "goal sanity: goal x - " << (hbArray[b]->getXMax () + 1 - hbArray[b]->getXMin ()) << " smaller than " << kGoalMinWidth;
		(*outfile) << " or goal y - " << (hbArray[b]->getYMax () + 1 - hbArray[b]->getYMin ()) << " smaller than " << kGoalMinHeight << endl;
#endif
#endif	    
            fireSanity(BLOB_TOO_SMALL);
            continue;
        }
        // a goal blob cannot be to the left of the left goal beacon
        // so we conclude it cannot be the goal
        //if (logoal.var < get95CF(40) && (logoal.x+(logoal.width/2)) > hbArray[b]->cx)
//        int rotatedLogoalx = (int) (cos_eroll *logoal.cx - sin_eroll * logoal.cy);
        int rotatedLogoalx = WIDTH/2 - (int) (cos_eroll *(WIDTH/2 - logoal.cx) - sin_eroll * (HEIGHT/2 - logoal.cy));
//        if (logoal.cf > 500 && (logoal.x + (logoal.width / 2)) > hbArray[b]->cx) {
        if (logoal.cf > 500 && rotatedLogoalx > hbArray[b]->cRx) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Goal Sanity A logoal threshold " << logoal.cx << " blob cx " << hbArray[b]->cx;
            (*outfile) << " rotated " << rotatedLogoalx << " blob cx " << hbArray[b]->cRx << endl;
#endif
#endif
            fireSanity(BLOB_LEFT_OF_LEFT_BEACON);
            continue;
        }
        // a goal blob cannot be to the right of the right beacon
        // so we conclude it cannot be the goal
        //if (rogoal.var < get95CF(40) && (rogoal.x+(rogoal.width/2)) < hbArray[b]->cx)
//        int rotatedRogoalx = (int) (cos_eroll *rogoal.cx - sin_eroll * rogoal.cy);
        int rotatedRogoalx = WIDTH/2 - (int) (cos_eroll *(WIDTH/2 - rogoal.cx) - sin_eroll * (HEIGHT/2 - rogoal.cy));
//        if (rogoal.cf > 500 && (rogoal.x + (rogoal.width / 2)) < hbArray[b]->cx) {
        if (rogoal.cf > 500 && rotatedRogoalx < hbArray[b]->cRx) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << rogoal.cx << "," << rogoal.cy << endl;
            (*outfile) << "Goal Sanity B rogoal threshold " << (rogoal.x + (rogoal.width / 2)) << " blob cx " << hbArray[b]->cx;
            (*outfile) << " rotated " << rotatedRogoalx << " blob cx " << hbArray[b]->cRx << endl;
#endif
#endif
            fireSanity(BLOB_RIGHT_OF_RIGHT_BEACON);
            continue;
        }

        // if there is too much white between the centroids of the blobs
        int blob_cy = (hbArray[b]->getYMin() + hbArray[b]->getYMax()) / 2;
        int blob_cx = (hbArray[b]->getXMin() + hbArray[b]->getXMax()) / 2;

        // too many background colours between the blobs.
	// shudnt be field in between either
	// also should use unrotated coordinates since accessing cplane
        if (goal_blob.area != 0
             && (drawLine ((uxmin+uxmax)/2, (uymin+uymax)/2, blob_cx, blob_cy, BACKGROUND) > 7  // ricky: experiment, remind
	     || drawLine ((uxmin+uxmax)/2, (uymin+uymax)/2, blob_cx, blob_cy, GREEN_FIELD) > 0)) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Goal sanity too  much bg color btwn blob" << endl;
            (*outfile) << (uxmin+uxmax)/2 << ", " << (uymin+uymax)/2 << " " << blob_cx << ", " <<  blob_cy << " " << endl;
#endif
#endif
            fireSanity(TOO_MUCH_BACKGROUND_COLOUR_BETWEEN_BLOBS);
            continue;
        }

      /*
        // if the bottom of this blob is too high from the current goal blob

        if (hbArray[b]->yRmax < goal_blob.getYMin() && goal_blob.getYMin() != 100000) {
            fireSanity(BLOB_BOTTOM_TOO_HIGH_FROM_CURRENT_GOAL_BLOB);
            continue;
        }

       */

/*        if (hbArray[b]->yRmin > goal_blob.yRmax) {
            goal_blob.setXMax(hbArray[b]->xRmax);
            goal_blob.setYMax(hbArray[b]->yRmax);
            goal_blob.setXMin(hbArray[b]->xRmin);
            goal_blob.setYMin(hbArray[b]->yRmin);
        } else { 
*/
        goal_blob.setXMax (MAX (goal_blob.getXMax (), hbArray[b]->xRmax));
        goal_blob.setYMax (MAX (goal_blob.getYMax (), hbArray[b]->yRmax));
        goal_blob.setXMin (MIN (goal_blob.getXMin (), hbArray[b]->xRmin));
        goal_blob.setYMin (MIN (goal_blob.getYMin (), hbArray[b]->yRmin));

        uxmin = MIN(uxmin, hbArray[b]->getXMin());
        uymin = MIN(uymin, hbArray[b]->getYMin());
        uxmax = MAX(uxmax, hbArray[b]->getXMax());
        uymax = MAX(uymax, hbArray[b]->getYMax());

        //}
        //cout << "hbArea: " << hbArray[b]->area << endl;
        goal_blob.area += hbArray[b]->area;

      /*
         // patch for Osaka game
         // if this passes it is the goal, so break
         if (goal_blob.area > 0)
         break;
       */
    }

    if (goal_blob.area < kGoalMinSize) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Goal area less than minimum goal size" << endl;
#endif
#endif
        fireSanity(BLOB_AREA_LESS_THAN_MINIMUM_GOAL_SIZE);
        return;
    }

    // take out goals of low density
    // remember that area is actually number of pixels
    double goal_density = (goal_blob.area * 1.0) / ((uymax - uymin + 1) * (uxmax - uxmin + 1));
    double goal_proportion = ((goal_blob.getXMax() - goal_blob.getXMin() + 1) * 1.0) / (goal_blob.getYMax() - goal_blob.getYMin() + 1);
    if (goal_density < 0.3 || goal_proportion < 0.2 || goal_proportion > 5.0) {
        fireSanity(BLOB_DENSITY_TOO_LOW);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Goal density too low " << goal_density << " or out of proportion " << goal_proportion << endl;
#endif
#endif
        return;
    } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "goal density passed " << goal_density << " proportion passed " << goal_proportion << endl;
#endif
#endif
    }        

    // a goal with small blob area not on the field? can't be!
    //if (NoGreenNearBy(goal_blob.getXMin (), goal_blob.getXMax (), goal_blob.getYMin (),
    //                  goal_blob.getYMax ()) && goal_blob.area < 7000) {
    if (NoGreenNearBy(uxmin, uxmax, uymin, uymax) && goal_blob.area < 7000) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "No green field around a small goal" << endl;
#endif
#endif
        fireSanity(NO_GREEN_NEAR_BLOB);
        return;
    }
    
    goal_blob.cx = (goal_blob.getXMax () + goal_blob.getXMin ()) / 2.0;
    goal_blob.cy = (goal_blob.getYMax () + goal_blob.getYMin ()) / 2.0;

    long xmin = goal_blob.getXMin ();
    long xmax = goal_blob.getXMax ();
    long ymin = goal_blob.getYMin ();
    long ymax = goal_blob.getYMax ();;
    double width = xmax + 1 - xmin;
    double height = ymax + 1 - ymin;

    // get rid of goals that has tiny width or height
    if (width < minGoalDim || height < minGoalDim) {
        goal.var = LARGE_VAL;
        goal.angleVar = LARGE_VAL;
        fireSanity(TINY_BLOB_WIDTH_OR_HEIGHT);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Goal sanity, dimension under minimum dimension" << endl;
#endif
#endif
        return;
    }
    
    int delta = 4;

    if (height <= 0)
    height = 1;

    if ((width / height) > kGoalDimensionRatio) {
        // need to expand the height
        double new_height = width / kGoalDimensionRatio;

        // what is delta??
        // give it some tolerance so that the edge of the
        // object doesnt have to be exactly on the edge of the screen
        if (ymax >= USABLE_HEIGHT - 1 - delta)        // object is on the bottom edge of the FOV
            ymax = ymin + (int)new_height;
        else if (ymin <= delta)        // object is on the top edge of the FOV
            ymin = ymax - (int)new_height;
        else {
// object's bbox is not in ratio?
            double offset;
            offset = (new_height - height) / 2.0;
            ymin -= (int)offset;
            ymax += (int)offset;
        }

        height = new_height;
    }
// commented out by Ben

//  else if (goal.width * 1.0 / goal.height < kGoalDimensionRatio)
//  {
//      // need to expand the width
//      double new_width = height * kGoalDimensionRatio;
//
//      if (xmax >= WIDTH - 1 - delta)      // object is on the right edge of the FOV
//          xmax = xmin + new_width;
//        else if (xmin <= delta)               // object is on the left edge of the FOV
//          xmin = xmax - new_width;
//        else {
//            // object's bbox is not in ratio?
//          double offset;
//          offset = (new_width - width) / 2.0;
//          xmin -= offset;
//          xmax += offset;
//          }
//
//          width = new_width;
//  }

  /*
   * Store all the visual object information regarding the goal in goal.
   */
    double lowest_goal_elev = PointToElevation(ymax);
    for (int k = vobBlueLeftBeacon; k <= vobYellowRightBeacon; k++) {
        if (vob[k].cf > 200) {
            double beacon_elev = PointToElevation (vob[k].cx, vob[k].cy, sin_eroll, cos_eroll);
            if (lowest_goal_elev > beacon_elev) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "Goal is too high " << lowest_goal_elev << " beacon " << k << " " << beacon_elev << "\n";
#endif
#endif	    
                goal.cf = 0;
                return;
            }
        }
    }

    goal.x = xmin;
    goal.y = ymin;
    goal.width = width;
    goal.height = height;

    goal.cx = (xmin + xmax) / 2.0;
    goal.cy = (ymin + ymax) / 2.0;

    // apply clippings
    long pxmin = (long) MAX (xmin, 0);
    long pxmax = (long) MIN (xmax, WIDTH);
    long pymin = (long) MAX (ymin, 0);
    long pymax = (long) MIN (ymax, HEIGHT);

    goal.projected_x = pxmin;
    goal.projected_y = pymin;
    goal.projected_w = pxmax - pxmin;
    goal.projected_h = pymax - pymin;

    goal.area = goal_blob.area;
    goal.tilt = 0;
    goal.misc = 0;

    double goalCF = 0;
    int sanities = 0;

    // the goal cannot be to the left of the left goal beacon
    // so we conclude it cannot be the goal
    if (logoal.cf > 500 && (logoal.x + (logoal.width / 2)) < goal_blob.cx) {
        //if (logoal.var < get95CF(20) && (logoal.x+(logoal.width/2)) < goal_blob.cx) {
        goalCF += logoal.cf;
        //goalVar += logoal.var;
        sanities++;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Goal sanity beacon" << endl;
#endif
#endif
        fireSanity(BLOB_LEFT_OF_LEFT_BEACON2);
    }
    // the goal cannot be to the right of the right blue beacon
    // so we conclude it cannot be the goal
    if (rogoal.cf > 500 && (rogoal.x + (rogoal.width / 2)) < goal_blob.cx) {
        //if (rogoal.var < get95CF(20) && (rogoal.x+(rogoal.width/2)) < goal_blob.cx) {
        goalCF += rogoal.cf;
        //goalVar += rogoal.var;
        sanities++;
        fireSanity(BLOB_RIGHT_OF_RIGHT_BEACON2);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Goal sanity beacon" << endl;
#endif
#endif
    }
    //goal.imgHead = PointToHeading(goal.cx);
    //goal.imgElev = PointToElevation(goal.cy);

    //goal.imgHead = PointToHeading (goal.cx, goal.cy, usin_eroll, ucos_eroll);
    //goal.imgElev = PointToElevation (goal.cx, goal.cy, usin_eroll, ucos_eroll);

    // rotating image back to normal
    int xOrigin = WIDTH/2;
    int yOrigin = HEIGHT/2;
    int uGoalX = (int)(xOrigin - (ucos_eroll*(xOrigin - goal.cx) - usin_eroll*(yOrigin - goal.cy)));
    int uGoalY = (int)(yOrigin - (usin_eroll*(xOrigin - goal.cx) + ucos_eroll*(yOrigin - goal.cy)));
    goal.imgHead  = PointToHeading(uGoalX);
    goal.imgElev  = PointToElevation(uGoalY);

    // Add the variance for the goal Tim Tam 13/03/2002
    goal.var = LARGE_VAL;
    goal.angleVar = LARGE_VAL;

#ifdef HIGH
    //David: new carlibration done, which also included variances
    //**********************************
    //hyperbola again for distance based on height of goal
    //whole goal is usually obsucred but height stays more constant
    //calibrated by martin and tom
    //plus dist_neck2snout to make distance neck relative
    //goal.dist = 4350.0/goal.height + 10;
    double expectedArea = goal.width * goal.height;

#ifdef SHAPE_CHALL

    int elev_height = (goal.y + goal.height);

    double offsetDeg = -0.0529528 + 0.0007467 * elev_height;

    double alpha2 = DEG2RAD (PointToElevation (elev_height));

    double beta = 0;

    beta = effective_tilt - alpha2 - (offsetDeg);

    // stops division by zero
    if (beta == 0)
        beta = 0.00001;
    goal.dist = 15.0 / tan (beta);

#endif
#ifndef SHAPE_CHALL
    //goal.dist = 5192.11 / (goal.height - 4.1947);
    goal.dist = (double)goalDistanceLookup[(int)CLIP(goal.height,143.0)] + 10;  //add 10 since the calibrated measurements go to the mouth of the goal which is 10cm from the start of the goal color itself. 


#endif
    //goal.var  = exp(2*exp(0.436724 + 0.0007573*goal.dist))/
    //            pow(goal.area/expectedArea,errorScale);
    goal.var = exp (2 * exp (0.436724 + 0.0007573 * goal.dist)) +
               SQUARE (varConst * (1 - goal.area / expectedArea));
    goal.angleVar = goal.var;

    //wout << "goal dist, var: " << goal.dist << ", " << goal.var << "\n";
    //wout << "x,y,width,height: " << goal.x << ", " << goal.y << ", " << goal.width << ", " << goal.height << "\n";

    // there are no beacons that we can cross-ref against, so we will have to reduce
    // our certainty by a bit
    if (!sanities) {
        goalCF = 666;
        goal.var *= 2;                // this is an adhoc value
        goal.angleVar *= 2;
    }
    // figure out our new confidence factor.
    if (sanities <= 0)
        sanities = 1;

    goal.cf = (int)(goalCF / sanities);
#endif

#ifdef MED
    int xmaybe, ymaybe;
    xmaybe = goal.x <= 0 || (goal.x + goal.projected_w) >= WIDTH - 1;
    ymaybe = goal.y <= 0 || (goal.y + goal.projected_h) >= USABLE_HEIGHT - 1;

    if (xmaybe) {
        if (ymaybe)
            goal.dist = imin (SizeToDistance (29, goal.projected_h),
                              SizeToDistance (60, goal.projected_w));
        else
            goal.dist = SizeToDistance (29, goal.projected_h);
    } else {
        if (ymaybe)
            goal.dist = SizeToDistance (60, goal.projected_w);
        else
            goal.dist = SizeToDistance (29, goal.projected_h);
    }                                // yesyes, optimisation opportunity neglected - it's easier to see this way
#endif

    // modified by Spencer confidence is capped at 1000
    goal.cf = MIN (goal.cf, 1000);

    if (goal.cf || goal.var != LARGE_VAL) {
        RobotRelative (goal);

        goal.lhead = goal.head + (PointToHeading (xmin) - PointToHeading (goal.cx));
        goal.rhead = goal.head + (PointToHeading (xmin + width) - PointToHeading (goal.cx));
        
        
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "goal found, elev " << goal.elev << " dist " << goal.d << endl;
#endif
#endif
//        wout << "left: " << goal.lhead << " mid: " << goal.head << " right: " << goal.rhead << "\n";

    }

    // check the HEIGHT of the bottom of goal above head, similar to checking for height of beacon
    VisualObject temp = goal;
    temp.imgElev = PointToElevation(ymax);
    if (tan (DEG2RAD (temp.elev)) * temp.dist > 25.0) {
        goal.cf = 0;
        goal.var = LARGE_VAL;
        goal.angleVar = LARGE_VAL;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "goal sanity, goal too high ";
        (*outfile) << " elev " << temp.elev << " dist " << temp.dist;
        (*outfile) << " height " << tan (DEG2RAD (temp.elev)) * temp.dist << endl;
#endif
#endif
    } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        (*outfile) << "goal sanity, passed ";
        (*outfile) << " elev " << temp.elev << " dist " << temp.dist;
        (*outfile) << " height " << tan (DEG2RAD (temp.elev)) * temp.dist << endl;
#endif
#endif
	}
}


void VisualCortex::Beacon::Set (const HalfBeacon & pink, const HalfBeacon & closest) {
	valid = true;
	cx = (pink.cx + closest.cx) / 2.0;
	cy = (pink.cy + closest.cy) / 2.0;
	area = pink.area + closest.area;
	halfmass = pink.area > closest.area ? pink.area : closest.area;
	//old method of centroid and beacon extention
	c2c =
	sqrt (SQUARE (pink.cx - closest.cx) + SQUARE (pink.cy - closest.cy));

	//new method of centroid
	c2c2 =
	sqrt (SQUARE (pink.cx2 - closest.cx2) + SQUARE (pink.cy2 - closest.cy2));

	//new method of centrod + beacon extention
	c2c3 =
	sqrt (SQUARE (pink.cx3 - closest.cx3) + SQUARE (pink.cy3 - closest.cy3));

	//!!ERIC centroid debugging stuff
	pinkXsum = pink.xsum;
	pinkYsum = pink.ysum;
	pinkarea = pink.area;
	closestXsum = closest.xsum;
	closestYsum = closest.ysum;
	closestarea = closest.area;


	angle =
	NormalizeAngle_0_360 (RAD2DEG
						  (atan2
						   (pink.cy - closest.cy, pink.cx - closest.cx)));
	width =
	(pink.getXMin () <
	 closest.getXMin ()) ? (closest.getXMax () - pink.getXMin () +
							1) : (pink.getXMax () -
								  closest.getXMin () + 1);
	height =
	pink.getYMin () <
	closest.getYMin () ? closest.getYMax () - pink.getYMin () +
	1 : pink.getYMax () - closest.getYMin () + 1;
}

void VisualObject::setDistance (double myX, double myY, double myZ) {
	objectPosition[0] = myX;
	objectPosition[1] = myY;
	objectPosition[2] = myZ;
	//dist=sqrt(SQUARE(myX) + SQUARE(myY) + SQUARE(myZ));
	dist = sqrt (SQUARE (myX) + SQUARE (myY) + SQUARE (myZ));
	d = sqrt (SQUARE (myX) + SQUARE (myZ));
	h = head = RAD2DEG (atan2 (myX, myZ));
	double temp1 = sqrt (SQUARE (myX) + SQUARE (myZ));
	double temp2 = dist;
	double temp3 = acos (temp1 / temp2);
	if (myY >= 0) {
		elev = RAD2DEG (temp3);
	} else {
		elev = -RAD2DEG (temp3);
	}
}

double VisualObject::getFlatDistance () const {
	return sqrt (SQUARE (objectPosition[2]) + SQUARE (objectPosition[0]));
}

void VisualObject::Print () {
	if (!cf) {
		cout << "\tVOB is invalid" << endl;
	} else {
		this->printPlayData ();
		cout << "\tBBox:       (" << x << "," << y << ")-(" << x +
		  width << "," << y + height << ")" << endl;
		cout << "\tBBox H:     " << height << endl;
		cout << "\tBBox W:     " << width << endl;

		cout << "\tProjBox:    (" << projected_x << "," << projected_y
		  << ")-(" << projected_x +
		  projected_w << "," << projected_y + projected_h << ")" << endl;
		cout << "\tProjBox H:  " << projected_h << endl;
		cout << "\tProjBox W:  " << projected_w << endl;

	  }
}
void VisualObject::printPlayData () {
	if (cf) {
		cout << "\tConfidence: " << cf << endl;
		cout << "\tCentroid:   (" << cx << "," << cy << ")" << endl;
		cout << "\tImage Head: " << imgHead << endl;
		cout << "\tImage Elev: " << imgElev << endl;
		cout << "\tArea:       " << area << endl;
		cout << "\tHeading:    " << head << endl;
		cout << "\tElevation:  " << elev << endl;
		cout << "\tDistance:   " << dist << endl;
	}
}

void VisualObject::Init () {
	cf = 0;
	var = LARGE_VAL;
	cx = 0;
	cy = 0;
	x = y = height = width = 0;
	projected_h = 0;
	projected_w = 0;
	area = 0;
	imgHead = 0;
	imgElev = 0;
	tilt = 0;
	head = 0;
	elev = 0;
	dist = 0;
	misc = 0;
	pad = 0;
	lhead = 0;
	rhead = 0;
}

bool VisualCortex::NoGreenNearBy (int xmin, int xmax, int ymin, int ymax) {
	int min_delta = 7;
	int delta_x = MAX (min_delta, (xmax - xmin) / 3);
	int delta_y = MAX (min_delta, (ymax - ymin) / 3);
	return NoColourXNearBy (cFIELD_GREEN, xmin, xmax, ymin, ymax,
							delta_x, delta_y, kFieldGreenThreshold);
}

double getAng (double dx, double dy) {
    double t;

    if (!dy)
        return (dx >= 0) ? 0 : 180;
    if (!dx)
        return (dy >= 0) ? 90 : 270;

    t = atan ((int) abs ((int) dy) / dx) * (double) (180 / 3.14159);
    if (dx < 0)
        t += 180;
    return (dy >= 0) ? t : 360 - t;
}

int imax (int a, int b) {
    if (a > b)
        return a;
    else
        return b;
}

int imin (int a, int b) {

    if (a < b)
        return a;
    else
        return b;
}

int pi_abs (int a) {
    if (a < 0)
        return -a;
    else
        return a;
}

// Calculates Image Heading
double
PointToHeading (double x) {
    return RAD2DEG (atan (((WIDTH / 2.0) - x) * 2.0 * tan (xField / 2.0) / WIDTH));
}

// Calculates Image Elevation
double PointToElevation (double y) {
    return RAD2DEG (atan (((HEIGHT / 2.0) - y) * 2.0 * tan (yField / 2.0) / HEIGHT));
}

double
PointToHeading (double x, double y, double sin_eroll, double cos_eroll) {
    return RAD2DEG (atan((
                         (WIDTH / 2.0 - x) * cos_eroll +
                         (HEIGHT / 2.0 - y) * sin_eroll) * 2.0 * tan (xField / 2.0) / WIDTH));
}

// Calculates Image Elevation
double PointToElevation (double x, double y, double sin_eroll, double cos_eroll) {
    return RAD2DEG (atan((
                         (WIDTH / 2.0 - x) * sin_eroll +
                         (HEIGHT / 2.0 - y) * cos_eroll) * 2.0 * tan (yField / 2.0) / HEIGHT));
}

// arguments are in centimetres and pixels
long SizeToDistance (long physicalSize, long apparentSize) {
// OK, when the chessboard was 495mm away, it appeared to be
// 60 by 54 pixels in size. The chessboard is IRL 325 by 330mm

// Time for averages here... chessboard is 33cm. Its image
// at a distance of 50cm appeared over 57 pixels.

// Say height is 1cm. Then at 50cm away it would take up
// 57/33 pixels. i.e. apparent = sz/z, or 57/33 = sz/50.
// So sz ('distance' to screen) is (50*57)/33.
// In general eqn is apparent = physical*sz/z.
// So now z = (physical*sz)/apparent

    if (!apparentSize)
        return 400;                        // a long way away

     return (long)((physicalSize * DIST_CONST) / apparentSize);
}

double
VisualCortex::NewBallDistance () {
    // assume the image is pretty much horizontally alligned
    double elev_of_top_of_ball = effective_tilt + DEG2RAD (-PointToElevation (vob[vobBall].misc));

    // ball above horizon... abnormal ball position
    if (elev_of_top_of_ball < 0)
        return 999;

    double height_snout = NECKBASE_HEIGHT + NECK_LENGTH * cos_tilt -
                          FACE_LENGTH * sin_tilt * cos_pan;
    double similar_triangle_height_at_centre_of_ball =
                            BALL_RADIUS / cos (elev_of_top_of_ball) + BALL_RADIUS;
    double distance_snout_2_ball = (height_snout - 
                                   similar_triangle_height_at_centre_of_ball) / 
                                   tan (elev_of_top_of_ball);
/*
    cout << "elev_of_top_of_ball: " << elev_of_top_of_ball << endl;
    cout << "height_snout: " << height_snout << endl;
    cout << "similar_triangle_height_at_centre_of_ball: " << similar_triangle_height_at_centre_of_ball << endl;
    cout << "distance_snout_2_ball: " << distance_snout_2_ball << endl;
    cout << "dist_neck2snout: " << dist_neck2snout << endl;
*/
    return distance_snout_2_ball;        //(distance_snout_2_ball + dist_neck2snout);
}

// convert VisualObject into robot relative
// note: rewritten for high resolution + new distance calcultion, MED might not work
// using 3D transformation to convert coordinate space.
// Using standard Rotation and Translation matrices.
// Reference for the Math Calucalation obtained from Computer Graphics comp3421 lecture.
void
VisualCortex::RobotRelative (VisualObject & object, bool dp) {
    double ihead = DEG2RAD (object.imgHead);
    double ielev = DEG2RAD (object.imgElev);
    double x1 = 0;
    double y1 = 0;
    double z1 = object.dist;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB    
    if (dp)
        (*outfile) << "1: " << x1 << " " << y1 << " " << z1 << endl;
#endif
#endif        

    // Rotation for Image Heading
    double x2 = x1 * cos (ihead) + z1 * sin (ihead);
    double y2 = y1;
    double z2 = -x1 * sin (ihead) + z1 * cos (ihead);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB    
    if (dp)
        (*outfile) << "2: " << x2 << " " << y2 << " " << z2 << endl;
#endif
#endif        

    // Rotation for Image Elevation
    double x3 = x2;
    double y3 = y2 * cos (ielev) + z2 * sin (ielev);
    double z3 = -y2 * sin (ielev) + z2 * cos (ielev);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB    
    if (dp)
        (*outfile) << "3: " << x3 << " " << y3 << " " << z3 << endl;
#endif
#endif        

    // Translation for Face Length
    double x4 = x3;
    double y4 = y3;
    double z4 = z3 + FACE_LENGTH;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB    
    if (dp)
        (*outfile) << "4: " << x4 << " " << y4 << " " << z4 << endl;
#endif
#endif        

    // Rotation for Head Pan
    double x5 = x4 * cos_pan + z4 * sin_pan;
    double y5 = y4;
    double z5 = -x4 * sin_pan + z4 * cos_pan;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB    
    if (dp)
        (*outfile) << "5: " << x5 << " " << y5 << " " << z5 << endl;
#endif
#endif        

    // Translation for Neck Length
    double x6 = x5;
    double y6 = y5 + NECK_LENGTH;
    double z6 = z5;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB    
    if (dp)
        (*outfile) << "6: " << x6 << " " << y6 << " " << z6 << endl;
#endif
#endif        

    // Rotation for Head Tilt
    double x7 = x6;
    double y7 = y6 * cos_tilt - z6 * sin_tilt;
    double z7 = y6 * sin_tilt + z6 * cos_tilt;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB    
    if (dp)
        (*outfile) << "7: " << x7 << " " << y7 << " " << z7 << endl;
#endif
#endif        
    // Calculate actual distance and robot relative heading and elevation
    object.setDistance (x7, y7, z7);
}

bool
VisualCortex::DarkBackgroundInBetween(int left, int right, int top, int bottom) {

    int num  = 0;
    int th   = 144;
    int xMin = 176, xMax = 0, yMin = 144, yMax = 0;
    for (int y = top; y >= bottom; y--) {
        for (int x = left; x <= right; x++) {
			//int index = y * WIDTH + x;
			//if(index >= PIXELS) {
			//	cout << "y=" << y << " x=" << x << " (" << index << ">=" << PIXELS << ")" << endl;
			//}
            //if (cplane[y * WIDTH + x] == cFIELD_GREEN) { this was old when i did an up - Ross
            
            // if background i.e. robot grey or black
            if (cplane[y * WIDTH + x] == ROBOT_GREY || cplane[y * WIDTH + x] == BLACK) {
                num++;
                
                xMin = MIN(xMin,x);
                xMax = MAX(xMax,x);
                yMin = MIN(yMin,y);
                yMax = MAX(yMax,y);
                
                double density = (double)num / ((xMax-xMin)*(yMax-yMin));

                if (num >= th && density > 0.8)
                    return true;
            }
        }
    }

    return false;
}


void
VisualCortex::FindColorRobot (int color) {
    VisualObject *robot;
    int vobNum;
    int numRobots = 0;
    int totalArea = 0;
    int minArea;
    int maxBlob = 0;

    // numOfBlobs associaled with a particular robot
    int numberOfBlobs = 0;

    if (color == cROBOT_RED) {
        minArea = kMinRedRobotSize;
        robot = &vob[vobRedDog];
        vobNum = vobRedDog;
    } else if (color == cROBOT_BLUE) {
        minArea = kMinBlueRobotSize;
        robot = &vob[vobBlueDog];
        vobNum = vobBlueDog;
    } else {
        cout << "Error: invalid visual object" << endl;
        return;
    }
    // find the bottom patch

    int buffer = 0;
    // finding the leftmost adjointing blob

    //wout << blobs->count(color) << " ";

    int xMin = 176, xMax = 0, yMin = 144, yMax = 0;
    int uxMin = 176, uxMax = 0, uyMin = 144, uyMax = 0;

    //wout << maxIndex << "\n";
	
	// Ross's archaelogical dig - this gathers blobs for each dog then processes each dog and
	// hence we must hold 4 times as much information, in future would like to gather blobs for
	// each dog and process the bunch of blobs before gathering next bunch
	int area[ROBOTS_EACH_COLOUR];
	double data[ROBOTS_EACH_COLOUR][3*MAX_BLOBS_PER_DOG+1];

    for (int i = 0; i < blobs->count (color); i++) {
        //wout << blobs->getBlob(color,i).area << " ";
        if (blobs->getBlob (color, i).area > 0) {
        
            int top    = MAX(yMax, blobs->getBlob (color, i).yRmax);
            int bottom = MIN(yMin, blobs->getBlob (color, i).yRmin);
	    
            int utop   = MAX(uyMax, blobs->getBlob (color, i).getYMax());
            int ubottom= MIN(uyMin, blobs->getBlob (color, i).getYMin());

            if (blobs->getBlob (color, i).xRmin <= xMin
                || blobs->getBlob (color, i).xRmin <= xMax + buffer
                || (infaredDistance/10000.0 < 15 && DarkBackgroundInBetween(blobs->getBlob (color, i).getXMin(), uxMax,
		utop, ubottom))
                ) {

                //BlobInfo& blob = blobs->getBlob(color,i);
                xMin = MIN (xMin, blobs->getBlob (color, i).xRmin);
                xMax = MAX (xMax, blobs->getBlob (color, i).xRmax);
                yMin = MIN (yMin, blobs->getBlob (color, i).yRmin);
                yMax = MAX (yMax, blobs->getBlob (color, i).yRmax);

                uxMin = MIN (uxMin, blobs->getBlob (color, i).getXMin());
                uxMax = MAX (uxMax, blobs->getBlob (color, i).getXMax());
                uyMin = MIN (uyMin, blobs->getBlob (color, i).getYMin());
                uyMax = MAX (uyMax, blobs->getBlob (color, i).getYMax());

                robot->cf = 1000;
            } else {
                xMin = blobs->getBlob (color, i).xRmin;
                xMax = blobs->getBlob (color, i).xRmax;
                yMin = blobs->getBlob (color, i).yRmin;
                yMax = blobs->getBlob (color, i).yRmax;

                uxMin = blobs->getBlob (color, i).getXMin();
                uxMax = blobs->getBlob (color, i).getXMax();
                uyMin = blobs->getBlob (color, i).getYMin();
                uyMax = blobs->getBlob (color, i).getYMax();

                if (color == cROBOT_RED) {
                    if (vobNum < 12) {

                        robot = &vob[++vobNum];
                        robot->cf = 1000;
                        maxBlob = 0;
                        numRobots++;
                        totalArea = 0;
                        numberOfBlobs = 0;
                    } else {
                        break;
                    }
                } else if (color == cROBOT_BLUE) {
                    if (vobNum < 16) {

                        robot = &vob[++vobNum];
                        robot->cf = 1000;
                        maxBlob = 0;
                        numRobots++;

                        totalArea = 0;
                        numberOfBlobs = 0;
                    } else {
                        break;
                    }
                }

            }

            robot->x = xMin;
            robot->y = yMin;
            robot->width = xMax - xMin;
            robot->height = yMax - yMin;
			
			if(numberOfBlobs >= MAX_BLOBS_PER_DOG)
				continue;

			// Ross - what the hell?
            totalArea += blobs->getBlob (color, i).area;
            area[numRobots] = totalArea;

            data[numRobots][numberOfBlobs * 3 + 1] =
                        (blobs->getBlob (color, i).getXMin () +
                         blobs->getBlob (color, i).getXMax ());
            data[numRobots][numberOfBlobs * 3 + 2] =
                        (blobs->getBlob (color, i).getYMin () +
                         blobs->getBlob (color, i).getYMax ());
            data[numRobots][numberOfBlobs * 3 + 3] =
                        blobs->getBlob (color, i).area;
            numberOfBlobs++;
            data[numRobots][0] = numberOfBlobs;

/*              if (maxBlob < blobs->getBlob(color,i).area) {
                info[numRobots][3] = blobs->getBlob(color,i).cRx;
                info[numRobots][4] = blobs->getBlob(color,i).cRy;
                maxBlob = blobs->getBlob(color,i).area;
            }*/

            double hor = 26.0;
            double vert = 18.0;
            buffer = (int) MAX (MIN (((double) robot->height / vert) * hor - robot->width,
                             20), 10);
        }
    }

    if (color == cROBOT_RED) {
        vobNum = vobRedDog;

    } else if (color == cROBOT_BLUE) {
        vobNum = vobBlueDog;

    }

	// Ross's archaelogical dig - if ROBOTS_EACH_COLOUR is not 4 things can go very wrong ....
	// I leave this to future diggers ...
    for (int i = vobNum, robotNum = 0; i < vobNum + ROBOTS_EACH_COLOUR; i++, robotNum++) {

        robot = &vob[i];

        if (robot->cf != 0) {
            robot->cx = robot->x + robot->width / 2.0;
            robot->cy = robot->y + robot->height / 2.0;
            robot->cf = (int)(robot->height * 1000 / 144.0);

            robot->projected_x = robot->x;
            robot->projected_y = robot->y;
            robot->projected_w = robot->width;
            robot->projected_h = robot->height;
            robot->tilt = 0;        // not being used
            // rotating the image back to original position for
            // robot relative to work correctly
            int xOrigin = WIDTH/2;
            int yOrigin = HEIGHT/2;
            int uRobotX = (int)(xOrigin - (ucos_eroll*(xOrigin - robot->cx) - usin_eroll*(yOrigin - robot->cy)));
            int uRobotY = (int)(yOrigin - (usin_eroll*(xOrigin - robot->cx) + ucos_eroll*(yOrigin - robot->cy)));
            robot->imgHead  = PointToHeading(uRobotX);
            robot->imgElev  = PointToElevation(uRobotY);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "robot unrotated (" << uRobotX << "," << uRobotY << ") imgElev " << robot->imgElev << endl;
#endif
#endif
            

            //robot->imgHead = PointToHeading (robot->cx, robot->cy, usin_eroll, ucos_eroll);
            //robot->imgElev = PointToElevation (robot->cx, robot->cy, usin_eroll, ucos_eroll);
            //robot->imgHead  = PointToHeading(robot->cx);
            //robot->imgElev  = PointToElevation(robot->cy);

            if (area[robotNum] != 0) {

                // normalize data
                for (int i = 0; i < (int) data[robotNum][0]; i++) {
                    data[robotNum][i * 3 + 1] =
                        (data[robotNum][i * 3 + 1] -
                         2.0 * robot->cx) / (double) (robot->width + 1);
                    data[robotNum][i * 3 + 2] =
                        (data[robotNum][i * 3 + 2] -
                         2.0 * robot->cy) / (double) (robot->height + 1);
                    data[robotNum][i * 3 + 3] =
                        data[robotNum][i * 3 + 3] / (double) area[robotNum];

                }

                double min_distance = 10000.0;
                int min_index = -1;
                double distance_to_node;

                // Only when we have enough data to learn
                if ((int) data[robotNum][0] > 1) {
                    for (int i = 0; i < trainingExpCount; i++) {
                        /* Pick a training example
                         * Calculate its matching to this pattern
                         */
                        double total_distance = 0.0;
                        for (int j = 0; j < (int) data[robotNum][0]; j++) {
                        // For each blob int the current robot
                            double node_min = 1000.0;
                            for (int k = 0; k < trainingData[i][1]; k++) {
                                // Scan all node in the training examples
                                       distance_to_node =
                                    ABS (data[robotNum][3 * j + 1] -
                                         trainingData[i][3 * k + 2]) +
                                    ABS (data[robotNum][3 * j + 2] -
                                         trainingData[i][3 * k + 3]) +
                                    ABS (data[robotNum][3 * j + 3] -
                                         trainingData[i][3 * k + 4]);

                                if (distance_to_node < node_min)
                                    node_min = distance_to_node;
                            }
                            total_distance += node_min;

                        }

                        if (total_distance < min_distance) {
                            min_distance = total_distance;
                            min_index = i;
                        }

                    }                // End  for loop - all training examples
                }
                if (min_index >= 0) {
                    robot->f = trainingData[min_index][0];

                } else {
                    robot->f = 0;
                }
            }

            int height = (int) (robot->y + robot->height);

            double offsetDeg = -0.0529528 + 0.0007467 * height;

            double alpha2 = DEG2RAD (PointToElevation (height));

            double beta = 0;

            beta = effective_tilt - alpha2 - (offsetDeg);

            // new equation as of 4/4
            //log(dist) = 0.727567 - 1.70424 log(beta) - 0.382294 log(beta)**2
            //robot->dist = pow(10.0,0.727567 - 1.70424*log10(beta) - 0.382294*pow(log10(beta),2 ));
//            wout << vobNum << " " << robot->x << " " << robot->y << " " << robot->width << " " << robot->height << "\n";
            //robot->misc = beta;

            // effective range of blue recongnition 134.357827
            // effective range of red recongnition 241.0810462

#ifdef  BAR_CHALL
            // calculate the distance to the end of the bar
            robot->dist = 15.0 / tan (beta);
            robot->var = 10;
#endif
#ifndef  BAR_CHALL

            if (color == cROBOT_RED) {
//                robot->var = pow(1.21742 - 2.20734 * beta,2);

                beta = MAX (0.007, beta);

                //recalibrated 5/4 using robot 7
                robot->dist =
                    pow (10.0,
                         0.890680 - 1.37435 * log10 (beta) -
                         0.316338 * pow (log10 (beta), 2));
                //robot->dist = closer + (further - closer)/2;
                //if (robot->dist <= 240) {
                double closer = pow (10.0,
                                     0.890680 - 1.37435 * log10 (beta +
                                                                 2 *
                                                                 yField /
                                                                 HEIGHT)
                                -
                                0.316338 *
                                pow (log10(beta + 2 * yField / HEIGHT), 2));
                double further = pow (10.0,
                                      0.890680 - 1.37435 * log10 (beta -
                                                                yField
                                                                /
                                                                HEIGHT)
                                -
                                0.316338 *
                                pow (log10 (beta - yField / HEIGHT), 2));
                robot->var = pow ((further - closer) / 2 + 10, 2);
                robot->angleVar = robot->var;
                    //wout << closer << " " << further << " ";
                    //wout << robot->var << " ";
		    /*
                } else if (robot->dist > 240) {
                    robot->var = LARGE_VAL;
                    robot->angleVar = robot->var;
                    fireSanity(RED_ROBOT_DISTANCE_GT240);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "red robot too far" << endl;
#endif
#endif
                }
		*/

            } else if (color == cROBOT_BLUE) {

                beta = MAX (0.017, beta);

                //robot->var = 1.21742 - 2.20734 * beta;
                //robot->var = pow(pow(10.0,0.142790 - 0.748878*log10(beta)),2);
                //recalibrated 5/4 using robot 7
                robot->dist = pow (10.0,
                                   0.879319 - 1.54566 * log10 (beta) -
                                   0.441522 * pow (log10 (beta), 2));
                //robot->dist = closer + (further - closer)/2;
                //if (robot->dist <= 165) {
                    double closer = pow (10.0,
                                         0.879319 - 1.54566 * log10 (beta +
                                                                   2 *
                                                                   yField /
                                                                   HEIGHT)
                                    -
                                    0.441522 *
                                    pow (log10(beta + 2 * yField / HEIGHT), 2));
                    double further = pow (10.0,
                                        0.879319 - 1.54566 * log10 (beta -
                                                                    yField
                                                                    /
                                                                    HEIGHT)
                                     -
                                     0.441522 * pow (log10 (beta - yField / HEIGHT), 2));
                    robot->var = pow ((further - closer) / 2 + 10, 2);
                    robot->angleVar = robot->var;
                    //wout << closer << " " << further << " ";
                /*
                } else if (robot->dist > 165) {
                    robot->var = LARGE_VAL;
                    robot->angleVar = LARGE_VAL;
                    fireSanity(BLUE_ROBOT_DISTANCE_GT165);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "blue robot too far" << endl;
#endif
#endif
                }
                */

            }
#endif
//            wout << vobNum << " " << robot->orient << " " << robot->dist << "\n";

            // distance using the height of the bounding box
            //robot->dist = (robot->dist + 859.18 * pow((double)robot->height,-0.768))/2;
            // old equation
            //robot->dist = pow(10.0,0.792326 - 1.5338*log10(beta) - 0.449760*pow(log10(beta),2));
            //wout << robot->orient << " ";
        }
    }
}

void
VisualCortex::RobotNoiseFilter () {

    for (int i = cROBOT_BLUE; i <= cROBOT_RED; i++) {
      //cout << "Filtering " << ColorName[i] << endl;

      for (int j = 0; j < blobs->count (i); j++) {
          BlobInfo & info = blobs->getBlob (i, j);
          VisualObject patch;
          VisualObject beacon;
          int flag = 0;

          //int cx = ((info.getXMin() + info.getXMax()) / 2.0) + 0.5;
          //int cy = ((info.getYMin() + info.getYMax()) / 2.0) + 0.5;

          //patch.imgHead = PointToHeading(cx, cy, sin_eroll, cos_eroll);
          //patch.imgElev = PointToElevation(cx, cy, sin_eroll, cos_eroll);
          patch.imgHead = PointToHeading (info.xRmax);
          patch.imgElev = PointToElevation (info.yRmax);

          //patch.dist = 100; // distance doesnt really matter here.

          //RobotRelative (patch);

          for (int k = vobBlueLeftBeacon; k <= vobYellowRightBeacon; k++) {
              if (vob[k].cf > 0) {
                  flag = 1;
                  beacon.imgHead =
                    PointToHeading (vob[k].cx, vob[k].cy, sin_eroll,
                                    cos_eroll);
                  beacon.imgElev =
                    PointToElevation (vob[k].cx, vob[k].cy, sin_eroll,
                                      cos_eroll);

                  beacon.dist = vob[k].dist;

                  // robot color pathces that are above the beacon
                  // kill the patch
                  //if (patch.elev > vob[k].elev)
                  //if (patch.imgElev > beacon.imgElev)
                  if (patch.imgElev >
                      PointToElevation (vob[k].x, vob[k].y, sin_eroll,
                                        cos_eroll)) {
                    if (i==cROBOT_BLUE) {
                        fireSanity(BLUE_BLOB_ABOVE_BEACON+k-vobBlueLeftBeacon);
                    } else if (i==cROBOT_RED) {
                        fireSanity(RED_BLOB_ABOVE_BEACON+k-vobBlueLeftBeacon);
                    }
                      info.ResetBlob ();
                    }
                }
            }

          for (int k = vobBlueGoal; k <= vobYellowGoal; k++) {
              if (vob[k].cf > 0) {
                  // robot color pathces that are above the beacon
                  // kill the patch
		  flag = 1;
                  if (patch.imgElev > PointToElevation (vob[k].y)) {
                    if (i==cROBOT_BLUE) {
                        fireSanity(BLUE_BLOB_ABOVE_GOAL+k-vobBlueGoal);
                    } else if (i==cROBOT_RED) {
                        fireSanity(RED_BLOB_ABOVE_GOAL+k-vobBlueGoal);
                    }
                      info.ResetBlob ();
                    }
                }
            }

          if (!flag) {
              // final check:
              //if doesnt see any beacon, then check its own elevation

              // check the elevation of the ball, if it is higher than the head.
              if (RAD2DEG (effective_tilt) - patch.imgElev < -5) {
                    if (i==cROBOT_BLUE) {
                        fireSanity(BLUE_BLOB_ABOVE_HEAD);
                    } else if (i==cROBOT_RED) {
                        fireSanity(RED_BLOB_ABOVE_HEAD);
                    }
                  info.ResetBlob ();
                }
            }

        }
    }
}

void
VisualCortex::FindRobot () {

  if (RAD2DEG (effective_tilt) > 45) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
      (*outfile) << "tilt too low" << endl;
#endif
#endif
      return;
    }
#ifdef BAR_CHALL
    // do not rotate blobs if it is bar challenge
    blobs->rotate (cROBOT_RED, 0, 1.0);
    blobs->rotate (cROBOT_BLUE, 0, 1.0);
#endif
#ifndef BAR_CHALL

    blobs->rotate (cROBOT_RED, sin_eroll, cos_eroll);
    blobs->rotate (cROBOT_BLUE, sin_eroll, cos_eroll);
#endif

    RobotNoiseFilter ();

	double bcx=0.0, bcy=0.0, ball_crx=0.0, ball_cry=0.0, ball_r=0.0, leftX=0.0, rightX=0.0, topY=0.0, bottomY=0.0, BUFFER=0.0;
	if(vob[vobBall].cf >0) {
    	// rotate the ball cx and cy
    	bcx = (WIDTH/2) - vob[vobBall].cx;
    	bcy = (HEIGHT/2) - vob[vobBall].cy;
    	ball_crx = (WIDTH/2) - (cos_eroll *bcx - sin_eroll * bcy);
    	ball_cry = (HEIGHT/2) - (sin_eroll *bcx + cos_eroll * bcy);
    	ball_r = vob[vobBall].radius;
    	leftX = ball_crx - ball_r;
    	rightX = ball_crx + ball_r;
    	topY = ball_cry - ball_r;
    	bottomY = ball_cry + ball_r;
    	BUFFER = MAX ((rightX - leftX) * 0.20, 6.0);
	}
    
    /* ricky: go through maybe dog blobs, and transform colour depending on whether there are any
       dog blobs nearby. find largest dog blob within limit and transform self into it */
#ifdef MAYBE
    for (int i = 0; i < blobs->count(MAYBE_DOG); i++) {
        HalfBeacon &maybedog = blobs->getBlob(MAYBE_DOG, i);
	    double xThreshold = (maybedog.getXMax() - maybedog.getXMin()) * 0.5;
	    double yThreshold = (maybedog.getYMax() - maybedog.getYMin()) * 0.5;
	    double greatestblue = 0;
	    double greatestred = 0;
	    double greatest;
	    for (int j = 0; j < blobs->count(cROBOT_RED); j++) {
	        HalfBeacon &red = (blobs->getBlob(cROBOT_RED, j));
	        if (red.area == 0) {
                    fireSanity(RED_BLOB_AREA_0);
                    continue;
                }
	        if (red.getXMax() >= maybedog.getXMin() - xThreshold && red.getXMin() <= maybedog.getXMax() + xThreshold &&
	            red.getYMax() >= maybedog.getYMin() - yThreshold && red.getYMin() <= maybedog.getYMax() + yThreshold) {
	            if (red.area > greatestred) greatestred = red.area;
	        }
	    }
	    for (int j = 0; j < blobs->count(cROBOT_BLUE); j++) {
	        HalfBeacon &blue = (blobs->getBlob(cROBOT_BLUE, j));
	        if (blue.area == 0) {
                    fireSanity(BLUE_BLOB_AREA_0);
                    continue;
                }
	        if (blue.getXMax() >= maybedog.getXMin() - xThreshold && blue.getXMin() <= maybedog.getXMax() + xThreshold &&
	            blue.getYMax() >= maybedog.getYMin() - yThreshold && blue.getYMin() <= maybedog.getYMax() + yThreshold) {
	            if (blue.area > greatestblue) greatestblue = blue.area;
	        }
	    }
	    if (greatestred > greatestblue && greatestred > maybedog.area) {
	        blobs->add(cROBOT_RED, maybedog);
	    } else if (greatestblue > greatestred && greatestblue > maybedog.area) {
	        blobs->add(cROBOT_BLUE, maybedog);
	    }
    }
#endif 
   
    blobs->qsort (cROBOT_RED);
    blobs->qsort (cROBOT_BLUE);

    // cleans up red noise inside ball and red blobs that are too close
    for (int j = 0; j < blobs->count (cROBOT_RED); j++) {
        BlobInfo & info = blobs->getBlob (cROBOT_RED, j);

#ifdef OFFLINE
#ifdef PRE_SANITY_BLOB
        if (sanityBlobDebugMsg) {
            cout << "PRE_SANITY_BLOB: " << vobRedDog << " " << info.xRmin << " " << info.yRmin << " " << (info.xRmax-info.xRmin) << " " << (info.yRmax-info.yRmin) << endl;
        }
#endif
#endif

        if (vob[vobBall].cf > 0) {
        /*  if (info.getXMin() >= leftX - BUFFER && info.getXMax() <= rightX + BUFFER &&
                info.getYMin() >= topY - BUFFER && info.getYMax() <= bottomY + BUFFER) 
        */
        // clean up all red around and below the ball
	    /*
            if (info.getXMin () >= leftX - BUFFER
                 && info.getXMax () <= rightX + BUFFER
                 && info.getYMin () >= topY - BUFFER * 2) {
	    */
            if (info.xRmin >= leftX - BUFFER && info.xRmax <= rightX + BUFFER && info.yRmin >= topY - BUFFER * 2) {
                fireSanity(RED_BLOB_NEAR_BALL);
                info.ResetBlob ();
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "red robot " << info.xRmin << "," << info.yRmin << " near ball, filter out" << endl;
#endif
#endif
            }
        }
    }


#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "start finding red robot" << endl;
#endif
#endif
    FindColorRobot (cROBOT_RED);
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "done finding red robot" << endl;
#endif
#endif    


    // kill all blue blobs inside red dogs and ball
    for (int j = 0; j < blobs->count (cROBOT_BLUE); j++) {
        BlobInfo & info = blobs->getBlob (cROBOT_BLUE, j);
        

#ifdef OFFLINE
#ifdef PRE_SANITY_BLOB
        if (sanityBlobDebugMsg) {
            cout << "PRE_SANITY_BLOB: " << vobBlueDog << " " << info.xRmin << " " << info.yRmin << " " << (info.xRmax-info.xRmin) << " " << (info.yRmax-info.yRmin) << endl;
        }
#else
        (*outfile) << "blue dog blob: " << info.xRmin << " " << info.yRmin << " " << info.xRmax << " " << info.yRmax << endl;
#endif
#endif

        if (vob[vobBall].cf > 0) {
	    /*
            if (info.getXMin () >= leftX - BUFFER
                 && info.getXMax () <= rightX + BUFFER
                 && info.getYMin () >= topY - BUFFER * 2) {
	    */
            if (info.xRmin >= leftX - BUFFER && info.xRmax <= rightX + BUFFER && info.yRmin >= topY - BUFFER * 2) {
                fireSanity(BLUE_BLOB_NEAR_BALL);
                info.ResetBlob ();
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "blue robot " << info.xRmin << "," << info.yRmin << " near ball, filter out" << endl;
#endif
#endif
            }

        }

	    // remove blue dog blob around red dog as well
	    // avoid id red dog's legs or area above bandana as blue dog
        for (int k = vobRedDog; k <= vobRedDog4; k++) {
            double rleftX = vob[k].x;
    		double rrightX = (vob[k].x + vob[k].width);
    		double rtopY = vob[k].y;
    		double rbottomY = vob[k].y + vob[k].height;
    		double rBUFFER = MAX ((rrightX - rleftX) * 0.20, 5.0);
            if (vob[k].cf > 0 && (rrightX - rleftX) * (rbottomY - rtopY) > (info.xRmax - info.xRmin) * (info.yRmax - info.yRmin)) {		
                if (info.cRx >= vob[k].x && info.cRx <= vob[k].x + vob[k].width) {
                    //&& info.cRy >= vob[k].y ) //&& info.cRy <= vob[k].y+vob[k].height)
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "blue dog x within red dog x" << endl;
#endif
#endif
                    fireSanity(BLUE_BLOB_AROUND_RED_ROBOT+k-vobRedDog);
                    info.ResetBlob ();
                } else if (info.xRmin >= rleftX - rBUFFER
                  && info.xRmax <= rrightX + rBUFFER
                  && info.yRmin >= rtopY - rBUFFER
		          && info.yRmax <= rbottomY + rBUFFER ) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "blue dog within red dog boundary" << endl;
#endif
#endif
                    fireSanity(RED_ROBOT_LEG_AREA_AS_BLUE_ROBOT+k-vobRedDog);
                   info.ResetBlob ();
		        } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "passed small blob" << endl;
                    (*outfile) << "red  (" << rleftX << "," << rtopY << ") (" << rrightX << "," << rbottomY << ")" << endl;
                    (*outfile) << "blue (" << info.getXMin() << "," << info.getYMin() << ") (" << info.getXMax() << "," << info.getYMax() << ")" << endl;
#endif
#endif
                }
		    } else if (vob[k].cf > 0) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "passed big blob" << endl;
                    (*outfile) << "red  (" << rleftX << "," << rtopY << ") (" << rrightX << "," << rbottomY << ")" << endl;
                    (*outfile) << "blue (" << info.getXMin() << "," << info.getYMin() << ") (" << info.getXMax() << "," << info.getYMax() << ")" << endl;
#endif
#endif
            } 
        }
    }
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "start to find blue robot" << endl;
#endif
#endif
    FindColorRobot (cROBOT_BLUE);

#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "done finding red and blue robot" << endl;
#endif
#endif

    // filter out red robots in blue, and vice versa
    for (int i = vobRedDog; i <= vobRedDog4; i++) {
        if (vob[i].cf > 0) {
            for (int j = vobBlueDog; j <= vobBlueDog4; j++) {
                if (vob[j].cf > 0) {
                    int rxmin = (int)(vob[i].x);
                    int rxmax = (int)(vob[i].x + vob[i].width);
                    int rymin = (int)(vob[i].y);
                    int rymax = (int)(vob[i].y + vob[i].height);

                    int bxmin = (int)(vob[j].x);
                    int bxmax = (int)(vob[j].x + vob[j].width);
                    int bymin = (int)(vob[j].y);
                    int bymax = (int)(vob[j].y + vob[j].height);
                                        
                    if (rxmin > bxmin && rxmax < bxmax && rymin > bymin && rymax < bymax) { // check red in blue
                        vob[i].cf = 0;
            			vob[i].var = LARGE_VAL;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                        (*outfile) << "red within blue" << endl;
#endif
#endif
                    } else if (bxmin > rxmin && bxmax < rxmax && bymin > rymin && bymax < rymax) { // check blue in red
                        vob[j].cf = 0;
			            vob[j].var = LARGE_VAL;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                        (*outfile) << "blue within red" << endl;
#endif
#endif
                    }
                }
            }
        }
    }

#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
    (*outfile) << "done checking robots in robots" << endl;
#endif
#endif

    for (int i = vobRedDog; i <= vobBlueDog4; i++) {
        int flag = 0;

        if (vob[i].cf <= 0) {
            vob[i].cf = 0;
            continue;
        } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            RobotRelative(vob[i], true);
#endif
#else
            RobotRelative(vob[i]);
#endif
            vob[i].lhead = vob[i].head + 
                (PointToHeading(vob[i].x) - PointToHeading(vob[i].cx));
            vob[i].rhead = vob[i].head + 
                (PointToHeading(vob[i].x + vob[i].width) - PointToHeading(vob[i].cx));
            
            //wout << i << " " << vob[i].orient << " " << vob[i].getFlatDistance() << "\n";
        }

        // A robot cannot have elevation lower than ball
        if (vob[vobBall].cf > 0 && vob[vobBall].elev > vob[i].elev + 10) {
            vob[i].cf = 0;
            vob[i].var = LARGE_VAL;
            vob[i].angleVar = LARGE_VAL;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Dog lower than ball " << i << " elev " << vob[i].elev << " ball elev " << vob[vobBall].elev << "\n";
#endif
#endif
        }        

	// attempt to filter out blue dog streaks from edges of white and background
	if (((vob[i].width * 1.0) / vob[i].height >= 7.5) || ((vob[i].height * 1.0) / vob[i].width >= 7.5)) {
            vob[i].cf = 0;
            vob[i].var = LARGE_VAL;
            vob[i].angleVar = LARGE_VAL;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
            (*outfile) << "Aspect Ratio screwed " << vob[i].width << "," << vob[i].height << " throw out dog " << i << "\n";
#endif
#endif
        }        
	


        for (int k = vobBlueLeftBeacon; k <= vobYellowRightBeacon; k++) {
            if (vob[k].cf > 200) {
                flag = 1;

                // beacons should be further away than robot
                // add 20 cm leniency due to crappy robot distance
                // future generations, look in to this, robot recognition needs major work
                // and is virtually untouched in 2003
                if (ABS (vob[k].head - vob[i].head) < 10 && vob[k].dist + 20 < vob[i].dist) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "Object " << i << "is too far" << vob[i].dist << " " << vob[k].dist << "\n";
#endif
#endif
                    vob[i].cf = 0;
                    vob[i].var = LARGE_VAL;
                    vob[i].angleVar = LARGE_VAL;
                    fireSanity(BEACON_TOO_CLOSE_TO_ROBOT_BLOB + k - vobBlueLeftBeacon);
                }
                // robot fly above the beacon
                // changed to
                //VisualObject beacon;

                //beacon.imgHead = PointToHeading(vob[k].cx, vob[k].cy, sin_eroll, cos_eroll);
                //beacon.imgElev = PointToElevation(vob[k].x + vob[k].width, vob[k].y + vob[k].height, sin_eroll, cos_eroll);

                //beacon.dist = vob[k].dist;

                //if (vob[i].imgElev > beacon.imgElev)
                if (PointToElevation(vob[i].cy) > PointToElevation(vob[k].cx, vob[k].cy, sin_eroll, cos_eroll) && vob[i].elev > 0) {
                    // 5 is the error threshold
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "Object " << i << "is too high " << PointToElevation(vob[i].cy)  << " " << PointToElevation(vob[k].cx, vob[k].cy, sin_eroll, cos_eroll) << endl;
#endif
#endif
                    vob[i].cf = 0;
                    vob[i].var = LARGE_VAL;
                    vob[i].angleVar = LARGE_VAL;
                    fireSanity(ROBOT_BLOB_TOO_HIGH_TO_BEACON + k - vobBlueLeftBeacon);
                } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "Dog elev " << PointToElevation(vob[i].cy) << " Beacon elev " << PointToElevation(vob[k].cx, vob[k].cy, sin_eroll, cos_eroll) << 
		    " Dog dist " << vob[i].dist << " Beacon dist " << vob[k].dist << " Dog head " << vob[k].h << " Beacon head " << vob[i].h << endl;
#endif
#endif
                }
            }
        }

        for (int k = vobBlueGoal; k <= vobYellowGoal; k++) {
            if (vob[k].cf > 200) {
                flag = 1;

                // goal should be further away than robot
                // add 20 cm leniency due to crappy robot distance
                if (ABS (vob[k].head - vob[i].head) < 5
                    && vob[k].dist + 20 < vob[i].dist) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "Object " << i << "is too far" << vob[i].dist << " " << vob[k].dist << "\n";
#endif
#endif
                    vob[i].cf = 0;
                    vob[i].var = LARGE_VAL;
                    vob[i].angleVar = LARGE_VAL;
                    fireSanity(GOAL_TOO_CLOSE_TO_ROBOT_BLOB + k - vobBlueGoal);
                }
                // robot fly above the beacon
                // changed to

                //if (vob[i].imgElev > PointToElevation(vob[k].x, vob[k].y, sin_eroll, cos_eroll))

                //if (vob[i].imgElev > PointToElevation(vob[k].y))
                if (vob[i].elev > vob[k].elev + 5) {
                    // 5 is the error threshold
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                    (*outfile) << "Object " << i << "is too high " << vob[i].elev << " " << vob[k].elev << "\n";
#endif
#endif
                    vob[i].cf = 0;
                    vob[i].var = LARGE_VAL;
                    vob[i].angleVar = LARGE_VAL;
                    fireSanity(ROBOT_BLOB_TOO_HIGH_TO_GOAL + k - vobBlueGoal);
                }
            }
        }

        if (!flag) {
            // final check:
            // if doesnt see any beacon, then check its own elevation
            //wout << vob[
            //if (vob[i].elev < 0)
            //  vob[i].cf = 0;
    	    // check the HEIGHT of the robot above head, similar to checking for height of beacon
            if (tan (DEG2RAD (vob[i].elev)) * vob[i].dist > 25.0) {
                vob[i].cf = 0;
                vob[i].var = LARGE_VAL;
                vob[i].angleVar = LARGE_VAL;
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "dog sanity, dog too high " << i;
                (*outfile) << " elev " << vob[i].elev << " dist " << vob[i].dist;
                (*outfile) << " height " << tan (DEG2RAD (vob[i].elev)) * vob[i].dist << endl;
#endif
#endif
            } else {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "dog sanity, passed " << i;
                (*outfile) << " elev " << vob[i].elev << " dist " << vob[i].dist;
                (*outfile) << " height " << tan (DEG2RAD (vob[i].elev)) * vob[i].dist << endl;
#endif
#endif
	        }
            if (RAD2DEG (effective_tilt) - vob[i].imgElev < -10 && tan (DEG2RAD (vob[i].elev)) * vob[i].dist > 5) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "robot " << i << " elevation too high" << endl;
#endif
#endif            
                vob[i].var = LARGE_VAL;
                vob[i].angleVar = LARGE_VAL;
                vob[i].cf = 0;
                fireSanity(ROBOT_TILT_TOO_HIGH);
            }
            // filter out robots too low
            if (RAD2DEG (effective_tilt) - vob[i].imgElev > 45) {
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
                (*outfile) << "robot " << i << " elevation too low" << endl;
#endif
#endif
                vob[i].var = LARGE_VAL;
                vob[i].angleVar = LARGE_VAL;
                vob[i].cf = 0;
                fireSanity(ROBOT_TILT_TOO_LOW);
            }	    
        }
	
#ifdef OFFLINE
#ifndef PRE_SANITY_BLOB
        if (vob[i].cf > 0) {
            (*outfile) << "dog " << i << " elev " << vob[i].elev;
            (*outfile) << " head " << vob[i].head << " dist " << vob[i].d << endl;
        }
#endif
#endif	
    }
}

int VisualCortex::drawLine (int Ax, int Ay, int Bx, int By, uchar color) {
    int numOfPixels = 0;

    //------------------------------------------------------------------------
    // INITIALIZE THE COMPONENTS OF THE ALGORITHM THAT ARE NOT AFFECTED BY THE
    // SLOPE OR DIRECTION OF THE LINE
    //------------------------------------------------------------------------
    int dX = abs (Bx - Ax);        // store the change in X and Y of the line endpoints
    int dY = abs (By - Ay);

    int CurrentX = Ax;             // store the starting point (just point A)
    int CurrentY = Ay;

    //------------------------------------------------------------------------
    // DETERMINE "DIRECTIONS" TO INCREMENT X AND Y (REGARDLESS OF DECISION)
    //------------------------------------------------------------------------
    int Xincr, Yincr;
    if (Ax > Bx) {
        Xincr = -1;
    } else {
        Xincr = 1;
    }                              // which direction in X?
    if (Ay > By) {
        Yincr = -1;
    } else {
        Yincr = 1;
    }                              // which direction in Y?

    //------------------------------------------------------------------------
    // DETERMINE INDEPENDENT VARIABLE (ONE THAT ALWAYS INCREMENTS BY 1 (OR -1) )
    // AND INITIATE APPROPRIATE LINE DRAWING ROUTINE (BASED ON FIRST OCTANT
    // ALWAYS). THE X AND Y'S MAY BE FLIPPED IF Y IS THE INDEPENDENT VARIABLE.
    //------------------------------------------------------------------------
    if (dX >= dY) {
                  // if X is the independent variable
        int dPr = dY << 1;        // amount to increment decision if right is chosen (always)
        int dPru = dPr - (dX << 1);        // amount to increment decision if up is chosen
        int P = dPr - dX;                // decision variable start value

        for (; dX >= 0; dX--) {
            // process each point in the line one at a time (just use dX) 
            //setPixel(CurrentX, CurrentY, Color);  // plot the pixel
#ifdef  SHAPE_CHALL
            splane[CurrentY][CurrentX] = color;
#endif

#ifndef SHAPE_CHALL
		int index = CurrentX + CurrentY * WIDTH;
		//if( index >= PIXELS) {
		//	cout << "CurrentX=" << CurrentX << " CurrentY=" << CurrentY << "  (" << index << " >= " << PIXELS << ")" << endl;
		//}
	    if (color == BACKGROUND && (cplane[CurrentX + CurrentY * WIDTH] == BACKGROUND || cplane[CurrentX + CurrentY * WIDTH] == 9))
	        numOfPixels++;
            else if (cplane[CurrentX + CurrentY * WIDTH] == color)
                numOfPixels++;
#endif
// is the pixel going right AND up?
            if (P > 0) {
                CurrentX += Xincr;        // increment independent variable
                CurrentY += Yincr;        // increment dependent variable
                P += dPru;        // increment decision (for up)
            } else {
// is the pixel just going right?
                CurrentX += Xincr;        // increment independent variable
                P += dPr;                // increment decision (for right)
            }
        }
    } else {
        // if Y is the independent variable 
        int dPr = dX << 1;        // amount to increment decision if right is chosen (always)
        int dPru = dPr - (dY << 1);        // amount to increment decision if up is chosen
        int P = dPr - dY;                // decision variable start value

// process each point in the line one at a time (just use dY)      
        for (; dY >= 0; dY--) {
            //SetPixel(CurrentX, CurrentY, Color);  // plot the pixel
#ifdef  SHAPE_CHALL
            splane[CurrentY][CurrentX] = color;
#endif

#ifndef SHAPE_CHALL
	    if (color == BACKGROUND && (cplane[CurrentX + CurrentY * WIDTH] == BACKGROUND || cplane[CurrentX + CurrentY * WIDTH] == 9))
	        numOfPixels++;
            else if (cplane[CurrentX + CurrentY * WIDTH] == color)
                numOfPixels++;
#endif
                // is the pixel going up AND right?
            if (P > 0) {
                    CurrentX += Xincr;        // increment dependent variable
                    CurrentY += Yincr;        // increment independent variable
                    P += dPru;        // increment decision (for up)
            } else {
                // is the pixel just going up?
                CurrentY += Yincr;        // increment independent variable
                P += dPr;                // increment decision (for right)
            }
        }
    }
  return numOfPixels;
}

#ifdef VISION_DEBUG
void
VisualCortex::changeCalibration (int type) {
    gcd.changeCalibration (type);
}

void
VisualCortex::switchFile (int suff) {
    gcd.switchFile (suff);
}

void
VisualCortex::updateCalibration (int col, int row, int newcolour) {
    gcd.updateCalibration (col, row, newcolour, yplane);
}

void
VisualCortex::freezeImageProcessing (int i) {
}
#endif

bool
VisualCortex::lookup_mc (unsigned char prev_colour, int height, int width) {
    return blobs->lookup_mc (prev_colour, height, width);
}

void
VisualCortex::set_mc (int x) {
    blobs->multi_colour = x;
}

int
VisualCortex::get_mc () {
    return blobs->multi_colour;
}

#ifdef SHAPE_CHALL

int
VisualCortex::findShape (int color) {
    blobs->rotate (color, 0, 1.0);

    VisualObject *shape = &vob[vobRedDog];
    int vobNum = vobRedDog;
    int width = 0;
    int height = 0;
    double minDist = 10000;

    // find the bottom patch
    for (int k = vobYellowGoal; k <= vobBlueDog4; k++) {

        if (k == vobYellowGoal || k == vobYellowRightBeacon
             || k == vobYellowLeftBeacon) {

            int startIndex = -1;
            int left = 0, right = 0, top = 0, bottom = 0;

            if (vob[k].cf > 0) {

                Queue *pq = new Queue ();

                double dist = 1000.0;
                double xDist = 1000.0;
                double yDist = 1000.0;
                double avg_area = 0;
                double numCheckBox = 0;

                for (int i = 0; i < blobs->count (color); i++) {
                    BlobInfo & info = blobs->getBlob (color, i);

                    if ((info.cRx - vob[k].cx) * (info.cRx - vob[k].cx) +
                        (info.cRy - vob[k].y) * (info.cRy - vob[k].y) <=
                        dist && info.getXMax () - info.getXMin () >= 3
                        && info.getYMax () - info.getYMin () >= 3) {
                        startIndex = i;
                        dist = (info.cRx - vob[k].cx) * (info.cRx - vob[k].cx) +
                               (info.cRy - vob[k].y) * (info.cRy - vob[k].y);
                    }

                }

                if (startIndex != -1) {
                    BlobInfo & info = blobs->getBlob (color, startIndex);

                    pq->enqueue (startIndex);
                    info.enqueued = true;

                    left = info.getXMin ();
                    top = info.getYMin ();

                    right = info.getXMax ();
                    bottom = info.getYMax ();

                    shape->cf = 1000;
                    shape->cx = info.cRx;
                    shape->cy = info.cRy;

                }

                while (!pq->isEmpty ()) {
                    int blobIndex = pq->dequeue ();
                    BlobInfo & cur = blobs->getBlob (color, blobIndex);

                    avg_area += cur.area;

                    int maxArea = 2.0 * cur.area;
                    int minArea = 0.5 * cur.area;

                    for (int i = 0; i < blobs->count (color); i++) {
                        BlobInfo & info = blobs->getBlob (color, i);

                        // if the blob has already been included
                        if (info.enqueued == true) {
                            //wout << "enqueued.\n";
                            continue;
                        }
                        // if the blob is too far away
                        if ((info.cRx - cur.cRx) * (info.cRx - cur.cRx) +
                             (info.cRy - cur.cRy) * (info.cRy - cur.cRy) >
                             (4 * 2 * (cur.xRmax - cur.xRmin) *
                             (cur.xRmax - cur.xRmin))) {
                            //wout << "too far\n";
                            fireSanity(BLOB_TOO_FAR);
                            continue;
                        }
                        // if the blob is too big
                        if (info.xRmax - info.xRmin >
                            cur.yRmax - cur.yRmin +
                            MIN (2.0 * (cur.xRmax - cur.xRmin), 10)) {
                            //wout << "too wide\n";
                            fireSanity(BLOB_TOO_WIDE);
                            continue;
                        }

                        if (info.yRmax - info.yRmin >
                             cur.yRmax - cur.yRmin +
                             MIN (2.0 * (cur.yRmax - cur.yRmin), 10)) {
                            //wout << "too tall\n";
                            fireSanity(BLOB_TOO_TALL);
                            continue;
                        }

                        if (info.xRmax - info.xRmin <
                             MAX (0.5 * (cur.xRmax - cur.xRmin), 2)) {
                            //wout << "too thin\n";
                            fireSanity(BLOB_TOO_THIN);
                            continue;
                        }
                        // if the blob is too small
                        if (info.yRmax - info.yRmin <
                             MAX (0.5 * (cur.yRmax - cur.yRmin), 2)) {
                            //wout << "too low\n";
                            fireSanity(BLOB_TOO_LOW);
                            continue;
                        }
                        // if the aspect ratios are wrong
                        if ((info.xRmax -
                             info.xRmin) / (double) (info.yRmax - info.yRmin) < 0.75
                             || 
                             (info.xRmax - info.xRmin) / 
                             (double) (info.yRmax - info.yRmin) > 1.5) {
                            //wout << "wrong aspect ratio\n";
                            fireSanity(WRONG_BLOB_ASPECT_RATIO);
                            continue;
                        }

                        if (minDist >
                            (info.cRx - cur.cRx) * (info.cRx - cur.cRx) +
                            (info.cRy - cur.cRy) * (info.cRy - cur.cRy)) {
                          minDist = (info.cRx - cur.cRx) * (info.cRx - cur.cRx) +
                                    (info.cRy - cur.cRy) * (info.cRy - cur.cRy);
                        }
                        /************************ bounding box stuff  */
                        // include the blob in the shape
                        if (info.getXMin () < left) {
                            left = info.getXMin ();
                        }
                        if (info.getXMax () > right) {
                            right = info.getXMax ();
                        }
                        if (info.getYMin () < top) {
                            top = info.getYMin ();
                        }
                        if (info.getYMax () > bottom) {
                            bottom = info.getYMax ();
                        }

                        pq->enqueue (i);
                        info.enqueued = true;
                    }
                }

                shape->x = left;
                shape->y = top;
                shape->width = abs (right - left);
                shape->height = abs (top - bottom);

                shape->projected_x = left;
                shape->projected_y = top;
                shape->projected_h = abs (left - right);
                shape->projected_w = abs (top - bottom);

                
				memset (splane, FILL_COLOUR, HEIGHT * WIDTH * sizeof (uchar));

                minDist = MAX (minDist, 30.0);

                // draws grid
                for (int i = 0; i < blobs->count (color); i++) {
                    BlobInfo & info = blobs->getBlob (color, i);
                    if (info.enqueued) {
                        for (int j = 0; j < blobs->count (color); j++) {
                            BlobInfo & cur = blobs->getBlob (color, j);
                            if (cur.enqueued && 
                                 (info.cRx - cur.cRx) * (info.cRx - cur.cRx) +
                                 (info.cRy - cur.cRy) * (info.cRy - cur.cRy) < 4 * minDist) {
                                drawLine (info.cRx - left, info.cRy - top,
                                          cur.cRx - left, cur.cRy - top,
                                          BORDER_COLOUR);
                            }
                        }

                    }
                }

                if (bottom - top > 0 && right - left > 0) {
                    shape->cf = 1000;
                    seedFill (shape->width + 1, shape->height + 1,
                              shape->width + 2, shape->height + 2,
                              BACKGROUND, BORDER_COLOUR);
                    calcMoments (BORDER_COLOUR, FILL_COLOUR, bottom - top,
                                 right - left);
                } else {
                    shape->cf = 0;
                }

                delete pq;
                pq = NULL;
                return shape->cf;
            }

        }
    }
}

void
VisualCortex::seedFill (int x, int y, int width, int height,
                        uchar fill, uchar border) {
    if (x >= 0 && y >= 0 && x <= width && y <= height &&
        splane[y][x] != fill && splane[y][x] != border) {
        splane[y][x] = fill;
        seedFill (x - 1, y, width, height, fill, border);
        seedFill (x + 1, y, width, height, fill, border);
        seedFill (x, y - 1, width, height, fill, border);
        seedFill (x, y + 1, width, height, fill, border);
    }
}

void
VisualCortex::runFill (int x, int y, int width, int height, uchar fill,
                       uchar border) {

}

void
VisualCortex::calcMoments (uchar label, uchar label2, int height, int width) {

    
	memset (ptr, 0, 7 * sizeof(double));
    memset (moment, 0, 2 * 2 * sizeof(double));
    memset (moment1, 0, 4 * 4 * sizeof(double));
    memset (moment2, 0, 4 * 4 * sizeof(double));

	int y,                        /* row index */
        x,                                /* column index */
        p, q, i, j;

    int rows,                        /* number of rows in image */
        cols;                        /* number of columns in image */

    double value, value2;

    rows = height;
    cols = width;

    for (i = 0; i < 7; i++)
        ptr[i] = 0.0;

    for (i = 0; i < 2; i++)
        for (j = 0; j < 2; j++)
        moment[i][j] = 0.0;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++) {
            moment1[i][j] = 0.0;
            moment2[i][j] = 0.0;
        }

    /*
     * compares the pixel value at the location (y,x)
     * with the threshold value. If it is greater than
     * the threshold value it writes 255 at the location
     * else 0
    **/
    //image = (int **)getData_Image(label_image, 0);

    //label = image[r][c];

    for (p = 0; p <= 1; p++) {
        for (q = 0; q <= 1; q++) {
            for (y = 1; y < rows; y++) {
                for (x = 1; x < cols; x++) {

                    if (splane[y][x] == label || splane[y][x] == label2) {

                        moment[p][q] = moment[p][q] + pow (x, p) * pow (y, q) * 1;        /* for binary image, cosider binary patterns. where f(x, y)= 1 or 0, here 1 is white, 0 is black */

                    }
                }
            }
        }
    }

// esapce division by zero
    if (moment[0][0] == 0) {
        wout << "moment[0][0] == 0\n";
        return;
    }

    for (p = 0; p <= 3; p++) {
        for (q = 0; q <= 3; q++) {
            for (y = 1; y < rows; y++) {
                for (x = 1; x < cols; x++) {

                    if (splane[y][x] == label || splane[y][x] == label2) {

                        value = moment[1][0] / moment[0][0];
                        value2 = moment[0][1] / moment[0][0];

                        moment1[p][q] = moment1[p][q] + pow ((x - (value)), p) * pow ((y - value2), q) * 1;        /* for binary image, cosider binary patterns. where f(x, y)= 1 or 0, here 1 is white, 0 is black */

                    }
                }
            }
        }
    }

    // escape division by zero
    if (pow (moment1[0][0], ((p + q) * 0.5) + 1) == 0) {
        wout << "moment1[0][0] == 0\n";
        return;
    }
    for (p = 0; p < 4; p++) {
        for (q = 0; q < 4; q++) {
            if ((p + q) == 2 || (p + q) == 3) {

                moment2[p][q] = moment1[p][q] / pow (moment1[0][0], ((p + q) * 0.5) + 1);
            }
        }
    }

    ptr[0] = moment2[2][0] + moment2[0][2];

    ptr[1] = (moment2[2][0] - moment2[0][2]) * (moment2[2][0] - moment2[0][2]) +
              4 * (moment2[1][1]) * (moment2[1][1]);

    ptr[2] = (moment2[3][0] - 3 * moment2[1][2]) * (moment2[3][0] - 3 * moment2[1][2]) +
             (3 * moment2[2][1] - moment2[0][3]) * (3 * moment2[2][1] - moment2[0][3]);

    ptr[3] = (moment2[3][0] + moment2[1][2]) * (moment2[3][0] + moment2[1][2]) +
             (moment2[2][1] + moment2[0][3]) * (moment2[2][1] + moment2[0][3]);

    ptr[4] = (moment2[3][0] - 3 * moment2[1][2]) * (moment2[3][0] + moment2[1][2]) *
             ((moment2[3][0] + moment2[1][2]) * (moment2[3][0] +  moment2[1][2]) -
              3 * (moment2[2][1] + moment2[0][3]) * (moment2[2][1] + moment2[0][3])) +
              (3 * moment2[2][1] - moment2[0][3]) * (moment2[2][1] + moment2[0][3]) * 
              (3 * (moment2[3][0] + moment2[1][2]) * (moment2[3][0] + moment2[1][2]) - 
              (moment2[2][1] + moment2[0][3]) * (moment2[2][1] + moment2[0][3]));

    ptr[5] = (moment2[2][0] - moment2[0][2]) * ((moment2[3][0] + moment2[1][2]) * 
             (moment2[3][0] + moment2[1][2]) - (moment2[2][1] + moment2[0][3]) * 
             (moment2[2][1] + moment2[0][3])) + 4 * (moment2[1][1]) * 
             (moment2[3][0] + moment2[1][2]) * (moment2[2][1] + moment2[0][3]);

    ptr[6] = (3 * moment2[2][1] - moment2[0][3]) * (moment2[3][0] + moment2[1][2]) *
             ((moment2[3][0] + moment2[1][2]) * (moment2[3][0] + moment2[1][2]) -
              3 * (moment2[2][1] + moment2[0][3]) * (moment2[2][1] + moment2[0][3])) - 
              (moment2[3][0] - 3 * moment2[1][2]) * (moment2[2][1] + moment2[0][3]) * 
              (3 * (moment2[3][0] + moment2 [1][2]) * (moment2 [3][0] + moment2 [1][2]) -
              (moment2 [2][1] + moment2[0][3]) * (moment2[2][1] + moment2[0][3]));

}

#endif



#ifdef EDGE_DETECT
//{{{


/**
 * Common stuff in the pipeline for both Hough and pattern methods. 
 * - setSelfPos <--- Removed
 * - runEdgeDetect
 * - initEdgeDetect
 * - initEdgeTable
 * - findFieldEdges
 */
//{{{
/**
 * Sets self position based on GPS. Note that this should really be handled 
 * in GPS but it's in here for the time being. 
 */
/*
void VisualCortex::setSelfPos(int inX, int inY, double inT, double inVarX, double inVarY, double inVarT)
{//{{{
  currentX = inX;
  currentY = inY;
  currentT = inT;
  varX = inVarX;
  varY = inVarY;
  varT = inVarT;
}//}}}
*/
void VisualCortex::runEdgeDetect(const WMObj& self)
{//{{{
  runEdgeDetectHough(self);
}//}}}

void VisualCortex::initEdgeDetect()
{//{{{
	fieldEdges = (uchar*)malloc(HEIGHT*WIDTH);
	edgeList = (uchar*)malloc(MAX_EDGE_PIXELS * ELEMENTS_PER_EDGELIST_ENTRY);
  numEdgePixels = origNumEdgePixels = 0;
  abortCount = 0;
	//This must be sufficient to hold at least the width 
	//(because width is greater than height). Note that if the 
	//horizon is more vertical, this array may not get filled. 
	horizonPointsGreen = (int*)malloc((WIDTH+1)*sizeof(int)*2);
	evilHackFactor1 = 0.3;//0.5;
	evilHackFactor2 = 0.1;//0.1;
	evilTiltOffset = 0;//-0.05;
	evilDistanceHack = 1.1;//1.0;
	maxVals = new int[HOUGH_NUM_LINES*2];
	initDegSinLookup();
	initEdgeTable();
	houghFrequencyCount = 0;
	HOUGH_SHIFT_HORIZON = 0.180;//0.250  //Note, add evilTiltOffset to this to get the actual horizon shift.
	HOUGH_GREEN_LOWER_LIMIT = 1.0472;
	HOUGH_CULL = 0.6;
	HOUGH_CULL2 = 10;
	HOUGH_CULL3 = 10;
  //Weightings for variance contributions based on 1/numPixels matched. 
  //Note that a match of 30 pixels is considered strong. 
  HOUGH_DISTVAR_NUMPIXELS = 12000.0;
  HOUGH_ANGVAR_NUMPIXELS = 4000.0;
  xVarCutoff = 50.0;
  yVarCutoff = 50.0;
  tVarCutoff = 10.0;
  altProject = true;
  groundPlane = 0;//-3
  sendDirField = 0;
  whiteAndBallThreshold = 5500;
  blueAndBallThreshold = 4500;
  yellowAndBallThreshold = 4500;
  whiteThreshold = 1000;
  blueThreshold = 1000;
  yellowThreshold = 1000;
  if (EDGE_DETECT_TYPE == EDGE_PATTERN)
  {
#ifndef PRE_SANITY_BLOB
    cout << "About to initialize gradient field\n";
#endif
    sendEdges = true;
    initCurrentPosition();
    currentXint = -1;
    currentYint = -1;
    currentTint = -1;
    edgeMatchQuality = 0;
    edgesValid = false;
#ifndef PRE_SANITY_BLOB
    cout << "Gradient field initialized\n";
#endif
    numTests = maxNumTests;
    enableAbortBorder = false;
    abortBorderSet = false;
    patternDx = 0;
    patternDy = 0;
    patternDt = 0;
    maxDiffX = 30;
    maxDiffY = 30;
    maxDiffT = 20;
    cfMatchedX = 50;
    cfMatchedY = 50;
    cfMatchedT = 50;
    cfNoMatchX = (int)LARGE_VAL;
    cfNoMatchY = (int)LARGE_VAL;
    cfNoMatchT = (int)LARGE_VAL;
    matchCutoffX = 20;
    matchCutoffY = 20;
    matchCutoffT = 30;
    isolationCutoff = 0;//10;
    isGoalie = false;
  }
}//}}}

void VisualCortex::initEdgeTable()
{//{{{
  //Initializes the edge type table of 256 values based on 
  //the initialization table (of 48 values). Values that don't 
  //exist in the init table get set to NOT_EDGE. 
  int i;
  for (i = 0;i < 256; i ++)
  {//{{{
    edgeTypeTable[i] = NOT_EDGE;
  }//}}}
  for (i = 0; i < 24; i ++)
  {//{{{
    edgeTypeTable[edgeTypeTableInit[2*i]] = (int)edgeTypeTableInit[2*i+1];
    if (edgeTypeTable[edgeTypeTableInit[2*i]] > 7)
    {//{{{
      cout << "Edge type table error index = " << i << " value = " << edgeTypeTable[edgeTypeTableInit[2*i]] << endl;
    }//}}}
  }//}}}
#ifndef PRE_SANITY_BLOB
  cout << "Edge type table initialized\n";
#endif
}//}}}

/**
 * int[][] VisualCortex::findFieldEdges()
 *
 * Finds the edges of the field, returns an array of 
 * edges size [HEIGHT*WIDTH]. Takes the cplane, finds the
 * edges of green field then applies the edge test to make sure
 * that the edges are feasible. At this stage, only checks
 * to see that there are between 4 and 6 blob pixels around, 
 * may do more advanced checking later. 
 *
 * NOTE: Also returns edgeListOut which is 3 times the return value
 * and stores [3*n] = x, [3*n+1] = y, [3*n+2] = direction as defined
 * in EdgeTypes.txt
 *
 */
int VisualCortex::findFieldEdges(
    uchar * retVal, 
    uchar* edgeListOut)
{//{{{
  int h,w,edgeCount = 0, type;
  uchar edgeIndex, edgeCross;
  int whiteAndBallCount = 0; //counters for facingFieldEdge
  int whiteCount = 0;
  int blueAndBallCount = 0;
  int blueCount = 0;
  int yellowAndBallCount = 0;
  int yellowCount = 0;
  avgEdgeX = 0;
  avgEdgeY = 0;
  //Stay away from edges to make life easier. NOTE: Blank edges are now numbered NOT_EDGE. 
#ifdef HOUGH_INCLUDE_EDGEPLANE
  for (w = 0; w < WIDTH; w ++)
  {//{{{
    retVal[w] = NOT_EDGE;
    retVal[(HEIGHT-1)*WIDTH+w] = NOT_EDGE;
    retVal[WIDTH+w] = NOT_EDGE;
    retVal[(HEIGHT-2)*WIDTH+w] = NOT_EDGE;
  }//}}}
  for (h = 0; h < HEIGHT; h ++)
  {//{{{
    retVal[h*WIDTH] = NOT_EDGE;
    retVal[h*WIDTH+WIDTH-1] = NOT_EDGE;
    retVal[h*WIDTH+1] = NOT_EDGE;
    retVal[h*WIDTH+WIDTH-2] = NOT_EDGE;
  }//}}}
#endif //HOUGH_INCLUDE_EDGEPLANE

#ifdef HOUGH_EDGEDETECT_USEHORIZON
  horizonCounter = 0;  //VERY IMPORTANT TO SET THIS! This is a counter that 
                      //lets the horizonStartX and horizonEndX routines work. 
                      //I should be setting this in those functions but 
                      //setting it once here is probably more efficient. 
                      //I may move this to the main init thingy just to 
                      //avoid confusion. 
#ifdef HOUGH_INCLUDE_EDGEPLANE
  for (h = 2; h < HEIGHT - 2; h ++)  //We count up to avoid confusion. Noise higher up 
                                    //should be eliminated by the horizon. 
  {//{{{
    int startX = horizonStartX(h);
    int endX = horizonEndX(h);
    for (w = 2; w < startX; w ++)
    {//{{{
      retVal[h*WIDTH+w] = NOT_EDGE;
    }//}}}
    for (w = endX; w < WIDTH-2; w ++)
    {//{{{
      retVal[h*WIDTH+w] = NOT_EDGE;
    }//}}}
    for (w = startX; w < endX; w ++)
#else   //HOUGH_INCLUDE_EDGEPLANE
  for (h = 2; h < HEIGHT-2; h += 2)
  {//{{{
    int startX = horizonStartX(h);
    int endX = horizonEndX(h);
    for (w = startX; w < endX; w += 2)
#endif  //HOUGH_INCLUDE_EDGEPLANE
#else //HOUGH_EDGEDETECT_USEHORIZON

#ifdef HOUGH_INCLUDE_EDGEPLANE
  for (h = HEIGHT - 3; h > 1; h --)  //We count down to make sure we get the closest pixels first 
                                    //(in case we run out of edge points due to noise higher up).
  {//{{{
    for (w = 2; w < WIDTH - 2; w ++)
#else   //HOUGH_INCLUDE_EDGEPLANE
  for (h = HEIGHT - 3; h > 1; h -= 2)
  {//{{{
    for (w = 2; w < WIDTH - 2; w += 2)
#endif  //HOUGH_INCLUDE_EDGEPLANE
#endif //HOUGH_EDGEDETECT_USEHORIZON
    {//{{{
#ifdef HOUGH_INCLUDE_EDGEPLANE
      //Check if either h or w AND'd with CROSS_SPACING_MASK
      //is equal to zero (so we only consider pixels on a grid, 
      //cuts down on computation but is lossy). 
      //!!!!!! Figure out why AND'ing with 0x03 or the like doesn't work!
      //
      if (h%2==0 && w%2==0)//(h%4 == 0 && w % 2 == 0) || (w % 4 == 0 && h % 2 == 0))
      {//{{{
#endif //HOUGH_INCLUDE_EDGEPLANE
        //counts for facingFieldEdge
        if (cplane[h*WIDTH+w] == FIELDWHITE || cplane[h*WIDTH+w] == BALL)
        {
          whiteAndBallCount ++;
        }
        if (cplane[h*WIDTH+w] == FIELDWHITE)
        {
          whiteCount ++;
        }
        if (cplane[h*WIDTH+w] == BLUE || cplane[h*WIDTH+w] == BALL)
        {
          blueAndBallCount ++;
        }
        if (cplane[h*WIDTH+w] == BLUE)
        {
          blueCount ++;
        }
        if (cplane[h*WIDTH+w] == YELLOW || cplane[h*WIDTH+w] == BALL)
        {
          yellowAndBallCount ++;
        }
        if (cplane[h*WIDTH+w] == YELLOW)
        {
          yellowCount ++;
        }
        //First check if we're on the right colored pixel.
#ifdef HOUGH_BEACON_AS_FIELD
        if (edgeCount <= MAX_EDGE_PIXELS && (cplane[h*WIDTH+w] == GREEN_FIELD || cplane[h*WIDTH+2] == GREEN))// || cplane[h*WIDTH+w] == YELLOW || cplane[h*WIDTH+w] == BLUE )) //Added yellow and blue to account for colored fringes due to chromatic abbheration
#else  //HOUGH_BEACON_AS_FIELD
        if (edgeCount <= MAX_EDGE_PIXELS && cplane[h*WIDTH+w] == GREEN_FIELD)
#endif //HOUGH_BEACON_AS_FIELD
        {//{{{
          //FOLDNOTE: Performs edge ID on current h/w -> retVal
          //Put the 8 surrounding pixels into a byte that we
          //can use to index into the edge types table.
          edgeIndex = 0;
          edgeCross = 0;
          uchar cplaneTemp;
          edgeIndex = 0xFF;
          //Set to all 1's, then set white pixels to zero. 
          //Green bordering on anything other than fieldwhite
          //will end up with all pixels set to 1 which is 
          //flagged as not an edge. 
          if ((cplaneTemp = cplane[(h-2)*WIDTH+w]) == FIELDWHITE)
          {//{{{
            edgeIndex = edgeIndex & (0xFF - 64);
            edgeCross ++;
          }//}}}
          if ((cplaneTemp = cplane[h*WIDTH+w+2]) == FIELDWHITE)
          {//{{{
            edgeIndex = edgeIndex & (0xFF - 16);
            edgeCross ++;
          }//}}}
          if ((cplaneTemp = cplane[(h+2)*WIDTH+w]) == FIELDWHITE)
          {//{{{
            edgeIndex = edgeIndex & (0xFF - 4);
            edgeCross ++;
          }//}}}
          if ((cplaneTemp = cplane[h*WIDTH+w-2]) == FIELDWHITE)
          {//{{{
            edgeIndex = edgeIndex & (0xFF - 1);
            edgeCross ++;
          }//}}}
          //Note, for a valid edge, at least 1 and no more
          //than 3 of the pixels above, below, left and right 
          //of the target pixel should be green field. 
          if (edgeCross >= 1 && edgeCross <= 3)        //Yes, this is a little strange in that it 
                                                      //tests for FIELDWHITE twice and all that ... 
                                                      //just ignore it for now ...
          {//{{{
            if ((cplaneTemp = cplane[(h-2)*WIDTH+w-2]) == FIELDWHITE)
            {//{{{
              edgeIndex = edgeIndex & (0xFF - 128);
            }//}}}
            if ((cplaneTemp = cplane[(h-2)*WIDTH+w+2]) == FIELDWHITE)
            {//{{{
              edgeIndex = edgeIndex & (0xFF - 32);
            }//}}}
            if ((cplaneTemp = cplane[(h+2)*WIDTH+w+2]) == FIELDWHITE)
            {//{{{
              edgeIndex = edgeIndex & (0xFF - 8);
            }//}}}
            if ((cplaneTemp = cplane[(h+2)*WIDTH+w-2]) == FIELDWHITE)
            {//{{{
              edgeIndex = edgeIndex & (0xFF - 2);
            }//}}}
          //Now edgeIndex has our index into the array that 
          //will return the desired angle. Follow it to find the type.
          if ((type = edgeTypeTable[(int)edgeIndex]) == NOT_EDGE)// && fieldWhiteCount > 0)
          {//{{{
            //ie. we're not on an edge that is valid
#ifdef HOUGH_INCLUDE_EDGEPLANE
            retVal[h*WIDTH+w] = NOT_EDGE;
#endif   //HOUGH_INCLUDE_EDGEPLANE
          }//}}}
          else
          {//{{{
            //add to output.
#ifdef HOUGH_INCLUDE_EDGEPLANE
            retVal[h*WIDTH+w] = type;
#endif  //HOUGH_INCLUDE_EDGEPLANE
            edgeListOut[edgeCount*3] = w;
            edgeListOut[edgeCount*3+1] = h;
            edgeListOut[edgeCount*3+2] = type;
            avgEdgeX += w;
            avgEdgeY += h;
            edgeCount ++;
          }//}}}
          }//}}}
#ifdef HOUGH_INCLUDE_EDGEPLANE
          else
          {//{{{
            //The pixels above, below, left and right didn't fit ...
            retVal[h*WIDTH+w] = NOT_EDGE;
          }//}}}
#endif  //HOUGH_INCLUDE_EDGEPLANE
        }//}}}
#ifdef HOUGH_INCLUDE_EDGEPLANE
        else
        {//{{{
          //We're not on a green pixel so set to zero.
          retVal[h*WIDTH+w] = NOT_EDGE;
        }//}}}
#endif  //HOUGH_INCLUDE_EDGEPLANE
#ifdef HOUGH_INCLUDE_EDGEPLANE  //yes we take off one level of bracketing here
                                //because we now skip one 'if' statement right at the 
                                //top (the one that takes the modulus to check gridlines). 
      }//}}}
      else
      {//{{{
        retVal[h*WIDTH+w] = NOT_EDGE;
      }//}}}
#endif //HOUGH_INCLUDE_EDGEPLANE
    }//}}}
  }//}}}
  if (edgeCount > 0)
  {
    avgEdgeX = avgEdgeX/edgeCount;
    avgEdgeY = avgEdgeY/edgeCount;
  }
  else
  {
    avgEdgeX = WIDTH/2;
    avgEdgeY = HEIGHT/2;
  }
  origNumEdgePixels = edgeCount;
  //Now thin edges if necessary.
  if (maxEdgeThin != 0 && edgeCount > maxEdgeThin)
  {
    //If the number of edges is greater than maxEdgeThin, only take samples every 
    //edgeCount/maxEdgeThin samples. 
    //Note, as we go through the edgeListOut array, we don't need a second array 
    //as we're moving things down the array anyway (ie. we're overwriting elements 
    //in the array that have already been processed). 
    double edgeJump = (double)edgeCount/(double)maxEdgeThin;
    for (int i = 0; i < maxEdgeThin; i ++)
    {
      edgeListOut[i*3+0] = edgeListOut[(int)((double)i*edgeJump)*3+0];
      edgeListOut[i*3+1] = edgeListOut[(int)((double)i*edgeJump)*3+1];
      edgeListOut[i*3+2] = edgeListOut[(int)((double)i*edgeJump)*3+2];
    }
    edgeCount = maxEdgeThin;
  }
  facingFieldEdge = (whiteAndBallCount > whiteAndBallThreshold && whiteCount > whiteThreshold);
  facingYellowGoal = (yellowAndBallCount > yellowAndBallThreshold && yellowCount > yellowThreshold);
  facingBlueGoal = (blueAndBallCount > blueAndBallThreshold && blueCount > blueThreshold);
  /*
  if (facingYellowGoal)
  {
    cout << "Facing yellow goal\n";
  }
  if (facingBlueGoal)
  {
    cout << "Facing blue goal\n";
  }
  */
  return edgeCount;
}//}}}
//}}}
//}}}
//}}}//these are to stop vim's folding from getting lost due to the #defines
//}}}




/**
 * Stuff in the Hough 'pipeline' between the cplane coming in to 
 * the identified edges coming out. Higher level stuff mainly.
 * - runEdgeDetectHough
 * - findHoughSpaceMultilevel
 *    - findHoughSpace (OLD, header comments only)
 * - setLineEnds
 * - setLineType
 * - projectPoints
 * - defineFieldLine
 * - runSanityCheck
 */
//{{{

#define TIME_INT(s,e) ((int)(((e) - (s)).useconds))
void VisualCortex::runEdgeDetectHough(const WMObj& self)
{//{{{
  currentX = (int)(self.pos.x);
  currentY = (int)(self.pos.y);
  currentT = self.h;
  varX = self.posVar;
  varY = self.posVar;
  varT = self.hVar;
#ifndef OFFLINE
  /*
  SystemTime startT, horizonT, findT, projectT, nightT;
  GetSystemTime(&startT);
  */
#endif
  //Find the pixels along the line that is HOUGH_SHIFT_HORIZON radians
  //below the actual horizon.
  generateHorizonPoints(HOUGH_SHIFT_HORIZON, horizonPointsGreen);
#ifndef OFFLINE
//  GetSystemTime(&horizonT);
#endif
  //Find the edges between green and white under the 'horizon'
  numEdgePixels = findFieldEdges(fieldEdges, edgeList);
#ifndef OFFLINE
//  GetSystemTime(&findT);
#endif
  //Project edge points onto the field plane then try to 'gravitate' 
  //towards best fit to the nearest known line. 
  //Firstly, note that the edgeList coming in has 3 values being w, h and type. 
  //However, to project these onto the field we need an array with only 2 so we need 
  //to rearrange things. 
  for (int i = 0; i < numEdgePixels; i ++)
  {
    tempFieldEdges[i*2+0] = (int)(edgeList[i*3+0] & 0xFF) - WIDTH/2;
    tempFieldEdges[i*2+1] = -((int)(edgeList[i*3+1] & 0xFF) - HEIGHT/2);
    //!!!!!! I don't know how useful the edge space type is given 
    //the angles in the visual plane will differ to those on the field ... 
    //perhaps we can still apply some angle stuff and transform them then 
    //use this info to limit which edges they gravitate to. 
    fieldEdgeSpaceType[i] = edgeList[i*3+2];
  }
  //Now we can project the points onto the field.
  projectPoints(tempFieldEdges, numEdgePixels, fieldEdgeSpace, groundPlane);
#ifndef OFFLINE
//  GetSystemTime(&projectT);
#endif
  //Then translate them to world relative. 
  localToGlobal((int)currentX, (int)currentY, (int)currentT, fieldEdgeSpace, numEdgePixels, fieldEdgeSpace);
  currentXint = (int)currentX;
  currentYint = (int)currentY;
  currentTint = (int)currentT;
  if (sendDebugNow)
  {
    cout << "current position is " << currentXint << "/" << currentYint << "/" << currentTint << endl;
  }
  findCurrentPosition(&currentXint, &currentYint, &currentTint);
  if (sendDebugNow)
  {
    cout << "New position is " << currentXint << "/" << currentYint << "/" << currentTint << endl;
  }
#ifndef OFFLINE
//  GetSystemTime(&nightT);
/*
  if (peakHorizonT < TIME_INT(startT,horizonT))
  {
    peakHorizonT = TIME_INT(startT,horizonT);
  }
  if (peakFindT < TIME_INT(horizonT,findT))
  {
    peakFindT = TIME_INT(horizonT,findT);
  }
  if (peakProjectT < TIME_INT(findT,projectT))
  {
    peakProjectT = TIME_INT(findT,projectT);
  }
  if (peakNightT < TIME_INT(projectT,nightT))
  {
    peakNightT = TIME_INT(projectT,nightT);
  }
  */
#endif    
}//}}}





/**
 * void VisualCortex::projectPoints(int * inPoints, int numPoints, int * outPoints)
 * 
 * Takes the input line endpoints, projects them 
 * through the 'lens' of the camera and, based on the head pan and tilt, finds the 
 * heading and direction of the equivalent lines on the field based on co-ordinates 
 * starting at the base of the neck. 
 * NOTE: ZERO IS IN THE MIDDLE OF THE IMAGE AND INCREASES UPWARDS (standard co-ords). 
 * You therefore need to convert from the co-ordinate system used elsewhere in this file
 * (zero is in the upper left corner and increases downwards). 
 * //{{{
 * Uses the following global vars and defines:
 * pan                Pan of head in radians
 * tilt                Tilt of head in radians
 * FACE_LENGTH        Horizontal distance from camera to neck pan point
 * NECK_LENGTH        Vertical distance from camera to neck tilt point
 * NECKBASE_HEIGHT    Vertical distance from neck tilt point to ground (based on stance)
 *
 * Note that internal multidimensional arrays are implemented as pointer-pointer arrays.
 * External multidimensional arrays are implemented as single pointer arrays.
 * //}}}
 */
void VisualCortex::projectPoints(int * inPoints, int numPoints, int * outPoints, int projectHeight)
{//{{{
  if (altProject)
  {
    projectPointsAlt(inPoints, numPoints, outPoints, projectHeight);
    return;
  }
  //Find co-ordinates of the end points. 
  double **endVectors, *origin, *newOrigin, **newEndVectors;
  int  *fieldPoints;
  //Matrices.
  //single letters are individual transform matrix elements. 
  double * xfrmMatrix, a, b, c, d, e, f, g, h, ii, jj, kk, u;// *camToPan, *removePan, *panToTilt, *removeTilt, *tiltToGround;
  int i,j, k;
  endVectors = (double **)malloc(numPoints*sizeof(double*));
  newEndVectors = (double**)malloc(numPoints * sizeof(double*));
  fieldPoints = (int*)malloc(numPoints * 2 * sizeof(int));  //array of numpoints times numbers per point (2)
  for (i = 0; i < numPoints; i ++)
  {//{{{
    endVectors[i] = (double*)malloc(4*sizeof(double));  //4 for homogenous vectors
    //Allocate the last dimension of newEndVectors later 
    //(when applying the matrix)
  }//}}}
  origin = (double*)malloc(4*sizeof(double));  //location of the origin. 
  for (i = 0; i < numPoints; i ++)
  {//{{{
    endVectors[i][0] = inPoints[i*2+0];
    endVectors[i][1] = inPoints[i*2+1];
    endVectors[i][2] = -(int) APERTURE_DISTANCE;  
    endVectors[i][3] = (int) 0;
  }//}}}


  //We now have the endpoints of the Hough lines in endVectors[i*4+k]. Form 
  //rays that go through the camera aperture. 
  origin[0] = 0;
  origin[1] = 0;
  origin[2] = 0;
  origin[3] = 1;

  //Initialize transformation matrices.
  a = -FACE_LENGTH;
  b = cosRAD(pan);
  c = sinRAD(pan);
  d = -c;
  e = b;
  f = NECK_LENGTH;
  g = cosRAD((-(tilt+evilTiltOffset)));
  h = -sinRAD((-(tilt+evilTiltOffset)));
  ii = -h;
  jj = g;
  kk = NECKBASE_HEIGHT - projectHeight;
  xfrmMatrix = createMatrix4x4(  b,    0,    c,    c*a,
                                d*h,  g,    h*e,  a*h*e+f*g+kk,
                                d*jj,  ii,    e*jj,  a*e*jj+ii*f,
                                0,    0,    0,    1);
  newOrigin = applyMatrix(xfrmMatrix, origin);
  for (i = 0; i < numPoints; i ++)
  {//{{{
#ifdef PROJECT_SENTINEL
    if (inPoints[i*2+0] != -1)
#endif //PROJECT_SENTINEL
    {//{{{
      newEndVectors[i] = applyMatrix(xfrmMatrix, endVectors[i]);
    }//}}}
  }//}}}
  //Now we have the rays through the lens which start at the point newOrigin and 
  //extend in the direction of each newEndVectors element by an amount u.
  //Now intersect it with the plane y=0. 
  for (i = 0; i < numPoints; i ++)
  {//{{{
      //NOTE: These vectors are STILL in the co-ordinate system centred about the spot 
      //on the ground under the robot's neck where X is to the right, Y is straight up 
      //and Z is pointing towards the BACK of the robot, where we have [X][Y][Z].
#ifdef PROJECT_SENTINEL      
      if (inPoints[i*2+0] != -1)
#endif //PROJECT_SENTINEL
      {//{{{
        u = newOrigin[1]/(newOrigin[1] - newEndVectors[i][1]);
//EVIL HACK:
if (u < 0)
{//{{{
//  u = -u;
}//}}}
      fieldPoints[i*2+0] = (int)(newOrigin[0] + u*(newEndVectors[i][0] - newOrigin[0]));
      fieldPoints[i*2+1] = -(int)(newOrigin[2] + u*(newEndVectors[i][2] - newOrigin[2]));

//WARNING: EVIL HACK!
//
//We're having a problem where points that are at an angle to the robot 
//(left or right) seem to have a depression value that is too high (and 
//the error is proportional to the angle). This can be seen where points 
//that are really in a line, after passing through all this transform, look 
//like they lie on an arc with the centre somewhere around the robot's tail. 
//This is invariant on head tilt or pan (you take a given point and move the 
//head around whilst keeping that point in view somehow and it doesn't get 
//much better or worse except for random noise). 
//
//To fix it for now, we're going to artificially increase the u value 
//proportional to the angle the vector to that point forms from 
//dead centre. This will cause the next process to extend the line more 
//and hopefully correct for this problem. 
//
//Measuring roughly, it appears that when points are dead centre, distances
//are right. When points are at 90 degrees either side, distances are 
//about 4/5 what they should be. 
/*
double correction;
//correction = (double)fieldEndPoints[i*4+j*2+0]/(double)fieldEndPoints[i*4+j*2+1];//was sine(atan()) of this
if (abs(fieldPoints[i*2+0]) < abs(fieldPoints[i*2+1]))
{//{{{
  correction = sin(atan((double)fieldPoints[i*2+0]/(double)fieldPoints[i*2+1]));
}//}}}
else
{//{{{
  correction = sin(3.14159/2.0 - atan((double)fieldPoints[i*2+1]/(double)fieldPoints[i*2+0]));
}//}}}
if (correction < 0)
{//{{{
  correction = -correction;
}//}}}

//cout << "correction for line " << i << " end " << j << " = " << correction << endl;

fieldPoints[i*2+0] = (int)((double)fieldPoints[i*2+0]*(1.0+evilHackFactor1*correction));
fieldPoints[i*2+1] = (int)((double)fieldPoints[i*2+1]*(1.0+evilHackFactor1*correction));
*/
//EVIL HACK NUMBER 2:
//
//After the previous hack was implemented, it was discovered that there was also
//a component of displacement forward that seems dependant on head pan (the angle of a 
//straight line is now correct but as the head pans to the left or right, it moves 
//closer to the robot). The following fudge is to correct for this (only in axis 
//1 which points out of the robot). 
/*
correction = sinRAD(pan);
if (correction < 0)
{//{{{
  correction = -correction;
}//}}}
fieldPoints[i*2+1] = (int)((double)fieldPoints[i*2+1]*(1.0+evilHackFactor2*correction));

//Now apply evil distance hack.
fieldPoints[i*2+1] = (int)((double)fieldPoints[i*2+1]*evilDistanceHack);
*/
      }//}}}
#ifdef PROJECT_SENTINEL
      else
      {//{{{
        //Line was deleted ...
        fieldPoints[i*2+0] = -1;
        fieldPoints[i*2+1] = -1;
      }//}}}
#endif //PROJECT_SENTINEL
  }//}}}
  //For now, we just copy this straight to the out array (for debugging!). 
  memcpy(outPoints, fieldPoints, numPoints*2*sizeof(int));
  free(newOrigin);
  free(origin);
  free(fieldPoints);
  for (i = 0; i < numPoints; i ++)
  {//{{{
    free(endVectors[i]);
#ifdef PROJECT_SENTINEL
    if (inPoints[i*2+0] != -1)
#endif //PROJECT_SENTINEL
    {
      //Remember, we didn't allocate this if inPoints was -1.
      free(newEndVectors[i]);
    }
  }//}}}
  free(endVectors);
  free(newEndVectors);
}//}}}

void VisualCortex::projectPointsAlt(int * inPoints, int numPoints, int * outPoints, int projectHeight)
{//{{{
  //Project points as per Bernhard's algorithm. 
  double hs = 7;
  double h = NECKBASE_HEIGHT;
  double b = 11;
  double s = 11.9;
  double T = (tilt - TILT_OFFSET + evilTiltOffset);
  double P = -pan;
  double c = FACE_LENGTH;
  double f = (double)APERTURE_DISTANCE;
  double n = NECK_LENGTH;
  double p = (double)projectHeight;
  for (int i = 0; i < numPoints; i ++)
  {
    double A, R, xc, yc, zc, xi, yi, zi, k, x, y, u, w;
    
    w = (double)(inPoints[i*2+0]);
    u = (double)(inPoints[i*2+1]);
    // Effective tilt = body tilt + head tilt
    A = asin((b-hs)/s);
    R = A+T;

    // Coordiates of camera point(c)
    xc = c*sin(P);
    yc = n*sin(R)+c*cos(R)*cos(P);
    zc = h-p-c*sin(R)*cos(P)+n*cos(R);

    // Coordinates of image point (i) to be projected
    xi = w*cos(P)+(c+f)*sin(P);
    yi = -w*sin(P)*cos(R)+(c+f)*cos(R)*cos(P)+(n+u)*sin(R);
    zi = h-p+w*sin(P)*sin(R)-(c+f)*sin(R)*cos(P)+(n+u)*cos(R);

//Raymond: Added: Check to make sure that zi-zc is greater than 0 (actually check 
//against 1E-5 just to make sure the FPU doesn't stuff up). If this is the 
//case then the points exist over the horizon so set the outpoints at 
//that point to equal infinity. 
    if (zi-zc > -(1E-5))
    {
      outPoints[i*2+0] = (int)LARGE_VAL;
      outPoints[i*2+1] = (int)LARGE_VAL;
    }
    else
    {
      // Ground plane z=0 intersection of line going through points c and i
      k = zi/(zi-zc);
      x = xi+k*(xc-xi);
      y = yi+k*(yc-yi);


/*    
    // Effective tilt = body tilt + head tilt;
    A = asin((b-h)/s);
    R = A+T;
    // Coordiates of camera point(c);
    xc = c*sin(P);
    yc = n*sin(R)+c*cos(R)*cos(P);
    zc = h-p-c*sin(R)*cos(P)+n*cos(R);
    // Coordinates of image point (i) to be projected;
    xi = w*cos(P)+(c+f)*sin(P);
    yi = -w*sin(P)*cos(R)+(c+f)*cos(R)*cos(P)+(n+u)*sin(R);
    zi = h-p+w*sin(P)*sin(R)-(c+f)*sin(R)*cos(P)+(n+u)*cos(R);
    // Ground plane z=0 intersection of line going through points c and i;
    k = zi/(zi-zc);
    x = xi+k*(xc-xi);
    y = yi+k*(yc-yi);
*/
      outPoints[i*2+0] = (int)x;
      outPoints[i*2+1] = (int)y;
    }
  }
}//}}}

//}}}


/**
 * Various horizon functions
 * - generateHorizonPoints
 * - generateHorizonDefinition
 * - imageDistanceToGround
 * - HOUGH_EDGEDETECT_USEHORIZON
 *  - horizonStartX
 *  - horizonEndX
 */
//{{{

/**
 * int * VisualCortex::generateHorizonPoints()
 *
 * Takes the current values for the head position and generates an array with the 
 * following data.
 * //{{{
 * [0]:    Equal to HORIZON_HORIZONTAL if the horizon is more horizontal than vertical.
 *         Equal to HORIZON_VERTICAL if the horizon is more vertical than horizontal.
 *
 * If horizon is more horizontal than vertical:
 *     [1] to [WIDTH]:   The Y co-ord of each horizon pixel where X+1 is the array index.
 *
 * If horizon is more vertical than horizontal:
 *     [1] to [HEIGHT]:   The X co-ord of each horizon pixel where Y+1 is the array index.
 *
 * Note that the origin for these purposes is in the upper left corner, Y increases downwards, 
 * X increases towards the right.
 *
 * inAngle specifies the adjustment in angle for the horizon (+ve pushes the horizon down 
 * relative to the real horizon). 
 *
 * NOTE:   Assumes SQUARE pixels. All linear units are in terms of pixels, all angular 
 *         units are in degrees. 
 * //}}}
 */
void VisualCortex::generateHorizonPoints(double inAngle, int * retVal)
{//{{{
  double * horizonDef = generateHorizonDefinition(inAngle);
  double m, tempX, tempY;
  int i;
  //Process angle
  while (horizonDef[1] < 0)
  {//{{{
    horizonDef[1] = horizonDef[1] + 360;
  }//}}}
  while (horizonDef[1] > 359)
  {//{{{
    horizonDef[1] = horizonDef[1] - 360;
  }//}}}
  if (horizonDef[0] > 179)
  {//{{{
    horizonDef[1] = horizonDef[1] - 180;
    horizonDef[0] = -horizonDef[0];
  }//}}}
  //Closer to horizontal?
if (sendDebugNow)
{//{{{
  cout << "horizonDef[1] = " << horizonDef[1] << endl;
}//}}}
  if (horizonDef[1] < 45 || (horizonDef[1] > 135 && horizonDef[1] < 225) || horizonDef[1] > 315)
  {//{{{//Horizon is closer to horizontal, do things in terms of X.
    double leftY;
    retVal[0] = HORIZON_HORIZONTAL;
    //Find slope of the line. Note that we negate it because 
    //Y increases downwards. 
    m = -tan(horizonDef[1]*3.141592/180.0);
    //Find any point on the line. Note that this point may well be 
    //off the edge of the picture. 
    tempX = (double)WIDTH/2.0 + ((double)horizonDef[0]*sinDEG(horizonDef[1]));
    tempY = (double)HEIGHT/2.0 + ((double)horizonDef[0]*cosDEG(horizonDef[1]));
    //Based on this point, find the left vertical intercept. Note that 
    //this point may be off the top or bottom of the picture but it will 
    //never be more than 1 image height above or below the image as 
    //the angle is limited to within 45 degrees of horizontal. 
    //Again be aware of the fact that Y increases downwards. 
    leftY = tempY - m * tempX;
    
    //Iterate across the image. !!!!! Note that we can make this more efficient
    //if we were to predefine the points that will be on the image (as opposed to 
    //setting them to 0 or HEIGHT-1 after the fact).
    for (i = 0; i < WIDTH; i ++)
    {//{{{
      retVal[i+1] = (int)(leftY + m * (double)i);
      if (retVal[i+1] < 0)
      {//{{{
        retVal[i+1] = 0;
      }//}}}
      else if (retVal[i+1] >= HEIGHT)
      {//{{{
        retVal[i+1] = HEIGHT-1;
      }//}}}
    }//}}}
  }//}}}
  else
  {//{{{//Horizon is closer to vertical, do things in terms of Y.
    double topX;
    retVal[0] = HORIZON_VERTICAL;
    //Find slope of the line. Note that we still negate it.
    //Y increases downwards, the co-ord system is now 'standard' 
    //(tilt your head to the right if you're lost! The angle here
    //is the angle of the line measured from a line 
    //increasing int the X direction ie. upwards if you rotate your 
    //head to the right)
    m = -1.0/tan(horizonDef[1]*3.141592/180.0);
    //Find any point on the line. Note that this point may well be 
    //off the edge of the picture. 
    tempX = (double)WIDTH/2.0 + ((double)horizonDef[0]*sinDEG(horizonDef[1]));
    tempY = (double)HEIGHT/2.0 + ((double)horizonDef[0]*cosDEG(horizonDef[1]));
    //Based on this point, find the top horizontal intercept. Note that 
    //this point may be off the left or right of the picture but it will 
    //never be more than 1 image width wide of the image as 
    //the angle is limited to within 45 degrees of vertical. 
    //Again be aware of the fact that Y increases downwards. 
    topX = tempX - m * tempY;
    //Iterate across the image. !!!!! Note that we can make this more efficient
    //if we were to predefine the points that will be on the image (as opposed to 
    //setting them to 0 or WIDTH-1 after the fact).
    for (i = 0; i < HEIGHT; i ++)
    {//{{{
      retVal[i+1] = (int)(topX + m * (double)i);
      if (retVal[i+1] < 0)
      {//{{{
        retVal[i+1] = 0;
      }//}}}
      else if (retVal[i+1] >= WIDTH)
      {//{{{
        retVal[i+1] = WIDTH-1;
      }//}}}
    }//}}}
  }//}}}
//cout << "Horizon generated\n";
  free(horizonDef);
}//}}}
/**
 * Generates r-theta pair describing the horizon. inAngle specifies 
 * the angle above or below the actual horizon to give the horizon definition to 
 * (+ve is downwards). 
 */
double * VisualCortex::generateHorizonDefinition(double inAngle)
{//{{{
  double * xfrmMatrix = generatePicToFieldXfrm();
  double * centrePoint, *directionPoint, *newCentre, *newDirection;
  double *projectedVector;
  double horizonElevation;
  double picHeightArbitary;
  double horizonDistArbitary;
  double *retVal;
  int i;
  centrePoint = (double*)malloc(4*sizeof(double));
  directionPoint = (double*)malloc(4*sizeof(double));
  newCentre = (double*)malloc(4*sizeof(double));
  newDirection = (double*)malloc(4*sizeof(double));
  projectedVector = (double*)malloc(3*sizeof(double));
  retVal = (double*)malloc(2*sizeof(double));
  for (i = 0; i < 4; i ++)
  {//{{{
    centrePoint[i] = 0;
    directionPoint[i] = 0;
  }//}}}
  directionPoint[2] = -10; //-10 chosen arbitrarily !!!! Make this a define!
  //Find the angle that the horizon makes with the picture horizontal
  retVal[1] = asin(sin(pan)*sin(tilt+offsetValue));
  retVal[1] = retVal[1] * 180.0/3.141592654;
  //Correct for what happens when tilt hits 90 degrees
  //(because of the asin, we end up going backwards!). 
  //Note, 1.570796=pi/2
  if (tilt+offsetValue > 1.570796)
  {//{{{
    retVal[1] = 180.0-retVal[1];
  }//}}}

  //Now translate all of these into world co-ords. 
  newCentre = applyMatrix(xfrmMatrix, centrePoint);
  newDirection = applyMatrix(xfrmMatrix, directionPoint);
  //Now find the angle that the line from Centre to Direction makes with 
  //the horizontal. The distance on the y-plane is in terms of x and z. 
  projectedVector[0] = newDirection[0] - newCentre[0];  //x
  projectedVector[1] = newDirection[1] - newCentre[1];  //y
  projectedVector[2] = newDirection[2] - newCentre[2];  //z
  //Angle formed is asin(ydist/10) where 10 is the hypotenuse (the original length arbitrarily chosen). 
  horizonElevation = asin(projectedVector[1]/10.0);
  //Alter the horizon elevation with the input factor.
  horizonElevation = horizonElevation + inAngle;
if (sendDebugNow)
{//{{{
  cout << "horizonElevation (after correction) = " << horizonElevation << endl;
}//}}}
//  if (HOUGH_SHIFT_HORIZON != 0)
//  {//{{{
//    horizonElevation = horizonElevation + HOUGH_SHIFT_HORIZON;
//  }//}}}
  //Translate to a distance on the visual plane. Note that we make the assumption that 
  //the pixels are square. We can do this because we know the original distance was 
  //10 (we arbitrarily chose it before). This distance will be adjacent to the 
  //horizonElevation we formed before. Before we do this however we need to know how many 
  //arbitrary units high the picture is. We use the field of view to get this. 
  picHeightArbitary = 10.0*tan(yField/2.0);
  horizonDistArbitary = 10.0*tan(horizonElevation);
  retVal[0] = (int)((horizonDistArbitary/picHeightArbitary)*((double)HEIGHT/2.0));
  free(xfrmMatrix);
  free(centrePoint);
  free(directionPoint);
  free(newCentre);
  free(newDirection);
  free(projectedVector);
//cout << "Horizon r=" << retVal[0] << " t=" << retVal[1] << endl;
  return retVal;
}//}}}
/**
 * double VisualCortex::imageDistanceToGround()
 *
 * Returns the distance along which a line projected out from the 
 * middle of the visual plane intercepts the field. Returns -1 
 * if such a line does not intercept the field.
 */
double VisualCortex::imageDistanceToGround()
{//{{{
  //Create a unit vector pointing out of the middle of the image. 
  double * imageCentre = (double*)malloc(4*sizeof(double));
  double * imageProjection = (double*)malloc(4*sizeof(double));
  double * imageCentreField, * imageProjectionField;
  double * picToFieldXfrm;
  double retVal; 
  for (int i = 0; i < 4; i ++)
  {//{{{
    imageCentre[i] = 0;
    imageProjection[i] = 0;
  }//}}}
  imageProjection[2] = -1;  //Remember, positive Z is into the robot's head.
  imageCentre[3] = 1;
  imageProjection[3] = 1;
  //Transform these into robot-centric co-ordinates. 
  picToFieldXfrm = generatePicToFieldXfrm();
  imageCentreField = applyMatrix(picToFieldXfrm, imageCentre);
  imageProjectionField = applyMatrix(picToFieldXfrm, imageProjection);
  //The unit vector is now the difference between these 2 vectors. 
  //The amount by which it needs to be multiplied (and hence the distance projected 
  //to the field) is equal to the y value of the centre divided by the y length 
  //of the unit vector. Note that if the y-value of the unit vector is positive, 
  //the line will never intersect the field (it's pointing up in the air). If 
  //the y-value is zero, the dog is looking parallel to the horizon, again no intersection. 
  //Check for this. 
  if (imageCentreField[1] - imageProjectionField[1] <= 0)
  {//{{{
    retVal = -1;
  }//}}}
  else
  {//{{{
    retVal = imageCentreField[1]/(imageCentreField[1]-imageProjectionField[1]);
  }//}}}
//DEBUG: Also cout the effective depression angle. Note that the 
//hypotenuse is 1. 
cout << "Depression of centre = " << 180.0*asin(imageCentreField[1] - imageProjectionField[1])/3.1415925 << endl;
//cout << "imageCentreField[1] = " << imageCentreField[1] << " imageProjectionField[1] = " << imageProjectionField[1] << endl;
  //Note that retVal is in arbitary units of 'pixels' What do we do about this?
  free(imageCentre);
  free(imageProjection);
  free(imageCentreField);
  free(picToFieldXfrm);
  return retVal;
}//}}}


#ifdef HOUGH_EDGEDETECT_USEHORIZON
/**
 * int VisualCortex::horizonStartX(int inH)
 *
 * Generates the required starting X position for a runline of 
 * below-horizon pixels given the current height in the visual plane 
 * (note, zero is at the top and increases positively downwards). 
 */
int 
VisualCortex::horizonStartX(int inH)
{//{{{
  int retVal;
  if (horizonPointsGreen[0] == HORIZON_VERTICAL)
  {//{{{
    //Easy case first, if horizon is more vertical than horizontal 
    //then simply figure out which side of the horizon we're supposed 
    //to be on and return the appropriate value. 
    if (pan < 0)
    {//{{{
      //Head is turned to the right, we start at the left. 
      retVal = 0;
    }//}}}
    else
    {//{{{
      //Head is turned to the left, we start at the horizon. 
      if (inH < HEIGHT)
      {
        retVal = horizonPointsGreen[inH+1];
      }
      else
      {
        retVal = 0;
        cout << "inH into horizonEndX is too large: " << inH << endl;
      }
    }//}}}
  }//}}}
  else
  {//{{{
    //Hard case. Horizon is more horizontal than vertical. Separate 
    //out if the horizon is sloping left or sloping right or dead 
    //straight. 
    if (horizonPointsGreen[1] == horizonPointsGreen[WIDTH])
    {//{{{
      //straight. We set this to zero, when above 
      //the horizon we also set the end to zero else we set the 
      //end to WIDTH. 
      retVal = 0;
    }//}}}
    else if (horizonPointsGreen[1] < horizonPointsGreen[WIDTH])
    {//{{{
      //left higher than right. Start point is 
      //always at zero. 
      retVal = 0;
    }//}}}
    else
    {//{{{
      //right higher than left. Start point is WIDTH when 
      //well above the horizon, 0 when well below the horizon
      //and the horizon itself in between. 
      if (horizonPointsGreen[WIDTH] > inH)
      {//{{{
        //Above the highest point in the horizon
        retVal = WIDTH;
      }//}}}
      else if (horizonPointsGreen[1] > inH)
      {//{{{
        //Between highest and lowest points in the horizon.
        //I'm using retVal as a temp variable here, watch out 
        //as it's supposed to be a WIDTH measurement but as a 
        //temp variable here it's being used as HEIGHT measurement. 
        //Note, horizonCounter starts at 0. 
        retVal = horizonPointsGreen[WIDTH-horizonCounter];
        while (retVal == horizonPointsGreen[WIDTH-horizonCounter])
        {//{{{
          horizonCounter ++;
        }//}}}
        //Note that whilst horizonCounter finishes up 1 past the 
        //desired point, because the horizon array starts at 1 and 
        //not zero, we skip the +1 at the end. 
        retVal = WIDTH-horizonCounter;
      }//}}}
      else
      {//{{{
        //Below the lowest point in the horizon
        retVal = 0;
      }//}}}
    }//}}}
  }//}}}
  //Now we need to make sure retVal stays at least 2 pixels away from the edges
  //(because of the nature of the edge detection used - otherwise we run out of 
  //pixels and overrun the arrays). 
  if (retVal < 2)
  {//{{{
    retVal = 2;
  }//}}}
  else if (retVal > WIDTH-3)  //Remember, last element is WIDTH-1
  {//{{{
    retVal = WIDTH-3;
  }//}}}
  return retVal;
}//}}}

/**
 * End point for above. Note that if the runline has no pixels above 
 * the horizon, horizonStartX and horizonEndX should return the same number. 
 */
int 
VisualCortex::horizonEndX(int inH)
{//{{{
  int retVal = -1;
  if (horizonPointsGreen[0] == HORIZON_VERTICAL)
  {//{{{
    //Easy case first, if horizon is more vertical than horizontal 
    //then simply figure out which side of the horizon we're supposed 
    //to be on and return the appropriate value. 
    if (pan < 0)
    {//{{{
      //Head is turned to the right, we end at the horizon. 
      if (inH < HEIGHT)
      {
        retVal = horizonPointsGreen[inH+1];
      }
      else
      {
        retVal = 0;
        cout << "inH into horizonEndX is too large: " << inH << endl;
      }
    }//}}}
    else
    {//{{{
      //Head is turned to the left, we end at the right. 
      retVal = WIDTH;
    }//}}}
  }//}}}
  else
  {//{{{
    //Hard case. Horizon is more horizontal than vertical. Separate 
    //out if the horizon is sloping left or sloping right or dead 
    //straight. 
    if (horizonPointsGreen[1] == horizonPointsGreen[WIDTH])
    {//{{{
      //straight. When above the horizon we set the end 
      //to zero else we set the end to WIDTH. 
      if (horizonPointsGreen[1] > inH)
      {//{{{
        //Above horizon
        retVal = 0;
      }//}}}
      else
      {//{{{
        //Below horizon
        retVal = WIDTH;
      }//}}}
    }//}}}
    else if (horizonPointsGreen[1] < horizonPointsGreen[WIDTH])
    {//{{{
      //left higher than right. End point is 0 well above the 
      //horizon, WIDTH well below the horizon, and the horizon
      //itself in between. 
      if (horizonPointsGreen[1] > inH)
      {//{{{
        //Above the highest point in the horizon
        retVal = 0;
      }//}}}
      else if (horizonPointsGreen[WIDTH] > inH)
      {//{{{
        //Between the highest and lowest points in the horizon. 
        //I'm using retVal as a temp variable here, watch out 
        //as it's supposed to be a WIDTH measurement but as a 
        //temp variable here it's being used as a HEIGHT measurement. 
        //Note, horizonCounter starts at 0. 
        retVal = horizonPointsGreen[horizonCounter+1];
        while (retVal == horizonPointsGreen[horizonCounter+1])
        {//{{{
          horizonCounter ++;
        }//}}}
        //Note that horizonCounter+1 finishes 1 past the desired point. 
        //Since the horizon array is already indexed starting from 1, 
        //we must subtract 2 to get an index starting from 0. 
        retVal = horizonCounter-2;  
      }//}}}
      else
      {
        //Below ... end point is against the right.
        retVal = WIDTH;
      }
    }//}}}
    else
    {//{{{
      //right higher than left. End point is always WIDTH. 
      retVal = WIDTH; 
    }//}}}
  }//}}}
  //Now we need to make sure retVal stays at least 2 pixels away from the edges
  //(because of the nature of the edge detection used - otherwise we run out of 
  //pixels and overrun the arrays). 
  if (retVal < 2)
  {//{{{
    retVal = 2;
  }//}}}
  else if (retVal > WIDTH-3)  //Remember, last element is WIDTH-1
  {//{{{
    retVal = WIDTH-3;
  }//}}}
  return retVal;
}//}}}

#endif //HOUGH_EDGEDETECT_USEHORIZON




//}}}



/**
 * Reasonably specific matrix functions
 * - generatePicToFieldXfrm
 * - createMatrix4x4
 * - applyMatrix
 */
//{{{

/**
 * double * VisualCortex::generatePicToFieldXfrm()
 * {{{
 * Generates the transformation matrix that allows 
 * points in the co-ordinate space of the visual plane
 * to be transformed into local field plane co-ordinates.
 *
 * !!!!! Make this execute only once per frame coming in!
 * }}}
 */
double * VisualCortex::generatePicToFieldXfrm()
{//{{{
  double a = -FACE_LENGTH;
  double b = cosRAD(pan);
  double c = sinRAD(pan);
  double d = -c;
  double e = b;
  double f = NECK_LENGTH;
  double g = cosRAD((-(tilt+evilTiltOffset)));
  double h = -sinRAD((-(tilt+evilTiltOffset)));
  double ii = -h;
  double jj = g;
  double kk = NECKBASE_HEIGHT;
  return createMatrix4x4(  b,    0,    c,    c*a,
                          d*h,  g,    h*e,  a*h*e+f*g+kk,
                          d*jj,  ii,    e*jj,  a*e*jj+ii*f,
                          0,    0,    0,    1);
}//}}}
double* VisualCortex::createMatrix4x4(double a1, double b1, double c1, double d1, 
                                      double a2, double b2, double c2, double d2,
                                      double a3, double b3, double c3, double d3,
                                      double a4, double b4, double c4, double d4)
{//{{{
  //Assigns memory for a new 4x4 matrix and initializes it with the input values.
  double * retVal = (double*)malloc(4*4*sizeof(double));
  retVal[0*4+0] = a1;
  retVal[0*4+1] = b1;
  retVal[0*4+2] = c1;
  retVal[0*4+3] = d1;
  retVal[1*4+0] = a2;
  retVal[1*4+1] = b2;
  retVal[1*4+2] = c2;
  retVal[1*4+3] = d2;
  retVal[2*4+0] = a3;
  retVal[2*4+1] = b3;
  retVal[2*4+2] = c3;
  retVal[2*4+3] = d3;
  retVal[3*4+0] = a4;
  retVal[3*4+1] = b4;
  retVal[3*4+2] = c4;
  retVal[3*4+3] = d4;
  return retVal;
}//}}}
double * VisualCortex::applyMatrix(double * inMatrix, double * inVector)
{//{{{
  double * retVal = (double *)malloc(4*sizeof(double));
  int i, j;
  for (i = 0; i < 4; i ++)
  {//{{{
    retVal[i] = 0;
    for (j = 0; j < 4; j ++)
    {//{{{
      retVal[i] += inMatrix[i*4+j] * inVector[j];
    }//}}}
  }//}}}
  return retVal;
}//}}}
//}}}



/**
 * Somewhat generic trig functions, move these into somewhere where they're 
 * accessible to all ASAP.
 * - initDegSinLookup
 * - degSinLookup
 * - sinRAD
 * - sinDEG
 * - degCosLookup
 * - cosRAD
 * - cosDEG
 * - rad
 * - zeroTo360
 */
//{{{

void VisualCortex::initDegSinLookup()
{//{{{
  //Initializes lookup table.
  //Note that the table only stores sin values from 
  //0 to 90 degrees in degree increments. Any
  //other angle can be found from this and cosine 
  //can be found likewise.
  int i;
  for (i = 0; i < 91; i ++)
  {//{{{
    degSinLookupTable[i] = sin(rad(i));
  }//}}}
#ifndef PRE_SANITY_BLOB
  cout << "Sin lookup table initialized\n";
#endif
}//}}}
double VisualCortex::degSinLookup(int t)
{//{{{
  
  //Performs lookup on input t.
  //Make sure angle is between 0 and 360.
  //Yes this is iterative, it shouldn't be called
  //too often though.
  
  while (t >= 360)
  {//{{{
    t = t - 360;
  }//}}}
  while (t < 0)
  {//{{{
    t = t + 360;
  }//}}}
  //Now separate into the 4 possibilities.
  if (t < 90)
  {//{{{
    return degSinLookupTable[t];
  }//}}}
  else if (t < 180)
  {//{{{
    return degSinLookupTable[180-t];
  }//}}}
  else if (t < 270)
  {//{{{
    return -(degSinLookupTable[t-180]);
  }//}}}
  else
  {//{{{
    return -(degSinLookupTable[360-t]);
  }//}}}
//  return sin(rad(t));
}//}}}
double VisualCortex::sinRAD(double t)
{//{{{
  return sin(t);//degSinLookup((int)(t*360.0/2*3.141592654));
}//}}}
double VisualCortex::sinDEG(double t)
{//{{{
  return sin(t*3.14159/180.0);//degSinLookup((int)t);
}//}}}
double VisualCortex::degCosLookup(int t)
{//{{{
  //Like degSinLookup but for cosine.
  t = 90-t;
  return degSinLookup(t);
}//}}}
double VisualCortex::cosRAD(double t)
{//{{{
  return cos(t);//degCosLookup((int)(t*360.0/2*3.141592654));
}//}}}
double VisualCortex::cosDEG(double t)
{//{{{
  return cos(t*3.15159/180.0);//degCosLookup((int)t);
}//}}}
double VisualCortex::rad(int t)
{//{{{
  double angle = (double)t;
  return angle/180.0 * 3.141592654;
}//}}}
/*
void zeroTo360(int * inAngle)
{
  while (*inAngle < 0)
  {
    *inAngle = *inAngle + 360;
  }
  while (*inAngle >= 360)
  {
    *inAngle = *inAngle - 360;
  }
}*/
double VisualCortex::zeroTo360(double inAngle)
{
  while (inAngle < 0.0)
  {
    inAngle = inAngle + 360.0;
  }
  while (inAngle >= 360.0)
  {
    inAngle = inAngle - 360.0;
  }
  return inAngle;
}


//}}}


/**
 * NightOwl code stuff
 */
//{{{
/**
 * Overview of algorithm:
 *
 * - Generate gradient field (done at init time)
 * - Project edge points onto the field
 * - Convert edge points into global co-ordinates based on current known position
 * + For each test desired
 *   - Generate list of transforms
 *   - Generate transforms required to transform original position to test positions
 *   - Transform all points by each transform
 *   - Zero accumulator array
 *   + For each transform
 *     + For each transformed point
 *       - Look up error function in gradient field
 *       - Add error function to accumulator array in this transform's position
 *     = Next point
 *   = Next transform
 *   - Find the transform with the greatest count
 *   - Replace robot's current position with the current position transformed by the winning transform
 *   - Replace the field point list with the transformed field point list (from above)
 * = Next test
 *
 * 
 * Required storage:
 * Note, we try a number of transforms equal to n. This is equal to 
 * numAng * numTransX * numTransY, for 10 each way n becomes 1000. 
 * - Gradient field, 270x480 of uchars, 129600 in total
 * - Current best projected edge points converted to global, hard limit to 300 ints?
 * - Candidate edge points, 300n or about 300,000 of ints
 * - Candidate transforms, each 2x2 for a total of 4n or 4000 doubles
 * - Accumulator per transform, 1n of int or 1000 ints
 *
 *
 * NightOwl todo:
 * 5: - Face in dog
 * 4: - Ball sanities
 * 3: + Test intelligent things to do when facing the edge
 * 2a - Intelligent positioning for goalie
 * 2b + lines for goalie
 * 1: + Ball in goals (with goal in face)
 *
 *
 *
 * 
 * - Improbability filtering
 * - Cutoff for head looking down
 * - Fix variances
 *
 *
 * 
 * - Raise horizon
 * - Add goalboxes
 * - Add display for robot location
 * - Intelligent way of thinning out edge points
 * - Intelligent way of dealing with far away points
 * - Offset edge detect lines
 * - Way of dealing with points that end up outside field?
 * - Incorporation of derivative in final result
 * - Incorporation into GPS
 * - Experiment with alternative gradient fields
 *
 * Ideas:
 * - Simplex search?
 * - Line searches (1 dimensional each time)?
 * - Give cost to changes in angle or position based on GPS certainty? (eg. if angle is certain then don't allow angle to change much)
 */



//#define GF(x,y) gradField[(y)*(fieldWidth+2*offsetX)+(x)]
//#define DF(x,y) dirField[(y)*(fieldWidth+2*offsetX)+(x)]
#define IF(x,y) isolationField[(y)*(fieldWidth+2*offsetX)+(x)]
#define IFG(x,y) isolationFieldGoalie[(y)*(fieldWidth+2*offsetX)+(x)]

void VisualCortex::initCurrentPosition()
{//{{{
  int maxErr = 252;
  //Allocate stuff. Note that the point arrays should rarely be full as they'll 
  //only have numEdgePixel sets in them instead fo MAX_EDGE_PIXELS sets.  
//  fieldPointsLocal = (int*)malloc(MAX_EDGE_PIXELS*2*sizeof(int));
  fieldPointsGlobalOrig = (int*)malloc(MAX_EDGE_PIXELS*2*sizeof(int));
  fieldPointsGlobal = (int*)malloc(MAX_EDGE_PIXELS*maxNumSamplesX*maxNumSamplesY*maxNumSamplesT*2*sizeof(int));
  transformAccs = (int*)malloc(maxNumSamplesX*maxNumSamplesY*maxNumSamplesT*sizeof(int));
  //Just generate the gradient field. Note that field lines are represented by SINGLE lines here
  //(not double lines). Also note that because we need to represent points up to 0.5m outside the beacons, 
  //we need to add a factor whenever we reference gradField. Therefore, to get point (x,y) in gradField, 
  //we need to go:
  //gradField[x*cfX+y*cfY+gOff]
  //where gOff = (cfX*offsetX+cfY*offsetY)

  gradField = (unsigned char*)malloc(((fieldWidth+2*offsetX)/incX)*((fieldHeight+2*offsetY)/incY)*sizeof(unsigned char));
//  dirField = (unsigned char*)malloc(((fieldWidth+2*offsetX)/incX)*((fieldHeight+2*offsetY)/incY)* sizeof(unsigned char));
  isolationField = (unsigned char*)malloc(((fieldWidth+2*offsetX)/incX)*((fieldHeight+2*offsetY)/incY)* sizeof(unsigned char));
  isolationFieldGoalie = (unsigned char*)malloc(((fieldWidth+2*offsetX)/incX)*((fieldHeight+2*offsetY)/incY)* sizeof(unsigned char));

  //Zero out gradField (SHOULDN'T NEED TO DO THIS)
  for (int i = 0; i < ((fieldWidth+2*offsetX)/incX)*((fieldHeight+2*offsetY)/incY); i ++)
  {
    gradField[i] = 0;
//    dirField[i] = 0;
    isolationField[i] = 0;
    isolationFieldGoalie[i] = 0;
  }



/*  Enable to get corner points instead of lines
  //corners:
  //Either end of centre line
  gradField[(int)(220*cfY+10*cfX)+gOff] = maxErr;
  gradField[(int)(220*cfY+(fieldWidth-10)*cfX)+gOff] = maxErr;
  //Corner cutouts
  for (int i = 0; i < 32; i ++)
  {
    //Bottom left
    gradField[(int)((10+32-i)*cfY+(10+i)*cfX)+gOff] = maxErr;
    //Bottom right
    gradField[(int)((10+i)*cfY+(fieldWidth-10-32+i)*cfX)+gOff] = maxErr;
    //Top left
    gradField[(int)((fieldHeight-10-32+i)*cfY+(10+i)*cfX)+gOff] = maxErr;
    //Top right
    gradField[(int)((fieldHeight-10-i)*cfY+(fieldWidth-10-32+i)*cfX)+gOff] = maxErr;
  }
  //Corners of goal boxes
  gradField[(int)(10*cfY+95*cfX)+gOff] = maxErr;
  gradField[(int)(60*cfY+95*cfX)+gOff] = maxErr;
  gradField[(int)(10*cfY+195*cfX)+gOff] = maxErr;
  gradField[(int)(60*cfY+195*cfX)+gOff] = maxErr;
  gradField[(int)((fieldHeight-10)*cfY+95*cfX)+gOff] = maxErr;
  gradField[(int)((fieldHeight-60)*cfY+95*cfX)+gOff] = maxErr;
  gradField[(int)((fieldHeight-10)*cfY+195*cfX)+gOff] = maxErr;
  gradField[(int)((fieldHeight-60)*cfY+195*cfX)+gOff] = maxErr;
  //Corners of the goal itself
  gradField[(int)(115*cfX+10*cfY)+gOff] = maxErr;
  gradField[(int)(115*cfX-25*cfY)+gOff] = maxErr;
  gradField[(int)(175*cfX+10*cfY)+gOff] = maxErr;
  gradField[(int)(175*cfX-25*cfY)+gOff] = maxErr;
  gradField[(int)(115*cfX+(fieldHeight-10)*cfY)+gOff] = maxErr;
  gradField[(int)(115*cfX+(fieldHeight+25)*cfY)+gOff] = maxErr;
  gradField[(int)(175*cfX+(fieldHeight-10)*cfY)+gOff] = maxErr;
  gradField[(int)(175*cfX+(fieldHeight+25)*cfY)+gOff] = maxErr;
*/
  

  






  
  
  //Center line.
  for (int x = 10; x < fieldWidth-10; x += incX)
  {
    //Center line
    gradField[(int)(220*cfY+x*cfX)+gOff] = maxErr;
  }
  //Horizontal edges. Note that because of the corners, these are set in by 32cm each side. 
  for (int x = 42; x < fieldWidth-42; x += incX)
  {
    //Home edge/line
    gradField[(int)(10*cfY+x*cfX)+gOff] = maxErr;
    //Target edge/line
    gradField[(int)(430*cfY+x*cfX)+gOff] = maxErr;
  }
  //Vertical edges
  for (int y = 42; y < fieldHeight-42; y += incY)
  {
    //Left edge
    gradField[(int)(y*cfY+10*cfX)+gOff] = maxErr;
    //Right edge
    gradField[(int)(y*cfY+280*cfX)+gOff] = maxErr;
  }
  //Corners. Note that each corner cuts in by 32cm. 
  for (int i = 0; i < 32; i ++)
  {
    //Bottom left
    gradField[(int)((10+32-i)*cfY+(10+i)*cfX)+gOff] = maxErr;
    //Bottom right
    gradField[(int)((10+i)*cfY+(fieldWidth-10-32+i)*cfX)+gOff] = maxErr;
    //Top left
    gradField[(int)((fieldHeight-10-32+i)*cfY+(10+i)*cfX)+gOff] = maxErr;
    //Top right
    gradField[(int)((fieldHeight-10-i)*cfY+(fieldWidth-10-32+i)*cfX)+gOff] = maxErr;
  }
  //Tops of the goal boxes
  for (int i = 0; i < 100; i ++)
  {
    //Home goalbox
    gradField[(int)((10+50)*cfY+(95+i)*cfX)+gOff] = maxErr;
    //Target goalbox
    gradField[(int)((fieldHeight-10-50)*cfY+(95+i)*cfX)+gOff] = maxErr;
  }
  //Sides of goal boxes
  for (int i = 0; i < 50; i ++)
  {
    //Left of home
    gradField[(int)((10+i)*cfY+(95)*cfX)+gOff] = maxErr;
    //Right of home
    gradField[(int)((10+i)*cfY+(195)*cfX)+gOff] = maxErr;
    //Left of target
    gradField[(int)((fieldHeight-10-i)*cfY+(95)*cfX)+gOff] = maxErr;
    //Right of target
    gradField[(int)((fieldHeight-10-i)*cfY+(195)*cfX)+gOff] = maxErr;
  }
  /*
  //Goal sides
  for (int i = 0; i < 35; i ++)
  {
    //Left of home
    gradField[(int)(115*cfX+(10-i)*cfY)+gOff] = maxErr;
    //Right of home
    gradField[(int)(175*cfX+(10-i)*cfY)+gOff] = maxErr;
    //Left of target
    gradField[(int)(115*cfX+(fieldHeight-10+i)*cfY)+gOff] = maxErr;
    //Right of target
    gradField[(int)(175*cfX+(fieldHeight-10+i)*cfY)+gOff] = maxErr;
  }
  //Goal ends
  for (int i = 0; i < 60; i ++)
  {
    //Home
    gradField[(int)((115+i)*cfX+(-25)*cfY)+gOff] = maxErr;
    //Target
    gradField[(int)((115+i)*cfX+(fieldHeight+25)*cfY)+gOff] = maxErr;
  }
  */
  
  //Centre circle
  double circleRad = 14.0;//circle is of radius 16cm 
  //Do every degree (much finer than we really need but it doesn't matter if we set the 
  //same pixel twice anyway)
  for (int i = 0; i < 360; i ++)
  {
    int x = (int)(circleRad*cos(rad(i)));
    int y = (int)(circleRad*sin(rad(i)));
    //Move to the centre of the field
    x = x + (fieldWidth/2);
    y = y + (fieldHeight/2);
    gradField[(int)(x*cfX)+(int)(y*cfY)+gOff] = maxErr;
  }
  //Now learn the gradient and direction field.
  int delta = 1;
  while (delta > 0)
  {
    delta = 0;
    for (int y = 1; y < fieldHeight+2*offsetY-1; y ++)
    {
      for (int x = 1; x < fieldWidth+2*offsetX-1; x ++)
      {
        if (gradField[y*(fieldWidth+2*offsetX)+x] >= maxErr)
        {
          //do nothing to gradient field
        }
        else
        {
          int max = 0;
          if (gradField[(y-1)*(fieldWidth+2*offsetX)+x] > max)
          {
            max = gradField[(y-1)*(fieldWidth+2*offsetX)+x];
          }
          if (gradField[(y+1)*(fieldWidth+2*offsetX)+x] > max)
          {
            max = gradField[(y+1)*(fieldWidth+2*offsetX)+x];
          }
          if (gradField[(y)*(fieldWidth+2*offsetX)+(x-1)] > max)
          {
            max = gradField[(y)*(fieldWidth+2*offsetX)+(x-1)];
          }
          if (gradField[(y)*(fieldWidth+2*offsetX)+(x+1)] > max)
          {
            max = gradField[(y)*(fieldWidth+2*offsetX)+(x+1)];
          }
          if (GRADFIELD_DIAGONAL == 1)
          {
            if (gradField[(y-1)*(fieldWidth+2*offsetX)+(x-1)] > max)
            {
              max = gradField[(y-1)*(fieldWidth+2*offsetX)+(x-1)];
            }
            if (gradField[(y-1)*(fieldWidth+2*offsetX)+(x+1)] > max)
            {
              max = gradField[(y-1)*(fieldWidth+2*offsetX)+(x+1)];
            }
            if (gradField[(y+1)*(fieldWidth+2*offsetX)+(x-1)] > max)
            {
              max = gradField[(y+1)*(fieldWidth+2*offsetX)+(x-1)];
            }
            if (gradField[(y+1)*(fieldWidth+2*offsetX)+(x+1)] > max)
            {
              max = gradField[(y+1)*(fieldWidth+2*offsetX)+(x+1)];
            }
          }
          if (max == 0)
          {
            max = 1;  //this way, max-1 = 0
          }
          delta += abs(gradField[y*(fieldWidth+2*offsetX)+x] - (max-1));
          gradField[y*(fieldWidth+2*offsetX)+x] = max-1;
        }
      }
    }
  }
  //Form direction field. At this stage, for each point, test points that are 
  //a certain distance away in each of the 8 directions and set the lookat direction to be 
  //the highest one. If given points put us off the field, ignore them (the idea being that 
  //we're too close to the edge of the field at that point anyway). 
  /*
  //{{{
  int ds = 10; //short for direction separation
  for (int y = 0; y < fieldHeight+2*offsetY; y ++)
  {
    for (int x = 0; x < fieldWidth+2*offsetX; x ++)
    {
      int max = 0;
      int maxDirection = 0;
      int cx, cy, cf;
      
      //top left
      cx = x-ds; 
      cy = y+ds;
      if (cx > 0 && cx < fieldWidth+2*offsetX && cy > 0 && cy < fieldHeight+2*offsetY && (cf = gradField[cy*(fieldWidth+2*offsetX)+cx]) > max)
      {
        max = cf;
        maxDirection = 0;
      }
      //top
      cx = x; 
      cy = y+ds;
      if (cx > 0 && cx < fieldWidth+2*offsetX && cy > 0 && cy < fieldHeight+2*offsetY && (cf = gradField[cy*(fieldWidth+2*offsetX)+cx]) >= max) //>= to favor up
      {
        max = cf;
        maxDirection = 1;
      }
      //top right
      cx = x+ds; 
      cy = y+ds;
      if (cx > 0 && cx < fieldWidth+2*offsetX && cy > 0 && cy < fieldHeight+2*offsetY && (cf = gradField[cy*(fieldWidth+2*offsetX)+cx]) > max)
      {
        max = cf;
        maxDirection = 2;
      }
      //right
      cx = x+ds; 
      cy = y;
      if (cx > 0 && cx < fieldWidth+2*offsetX && cy > 0 && cy < fieldHeight+2*offsetY && (cf = gradField[cy*(fieldWidth+2*offsetX)+cx]) > max)
      {
        max = cf;
        maxDirection = 3;
      }
      //bottom right
      cx = x+ds; 
      cy = y-ds;
      if (cx > 0 && cx < fieldWidth+2*offsetX && cy > 0 && cy < fieldHeight+2*offsetY && (cf = gradField[cy*(fieldWidth+2*offsetX)+cx]) > max)
      {
        max = cf;
        maxDirection = 4;
      }
      //bottom
      cx = x; 
      cy = y-ds;
      if (cx > 0 && cx < fieldWidth+2*offsetX && cy > 0 && cy < fieldHeight+2*offsetY && (cf = gradField[cy*(fieldWidth+2*offsetX)+cx]) >= max) //>= to favor down
      {
        max = cf;
        maxDirection = 5;
      }
      //bottom left
      cx = x-ds; 
      cy = y-ds;
      if (cx > 0 && cx < fieldWidth+2*offsetX && cy > 0 && cy < fieldHeight+2*offsetY && (cf = gradField[cy*(fieldWidth+2*offsetX)+cx]) > max)
      {
        max = cf;
        maxDirection = 6;
      }
      //left
      cx = x-ds; 
      cy = y;
      if (cx > 0 && cx < fieldWidth+2*offsetX && cy > 0 && cy < fieldHeight+2*offsetY && (cf = gradField[cy*(fieldWidth+2*offsetX)+cx]) > max)
      {
        max = cf;
        maxDirection = 7;
      }
      dirField[y*(fieldWidth+2*offsetX)+x] = maxDirection*30;//x10 to make it show up when we test
    }
  }
//}}}
*/
 
/*
  //Alternative direction field method:
  //1: Project rays out in each of the 8 directions for a max distance of maxRayDist on the gradient field, ignoring anything 
  //    closer than minRayDist.
  //2: If there is any point between minRayDist and maxRayDist along that ray that is over rayMaximaCutoff then mark that direction as valid. Note that 
  //    we need to be moving UP a slope (this stops us from reporting 
  //Note that this may result in some places where we've got no direction to look at!
  int maxRayDist = 100;
  int minRayDist = 20;
  int rayMaximaCutoff = 250;
  for (int y = 1; y < fieldHeight+2*offsetY-1; y ++)
  {
    for (int x = 1; x < fieldWidth+2*offsetX-1; x ++)
    {


      int cx,cy,px,py;
      DF(x,y) = 0;
      int dx[] = {-1, 0, 1, 1, 1, 0,-1,-1};
      int dy[] = { 1, 1, 1, 0,-1,-1,-1, 0};
      uchar m[] = {1,2,4,8,16,32,64,128};
      
      for (int z = 0; z < 7; z ++)
      {
        for (int i = minRayDist; i < maxRayDist; i ++)
        {
          cx = x + i*dx[z];
          cy = y + i*dy[z];
          px = x + (i-1)*dx[z];  //Note, this will always be within the gradient field as long as cx and cy are. 
          py = y + (i-1)*dy[z];
          if (cx > 0 && cy > 0 && cx < fieldWidth+2*offsetX && cy < fieldHeight+2*offsetY   //we're still on the field
              &&
              GF(cx,cy) > rayMaximaCutoff //we're high enough
              &&
              GF(px,py) <= GF(cx,cy) //we're going uphill or we're flat (ie. looking along a line)
             )
          {
            //Mark this direction as valid and break out of the 'i' loop.
            DF(x,y) = DF(x,y) | m[z];
            i = maxRayDist;
          }
        }
      }
*/


      /*
      //dirField bits are numbered clockwise starting with the upper left point, a bit like edge detect. 
      //Max at end of ray (may not be needed if we're not stuck in the middle of nowhere)
      int endMax = 0;
      int endMaxIndex = -1;
      int currentMax = 0;
      int cx,cy,dx,dy,px,py;  //cx,cy = current, dx,dy = difference, px,py = previous
      
      //top left
      dx = -1;
      dy = 1;
      cx = x + minRayDist*dx;
      cy = y + minRayDist*dy;
      if (cx > 0 && cy > 0 && cx < fieldWidth+2*offsetX && cy < fieldHeight+2*offsetY)  //Only proceed if the starting point is on the field!
      {
        currentMax = 0;
        for (int i = minRayDist;+1 i < maxRayDist+1; i ++)  // +1 to test the final maximum idea
        {
          cx = x + i*dx; d
          cy = y + i*dy;
          px = x + (i-1)*dx;
          py = y + (i-1)*dy;
          if (cx > 0 && cy > 0 && cx < fieldWidth+2*offsetX && cy < fieldHeight+2*offsetY   //we're still on the field
              &&
              GF(cx,cy) > GF(px,py) //We're larger than the previous one (to make sure we're heading for real maxima)
              &&
              GF(cx,cy) > currentMax  //We're larger than the global max
              )
          {
          }
        }
      }
    }
  }
*/

  
  
  //Now deliberately de-emphasise anything lower than 240.
  for (int y = 1; y < fieldHeight+2*offsetY-1; y ++)
  {
    for (int x = 1; x < fieldWidth+2*offsetX-1; x ++)
    {
      if (gradField[y*(fieldWidth+2*offsetX)+x] < 250)
      {
        /*
        if (gradField[y*(fieldWidth+2*offsetX)+x] > 10)
        {
          gradField[y*(fieldWidth+2*offsetX)+x] = gradField[y*(fieldWidth+2*offsetX)+x] - 10;
        }
        else
        {
          gradField[y*(fieldWidth+2*offsetX)+x] = 0;
        }*/
        gradField[y*(fieldWidth+2*offsetX)+x] = gradField[y*(fieldWidth+2*offsetX)+x]/2;
      }
    }
  }



  //Set isolation regions to 1. This indicates that if more than isolationCutoff pixels fall into 
  //isolation regions then don't proceed any further. This is to prevent 'T' shapes from matching 
  //the centre circle, to prevent corners from matching the edges and the like. 
  
  //NOTE: For now, we'll set the entire centre of the field to isolation for forwards because we're having too many problems 
  //with things matching the centre circle and the goal boxes. For goalie (isolationFieldGoalie, IFG), we'll remove the isolation field from the 
  //goal box itself (but keep it there for the lines across the goal mouth). 
  
  //Area around centre circle.
  for (int h = offsetY; h < fieldHeight+offsetY; h ++)//(fieldHeight/2+offsetY) - 150; h < (fieldHeight/2+offsetY) + 150; h ++)
  {
    for (int w = 85+offsetX; w < 205+offsetX; w ++) //(fieldWidth/2+offsetX)-100; w < (fieldWidth/2+offsetX)+100; w ++)
    {
      //if (GF(w,h) < 200)
      {
        IF(w,h) = 1;
      }
    }
  }
  //thats all in the isolation region for now
  
/*
  //Isolate out everything except the regions around the goal box
  for (int h = offsetY + 55; h < fieldHeight+offsetY-55; h ++)
  {
    for (int w = 85+offsetX; w < 205+offsetX; w ++) //(fieldWidth/2+offsetX)-100; w < (fieldWidth/2+offsetX)+100; w ++)
    {
      IFG(w,h) = 1;
    }
  }
  //Now isolate the baseline (we shouldn't need the far baseline, if our goalie has snapped in THERE then we've got 
  //bigger problems!). 
  for (int h = offsetY; h < offsetY + 50; h ++)
  {
    for (int w = 100+offsetX; w < 190+offsetX; w ++)
    {
      IFG(w,h) = 1;
    }
  }
 */

  // We want the goalie to ONLY use the home baseline and the goalbox. We also want to only use lines
  // if we've got a really good match. 
  // 
  // Set everything
  for (int h = 1; h < fieldHeight+2*offsetY-1; h ++)
  {
    for (int w = 1; w < fieldWidth+2*offsetX-1; w ++)
    {
      IFG(w,h) = 1;
    }
  }
  // Mask out just the region around the goal
  for (int h = 1; h < offsetY + 70; h ++)
  {
    for (int w = offsetX + 50; w < fieldWidth + offsetX - 50; w ++)
    {
      if (GF(w,h) > 240)
      {
        IFG(w,h) = 0;
      }
    }
  }
  // Remove the goalmouth area
  for (int h = offsetY - 10; h < offsetY + 20; h ++)
  {
    for (int w = offsetX+100; w < offsetX+190; w ++)
    {
      IFG(w,h) = 1;
    }
  }

  
}//}}}

void VisualCortex::findCurrentPosition(int * inX, int * inY, int * inT)
{
  //A wrapper for the new findCurrentPosition
  findCurrentPositionNew(inX, inY, inT, fieldEdgeSpace, numEdgePixels);  
  //findCurrentPositionOld(inX, inY, inT);  //Uncomment this to re-enable the old version. 
}

void VisualCortex::findCurrentPositionNew(int * inX, int * inY, int * inT, int * inFieldEdgeSpace, int inNumEdgePixels)
{//{{{
  int fieldPointsNew[MAX_EDGE_PIXELS*2];

  //Stuff moved from runEdgeDetectHough
//  localToGlobal(*inX, *inY, *inT, inFieldEdgeSpace, inNumEdgePixels, inFieldEdgeSpace);
//  localToGlobal(145, 0, 90, inFieldEdgeSpace, inNumEdgePixels, inFieldEdgeSpace);
  
  int currentMaxCount = 0;
  int currentMaxIndex = -1;
//  if (sendDebugNow)
//cout << "Input to findCurrentPosition: " << *inX << "/" << *inY << "/" << *inT << endl;
// - Generate gradient field (done at init time)
// - Project edge points onto the field. Note that projectPoints takes integer parameters with y increasing upwards 
// and an origin in the centre of the screen. 
  if (inNumEdgePixels == 0)
  {
    edgesValid = false;
    return;
  }

  edgesValid = true;
  if (sendDebugNow)
  {
    cout << "About to pattern match\n";
  }
  memcpy(fieldPointsGlobalOrig, inFieldEdgeSpace, inNumEdgePixels*2*sizeof(int));
    
// + For each test desired "i"
  for (int i = 0; i < numTests; i ++)
  {

    //NOTE: We can find the transform given the index now anyway. 
//   - Generate list of transforms, Generate transforms required to transform current best position to test positions, Transform all points by each transform
    abortBorderSet = false;
//   - Zero accumulator array
    for (int j = 0; j < numSamplesX[i]*numSamplesY[i]*numSamplesT[i]; j ++)
    {
      transformAccs[j] = 0;
    }
//   + For each transform "j"
    for (int j = 0; j < numSamplesX[i]*numSamplesY[i]*numSamplesT[i]; j ++)
    {
      applyTransformNew(fieldPointsGlobalOrig, *inX, *inY, inNumEdgePixels, fieldPointsNew, i, j);
//     + For each transformed point "k"
      for (int k = 0; k < inNumEdgePixels; k ++)
      {
//       - Look up error function in gradient field and add error function to acc array in this transform's position.
//       NOTE: FIELDPOINTSGLOBAL HAS 3 DIMENSIONS.
//             1: Each test transformed set j, has size of inNumEdgePixels pixels. THIS IS VARIABLE.
//             2: Each pixel k, has size of 2.
//             3: Each co-ordinate, has size of int. 
        int x = fieldPointsNew[k*2+0];
        int y = fieldPointsNew[k*2+1];
        //Note, we need to make sure that we don't go outside the gradient field. If we do, set this transformAccs[j] to zero.
        //If all the transforms are zero (ie. however we move the points we're still outside the gradient field), just return. 
        //WARNING WARNING WARNING
        //THE ELEMENTS IN FIELDPOINTSGLOBAL MAY WELL BE OFF THE EDGE OF THE GRADIENT FIELD. MAKE SURE YOU CHECK FOR THIS IF YOU USE IT. 
        //this is NOT a bug - if we're looking over the edge of the field we may well get points out there but we won't know about it until 
        //we project them. 

        if (x > -offsetX/incX && x < (fieldWidth+offsetX)/incX && y > -offsetY/incY && y < (fieldHeight+offsetY)/incY)
        {
          transformAccs[j] += gradField[(int)(y*cfY+x*cfX)+gOff];
        }
//     = Next point
      }
      if (currentMaxCount <= transformAccs[j])
      {
        currentMaxCount = transformAccs[j];
        currentMaxIndex = j;
      }
//   = Next transform
    }
/*
//Print out transform grid. 
int blahindex = 0;
for (int y = 0; y < numSamplesYLC[i]; y ++)
{
  for (int x = 0; x < numSamplesXLC[i]; x ++)
  {
    fprintf(stdout, "%4.d ", transformAccs[blahindex]);
    blahindex += numSamplesTLC[i];
  }
  fprintf(stdout, "\n");
}
cout <<  "Winning transform has count " << currentMaxCount << endl;
  */  
    
//   - Find the transform with the greatest count, Replace robot's current position with the current position transformed by the winning transform
    if (currentMaxCount > 0)
    {
      applyTransformNew(fieldPointsGlobalOrig, *inX, *inY, inNumEdgePixels, fieldPointsNew, i, currentMaxIndex);
//   - Replace the field point list with the transformed field point list (from above)
    //NOTE: Is it cheaper to change pointers then reallocate or is it cheaper to copy over (as we are doing)?
    //WARNING WARNING WARNING
    //THE ELEMENTS IN FIELDPOINTSGLOBAL MAY WELL BE OFF THE EDGE OF THE GRADIENT FIELD. MAKE SURE YOU CHECK FOR THIS IF YOU USE IT. 
    //this is NOT a bug - if we're looking over the edge of the field we may well get points out there but we won't know about it until 
    //we project them. 
      for (int j = 0; j < inNumEdgePixels; j ++)
      {
        fieldPointsGlobalOrig[j*2+0] = fieldPointsNew[j*2+0];
        fieldPointsGlobalOrig[j*2+1] = fieldPointsNew[j*2+1];
      }
      transformXYT(inX, inY, inT, currentMaxIndex); //<--------

      //NOTE: this stuff can probably move to the bottom (outside the transform loop). 

      //See how many points fall into isolation regions in isolationField. If more than isolationCutoff of them do then discard
      //this one (by setting currentMaxCount to zero). Note that this is the new fieldPointsGlobalOrig as we've copied the final match 
      //into this array just above. 
      int isolationCount = 0;
      for (int j = 0; j < inNumEdgePixels; j ++)
      {
        int x = fieldPointsGlobalOrig[j*2+0];
        int y = fieldPointsGlobalOrig[j*2+1];
//        isolationCount += isolationField[(int)(y*cfY+x*cfX)+gOff];
/*  THIS IS FOR LOCALISATION CHALLENGE ONLY!        
        if (isGoalie)
        {
          isolationCount += isolationFieldGoalie[(int)(y*cfY+x*cfX)+gOff];
        }
        else
        {
          
          isolationCount += isolationField[(int)(y*cfY+x*cfX)+gOff];
        }
*/
        if (y > -offsetY/incY && y < (FIELD_LENGTH+2*offsetY)/incY && x > -offsetX/incX && x < (FIELD_WIDTH+2*offsetX)/incX)
        {
          isolationCount += isolationField[(int)(y*cfY+x*cfX)+gOff];
        }
        else
        {
          //we're off the isolation field!!!!!
          isolationCount += 1;
        }
      }
      if (isolationCount > isolationCutoff)
      {
        if (sendDebugNow) cout << "isolationCount = " << isolationCount << " is too high, original max count=" << currentMaxCount << endl;
        currentMaxCount = 0;
        //cout << "Isolation, setting currentMaxCount to zero\n";
      }
      if (sendDebugNow)
      cout << "IsolationCount = " << isolationCount << endl;
        
      //calculate the dx/dy/dt values (the minimal difference in each direction). Gives a rough 
      //guide as to how 'unique' this match is. 
      //Note:
      //To get to the next grid position in t, increment by 1.      
      //To get to the next grid position in y, increment by numSamplesT[iteration]
      //To get to the next grid position in x, increment by numSamplesT[iteration]*numSamplesY[iteration]
      //and vice versa. Note that we need to make sure we're never outside the grid. 
      patternDx = LARGE_VAL;
      patternDy = LARGE_VAL;
      patternDt = LARGE_VAL;
      int dxAmount = 1;
      //check either side in t
      if (currentMaxIndex + dxAmount < numSamplesT[i]*numSamplesY[i]*numSamplesX[i] && abs(transformAccs[currentMaxIndex + dxAmount] - transformAccs[currentMaxIndex]) < patternDt)
      {
        patternDt = abs(transformAccs[currentMaxIndex + dxAmount] - transformAccs[currentMaxIndex]);
      }
      if (currentMaxIndex - dxAmount > 0 && abs(transformAccs[currentMaxIndex - dxAmount] - transformAccs[currentMaxIndex]) < patternDt)
      {
        patternDt = abs(transformAccs[currentMaxIndex - dxAmount] - transformAccs[currentMaxIndex]);
      }
      //either side of y
      if (currentMaxIndex + dxAmount*numSamplesT[i] < numSamplesT[i]*numSamplesY[i]*numSamplesX[i] && abs(transformAccs[currentMaxIndex + dxAmount*numSamplesT[i]] - transformAccs[currentMaxIndex]) < patternDy)
      {
        patternDy = abs(transformAccs[currentMaxIndex + dxAmount*numSamplesT[i]] - transformAccs[currentMaxIndex]);
      }
      if (currentMaxIndex - dxAmount*numSamplesT[i] > 0 && abs(transformAccs[currentMaxIndex - dxAmount*numSamplesT[i]] - transformAccs[currentMaxIndex]) < patternDy)
      {
        patternDy = abs(transformAccs[currentMaxIndex - dxAmount*numSamplesT[i]] - transformAccs[currentMaxIndex]);
      }
      //either side of x
      if (currentMaxIndex + dxAmount*numSamplesT[i]*numSamplesY[i] < numSamplesT[i]*numSamplesY[i]*numSamplesX[i] && abs(transformAccs[currentMaxIndex + dxAmount*numSamplesT[i]*numSamplesY[i]] - transformAccs[currentMaxIndex]) < patternDx)
      {
        patternDx = abs(transformAccs[currentMaxIndex + dxAmount*numSamplesT[i]*numSamplesY[i]] - transformAccs[currentMaxIndex]);
      }
      if (currentMaxIndex - dxAmount*numSamplesT[i]*numSamplesY[i] > 0 && abs(transformAccs[currentMaxIndex - dxAmount*numSamplesT[i]*numSamplesY[i]] - transformAccs[currentMaxIndex]) < patternDx)
      {
        patternDx = abs(transformAccs[currentMaxIndex - dxAmount*numSamplesT[i]*numSamplesY[i]] - transformAccs[currentMaxIndex]);
      }
      
      if (patternDx == LARGE_VAL || patternDy == LARGE_VAL || patternDt == LARGE_VAL)
      {
        cout << "WARNING: patternDx = " << patternDx << " patternDy = " << patternDy << " patternDt = " << patternDt << endl;
      }
    }
    //Print out the accumulators.
    if (sendDebugNow)
    {
      int maxx = 0, maxy = 0, maxt = 0, maxCount = 0;
      fprintf(stdout, "Accumulators: \n");
      for (int y = numSamplesY[i]-1; y >= 0; y --)
      {
        cout << "y=" << y << "  ";
        for (int x = 0; x < numSamplesX[i]; x ++)
        {
          //Print out only the one for zero angle for now. 
          for (int t = 0; t < numSamplesT[i]; t ++)
          {
            if (t == (numSamplesT[i]-1)/2)
            {
              fprintf(stdout, "%.3d ", transformAccs[t+y*numSamplesT[i]+x*numSamplesT[i]*numSamplesY[i]]);
              if (maxCount < transformAccs[t+y*numSamplesT[i]+x*numSamplesT[i]*numSamplesY[i]])
              {
                maxCount = transformAccs[t+y*numSamplesT[i]+x*numSamplesT[i]*numSamplesY[i]];
                maxx = x;
                maxy = y;
                maxt = t;
              }
            }
          }
        }
        fprintf(stdout, "\n");
      }
      maxx = (maxx - (numSamplesX[i]-1)/2)*distSamplesX[i];
      maxy = (maxy - (numSamplesY[i]-1)/2)*distSamplesY[i];
      maxt = (maxt - (numSamplesT[i]-1)/2)*distSamplesT[i];
      cout << "\nBest match is with x=" << maxx << " y=" << maxy << " t=" << maxt << " count=" << maxCount << endl;
      cout << "Best match according to NightOwl is at index " << currentMaxIndex << " value=" << transformAccs[currentMaxIndex] << endl;
      cout << "This has the transform x=" << originTransform[currentMaxIndex*3+0] << " y=" << originTransform[currentMaxIndex*3+1] << " t=" << originTransform[currentMaxIndex*3+2] << endl;
    }

// = Next test
  }

  //now set the global edgeMatchQuality variable based on the final currentMaxCount
  //(divided by the number of edge pixels).
  edgeMatchQuality = currentMaxCount/inNumEdgePixels;
  if (sendDebugNow)
  cout << " Output of findCurrentPosition: " << *inX << "/" << *inY << "/" << *inT << endl;

}//}}}


void VisualCortex::applyTransformNew(int * inOrig, int inX, int inY, int inNumPoints, int * outFinal, int iteration, int inTransformIndex)
{//{{{
  //We loop in the order X, Y, T. Note that we loop both sides of zero (for instance, if numSamplesX[iteration] = 5 and distSamplesX[iteration] = 5 
  //then we go -10, -5, 0, 5, 10). This is why we MUST have odd numbers for numSamples![iteration]
  //Note that we don't add current angle into the equation because the points we've got are, by definition, already rotated!
  
  int x, y, t;
  x = inTransformIndex/(numSamplesY[iteration]*numSamplesT[iteration]);
  x = x - (numSamplesX[iteration]/2);
  x = x*distSamplesX[iteration];

  y = inTransformIndex%(numSamplesY[iteration]*numSamplesT[iteration]);
  y = y/(numSamplesT[iteration]);
  y = y - (numSamplesY[iteration]/2);
  y = y*distSamplesY[iteration];

  t = inTransformIndex%(numSamplesT[iteration]);
  t = t - (numSamplesT[iteration]/2);
  t = t*distSamplesT[iteration];
  double sinT = sinDEG(t);
  double cosT = cosDEG(t);
  originTransform[inTransformIndex*3+0] = x;      //We retain this so we set the origin transform array. 
  originTransform[inTransformIndex*3+1] = y;      //NOTE: this means this function MUST be called for EVERY 
  originTransform[inTransformIndex*3+2] = t;      //      possible value of inTransformIndex. 
  for (int j = 0; j < inNumPoints; j ++)
  {

    //First, translate such that inX/inY are the origin.
    outFinal[j*2+0] = inOrig[j*2+0] - inX;
    outFinal[j*2+1] = inOrig[j*2+1] - inY;
    //Now rotate the points about the new origin
    outFinal[j*2+0] = (int)((double)outFinal[j*2+0]*cosT-(double)outFinal[j*2+1]*sinT);
    outFinal[j*2+1] = (int)((double)outFinal[j*2+0]*sinT+(double)outFinal[j*2+1]*cosT);
    //Translate the points back to the original point and add the offset.
    outFinal[j*2+0] += (inX+x);
    outFinal[j*2+1] += (inY+y);
  }
  return;
  


  /*
  int transformIndex = 0;
  for (int x = -((numSamplesX[iteration] - 1)*distSamplesX[iteration])/2; x <= ((numSamplesX[iteration] - 1)*distSamplesX[iteration])/2; x += distSamplesX[iteration])
  {
    for (int y = -((numSamplesY[iteration] - 1)*distSamplesY[iteration])/2; y <= ((numSamplesY[iteration] - 1)*distSamplesY[iteration])/2; y += distSamplesY[iteration])
    {
      for (int t = -((numSamplesT[iteration] - 1)*distSamplesT[iteration])/2; t <= ((numSamplesT[iteration] - 1)*distSamplesT[iteration])/2; t += distSamplesT[iteration])
      {

        if (inTransformIndex == transformIndex)
        {
//          cout << "Iteration " << iteration << "Transform " << transformIndex << " x=" << x << " y=" << y << " t=" << t << endl;
          //Now take each of the inNumPoints points in inOrig, rotate them t degrees about the point (inX, inY) then 
          //translates them by x and y. It then stores them starting from index i in outFinal (note that i increases 
          //in jumps of inNumPoints!). 
          double sinT = sinDEG(t);
          double cosT = cosDEG(t);
          originTransform[transformIndex*3+0] = x;      //We retain this so we set the origin transform array. 
          originTransform[transformIndex*3+1] = y;      //NOTE: this means this function MUST be called for EVERY 
          originTransform[transformIndex*3+2] = t;      //      possible value of inTransformIndex. 
          for (int j = 0; j < inNumPoints; j ++)
          {

            //First, translate such that inX/inY are the origin.
            outFinal[j*2+0] = inOrig[j*2+0] - inX;
            outFinal[j*2+1] = inOrig[j*2+1] - inY;
            //Now rotate the points about the new origin
            outFinal[j*2+0] = (int)((double)outFinal[j*2+0]*cosT-(double)outFinal[j*2+1]*sinT);
            outFinal[j*2+1] = (int)((double)outFinal[j*2+0]*sinT+(double)outFinal[j*2+1]*cosT);
            //Translate the points back to the original point and add the offset.
            outFinal[j*2+0] += (inX+x);
            outFinal[j*2+1] += (inY+y);
          }
          return;
        }
        transformIndex  ++;
      }
    }
  }
  */
}//}}}



void VisualCortex:: findCurrentPositionOld(int * inX, int * inY, int * inT)
{//{{{
  int currentMaxCount = 0;
  int currentMaxIndex = -1;
  sendDebugNow = false;
  /*
#ifndef OFFLINE
  SystemTime curTime;
#endif
*/
// - Generate gradient field (done at init time)
// - Project edge points onto the field. Note that projectPoints takes integer parameters with y increasing upwards 
// and an origin in the centre of the screen. 
  if (numEdgePixels == 0)
  {
    edgesValid = false;
    return;
  }
  edgesValid = true;
  //Zero out arrays
  for (int i = 0; i < MAX_EDGE_PIXELS*maxNumSamplesX*maxNumSamplesY*maxNumSamplesT*2; i ++)
  {
    fieldPointsGlobal[i] = 0;
  }
  if (sendDebugNow)
  {
    cout << "About to pattern match\n";
  }
  if (numEdgePixels > MAX_EDGE_PIXELS)
  {
    numEdgePixels = MAX_EDGE_PIXELS;
  }
  memcpy(fieldPointsGlobalOrig, fieldEdgeSpace, numEdgePixels*2*sizeof(int));
// + For each test desired "i"
  for (int i = 0; i < numTests; i ++)
  {
    //NOTE: We can find the transform given the index now anyway. 
//   - Generate list of transforms, Generate transforms required to transform current best position to test positions, Transform all points by each transform
    abortBorderSet = false;
    transformFieldPoints(fieldPointsGlobalOrig, *inX, *inY, numEdgePixels, fieldPointsGlobal, i);
//   - Zero accumulator array
    for (int j = 0; j < numSamplesX[i]*numSamplesY[i]*numSamplesT[i]; j ++)
    {
      transformAccs[j] = 0;
    }
//   + For each transform "j"
    for (int j = 0; j < numSamplesX[i]*numSamplesY[i]*numSamplesT[i]; j ++)
    {
      //Check to see that we're not overtime. If we are, set edgesValid to false and break.
      /*
#ifndef OFFLINE
      GetSystemTime(&curTime);
      int timediff;
      if (curTime.useconds > timeStart.useconds)
      {
        timediff = (int)(curTime.useconds - timeStart.useconds);
      }
      else
      {
        //current time has wrapped to the next second. 
        timediff = (int)(curTime.useconds + 1000000 - timeStart.useconds);
      }
      if (abortTime < timediff)
      {
        edgesValid = false;
        if (VISUALCORTEX_TIME_PRINT)
        {
//          cout << "Aborting edge detect at test " << i << " transform " << j << ", we are " << ((curTime-timeStart).useconds) << " into the frame.\n";
          abortCount ++;
        }
        return;
      }
#endif
*/
//     + For each transformed point "k"
      for (int k = 0; k < numEdgePixels; k ++)
      {
//       - Look up error function in gradient field and add error function to acc array in this transform's position.
//       NOTE: FIELDPOINTSGLOBAL HAS 3 DIMENSIONS.
//             1: Each test transformed set j, has size of numEdgePixels pixels. THIS IS VARIABLE.
//             2: Each pixel k, has size of 2.
//             3: Each co-ordinate, has size of int. 
        int x = fieldPointsGlobal[(j*numEdgePixels+k)*2+0];
        int y = fieldPointsGlobal[(j*numEdgePixels+k)*2+1];
        //Note, we need to make sure that we don't go outside the gradient field. If we do, set this transformAccs[j] to zero.
        //If all the transforms are zero (ie. however we move the points we're still outside the gradient field), just return. 
        if (x > -offsetX && x < fieldWidth+offsetX && y > -offsetY && y < fieldHeight+offsetY)
        {
          transformAccs[j] += gradField[(int)(y*cfY+x*cfX)+gOff];
        }
//     = Next point
      }
      if (currentMaxCount <= transformAccs[j])
      {
        currentMaxCount = transformAccs[j];
        currentMaxIndex = j;
      }
//   = Next transform
    }
//   - Find the transform with the greatest count, Replace robot's current position with the current position transformed by the winning transform
    if (currentMaxCount > 0)
    {
//   - Replace the field point list with the transformed field point list (from above)
    //NOTE: Is it cheaper to change pointers then reallocate or is it cheaper to copy over (as we are doing)?
      for (int j = 0; j < numEdgePixels; j ++)
      {
        fieldPointsGlobalOrig[j*2+0] = fieldPointsGlobal[(currentMaxIndex*numEdgePixels+j)*2+0];
        fieldPointsGlobalOrig[j*2+1] = fieldPointsGlobal[(currentMaxIndex*numEdgePixels+j)*2+1];
      }
      transformXYT(inX, inY, inT, currentMaxIndex); //<--------

      //See how many points fall into isolation regions in isolationField. If more than isolationCutoff of them do then discard
      //this one (by setting currentMaxCount to zero). Note that this is the new fieldPointsGlobalOrig as we've copied the final match 
      //into this array just above. 
      int isolationCount = 0;
      for (int j = 0; j < numEdgePixels; j ++)
      {
        int x = fieldPointsGlobalOrig[j*2+0];
        int y = fieldPointsGlobalOrig[j*2+1];
//        isolationCount += isolationField[(int)(y*cfY+x*cfX)+gOff];
        if (isGoalie)
        {
          isolationCount += isolationFieldGoalie[(int)(y*cfY+x*cfX)+gOff];
        }
        else
        {
          
          isolationCount += isolationField[(int)(y*cfY+x*cfX)+gOff];
        }

      }
      if (isolationCount > isolationCutoff)
      {
        currentMaxCount = 0;
        //cout << "Isolation, setting currentMaxCount to zero\n";
      }
      if (sendDebugNow)
      cout << "IsolationCount = " << isolationCount << endl;
        
      //calculate the dx/dy/dt values (the minimal difference in each direction). Gives a rough 
      //guide as to how 'unique' this match is. 
      //Note:
      //To get to the next grid position in t, increment by 1.      
      //To get to the next grid position in y, increment by numSamplesT[iteration]
      //To get to the next grid position in x, increment by numSamplesT[iteration]*numSamplesY[iteration]
      //and vice versa. Note that we need to make sure we're never outside the grid. 
      patternDx = INT_MAX;
      patternDy = INT_MAX;
      patternDt = INT_MAX;
      //check either side in t
      if (currentMaxIndex + 1 < numSamplesT[i]*numSamplesY[i]*numSamplesX[i] && abs(transformAccs[currentMaxIndex + 1] - transformAccs[currentMaxIndex]) < patternDt)
      {
        patternDt = abs(transformAccs[currentMaxIndex + 1] - transformAccs[currentMaxIndex]);
      }
      if (currentMaxIndex - 1 > 0 && abs(transformAccs[currentMaxIndex - 1] - transformAccs[currentMaxIndex]) < patternDt)
      {
        patternDt = abs(transformAccs[currentMaxIndex - 1] - transformAccs[currentMaxIndex]);
      }
      //either side of y
      if (currentMaxIndex + numSamplesT[i] < numSamplesT[i]*numSamplesY[i]*numSamplesX[i] && abs(transformAccs[currentMaxIndex + numSamplesT[i]] - transformAccs[currentMaxIndex]) < patternDy)
      {
        patternDy = abs(transformAccs[currentMaxIndex + numSamplesT[i]] - transformAccs[currentMaxIndex]);
      }
      if (currentMaxIndex - numSamplesT[i] > 0 && abs(transformAccs[currentMaxIndex - numSamplesT[i]] - transformAccs[currentMaxIndex]) < patternDy)
      {
        patternDy = abs(transformAccs[currentMaxIndex - numSamplesT[i]] - transformAccs[currentMaxIndex]);
      }
      //either side of x
      if (currentMaxIndex + numSamplesT[i]*numSamplesY[i] < numSamplesT[i]*numSamplesY[i]*numSamplesX[i] && abs(transformAccs[currentMaxIndex + numSamplesT[i]*numSamplesY[i]] - transformAccs[currentMaxIndex]) < patternDx)
      {
        patternDx = abs(transformAccs[currentMaxIndex + numSamplesT[i]*numSamplesY[i]] - transformAccs[currentMaxIndex]);
      }
      if (currentMaxIndex - numSamplesT[i]*numSamplesY[i] > 0 && abs(transformAccs[currentMaxIndex - numSamplesT[i]*numSamplesY[i]] - transformAccs[currentMaxIndex]) < patternDx)
      {
        patternDx = abs(transformAccs[currentMaxIndex - numSamplesT[i]*numSamplesY[i]] - transformAccs[currentMaxIndex]);
      }
      
      if (patternDx >= LARGE_VAL)
      {
        cout << "WARNING: patternDx = infinity\n";
        patternDx = 0;
      }
      if (patternDy >= LARGE_VAL)
      {
        cout << "WARNING: patternDy = infinity\n";
        patternDy = 0;
      }
      if (patternDt >= LARGE_VAL)
      {
        cout << "WARNING: patternDt = infinity\n";
        patternDt = 0;
      }
    }
// = Next test
  }
  //now set the global edgeMatchQuality variable based on the final currentMaxCount
  //(divided by the number of edge pixels).
  edgeMatchQuality = currentMaxCount/numEdgePixels;
  
}//}}}



//Takes inNumPoints points as local co-ordinate pairs in inLocal and transforms them into
//global co-ordinate pairs in outGlobal based on inX, inY and inT. NOTE: outGlobal must be 
//twice inNumPoints in size (as must inLocal) or we'll have problems. This routine 
//works even if inLocal and outGlobal are the same. 
void VisualCortex::localToGlobal(int inX, int inY, int inT, int * inLocal, int inNumPoints, int * outGlobal)
{//{{{
  //First, we rotate all the points about the origin by an amount equal to inT-90 where
  //positive angle is COUNTERCLOCKWISE. For instance, if we're facing left (inT = 180) we 
  //need to rotate all the points counterclockwise by 90 degrees. The transformation matrix 
  //for this operation is [[cos -sin] [sin cos]]. Then add inX and inY. 
  double sinT = sin(rad(inT - 90));
  double cosT = cos(rad(inT - 90));
  for (int i = 0; i < inNumPoints; i ++)
  {
    double in0 = (double)inLocal[i*2+0];
    double in1 = (double)inLocal[i*2+1];
    if (inLocal[i*2+0] == LARGE_VAL || inLocal[i*2+1] == LARGE_VAL)
    {
      cout << "WARNING: point " << i << " in localToGlobal is at infinity.\n";
    }
    outGlobal[i*2+0] = (int)(in0*cosT-in1*sinT) + inX;
    outGlobal[i*2+1] = (int)(in0*sinT+in1*cosT) + inY;
  }
}//}}}  

//Generates a series of transforms based on the static consts at the start then 
//applies them to inOrig to get outFinal. NOTE THAT WE ASSUME OUTFINAL HAS 
//BEEN DEFINED AS BEING LARGE ENOUGH OTHERWISE WE'LL SEGFAULT!
void VisualCortex::transformFieldPoints(int * inOrig, int inX, int inY, int inNumPoints, int * outFinal, int iteration)
{//{{{
  //We loop in the order X, Y, T. Note that we loop both sides of zero (for instance, if numSamplesX[iteration] = 5 and distSamplesX[iteration] = 5 
  //then we go -10, -5, 0, 5, 10). This is why we MUST have odd numbers for numSamples![iteration]
  int i = 0; //Keep our own counter on outFinal for simplicity ... we could recalculate this based on x, y and t. 
  int transformIndex = 0;
  double sinTArray[numSamplesT[iteration]];
  double cosTArray[numSamplesT[iteration]];
  int tCount = 0;
  for (int t = -((numSamplesT[iteration] - 1)*distSamplesT[iteration])/2; t <= ((numSamplesT[iteration] - 1)*distSamplesT[iteration])/2; t += distSamplesT[iteration])
  {
    sinTArray[tCount] = sinDEG(t);
    cosTArray[tCount] = cosDEG(t);
    tCount ++;
  }
  for (int x = -((numSamplesX[iteration] - 1)*distSamplesX[iteration])/2; x <= ((numSamplesX[iteration] - 1)*distSamplesX[iteration])/2; x += distSamplesX[iteration])
  {
    for (int y = -((numSamplesY[iteration] - 1)*distSamplesY[iteration])/2; y <= ((numSamplesY[iteration] - 1)*distSamplesY[iteration])/2; y += distSamplesY[iteration])
    {
      tCount = 0;
      for (int t = -((numSamplesT[iteration] - 1)*distSamplesT[iteration])/2; t <= ((numSamplesT[iteration] - 1)*distSamplesT[iteration])/2; t += distSamplesT[iteration])
      {
        //NOTE: WE CAN MAKE SUBSTANTIAL SAVINGS BY MOVING THIS 'T' LOOP OUTSIDE SO WE ONLY DO THE DOUBLE 
        //ARITHMETIC ONCE PER ANGLE. 
        //
        //Now take each of the inNumPoints points in inOrig, rotate them t degrees about the point (inX, inY) then 
        //translates them by x and y. It then stores them starting from index i in outFinal (note that i increases 
        //in jumps of inNumPoints!). 
        double sinT = sinTArray[tCount];
        double cosT = cosTArray[tCount];
        tCount ++;
        originTransform[transformIndex*3+0] = x;
        originTransform[transformIndex*3+1] = y;
        originTransform[transformIndex*3+2] = t;
        for (int j = 0; j < inNumPoints; j ++)
        {
          //First, translate such that inX/inY are the origin.
          outFinal[(i+j)*2+0] = inOrig[j*2+0] - inX;
          outFinal[(i+j)*2+1] = inOrig[j*2+1] - inY;
          //Now rotate the points about the new origin
          outFinal[(i+j)*2+0] = (int)((double)outFinal[(i+j)*2+0]*cosT-(double)outFinal[(i+j)*2+1]*sinT);
          outFinal[(i+j)*2+1] = (int)((double)outFinal[(i+j)*2+0]*sinT+(double)outFinal[(i+j)*2+1]*cosT);
          //Translate the points back to the original point and add the offset.
          outFinal[(i+j)*2+0] += (inX+x);
          outFinal[(i+j)*2+1] += (inY+y);
        }
        i += inNumPoints;
        transformIndex ++;
      }
    }
  }
}//}}}

//Transforms a xyt (t=theta) triple by the number of the input transform (note that 
//the number of the transform is as defined by the static consts and as used by 
//transformFieldPoints and as in the outFinal array). 
void VisualCortex::transformXYT(int * inX, int * inY, int * inT, int inIndex)
{//{{{
  *inX += originTransform[inIndex*3+0];
  *inY += originTransform[inIndex*3+1];
  *inT += originTransform[inIndex*3+2];
}//}}}

//}}}


/**
 * End of NightOwl code stuff
 */




//}}}
#endif 
//EDGE_DETECT

